﻿/*---------------------------------------------------------------------------*
  (C) Copyright 1998-2014, Technologies Quazal Inc. All rights reserved.
  Quazal Net-Z, Quazal Rendez-Vous and Quazal are trademarks of Quazal.

  Copyright (C) 2011-2014 Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo and/or its licensors and are
  protected by national and international copyright laws.
  They may not be disclosed to third parties or copied or duplicated
  in any form, in whole or in part, without the prior written consent
  of Nintendo.
  The content herein is highly confidential and should be handled
  accordingly.
 *---------------------------------------------------------------------------*/
/*
 * NEX Sample
 *
 * AutoMatchmake
 *
 * Description:
 *
 *  This sample illustrates how auto-matchmaking works.
 *  During the auto-matchmaking process, the sample searches for a session matching the specified minimum and maximum number of participants, game mode, and matchmaking system type. If a matching session is found, the local user joins that session.
 *  After that, the sample creates a P2P session so that participants can communicate with each other. The P2P communication session persists for a specified period of time, after which the host (session master) notifies the other participants that the game is over and ends P2P communication.
 *  Finally, the local user leaves the matchmaking session.
 *  
 *  In this sample, the game starts as soon as the maximum number of participants (3 here) join the P2P session. The sample then closes the matchmaking session to further participation.
 *  Although more players than the maximum will never be able to join the session even if it is not explicitly closed, closing the session ensures that no new players will be able to join the game in progress if one of the existing participants leaves the session.
 *  Should you like new players to be able to join the session during an ongoing game, it is not necessary to close the session.
 *  
 *  This sample creates a single matchmaking session and uses it both for the auto-matchmaking search and for managing the P2P communication session.
 *  The sample associates the matchmaking session which the local user joins with a P2P session and then uses the same user as both the owner of the matchmaking session as well as the host of the P2P session (session master).
 *  Should the owner/host disconnect from the session for any reason, both the matchmaking session and the P2P session continue by giving the owner/host role to one of the remaining participants.
 *  The host migration extension feature is used to associate the P2P session with the matchmaking session.
 *  Should the host of the P2P session unintentionally disconnect, the sample selects one of the remaining participants as the new host and automatically modifies the host information and owner of the associated matchmaking session.
 *  
 *  Both host and non-host players end P2P communication safely and leave the matchmaking session if an unexpected error occurs or if they detect that they have become orphaned.
 *  The AutoMatchmakeSample::CheckCondition function contains all of the error handling required for P2P communication.
 *  
 *  If all the participants of the matchmaking session leave, the sample automatically deletes the session.
 *  As a result, the MatchMakingClient::UnregisterGathering function does not get called.
 *  
 *  Calls to the MatchmakeExtensionClient::AutoMatchmake and MatchmakeExtensionClient::EndParticipation functions are processed sequentially, so the state of the game does not actually change until they finish running.
 *  However, the MatchmakeExtensionClient::CloseParticipation function is designed to be callable at any time as long as the game is in a state in which P2P communication is possible.
 *  If during this asynchronous process an error occurs or the instance of AutoMatchmakeSample must be destroyed for any reason, all unfinished asynchronous processes are cancelled.
 *  Note that while you don't need to explicitly wait for processes started by asynchronous functions (which were designed specifically so that the above would work), you do need to make sure that all asynchronous processes are either finished or canceled before the instance of the AutoMatchmakeSample class is destroyed due to the way callback functions are called.
 *  
 *  The following sequences are implemented in the sample.
 * 
 * 1. Log in to server (in nex_util)
 * 2. Start auto-matchmaking and then join a matchmaking session
 * 3. Initialize and start P2P communication
 * 4. Allow players to use P2P communication to communicate with each other
 * 5. (Close the matchmaking session if the maximum number of players join)
 * 6. End P2P communications
 * 7. Leave the matchmaking session
 * 8. Log out from the server (in nex_util)
 * 
 * Example of how to embed this in an application:
 * 
 *  The AutoMatchmakeGameLoop function mimics the way the game loop works in an actual game. Implement this process in the game loop.
 *  We recommend calling the nn::nex::Scheduler::DispatchAll function and each of the calculation processes in the order they are called in this sample.
 *  Call functions in the AutoMatchmakeSample class from the main game loop to advance the sequential processes.
 *  Sequential processes are handled separately from the P2P communication processes. Editing the Communicate function allows you to use any kind of P2P communication desired.
 *  
 *  See P2PApiChain.h for more information on the initialization and finalization of P2P communication.
 *  Revise the AutoMatchmakeSample::AutoMatchmake function as necessary to make sure the necessary matchmaking parameters are passed to the MatchmakeExtension::AutoMatchmake function.
 *  
 *  In this sample, all errors that occur during sequential processes or P2P communication are handled appropriately within the range of the functions and sequences that are implemented herein.
 *  If you decide to add other sequences or modify which API functions are used, however, the errors that can occur in the NEX library may change and you may need additional user-defined result codes.
 *  In this case, you can use the AutoMatchmakeSample::GetLastResult function to get the last error that occurred and display it on screen in the application to notify the players or you can handle the error appropriately in the application.
 




*/

#include <nex.h>

#include "../common/NexUtil.h"
#include "../common/NexUtilNetZ.h"

#include "P2PApiChain.h"

//-----------------------------------------------------------------------------
// Constant Definitions
//-----------------------------------------------------------------------------
// Matchmaking mode
static const nn::nex::qUnsignedInt32 MATCHMAKE_GAME_MODE = 1002;

// Number of participants.
static const nn::nex::qUnsignedInt16 MATCHMAKE_MAX_PARTICIPANTS = 3;

// Duration (in frames) of P2P communication.
static const nn::nex::qUnsignedInt32 COMMUNICATION_DURATION = 60 * 60;  // 60 seconds

// User-defined result codes.
enum UserDefinedResultCode
{
    USER_DEFINED_RESULT_CODE_SUCCESS                         =  0,  // Process completed successfully, no errors.
    USER_DEFINED_RESULT_CODE_GAME_OVER                       = -1,  // Game over notification (game ends normally in this case).
    USER_DEFINED_RESULT_CODE_NETWORK_DISCONNECTED            = -2,  // Detected AC network disconnection.
    USER_DEFINED_RESULT_CODE_P2P_SESSION_ORPHAN              = -3,  // Orphan player in P2P session.
    USER_DEFINED_RESULT_CODE_MATCHMAKESESSION_UNREGISTERED   = -4,  // Matchmaking session unintentionally deleted.
    USER_DEFINED_RESULT_CODE_UNEXPECTED_HOST_CHANGE          = -5,  // Host changed unexpectedly.
    USER_DEFINED_RESULT_CODE_UNEXPECTED_RMC_FAILURE          = -6   // Unexpected RMC failure.
};

// Callback for when receiving a session join request.
static void CheckJoiningProcesses(nn::nex::JoinSessionOperation* pOperation)
{
    // Here you can deny participation in the P2P session, but denying participation at this stage may result in fewer successful matchmaking attempts.
    // We strongly recommend avoiding denial of participation based on application design whenever possible.
    // In this sample, we check to see whether the number of maximum players for a matchmaking session matches the specified value.
    // Although more than the specified maximum number of players will almost never be able to join the P2P session, it can happen in rare cases due to offsets/delays in the timing of processes.
    // 

    // Get the number of stations that have already been approved to join the session. (This number does not include stations that have only requested to join.)
    qUnsignedInt32 uiApprovedPlayerCount = nn::nex::Station::GetApprovedStationNum();

    // To check that a given matchmaking session meets the specified criteria, get the value set with the Station::SetLocalIdentificationToken function in advance and check that the values match for the local user and other participants in the session.
    // 
    // The Station::SetLocalIdentificationToken function is called from within the P2PStartApiChain::Call function.
    nn::nex::String password = _T("");
    if(nn::nex::Station::GetLocalInstance() != NULL)
    {
        password = nn::nex::Station::GetLocalInstance()->GetIdentificationToken();
    }

    NEX_DEMO_LOG("CheckJoiningProcesses Master-Passord: " STRING_FORMAT " JoiningStation-Passord: " STRING_FORMAT "\n",
        password.CStr(), pOperation->GetJoiningStation()->GetIdentificationToken().CStr());

    if (uiApprovedPlayerCount < MATCHMAKE_MAX_PARTICIPANTS && pOperation->GetJoiningStation()->GetIdentificationToken() == password)
    {
        pOperation->Approve();
    }
    else
    {
        // The sessions attempting to join fail with the asynchronous process of the Session::JoinSession function's QERROR(DOCore, JoinDenied)
        pOperation->Deny();
    }
}

// Operation callback used for P2P communication (used to start a callback function when a station joins or leaves a session).
static void StationOperationBeginCallback(nn::nex::StationOperation* pOperation)
{
    (void)pOperation;
    NEX_DEMO_LOG("Begin operation: " STRING_FORMAT ", StationID: %d\n", pOperation->GetClassNameString(), pOperation->GetTargetStation()->GetStationID());
}

// Operation callback used for P2P communication (used to start a callback function when a station joins or leaves a session).
static void StationOperationEndCallback(nn::nex::StationOperation* pOperation)
{
    (void)pOperation;
    NEX_DEMO_LOG("End operation: " STRING_FORMAT ", sid: %d, State: %d\n",
        pOperation->GetClassNameString(), pOperation->GetTargetStation()->GetStationID(), pOperation->GetTargetStation()->GetState());
}

//-----------------------------------------------------------------------------
// State machine for setting up auto-matchmaking and P2P communication.
//-----------------------------------------------------------------------------
class AutoMatchmakeSample
{
public:
    AutoMatchmakeSample(nn::nex::NgsFacade* pNgsFacade)
        : m_state(STATE_AUTOMATCHMAKE)
        , m_finished(false)
        , m_errorOccurred(false)
        , m_unexpectedHostChangeOccurred(false)
        , m_unexpectedRmcFailureOccurred(false)
        , m_lastResult()
        , m_myPid(INVALID_PRINCIPALID)
        , m_ownerPid(INVALID_PRINCIPALID)
        , m_gid(INVALID_GATHERINGID)
        , m_pNgsFacade(pNgsFacade)
        , m_pNetZ(NULL)
        , m_pHostMigrationCallback(NULL)
        , m_pMatchMakingClient(NULL)
        , m_pNATTraversalClient(NULL)
        , m_pNotificationHandler(NULL)
        , m_joinedMatchmakeSession()
        , m_pP2pInitializeApiChain(NULL)
        , m_pP2pStartApiChain(NULL)
        , m_pP2pShutdownApiChain(NULL)
    {
        Bind();
    };
    virtual ~AutoMatchmakeSample()
    {
        // Cancel all asynchronous processes that are not finished.
        m_callContextStore.CancelAll();
        Unbind();
    };
private:
    enum State
    {
        STATE_AUTOMATCHMAKE,
        STATE_P2PINITIALIZE,
        STATE_P2PINITIALIZE_WAITING,
        STATE_P2PSTART,
        STATE_P2PSTART_WAITING,
        STATE_COMMUNICATE,
        STATE_P2PSHUTDOWN,
        STATE_P2PSHUTDOWN_WAITING,
        STATE_ENDPARTICIPATION,
        STATE_RMC_RESULT_WAITING,
        STATE_FINISH
    };

    // Event handler to process notifications from the NGS
    class MyNotificationEventHandler : public nn::nex::NotificationEventHandler
    {
    public:
        MyNotificationEventHandler(AutoMatchmakeSample* sample):m_pSample(sample){};
        virtual ~MyNotificationEventHandler(){};

        void ProcessNotificationEvent(const nn::nex::NotificationEvent &oEvent)
        {
            nn::nex::StringStream ss;
            ss << oEvent;
            NEX_DEMO_LOG("[Notified] " STRING_FORMAT "\n", ss.CStr());

            // Process for each notification.
            // If necessary, temporarily add the notification to the queue and then take it back out and start the associated process at a better time later.
            switch (oEvent.GetType())
            {
            case nn::nex::NotificationEvents::ParticipationEvent:
                switch (oEvent.GetSubType())
                {
                case nn::nex::ParticipationEvents::Participate:
                    NEX_DEMO_LOG("  ParticipationEvents::Participate: [PID %u] join [gathering ID %u]\n", oEvent.GetParam2(), oEvent.GetParam1());
                    break;
                case nn::nex::ParticipationEvents::EndParticipation:
                    NEX_DEMO_LOG("  ParticipationEvents::EndParticipation: [PID %u] end [gathering ID %u]\n", oEvent.GetParam2(), oEvent.GetParam1());
                    break;
                case nn::nex::ParticipationEvents::Disconnect:
                    NEX_DEMO_LOG("  ParticipationEvents::Disconnect: [PID %u] disconnect [gathering ID %u]\n", oEvent.GetParam2(), oEvent.GetParam1());
                    break;
                default:
                    NEX_DEMO_LOG("  ParticipationEvents: unsupported participation event\n");
                    break;
                }
                break;
            case nn::nex::NotificationEvents::OwnershipChangeEvent:
                NEX_DEMO_LOG("  NotificationEvents::OwnershipChangeEvent: [gathering ID %u] changed owner to [PID %u]\n", oEvent.GetParam1(), oEvent.GetParam2());
                m_pSample->SetOwnerPid(oEvent.GetParam2());
                break;
            case nn::nex::NotificationEvents::GatheringUnregistered:
                NEX_DEMO_LOG("  NotificationEvents::GatheringUnregistered: [gathering ID %u] unregistered\n", oEvent.GetParam1());
                // In this sample, the matchmaking session will not be deleted as long as the local user is still participating in it.
                // Throw an error if the matchmaking session is deleted unexpectedly.
                m_pSample->SetGatheringID(INVALID_GATHERINGID);
                break;
            case nn::nex::NotificationEvents::HostChangeEvent:
                NEX_DEMO_LOG("  NotificationEvents::HostChangeEvent: [gathering ID %u] host information has been changed. New host is %u\n", oEvent.GetParam1(), oEvent.GetSource());
                // The NotificationEvents::HostChangeEvent notification is sent to the previous host.
                // In this sample, the host information is modified using only the host migration extension feature.
                // If the host receives this notification while in a state where it can still communicate, it determines that its own communication environment is poor and that it has been separated from the P2P session and throws an error.
                if (oEvent.GetParam1() == m_pSample->GetGatheringID())
                {
                    m_pSample->UnexpectedHostChangeOccurred();
                }
                break;
            default:
                NEX_DEMO_LOG("  NotificationEvents: unsupported notification event\n");
                break;
            }
        }
    private:
        AutoMatchmakeSample* m_pSample;
    };

public:
    qBool ProcessState();
    qBool CanCommunicate() const { return m_state == STATE_COMMUNICATE; }
    void FinishCommunication();
    nex_demo_util::ResultContainer GetLastResult() const { return m_lastResult; }
    void CloseParticipation();
private:
    void NotifyGameOverToAllParticipants();
    void UnexpectedHostChangeOccurred() { m_unexpectedHostChangeOccurred = true; }
    void UnexpectedRmcFailureOccurred() { m_unexpectedRmcFailureOccurred = true; }
    nex_demo_util::ResultContainer CheckCondition();
    void ErrorOccurred(const nex_demo_util::ResultContainer& result);
    void SetState(State state){m_state = state;};
    State GetState() const {return m_state;};
    nn::nex::PrincipalID GetMyPid() const {return m_myPid;};
    void SetOwnerPid(nn::nex::PrincipalID pid){m_ownerPid = pid;};
    nn::nex::PrincipalID GetOwnerPid() const {return m_ownerPid;};
    qBool IsOwner() const {return GetOwnerPid() == GetMyPid();};
    void SetGatheringID(nn::nex::GatheringID gid){m_gid = gid;};
    nn::nex::GatheringID GetGatheringID() const {return m_gid;};
    void Bind();
    void Unbind();
    void AutoMatchmake();
    void OnCompletionAutoMatchmake(nn::nex::CallContext* pCallContext);
    void EndParticipation();
    void OnCompletionEndParticipation(nn::nex::CallContext* pCallContext);
    void OnCompletionCloseParticipation(nn::nex::CallContext* pCallContext);

    State m_state;                                      // The state of sequential processes.
    qBool m_finished;                                   // Whether sequential processes are complete.
    qBool m_errorOccurred;                              // Whether in a state in which an error occurred.
    qBool m_unexpectedHostChangeOccurred;               // Whether there was an unexpected host change.
    qBool m_unexpectedRmcFailureOccurred;               // Whether there was an unexpected RMC failure.
    nex_demo_util::ResultContainer m_lastResult;        // Final result of sequential process.
    nex_demo_util::CallContextStore m_callContextStore; // Container that saves CallContext while an asynchronous process is running. Only stores processes that may need to be canceled.
    nn::nex::PrincipalID m_myPid;                       // Local user's principal ID.
    nn::nex::PrincipalID m_ownerPid;                    // The principal ID of the owner of the matchmaking session.
    nn::nex::GatheringID m_gid;                         // Gathering ID of the gathering the user is currently participating in.
    nn::nex::NgsFacade* m_pNgsFacade;                   // Pointer to NgsFacade object currently in a logged in state.
    nn::nex::NetZ* m_pNetZ;                             // Pointer that stores a Net-Z object.
    nn::nex::HostMigrationCallback* m_pHostMigrationCallback;   // Pointer to object on which the host migration extension feature was used.
    nn::nex::MatchmakeExtensionClient* m_pMatchMakingClient;    // Pointer to object used to call matchmaking functions.
    nn::nex::NATTraversalClient* m_pNATTraversalClient;         // Pointer to object required to use the NAT traversal feature.
    MyNotificationEventHandler* m_pNotificationHandler;         // Pointer to object that handles notifications from NGS.
    nn::nex::MatchmakeSession m_joinedMatchmakeSession;         // The matchmaking session that was joined.

    P2PInitializeApiChain* m_pP2pInitializeApiChain;    // Pointer to object for initializing P2P communications.
    P2PStartApiChain* m_pP2pStartApiChain;              // Pointer to object for starting P2P communications.
    P2PShutdownApiChain* m_pP2pShutdownApiChain;        // Pointer to object for ending P2P communications.
};

// Initialize each server feature.
void AutoMatchmakeSample::Bind()
{
    nn::nex::Credentials* pCred = m_pNgsFacade->GetCredentials();
    NEX_DEMO_ASSERT(pCred != NULL);
    m_myPid = pCred->GetPrincipalID();
    m_pMatchMakingClient = new nn::nex::MatchmakeExtensionClient();
    // Each Bind process always succeeds as long as pCred is not NULL.
    NEX_DEMO_VERIFY(m_pMatchMakingClient->Bind(pCred));
    m_pNATTraversalClient = new nn::nex::NATTraversalClient();
    NEX_DEMO_VERIFY(m_pNATTraversalClient->Bind(pCred));
    m_pNotificationHandler = new MyNotificationEventHandler(this);
    m_pNgsFacade->RegisterNotificationEventHandler(m_pNotificationHandler);
}

// Destroy each server feature.
void AutoMatchmakeSample::Unbind()
{
    m_pNgsFacade->UnregisterNotificationEventHandler(m_pNotificationHandler);
    delete m_pNotificationHandler;
    m_pNotificationHandler = NULL;
    m_pNATTraversalClient->Unbind();
    delete m_pNATTraversalClient;
    m_pNATTraversalClient = NULL;
    m_pMatchMakingClient->Unbind();
    delete m_pMatchMakingClient;
    m_pMatchMakingClient = NULL;
}

// In the P2P communication session, the session master notifies all participants that the game is over.
void AutoMatchmakeSample::NotifyGameOverToAllParticipants()
{
    if (nn::nex::Session::GetRole() == nn::nex::SESSION_ROLE_MASTER)
    {
        // The GameOver notification is sent to all participants asynchronously.
        NEX_DEMO_LOG("Notify GameOver\n");
        nn::nex::Session::GetInstance()->GameOver();
    }
}

// End P2P communication.
void AutoMatchmakeSample::FinishCommunication()
{
    NEX_DEMO_ASSERT(CanCommunicate());
    if (!CanCommunicate())
    {
        return;
    }

    // Notify all participants that the game is over.
    NotifyGameOverToAllParticipants();

    SetState(STATE_P2PSHUTDOWN);
}

// Check connection state. Only used for STATE_COMMUNICATE. Returns a user-defined error or a qResult.
nex_demo_util::ResultContainer AutoMatchmakeSample::CheckCondition()
{
    // Whether disconnected from the game server.
    nn::nex::qResult ngsResult = m_pNgsFacade->IsConnected();
    if(!ngsResult)
    {
        NEX_DEMO_LOG("Disconnected from NGS.\n");
        return nex_demo_util::ResultContainer(ngsResult);
    }

    // Whether the matchmaking session has been deleted.
    // In this sample matchmaking sessions are never actively deleted, so if this happens we need to throw an unexpected deletion error.
    if (GetState() > STATE_P2PINITIALIZE && 
        GetState() < STATE_P2PSHUTDOWN_WAITING && 
        GetGatheringID() == INVALID_GATHERINGID)
    {
        NEX_DEMO_LOG("MatchmakeSession is unregistered.\n");
        return nex_demo_util::ResultContainer(USER_DEFINED_RESULT_CODE_MATCHMAKESESSION_UNREGISTERED);
    }


    // Check whether currently using P2P communications.
    if (CanCommunicate())
    {
        // Net-Z error checking.
        if (m_pNetZ)
        {
            nn::nex::qResult netZError = m_pNetZ->GetFatalError();
            if (!netZError)
            {
                // An error from which restoration is not possible occurs
                NEX_DEMO_LOG("Fatal error occurred in NetZ.\n");
                netZError.Trace(TRACE_ALL);
                return nex_demo_util::ResultContainer(netZError);
            }
        }

        // Check that P2P session has not been orphaned.
        // Typically, when the local user is not the session master he/she can only be orphaned if network connectivity is poor and is disconnected from the P2P session.
        // Only checked during P2P communication (from when the P2P session is established to when NetZ::Terminate is called).
        if (m_pHostMigrationCallback && m_pHostMigrationCallback->SessionHostIsOrphan())
        {
            // If orphaned, handle it as an error and end the matchmaking session.
            NEX_DEMO_LOG("I'm an orphan.\n");
            return nex_demo_util::ResultContainer(USER_DEFINED_RESULT_CODE_P2P_SESSION_ORPHAN);
        }

        // If the host is unexpectedly changed, assume that the host has been separated from the P2P session and throw an error.
        // Typically, when the local user is the session master this occurs when network connectivity is poor and the user was disconnected from the P2P session.
        if (m_unexpectedHostChangeOccurred)
        {
            // This doesn't necessarily mean that the session master alone was orphaned, so all participants must be notified that the game is over.
            NEX_DEMO_LOG("Detected unexpected host change.\n");
            NotifyGameOverToAllParticipants();
            return nex_demo_util::ResultContainer(USER_DEFINED_RESULT_CODE_UNEXPECTED_HOST_CHANGE);
        }

        // Check whether the session is over.
        // This check is only necessary when using Session::GameOver.
        if (nn::nex::Session::GetInstance()->CheckGameIsOver())
        {
            NEX_DEMO_LOG("Notified GameOver.\n");
            return nex_demo_util::ResultContainer(USER_DEFINED_RESULT_CODE_GAME_OVER);
        }

        // Check for unexpected failure of any asynchronous processes.
        if (m_unexpectedRmcFailureOccurred)
        {
            NEX_DEMO_LOG("Detected unexpected failure.\n");
            return nex_demo_util::ResultContainer(USER_DEFINED_RESULT_CODE_UNEXPECTED_RMC_FAILURE);
        }
    }

    // Check the network connectivity.
    if(!nex_demo_util::IsNetworkConnected())
    {
        NEX_DEMO_LOG("Network connection unavailable.\n");
        return nex_demo_util::ResultContainer(USER_DEFINED_RESULT_CODE_NETWORK_DISCONNECTED);
    }
    return nex_demo_util::ResultContainer(USER_DEFINED_RESULT_CODE_SUCCESS);
}

// Start auto-matchmaking. Join or create a matchmaking session.
void AutoMatchmakeSample::AutoMatchmake()
{
    NEX_DEMO_LOG("\nCall AutoMatchmake().\n");

    // Set up parameters for matchmaking session if creating one.
    nn::nex::MatchmakeSession matchmakeSession;

    // Set the condition for closing.
    matchmakeSession.SetOpenParticipation(true);

    // The minimum and maximum number of participants.
    matchmakeSession.SetMinParticipants(1);
    matchmakeSession.SetMaxParticipants(MATCHMAKE_MAX_PARTICIPANTS);

    // Game mode.
    matchmakeSession.SetGameMode(MATCHMAKE_GAME_MODE);

    // Matchmaking system type.
    matchmakeSession.SetMatchmakeSystemType(nn::nex::MATCHMAKE_SYSTEM_TYPE_ANYBODY);

    // Whether to migrate ownership of the matchmaking session when the current owner leaves.
    matchmakeSession.SetFlag(nn::nex::GatheringFlags::MigrateOwner);

    // Whether another user can take ownership. Necessary because we are using the MatchmakingClient::StaticUpdateSessionHostWithOwner function.
    matchmakeSession.SetFlag(nn::nex::GatheringFlags::ChangeOwnerByOtherHost);

    // Set this flag if you want notifications such as participation notifications to be sent to all participants in a matchmaking session.
    // Use nn::nex::GatheringFlags::NotifyParticipationEventsToAllParticipantsReproducibly if you only want users that are already participating to receive the nn::nex::ParticipationEvents::Participate notification.
    // 
    // This can be effective when you want to manage the participants in a matchmaking session without using P2P communication.
    //matchmakeSession->SetFlag(nn::nex::GatheringFlags::NotifyParticipationEventsToAllParticipants);

    // Set up auto-matchmaking parameters.
    // Because we are not using the AutoMatchmakeParam::SetSearchCriteria or AutoMatchmakeParam::SetTargetGatheringId function, the search conditions are created automatically from the parameters specified with the AutoMatchmakeParam::SetSourceMatchmakeSession function.
    // 
    nn::nex::AutoMatchmakeParam autoMatchmakeParam;
    autoMatchmakeParam.SetSourceMatchmakeSession(matchmakeSession);
    autoMatchmakeParam.SetJoinMessage(_T("auto matchmake session"));

    // Change state ahead of time because the completion callback may be called when an RMC is called.
    SetState(STATE_RMC_RESULT_WAITING);

    // Run automatic matchmaking.
    nn::nex::ProtocolCallContext* pCallContext = new nn::nex::ProtocolCallContext();
    nex_demo_util::RegisterCallContextCallback(this, &AutoMatchmakeSample::OnCompletionAutoMatchmake, pCallContext);
    m_pMatchMakingClient->AutoMatchmake(pCallContext, autoMatchmakeParam, &m_joinedMatchmakeSession);
}

// Callback called when auto-matchmaking is complete.
void AutoMatchmakeSample::OnCompletionAutoMatchmake(nn::nex::CallContext* pCallContext)
{
    nn::nex::qResult result = pCallContext->GetOutcome();
    if (!result)
    {
        NEX_DEMO_LOG("     AutoMatchmake() failed.\n");
        result.Trace(TRACE_ALL);
        ErrorOccurred(nex_demo_util::ResultContainer(result));
        return;
    }

    NEX_DEMO_LOG("     AutoMatchmake() succeeded.\n");
    SetGatheringID(m_joinedMatchmakeSession.GetID());
    SetOwnerPid(m_joinedMatchmakeSession.GetOwnerPID());
    NEX_DEMO_LOG("[PID Info >> JoinedGID:%d OwnerPID:%d MyPID:%d]\n",
                GetGatheringID(), 
                GetOwnerPid(),
                GetMyPid());

    // Set encryption key for P2P communication which was generated on the server.
    nn::nex::Network::GetInstance()->SetP2PDataPacketSessionSignatureKey(m_joinedMatchmakeSession.GetSessionKey());

    SetState(STATE_P2PINITIALIZE);
}

// Leave matchmaking session.
void AutoMatchmakeSample::EndParticipation()
{
    NEX_DEMO_LOG("\nCall EndParticipation().\n");

    // Change state ahead of time because the completion callback may be called when an RMC is called.
    SetState(STATE_RMC_RESULT_WAITING);
    nn::nex::ProtocolCallContext* pCallContext = new nn::nex::ProtocolCallContext();
    nex_demo_util::RegisterCallContextCallback(this, &AutoMatchmakeSample::OnCompletionEndParticipation, pCallContext);
    m_pMatchMakingClient->EndParticipation(pCallContext, GetGatheringID());
    SetGatheringID(INVALID_GATHERINGID);
}

// Callback called when a player leaves the matchmaking session.
void AutoMatchmakeSample::OnCompletionEndParticipation(nn::nex::CallContext* pCallContext)
{
    nn::nex::qResult result = pCallContext->GetOutcome();
    if (!result && !(result == QERROR(RendezVous, SessionVoid)))
    {
        // Fails if the user is already disconnected from the server.
        NEX_DEMO_LOG("     EndParticipation() failed.\n");
        result.Trace(TRACE_ALL);
    }
    else
    {
        NEX_DEMO_LOG("     EndParticipation() succeeded.\n");
    }
    SetState(STATE_FINISH);
}

// Close matchmaking session.
// Called in a way that this doesn't depend on a state change. This process is automatically canceled if running when the AutoMatchmakeSample object is destroyed.
void AutoMatchmakeSample::CloseParticipation()
{
    NEX_DEMO_LOG("\nCall CloseParticipation().\n");

    nn::nex::ProtocolCallContext* pCallContext = new nn::nex::ProtocolCallContext();
    nex_demo_util::RegisterCallContextCallback(this, &AutoMatchmakeSample::OnCompletionCloseParticipation, pCallContext);

    // Only save the generated CallContext for asynchronous processes that are currently running in m_callContextStore so that they can be canceled later if necessary.
    nex_demo_util::CallContextStoreAssociator::Register(&m_callContextStore, pCallContext);
    m_pMatchMakingClient->CloseParticipation(pCallContext, GetGatheringID());
}

// Callback for closing the matchmaking session.
void AutoMatchmakeSample::OnCompletionCloseParticipation(nn::nex::CallContext* pCallContext)
{
    switch (pCallContext->GetState())
    {
    case nn::nex::CallContext::CallSuccess:
        NEX_DEMO_LOG("     CloseParticipation() succeeded.\n");
        break;
    case nn::nex::CallContext::CallCancelled:
        // Don't do anything here because the processes that will be canceled will be canceled from the ~AutoMatchmakeSample destructor.
        NEX_DEMO_LOG("     CloseParticipation() cancelled.\n");
        break;
    case nn::nex::CallContext::CallFailure:
        NEX_DEMO_LOG("     CloseParticipation() failed.\n");
        pCallContext->GetOutcome().Trace(TRACE_ALL);
        UnexpectedRmcFailureOccurred();
        break;
    default:
        NEX_DEMO_ASSERT(false);
        UnexpectedRmcFailureOccurred();
        break;
    }
}

// Transition to error state the first time an error occurs.
void AutoMatchmakeSample::ErrorOccurred(const nex_demo_util::ResultContainer& result)
{
    // Don't do anything if already in the error state.
    if (m_errorOccurred)
    {
        return;
    }

    m_errorOccurred = true;
    m_lastResult = result;

    // After P2P communication is initialized.
    if (m_pNetZ)
    {
        SetState(STATE_P2PSHUTDOWN);
    }
    // After joining the matchmaking session.
    else if (GetGatheringID() != INVALID_GATHERINGID)
    {
        SetState(STATE_ENDPARTICIPATION);
    }
    // Before joining the matchmaking session.
    else
    {
        SetState(STATE_FINISH);
    }
}

// Transition the state.
qBool AutoMatchmakeSample::ProcessState()
{
    switch(GetState())
    {
    // Auto-matchmaking.
    case STATE_AUTOMATCHMAKE:
        AutoMatchmake();
        break;
    // Initialize P2P communication.
    case STATE_P2PINITIALIZE:
        m_pP2pInitializeApiChain = new P2PInitializeApiChain(&m_pNetZ, &m_pHostMigrationCallback, m_pMatchMakingClient);
        m_pP2pInitializeApiChain->Call();
        SetState(STATE_P2PINITIALIZE_WAITING);
        break;
    // Wait for P2P communication to be initialized.
    case STATE_P2PINITIALIZE_WAITING:
        if (m_pP2pInitializeApiChain->IsCompleted())
        {
            nn::nex::qResult result = m_pP2pInitializeApiChain->GetResult();
            if (result)
            {
                SetState(STATE_P2PSTART);
            }
            else
            {
                ErrorOccurred(nex_demo_util::ResultContainer(result));
            }
            delete m_pP2pInitializeApiChain;
            m_pP2pInitializeApiChain = NULL;
        }
        break;
    // Start P2P communication.
    case STATE_P2PSTART:
        m_pP2pStartApiChain = new P2PStartApiChain(m_pMatchMakingClient, m_pHostMigrationCallback, IsOwner(), GetGatheringID());
        m_pP2pStartApiChain->Call();
        SetState(STATE_P2PSTART_WAITING);
        break;
    // Wait for P2P communication to start.
    case STATE_P2PSTART_WAITING:
        if (m_pP2pStartApiChain->IsCompleted())
        {
            nn::nex::qResult result = m_pP2pStartApiChain->GetResult();
            if (result)
            {
                SetState(STATE_COMMUNICATE);
            }
            else
            {
                ErrorOccurred(nex_demo_util::ResultContainer(result));
            }
            delete m_pP2pStartApiChain;
            m_pP2pStartApiChain = NULL;
        }
        break;
    // State in which P2P communication is possible.
    case STATE_COMMUNICATE:
        {
            nex_demo_util::ResultContainer result = CheckCondition();
            if (!result)
            {
                if (result == nex_demo_util::ResultContainer(USER_DEFINED_RESULT_CODE_GAME_OVER))
                {
                    // Go to finalization process if GAME_OVER.
                    SetState(STATE_P2PSHUTDOWN);
                }
                else
                {
                    // Go to error state if an error other than GAME_OVER occurs.
                    ErrorOccurred(result);
                }
            }
        }
        break;
    // Finalize P2P communication.
    case STATE_P2PSHUTDOWN:
        m_pP2pShutdownApiChain = new P2PShutdownApiChain(&m_pNetZ, &m_pHostMigrationCallback, GetGatheringID());
        m_pP2pShutdownApiChain->Call();
        SetState(STATE_P2PSHUTDOWN_WAITING);
        break;
    // Wait for P2P communication to be finalized.
    case STATE_P2PSHUTDOWN_WAITING:
        if (m_pP2pShutdownApiChain->IsCompleted())
        {
            // P2PShutdownApiChain does not fail.
            NEX_DEMO_ASSERT(m_pP2pShutdownApiChain->GetResult());
            SetState(STATE_ENDPARTICIPATION);
            delete m_pP2pShutdownApiChain;
            m_pP2pShutdownApiChain = NULL;
        }
        break;
    // Leave the gathering.
    case STATE_ENDPARTICIPATION:
        EndParticipation();
        break;
    // Wait for the result of the RMC processes (AutoMatchmake, EndParticipation).
    case STATE_RMC_RESULT_WAITING:
        break;
    // All processes are complete.
    case STATE_FINISH:
        m_finished = true;
        break;
    }

    return m_finished;
}

// Communicate using P2P communication.
// In this sample we use DirectStream. You can also share data using RMC and DOs.
static void Communicate(nn::nex::qUnsignedInt32 communicationFrame)
{
    nn::nex::DirectStream* pDirectStream = nn::nex::DirectStream::GetInstance();
    const qChar8 helloMessage[] = "Hi everyone.";
    qBool bReceivedHello = false;
    nn::nex::StationID helloSender = INVALID_DOHANDLE_VALUE;

    // Receive.
    {
        nn::nex::qVector<nn::nex::DirectStream::Data> recvData; // Vector for receiving DirectStream.
        pDirectStream->GetReceivedData(recvData);
        nn::nex::qVector<nn::nex::DirectStream::Data>::iterator it = recvData.begin();
        for (;it != recvData.end(); ++it)
        {
            nn::nex::RefBuffer& buffer = it->buffer;
            buffer[buffer.GetSize() - 1];   // Terminate buffer just in case.
            NEX_DEMO_LOG("Received from: %d, message: %s\n", it->id, buffer.Get());

            if (std::strcmp(reinterpret_cast<char*>(buffer.Get()), helloMessage) == 0)
            {
                helloSender = it->id;
                bReceivedHello = true;
            }
        }
    }

    // Send.
    {
        // If a ping is received, send a reply to the sender.
        if (bReceivedHello)
        {
            nn::nex::qChar8 responseMessage[64];
            snprintf(responseMessage, sizeof(responseMessage) - 1, "Hello %u", helloSender);
            nn::nex::qBool result = pDirectStream->SendUnreliable(helloSender, reinterpret_cast<const qByte*>(responseMessage), std::strlen(responseMessage) + 1);
            if (!result)
            {
                NEX_DEMO_LOG("Send Unreliable error. dstID:%d\n", helloSender);
            }
        }

        // Host (session master) pings other participants every 60 frames.
        if (communicationFrame % 60 == 0 && nn::nex::Session::GetRole() == nn::nex::SESSION_ROLE_MASTER)
        {
            nn::nex::Station::SelectionIterator itStations;
            for (itStations.GotoStart(); !itStations.EndReached(); ++itStations)
            {
                if (!itStations->IsLocal() && itStations->IsConnected())
                {
                    // Send to stations other than the local host
                    nn::nex::StationID dstID = itStations->GetStationID();
                    nn::nex::qBool result = pDirectStream->SendUnreliable(dstID, reinterpret_cast<const qByte*>(helloMessage), sizeof(helloMessage));
                    if (!result)
                    {
                        NEX_DEMO_LOG("Send Unreliable error. dstID:%u\n", dstID);
                    }
                }
            }
        }
    }
}

// Auto-matchmaking game loop.
static qBool AutoMatchmakeGameLoop(nn::nex::NgsFacade* pNgsFacade)
{
    AutoMatchmakeSample sample(pNgsFacade);
    nn::nex::qUnsignedInt32 communicationFrame = 0;
    qBool isMatchmakeSessionOpen = true;

    for(;;)
    {
        // Starting game frame.

        // When DispatchAll is called here, packets are mainly received.
        nn::nex::Scheduler::GetInstance()->DispatchAll();

        // Handle game key input here.
        // Perform game calculations here.

        // State transitions.
        qBool finished = sample.ProcessState();

        // Check whether finished.
        if(finished)
        {
            break;
        }

        // Process for when game is in state where P2P communication is possible.
        if (sample.CanCommunicate())
        {
            // Close the session to further participation once the maximum number of players join. Although more players than the maximum will not be able to join the session even if it is not explicitly closed, no new players will be able to join the game in progress if one of the existing participants leaves the session either.
            if (isMatchmakeSessionOpen && nn::nex::Session::GetRole() == nn::nex::SESSION_ROLE_MASTER)
            {
                if (nn::nex::Station::GetApprovedStationNum() >= MATCHMAKE_MAX_PARTICIPANTS)
                {
                    // In this sample we can call the CloseParticipation function because the host of the P2P session and owner of the matchmaking session are the same user.
                    // It takes a short amount of time to actually join the P2P session, so if you close the matchmaking session before the maximum number of players have joined, there is a small chance that one or more extra players will still get into the P2P session.
                    // 
                    sample.CloseParticipation();
                    isMatchmakeSessionOpen = false;
                }
            }

            // Start P2P communication.
            Communicate(communicationFrame);
            if (++communicationFrame > COMMUNICATION_DURATION)
            {
                // After the fixed period of time, end P2P communications.
                sample.FinishCommunication();
            }
        }

        // Ideally, the Scheduler::DispatchAll function should be called two times: before and after calculating each game frame. This reduces the delay in packet processes.
        // When DispatchAll is called here, packets are mainly sent.
        nn::nex::Scheduler::GetInstance()->DispatchAll();

        // Perform game rendering here.

        // Synchronize the game in an appropriate frame.
        nn::nex::Platform::Sleep(nex_demo_util::SLEEP_DURATION);
    }

    // When an error occurs, display the error message.
    nex_demo_util::ResultContainer result = sample.GetLastResult();
    if (!result)
    {
        // Use the EULA applet/error viewer as necessary to display the error message.
        result.ReportError("An error occurred in AutoMatchmakeSample.\n");
    }

    return true;
}

qInt NEXSampleMain(qInt intArg, void* ptrArg)
{
    (void)intArg;
    (void)ptrArg;

    // Connect to the Internet
    if (!nex_demo_util::NetInit())
    {
        NEX_DEMO_LOG("net init failed...\n");
        NEX_DEMO_LOG("Leaving application...\n");
        return 1;
    }

    // Prepare to use tools
    nex_demo_util::Initialize();

#if defined(CTR_TARGET)
    // Initialize the friend library.
    nn::friends::Initialize();
#endif

    // Initialization
    nn::nex::GlobalVariables::AcquireInstance();

    // Set to thread unsafe transport buffer thread mode
    nn::nex::Core::SetThreadMode(nn::nex::Core::ThreadModeUnsafeTransportBuffer);

    // Sets level of log.
    nn::nex::EventLog::GetInstance()->SetLogLevel(nn::nex::EventLog::Debug);
    nn::nex::TraceLog::GetInstance()->SetFlag(TRACE_ALWAYS);

    // If P2P communication is not working well, uncomment the following line and output the NAT traversal log.
    // 
    //nn::nex::TraceLog::GetInstance()->SetFlag(TRACE_ALWAYS|TRACE_NAT);

    // Set the callback for when receiving a session join request
    nn::nex::Session::RegisterJoinApprovalCallback(CheckJoiningProcesses);

    // Configure operation callbacks used for P2P communication (used to start a callback function when a station joins or leaves a session).
    nn::nex::Station::RegisterOperationBeginCallback(StationOperationBeginCallback);
    nn::nex::Station::RegisterOperationEndCallback(StationOperationEndCallback);

    // Set the network topology.
    nn::nex::NetZ::SetNetworkTopology(nn::nex::NetworkTopology::Peer2Peer);

    nn::nex::NgsFacade* pNgsFacade = new nn::nex::NgsFacade();

    // Login game loop.
    if (nex_demo_util::LoginGameLoop(pNgsFacade))
    {
        // Auto-matchmaking game loop.
        if (!AutoMatchmakeGameLoop(pNgsFacade))
        {
            NEX_DEMO_LOG("AutoMatchmakeGameLoop finished with error.\n");
        }

        // Logout game loop.
        NEX_DEMO_VERIFY(nex_demo_util::LogoutGameLoop(pNgsFacade));
    }

    delete pNgsFacade;

#if defined(CTR_TARGET)
    // Finalize the friend library.
    nn::friends::Finalize();
#endif

    // Finalize.
    nn::nex::GlobalVariables::ReleaseInstance();

    // Shut down utilities
    nex_demo_util::Finalize();

    // Disconnect from the Internet.
    nex_demo_util::NetTerm();

    NEX_DEMO_LOG("Leaving application...\n");

    return 0;
}

//-----------------------------------------------------------------------------
// Main Function
//-----------------------------------------------------------------------------
qInt GenericMain(qInt argc, qChar** argv)
{
    NEX_DEMO_LOG("*******************************\n");
    NEX_DEMO_LOG("       AutoMatchmake sample\n");
    NEX_DEMO_LOG("*******************************\n\n");

#if defined(CAFE_TARGET)

#else
    NEXSampleMain(argc, argv);
#endif

    NEX_DEMO_LOG("Leaving application...\n");

    return 0;
}

NEX_MAIN_WRAPPER(GenericMain)
