﻿/*---------------------------------------------------------------------------*
  (C) Copyright 1998-2014, Technologies Quazal Inc. All rights reserved.
  Quazal Net-Z, Quazal Rendez-Vous and Quazal are trademarks of Quazal.

  Copyright (C) 2011-2014 Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo and/or its licensors and are
  protected by national and international copyright laws.
  They may not be disclosed to third parties or copied or duplicated
  in any form, in whole or in part, without the prior written consent
  of Nintendo.
  The content herein is highly confidential and should be handled
  accordingly.
 *---------------------------------------------------------------------------*/
/*
 * NEX Sample
 *
 * Description:
 *
 * This file defines P2PInitializeApiChain, P2PStartApiChain, and P2PShutdownApiChain.
 * Each of these functions executes a series of fixed processes. They are used in the following order.
 * 
 * 1. P2PInitializeApiChain - Initializes P2P communication.
 * 2. (P2PStartApiChain)    Joins the P2P session and puts the game in a state where P2P communications can be sent and received.
 * 3. P2PShutdownApiChain - Finalizes P2P communication.
 * 
 * After using P2PInitializeApiChain, you can use P2PShutdownApiChain even if you haven't called P2PStartApiChain.
 * After calling P2PStartApiChain, you can use each of the communication features, such as DirectStream, sharing of DO's, and RMC.
 * You can also use the host migration extension feature to automatically migrate the host or owner of a session when necessary.
 * When the host (session master) disconnects from the P2P session, the host information and owner of the associated matchmaking session are modified automatically.
 * 
 * This feature can be used in applications as-is or revised as necessary for use in the application.
 */

#include <nex.h>

#include "../common/NexUtil.h"

// Declaration for the SetDOClassInitializationCallback function.
namespace nn { namespace nex {
extern void InitDOClasses();
}}


// Initialize P2P-related items.
// Initialize NAT traversal process, generate a Net-Z object, and initialize the host migration extension feature.
class P2PInitializeApiChain : public nex_demo_util::ApiChain
{
public:
    /*
     

        

     
     
     
     */
    P2PInitializeApiChain(nn::nex::NetZ** ppNetZ,
                     nn::nex::HostMigrationCallback** ppHostMigrationCallback,
                     nn::nex::MatchmakeExtensionClient* pMatchMakingClient)
        : m_ppNetZ(ppNetZ)
        , m_ppHostMigrationCallback(ppHostMigrationCallback)
        , m_pMatchMakingClient(pMatchMakingClient)
    {
        SYSTEMCHECK(m_ppNetZ != NULL);
        SYSTEMCHECK(*m_ppNetZ == NULL);
        SYSTEMCHECK(m_ppHostMigrationCallback != NULL);
        SYSTEMCHECK(*m_ppHostMigrationCallback == NULL);
        SYSTEMCHECK(m_pMatchMakingClient != NULL);
    }

    /*
     
     */
    void Call()
    {
        NEX_DEMO_LOG("\nCall P2PInitializeApiChain.\n");

        // Enable NAT traversal.
        nn::nex::ProtocolCallContext* pCallContext = new nn::nex::ProtocolCallContext();
        new nex_demo_util::CallContextCallback<P2PInitializeApiChain>(this, &P2PInitializeApiChain::OnCompletionStartNATSession, pCallContext);
        QVERIFY(nn::nex::ConnectivityManager::GetInstance()->StartNATSession(pCallContext));
    }

private:
    void OnCompletionStartNATSession(nn::nex::CallContext* pCallContext)
    {
        nn::nex::qResult result = pCallContext->GetOutcome();
        if (!result)
        {
            NEX_DEMO_LOG("     StartNATSession() failed.\n");
            result.Trace(TRACE_ALL);
            Complete(result);
            return;
        }

        // This process is necessary when you want to use DO's or the host migration extension feature.
        nn::nex::NetZ::SetDOClassInitializationCallback(&nn::nex::InitDOClasses);

        // Create a Net-Z object
        *m_ppNetZ = new nn::nex::NetZ;
        QVERIFY((*m_ppNetZ)->IsValid());

        // Set up name displayed in station.
        nex_demo_util::SetupLocalPlayerName();

        // Enable sending and receiving of DirectStream.
        // Note: The library discards any packets received before calling the DirectStream::Enable function.
        nn::nex::DirectStream::GetInstance()->Enable();

        // Enable host migration extension feature.
        // The host migration extension feature automatically updates the host information for a matchmaking session on the server when the P2P host leaves the P2P session.
        // In this sample, the host and owner are the same user.
        // You can also use the StaticUpdateSessionHostWithOwner function to change the owner at the same time you update the host information.
        // The user that becomes the host of the P2P session also becomes the owner of the matchmaking session.
        // HostMigrationExtension automatically calls the Unregister function when the Net-Z object is destroyed.
        nn::nex::HostMigrationExtension::Register();
        *m_ppHostMigrationCallback = new nn::nex::HostMigrationCallback(m_pMatchMakingClient, nn::nex::MatchMakingClient::StaticUpdateSessionHostWithOwner);
        Complete(nn::nex::qResult(true));
    }

private:
    nn::nex::NetZ** m_ppNetZ;
    nn::nex::HostMigrationCallback** m_ppHostMigrationCallback;
    nn::nex::MatchmakeExtensionClient* m_pMatchMakingClient;
};


// Joins the P2P session and puts the game in a state where P2P communications can be sent and received.
// If the local user is the owner of the matchmaking session, the local user also becomes host of the P2P session. If the local user is not the owner, the local user gets the host information from the matchmaking session and then joins the associated P2P session.
// 
class P2PStartApiChain : public nex_demo_util::ApiChain
{
public:
    /*
     

     
     
     
     
     */
    P2PStartApiChain(nn::nex::MatchmakeExtensionClient* pMatchMakingClient,
                     nn::nex::HostMigrationCallback* pHostMigrationCallback,
                     qBool isOwner,
                     nn::nex::GatheringID gid)
        : m_pMatchMakingClient(pMatchMakingClient)
        , m_pHostMigrationCallback(pHostMigrationCallback)
        , m_isOwner(isOwner)
        , m_gid(gid)
        , m_lstSessionHostURL()
    {
        SYSTEMCHECK(m_pMatchMakingClient != NULL);
        SYSTEMCHECK(m_pHostMigrationCallback != NULL);
        SYSTEMCHECK(m_gid != INVALID_GATHERINGID);
    }

    /*
     
     
        

     */
    void Call()
    {
        NEX_DEMO_LOG("\nCall P2PStartApiChain(). m_isOwner: %d, m_gid: %u\n", m_isOwner, m_gid);

        // Use the SetLocalIdentificationToken function to set the ID of the gathering we joined, so that clients in a different matchmaking session do not mistakenly join P2P communication.
        // 
        // Check the set value here with the CheckJoiningProcesses function
        nn::nex::StringStream password;
        password << m_gid;
        nn::nex::Station::SetLocalIdentificationToken(password.CStr());

        if (m_isOwner)
        {
            // If owner of the matchmaking session, create a P2P session and take the host role.

            // Create a P2P Session.
            QVERIFY(nn::nex::Session::CreateSession(NULL));

            // Register the host migration callback.
            m_pHostMigrationCallback->Register(m_gid);

            // Register the fact that the local user is the P2P host to the matchmaking session.
            nn::nex::ProtocolCallContext* pCallContext = new nn::nex::ProtocolCallContext();
            new nex_demo_util::CallContextCallback<P2PStartApiChain>(this, &P2PStartApiChain::OnCompletionUpdateSessionHost, pCallContext);
            m_pMatchMakingClient->UpdateSessionHost(pCallContext, m_gid);
            return;
        }
        else
        {
            // If not owner of the matchmaking session, join the P2P session.

            // Get the P2P host information required to join the P2P session from the matchmaking session.
            nn::nex::ProtocolCallContext* pCallContext = new nn::nex::ProtocolCallContext();
            new nex_demo_util::CallContextCallback<P2PStartApiChain>(this, &P2PStartApiChain::OnCompletionGetSessionURLs, pCallContext);
            m_pMatchMakingClient->GetSessionURLs(pCallContext, m_gid, &m_lstSessionHostURL);
            return;
        }
    }

private:
    void OnCompletionUpdateSessionHost(nn::nex::CallContext* pCallContext)
    {
        nn::nex::qResult result = pCallContext->GetOutcome();
        if (!result)
        {
            NEX_DEMO_LOG("     UpdateSessionHost() failed.\n");
            result.Trace(TRACE_ALL);
            Complete(result);
            return;
        }
        Complete(nn::nex::qResult(true));
    }

    void OnCompletionGetSessionURLs(nn::nex::CallContext* pCallContext)
    {
        nn::nex::qResult result = pCallContext->GetOutcome();
        if (!result)
        {
            NEX_DEMO_LOG("     GetSessionURLs() failed.\n");
            result.Trace(TRACE_ALL);
            Complete(result);
            return;
        }

        // Join the P2P session.
        nn::nex::CallContext* pNewCallContext = new nn::nex::CallContext();
        new nex_demo_util::CallContextCallback<P2PStartApiChain>(this, &P2PStartApiChain::OnCompletionJoinSession, pNewCallContext);
        nn::nex::Session::JoinSession(pNewCallContext, m_lstSessionHostURL);
    }

    void OnCompletionJoinSession(nn::nex::CallContext* pCallContext)
    {
        nn::nex::qResult result = pCallContext->GetOutcome();
        if (!result)
        {
            NEX_DEMO_LOG("     JoinSession() failed.\n");
            result.Trace(TRACE_ALL);
            Complete(result);
            return;
        }
        // Register the host migration callback.
        m_pHostMigrationCallback->Register(m_gid);
        Complete(nn::nex::qResult(true));
    }
    nn::nex::MatchmakeExtensionClient* m_pMatchMakingClient;
    nn::nex::HostMigrationCallback* m_pHostMigrationCallback;
    qBool m_isOwner;
    nn::nex::GatheringID m_gid;
    nn::nex::qList<nn::nex::StationURL> m_lstSessionHostURL;
};


// Finalize P2P communication.
// Finalize the host migration extension feature, destroy the Net-Z object, and finalize the NAT traversal process.
// This is required if you used P2PInitializeApiChain. Do not call this function unless you already called P2PInitializeApiChain.
// This asynchronous process will never fail.
class P2PShutdownApiChain : public nex_demo_util::ApiChain
{
public:
    /*
     

        

     
     
     
     */
    P2PShutdownApiChain(nn::nex::NetZ** ppNetZ,
                        nn::nex::HostMigrationCallback** ppHostMigrationCallback,
                        nn::nex::GatheringID gid)
        : m_ppNetZ(ppNetZ)
        , m_ppHostMigrationCallback(ppHostMigrationCallback)
        , m_gid(gid)
    {
        SYSTEMCHECK(m_ppNetZ != NULL);
        SYSTEMCHECK(*m_ppNetZ != NULL);
        SYSTEMCHECK(m_ppHostMigrationCallback != NULL);
        SYSTEMCHECK(*m_ppHostMigrationCallback != NULL);
        SYSTEMCHECK(m_gid != INVALID_GATHERINGID);
    }

    /*
     
     */
    void Call()
    {
        NEX_DEMO_LOG("\nCall P2PShutdownApiChain().\n");

        // Disable the host migration extension feature. You can call the Unregister function even if you didn't call the HostMigrationCallback::Register function before.
        (*m_ppHostMigrationCallback)->Unregister(m_gid);
        delete (*m_ppHostMigrationCallback);
        *m_ppHostMigrationCallback = NULL;

        // Disable sending and receiving of packets.
        nn::nex::DirectStream::GetInstance()->Disable();

        // Finalize Net-Z.
        nn::nex::ProtocolCallContext* pCallContext = new nn::nex::ProtocolCallContext();
        new nex_demo_util::CallContextCallback<P2PShutdownApiChain>(this, &P2PShutdownApiChain::OnCompletionTerminate, pCallContext);
        (*m_ppNetZ)->Terminate(pCallContext);
    }

private:
    void OnCompletionTerminate(nn::nex::CallContext* pCallContext)
    {
        nn::nex::qResult result = pCallContext->GetOutcome();
        if (!result)
        {
            NEX_DEMO_LOG("     Terminate() failed.\n");
            result.Trace(TRACE_ALL);
            // Assume that NetZ::Terminate never fails. Continue the finalization process.
        }

        // The Net-Z object cannot be reused for other P2P sessions, so delete it after every P2P session.
        delete *m_ppNetZ;
        *m_ppNetZ = NULL;

        // Disable NAT traversal.
        nn::nex::ConnectivityManager::GetInstance()->StopNATSession();
        Complete(nn::nex::qResult(true));
    }
    nn::nex::NetZ** m_ppNetZ;
    nn::nex::HostMigrationCallback** m_ppHostMigrationCallback;
    nn::nex::GatheringID m_gid;
};
