﻿/*---------------------------------------------------------------------------*
  (C) Copyright 1998-2014, Technologies Quazal Inc. All rights reserved.
  Quazal Net-Z, Quazal Rendez-Vous and Quazal are trademarks of Quazal.

  Copyright (C) 2011-2014 Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo and/or its licensors and are
  protected by national and international copyright laws.
  They may not be disclosed to third parties or copied or duplicated
  in any form, in whole or in part, without the prior written consent
  of Nintendo.
  The content herein is highly confidential and should be handled
  accordingly.
 *---------------------------------------------------------------------------*/
/*
 * NEX Sample
 *
 * CreateAndBrowse
 *
 * Description:
 * This sample creates, searches for and joins in matchmaking sessions.
 *
 * This sample does not use P2P communication.
 * See the AutoMatch sample for a combination of server and P2P communication.
 *
 * This sample runs in thread unsafe transport buffer thread mode.
 *
 */

#include <nex.h>
#include <iterator>
#include "../common/NexUtil.h"
#include "../common/DemoPad.h"

//-----------------------------------------------------------------------------
// Constant Definitions
//-----------------------------------------------------------------------------
// Number of connected persons
#define MATCHMAKE_MAX_PARTICIPANTS 2
// Amount of time to sleep
#define DISPATCH_PERIOD 100
// Time to wait for a dispatch.
#define DISPATCH_TIMEOUT 3
// Application-specific game mode.
#define MATCHMAKE_GAME_MODE 1001

//-----------------------------------------------------------------------------
// Function definitions
//-----------------------------------------------------------------------------
void RunMatchmakingLoop(nn::nex::NgsFacade *pNgsFacade,
                        nn::nex::MatchmakeExtensionClient *pMatchMakingClient);

//-----------------------------------------------------------------------------
// Event handler to process notifications from the NGS
//-----------------------------------------------------------------------------

class MyNotificationEventHandler : public nn::nex::NotificationEventHandler
{
public:
    void ProcessNotificationEvent(const nn::nex::NotificationEvent &oEvent)
    {
        nn::nex::StringStream ss;
        ss << oEvent;
        NEX_DEMO_LOG("Notification %ls\n\n",ss.CStr());

        // Per-message processing
        switch (oEvent.GetType())
        {
        case nn::nex::NotificationEvents::ParticipationEvent:
            switch (oEvent.GetSubType())
            {
            case nn::nex::ParticipationEvents::Participate:
                NEX_DEMO_LOG("PARTICIPATION_EVENT PARTICIPATE: [PID %u] join [session %u]\n", oEvent.GetParam2(), oEvent.GetParam1());
                break;

            case nn::nex::ParticipationEvents::EndParticipation:
                NEX_DEMO_LOG("PARTICIPATION_EVENT END: [PID %u] end [session %u]\n", oEvent.GetParam2(), oEvent.GetParam1());
                break;

            case nn::nex::ParticipationEvents::Disconnect:
                NEX_DEMO_LOG("PARTICIPATION_EVENT DISCONNECT: [PID %u] disconnect [session %u]\n", oEvent.GetParam2(),
                             oEvent.GetParam1());
                break;

            default:
                NEX_DEMO_LOG("PARTICIPATION_EVENT: unsupported participation event\n");
                break;
            }
            break;

        case nn::nex::NotificationEvents::OwnershipChangeEvent:
            NEX_DEMO_LOG("NOTIFICATION_EVENT OWNERSHIPCHANGE: [session %u] changed owner to [PID %u]\n",
                         oEvent.GetParam1(), oEvent.GetParam2());
            break;

        case nn::nex::NotificationEvents::GatheringUnregistered:
            NEX_DEMO_LOG("NOTIFICATION_EVENT GATHERINGUNREGISTERED: [session %u] unregistered\n",
                         oEvent.GetParam1());
            break;

        default:
            NEX_DEMO_LOG("NOTIFICATION_EVENT: unsupported notification event\n");
            break;
        }
    }
};


//-----------------------------------------------------------------------------
// Variable Definitions
//-----------------------------------------------------------------------------
// Gathering ID to use for matchmaking
static nn::nex::GatheringID s_JoinedGatheringID = INVALID_GATHERINGID;

// Event handler to process notifications
static MyNotificationEventHandler s_MyNotificationEventHandler;


qInt NEXSampleMain(qInt intArg, void* ptrArg)
{
    (void)intArg;
    (void)ptrArg;

    // Connect to the Internet
    if (!nex_demo_util::NetInit())
    {
        NEX_DEMO_LOG("net init failed...\n");
        NEX_DEMO_LOG("Leaving application...\n");

        return 0;
    }

    // Prepare to use tools
    nex_demo_util::Initialize();

    // Initialization
    nn::nex::GlobalVariables::AcquireInstance();

    // Set to unsafe transport buffer thread mode
    nn::nex::Core::SetThreadMode(nn::nex::Core::ThreadModeUnsafeTransportBuffer);

    nn::nex::EventLog::GetInstance()->SetLogLevel(nn::nex::EventLog::Debug);
    nn::nex::TraceLog::GetInstance()->SetFlag(TRACE_ALWAYS);

    nn::nex::Credentials              *pCredentials        = NULL;
    nn::nex::NgsFacade                *pNgsFacade          = new nn::nex::NgsFacade();
    nn::nex::MatchmakeExtensionClient *pMatchMakingClient  = new nn::nex::MatchmakeExtensionClient();

    // Log in to the game server.
    pCredentials = nex_demo_util::GameServerLogin(pNgsFacade, NEXDEMO_GAME_SERVER_ID, NEXDEMO_ACCESS_KEY, 60000);

    // Bind service(s)
    if (pCredentials)
    {
        if (!pMatchMakingClient->Bind(pCredentials))
        {
            NEX_DEMO_LOG("Cannot bind to Matchmaking Service\n");
            nex_demo_util::GameServerLogout(pNgsFacade);
            goto END;
        }
        else
        {
            NEX_DEMO_LOG("Bind to Matchmaking service\n");
        }
    }
    else
    {
        goto END;
    }

    // Register an event handler to process notifications from the NGS.
    pNgsFacade->RegisterNotificationEventHandler(&s_MyNotificationEventHandler);

    // Main loop.
    RunMatchmakingLoop(pNgsFacade, pMatchMakingClient);

    // Delete the event handler.
    pNgsFacade->UnregisterNotificationEventHandler(&s_MyNotificationEventHandler);

END:
    // Delete each object
    pMatchMakingClient->Unbind();
    delete pMatchMakingClient;

    // Log out
    nex_demo_util::GameServerLogout(pNgsFacade);

    // Delete the login client
    delete pNgsFacade;

    // Finalize.
    nn::nex::GlobalVariables::ReleaseInstance();

    // Shut down utilities
    nex_demo_util::Finalize();

    // Disconnect from the Internet.
    nex_demo_util::NetTerm();

    return 0;
}

//-----------------------------------------------------------------------------
// Main Function
//-----------------------------------------------------------------------------
qInt GenericMain(qInt argc, qChar** argv)
{
    NEX_DEMO_LOG("*******************************\n");
    NEX_DEMO_LOG("       CreateAndBrowse sample\n");
    NEX_DEMO_LOG("*******************************\n\n");

#if defined(CAFE_TARGET)

#else
    NEXSampleMain(argc, argv);
#endif

    NEX_DEMO_LOG("Leaving application...\n");

    return 0;
}

NEX_MAIN_WRAPPER(GenericMain)

//-----------------------------------------------------------------------------
// Matchmaking
//-----------------------------------------------------------------------------
void RunGameMainLoop()
{
    NEX_DEMO_LOG("start session\n");
    while(true)
    {
        nex_demo_util::PadStatus oPadStatus;
        nex_demo_util::DemoPad::Update(&oPadStatus);
        if(oPadStatus.trigger == DEMO_PAD_BUTTON_B)
        {
            break;
        }
        nn::nex::Scheduler::GetInstance()->Dispatch(DISPATCH_TIMEOUT);
        nn::nex::Platform::Sleep(DISPATCH_PERIOD);
    }
}



//-----------------------------------------------------------------------------
// Display the list of sessions
//-----------------------------------------------------------------------------
void RenderSessions(nn::nex::qList<nn::nex::GatheringHolder> *pLstGatheringHolder, int index)
{
    (void)index;
    nn::nex::qList<nn::nex::GatheringHolder>::iterator it(pLstGatheringHolder->begin());
    while(it != pLstGatheringHolder->end())
    {
#ifndef NN_BUILD_RELEASE
        nn::nex::MatchmakeSession *session = static_cast<nn::nex::MatchmakeSession*>(it->Get());
        NEX_DEMO_LOG("%ls gid:%d, desc:%ls owner:%d host:%d %d/%d\n",
                     // The currently selected session
                     std::distance(pLstGatheringHolder->begin(), it) == index ? _T("*") : _T(" "),
                     // GatheringID
                     session->GetID(),
                     // Description
                     session->GetDescription().CStr(),
                     // The session owner's PID
                     session->GetOwnerPID(),
                     // The session host's PID
                     session->GetHostPID(),
                     // The current number of participants
                     session->GetParticipantCount(),
                     // The maximum number of participants.
                     session->GetMaxParticipants()
                    );
#endif
        ++it;
    }
}

//-----------------------------------------------------------------------------
// Search for a session
//-----------------------------------------------------------------------------
bool BrowseSession(nn::nex::MatchmakeExtensionClient *pMatchMakingClient,
                   nn::nex::qList<nn::nex::GatheringHolder> *pLstGatheringHolder)
{
    NEX_DEMO_LOG("Browse Session\n");

    nn::nex::ProtocolCallContext     oContext;

    // Search condition settings
    nn::nex::MatchmakeSessionSearchCriteria criteria;

    // Application-specific game mode.
    criteria.SetGameMode(MATCHMAKE_GAME_MODE);

    // Search attributes that the application can freely define.
    criteria.SetAttribute(0, 3333);

    // Matchmaking system type.
    criteria.SetMatchmakeSystemType(nn::nex::MATCHMAKE_SYSTEM_TYPE_ANYBODY);

    // Search only for sessions that can be joined.
    criteria.SetExcludeLocked(true);

    // Exclude full sessions.
    criteria.SetVacantOnly(true);

    // Search range, by default the offset is 0 and the maximum number of entries, 20
    nn::nex::ResultRange range;

    // Search for sessions
    if (!pMatchMakingClient->BrowseMatchmakeSession(&oContext, criteria, range, pLstGatheringHolder))
    {
        NEX_DEMO_LOG("Failed to browse matchmake session\n");
        return false;
    }
    oContext.Wait();
    if(!oContext.GetOutcome())
    {
        nex_demo_util::ReportPCallError(oContext, "Failed to browse matchmake session\n");
        return false;
    }

    return true;
}

//-----------------------------------------------------------------------------
// Join a session
//-----------------------------------------------------------------------------
bool JoinSession(nn::nex::MatchmakeExtensionClient *pMatchMakingClient,
                 nn::nex::GatheringID &gid)
{
    nn::nex::ProtocolCallContext oContext;
    nn::nex::qVector<nn::nex::qByte> sessionKey;
    // The local host won't join when users registered to the blacklist are participating (by default)
    if(!pMatchMakingClient->JoinMatchmakeSession(&oContext, gid, &sessionKey, _T("")))
    {
        NEX_DEMO_LOG("Failed to join matchmake session\n");
        return false;
    }
    oContext.Wait();
    if(!oContext.GetOutcome())
    {
        if(oContext.GetOutcome() == QERROR(RendezVous, SessionFull))
        {
            NEX_DEMO_LOG("Session is already full.\n");
        }
        if(oContext.GetOutcome() == QERROR(RendezVous, SessionVoid))
        {
            NEX_DEMO_LOG("Session is already unregistered.\n");
        }
        if(oContext.GetOutcome() == QERROR(RendezVous, DeniedByParticipants))
        {
            NEX_DEMO_LOG("Denid by Participants. You are in participant's blacklist.\n");
        }
        if(oContext.GetOutcome() == QERROR(RendezVous, ParticipantInBlackList))
        {
            NEX_DEMO_LOG("Participant is in your blacklist.\n");
        }
        nex_demo_util::ReportPCallError(oContext, "Failed to join matchmake session\n");
        return false;
    }
    s_JoinedGatheringID = gid;
    return true;
}

//-----------------------------------------------------------------------------
// Select a session.
//-----------------------------------------------------------------------------
bool SelectSession(nn::nex::qList<nn::nex::GatheringHolder> *pLstGatheringHolder,
                   nn::nex::GatheringID &gid)
{
    int index = 0;
    int gatheringsSize = pLstGatheringHolder->size();

    enum {None = 0, Redraw = 1, Quit = 2} state = None;

    while(state != Quit)
    {
        RenderSessions(pLstGatheringHolder, index);
        NEX_DEMO_LOG("[UP/DOWN]:select session [A]:join session [B]:cancel\n");

        //Waiting for input
        do
        {
            state = None;
            nex_demo_util::PadStatus oPadStatus;
            nex_demo_util::DemoPad::Update(&oPadStatus);
            switch(oPadStatus.trigger)
            {
            case DEMO_PAD_BUTTON_DOWN:
                index = (index + 1) % gatheringsSize;
                state = Redraw;
                break;
            case DEMO_PAD_BUTTON_UP:
                index = (index + gatheringsSize - 1) % gatheringsSize;
                state = Redraw;
                break;
            case DEMO_PAD_BUTTON_A:
            {
                nn::nex::qList<nn::nex::GatheringHolder>::iterator it(pLstGatheringHolder->begin());
                std::advance(it, index);
                gid = (*it)->GetID();
                return true;
            }
            case DEMO_PAD_BUTTON_B:
                state = Quit;
                break;
            }
            nn::nex::Scheduler::GetInstance()->Dispatch(DISPATCH_TIMEOUT);
            nn::nex::Platform::Sleep(DISPATCH_PERIOD);
        }
        while(state == None);
    }
    return false;
}

//-----------------------------------------------------------------------------
// Search for and join a session
//-----------------------------------------------------------------------------
bool BrowseAndJoinSession(nn::nex::MatchmakeExtensionClient *pMatchMakingClient)
{
    while(true)
    {
        nn::nex::qList<nn::nex::GatheringHolder> lstGatheringHolder;
        nn::nex::GatheringID gid(INVALID_GATHERINGID);
        if(!BrowseSession(pMatchMakingClient, &lstGatheringHolder))
        {
            return false;
        }


        NEX_DEMO_LOG("%d sessions found\n", lstGatheringHolder.size());

        if(lstGatheringHolder.size() == 0)
        {
            break;
        }
        if(!SelectSession(&lstGatheringHolder, gid))
        {
            break;
        }
        if(gid != INVALID_GATHERINGID)
        {
            return JoinSession(pMatchMakingClient, gid);
        }
    }
    return false;
}

//-----------------------------------------------------------------------------
// Creates the session.
//-----------------------------------------------------------------------------
bool CreateSession(nn::nex::MatchmakeExtensionClient *pMatchMakingClient)
{
    NEX_DEMO_LOG("Create Session\n");
    nn::nex::MatchmakeSession *matchmakeSession = new nn::nex::MatchmakeSession();

    // Description string.
    matchmakeSession->SetDescription(_T("description"));

    // Minimum number of participants
    matchmakeSession->SetMinParticipants(1);

    // The maximum number of participants.
    matchmakeSession->SetMaxParticipants(MATCHMAKE_MAX_PARTICIPANTS);

    // Application-specific game mode.
    matchmakeSession->SetGameMode(MATCHMAKE_GAME_MODE);

    // Search attributes that the application can freely define.
    matchmakeSession->SetAttribute(0, 3333);
    matchmakeSession->SetAttribute(1, 4444);

    // Matchmaking system type.
    matchmakeSession->SetMatchmakeSystemType(nn::nex::MATCHMAKE_SYSTEM_TYPE_ANYBODY);

    // Application-defined data
    nn::nex::qVector<nn::nex::qUnsignedInt8> applicationBuffer;
    applicationBuffer.push_back(1);
    applicationBuffer.push_back(2);
    applicationBuffer.push_back(3);
    matchmakeSession->SetApplicationBuffer(applicationBuffer);

    // Set to joinable from the moment the session was created
    // Set this value to False when using in combination with P2P communications,
    // and call OpenParticipation when P2P communication preparations are complete.
    matchmakeSession->SetOpenParticipation(true);

    // Matchmaking system type.
    matchmakeSession->SetMatchmakeSystemType(nn::nex::MATCHMAKE_SYSTEM_TYPE_ANYBODY);

    // Set this flag if you want notifications such as participation notifications to be sent to all participants in a matchmaking session.
    // Use nn::nex::GatheringFlags::NotifyParticipationEventsToAllParticipantsReproducibly if you only want users that are already participating to receive the nn::nex::ParticipationEvents::Participate notification.
    // 
    //matchmakeSession->SetFlag(nn::nex::GatheringFlags::NotifyParticipationEventsToAllParticipants);

    nn::nex::GatheringHolder oGatheringHolder(matchmakeSession);
    nn::nex::GatheringID gid(INVALID_GATHERINGID);
    nn::nex::ProtocolCallContext oContext;
    nn::nex::qVector<nn::nex::qByte> sessionKey;
    if(!pMatchMakingClient->CreateMatchmakeSession(&oContext, oGatheringHolder, &gid, &sessionKey))
    {
        NEX_DEMO_LOG("Failed to create matchmake session\n");
        return false;
    }
    oContext.Wait();
    if(!oContext.GetOutcome())
    {
        nex_demo_util::ReportPCallError(oContext, "Failed to create matchmake session\n");
        return false;
    }

    s_JoinedGatheringID = gid;

    return true;
}
//-----------------------------------------------------------------------------
// Mode selection
//-----------------------------------------------------------------------------
bool SelectMatchmakeMode(qBool *bCreateSessionRequired)
{
    // Select the mode
    NEX_DEMO_LOG("[A]:Create session [X]:Browse session [B]:QUIT\n");
    while(true)
    {
        nex_demo_util::PadStatus oPadStatus;
        nex_demo_util::DemoPad::Update(&oPadStatus);
        switch(oPadStatus.trigger)
        {
        case DEMO_PAD_BUTTON_A:
            *bCreateSessionRequired = true;
            return true;
        case DEMO_PAD_BUTTON_X:
            *bCreateSessionRequired = false;
            return true;
        case DEMO_PAD_BUTTON_B:
            return false;
        }
        nn::nex::Scheduler::GetInstance()->Dispatch(DISPATCH_TIMEOUT);
        nn::nex::Platform::Sleep(DISPATCH_PERIOD);
    }
}

//-----------------------------------------------------------------------------
// Invoke matchmaking and the main loop
//-----------------------------------------------------------------------------
void RunMatchmakingLoop(nn::nex::NgsFacade *pNgsFacade,
                        nn::nex::MatchmakeExtensionClient *pMatchMakingClient)
{
    // HID initialization
    nex_demo_util::DemoPad::Initialize();

    while(true)
    {
        //Exit if you are disconnected from the game  server
        if(!pNgsFacade->IsConnected())
        {
            NEX_DEMO_LOG("disconnected\n");
            break;
        }

        // Mode selection
        qBool bCreateSessionRequired;
        if(!SelectMatchmakeMode(&bCreateSessionRequired))
        {
            break;
        }

        // Create/search for the session in conjunction with the mode
        if(bCreateSessionRequired)
        {
            // Create a session
            if(!CreateSession(pMatchMakingClient))
            {
                continue;
            }
        }
        else
        {
            // Search for and join a session
            if(!BrowseAndJoinSession(pMatchMakingClient))
            {
                continue;
            }
        }

        // Joining a session
        RunGameMainLoop();

        // Leave a session
        // - When the number of participants is 0 as a result of having called the EndParticipation function, that session is automatically deleted
        // - The session owner can call UnregisterGathering instead of EndParticipation.
        //   In this case, the session will be deleted immediately, but the remaining participants will be notified with `NotificationEvents::GatheringUnregistered`.
        //   
        //   Matchmaking can be ended by handling this event, depending on the type of game.
        // - When used in combination with P2P communication, host synchronization may be needed when the session host leaves.
        //   For more information, see the reference for the HostMigrationExtension class.
        NEX_DEMO_LOG("leave from: %d\n", s_JoinedGatheringID);
        nn::nex::ProtocolCallContext oContext;
        if(!pMatchMakingClient->EndParticipation(&oContext, s_JoinedGatheringID))
        {
            NEX_DEMO_LOG("Failed to end participation\n");
            continue;
        }
        oContext.Wait();
        if(!oContext.GetOutcome())
        {
            nex_demo_util::ReportPCallError(oContext, "Failed to end participation\n");
        }
    }

    // Finalize HID
    nex_demo_util::DemoPad::Finalize();
}
