/*---------------------------------------------------------------------------*
  (C) Copyright 1998-2014, Technologies Quazal Inc. All rights reserved.
  Quazal Net-Z, Quazal Rendez-Vous and Quazal are trademarks of Quazal.

  Copyright (C)2011-2014 Nintendo  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo and/or its licensors and are
  protected by national and international copyright laws.
  They may not be disclosed to third parties or copied or duplicated
  in any form, in whole or in part, without the prior written consent
  of Nintendo.
  The content herein is highly confidential and should be handled
  accordingly.
 *---------------------------------------------------------------------------*/
#include "Avatar.h"
#include "World.h"
#include "../common/NexUtil.h"
//-----------------------------------------------------------------------------
// Constructor
//-----------------------------------------------------------------------------
Avatar::Avatar()
{
    SetPosition(0.0f, 0.0f);
    UpdateMoveDir();
}
Avatar::Avatar(qReal x, qReal y)
{
    SetPosition(x, y);
    UpdateMoveDir();
}

//-----------------------------------------------------------------------------
// Destructor
//-----------------------------------------------------------------------------
Avatar::~Avatar()
{
    NEX_DEMO_LOG("Avatar Destruct!!\n");
}

//-----------------------------------------------------------------------------
// Initialization
//-----------------------------------------------------------------------------
void Avatar::InitDO()
{
    if (!IsADuplicationMaster())
    {
        return;
    }

    // Gets own station ID
    m_dsInfo.stationID = Station::GetLocalInstance()->GetStationID();

    // Set extrapolation delay time
    m_dsExtrapolatedLoopbackPosition.GetPHBDRParameters()->SetExtrapolationDelay(Avatar::EXTRAPOLATION_DELAY_TIME);
    m_dsExtrapolatedLoopbackPosition.SetLoopbackDelay(Avatar::LOOPBACK_DELAY_TIME);

    //Sends `m_dsPosition` via `SUBSTREAM_ID`.
    m_dsPosition.SetSubStreamID(SUBSTREAM_ID);

    NEX_DEMO_LOG("*** Avatar::InitDO!!\n");
}

//-----------------------------------------------------------------------------
// Set the position
//-----------------------------------------------------------------------------
void Avatar::SetPosition(qReal x, qReal y)
{
    m_dsPosition.Set(x, y);
    m_dsExtrapolatedPosition.Set(m_dsPosition.x, m_dsPosition.y);
    m_dsExtrapolatedLoopbackPosition.Set(m_dsPosition.x, m_dsPosition.y);
}

//-----------------------------------------------------------------------------
// Updates the position
//-----------------------------------------------------------------------------
void Avatar::MovePosition(qReal x, qReal y)
{
    m_dsPosition.Move(x, y);
    m_dsExtrapolatedPosition.Set(m_dsPosition.x, m_dsPosition.y);
    m_dsExtrapolatedLoopbackPosition.Set(m_dsPosition.x, m_dsPosition.y);
}

//-----------------------------------------------------------------------------
// Update member and duplicated object
//-----------------------------------------------------------------------------
void Avatar::DoPhysic()
{
    // Update only when oneself is the duplication master
    if (IsADuplicationMaster())
    {
        // Movement update
        (void)MovePosition(mMove[0], mMove[1]);

        // Checks time and updates movement parameter if time has passed
        if (nn::nex::Time::GetTime() - mTime > Avatar::MOVE_TIME)
        {
            UpdateMoveDir();
        }

        // Collision detection with World
        UpdateCollisionWithWorld();

        // Update and transmit own information to duplica
        Update();

        // Loopback renewal
        m_dsExtrapolatedLoopbackPosition.UpdateLoopback();
        m_dsExtrapolatedLoopbackPosition.RefreshLoopback();
    }
    else
    {
        // Update interpolation for extrapolation filter
        Refresh();
    }
}

//-----------------------------------------------------------------------------
// Rendering
//-----------------------------------------------------------------------------
void Avatar::Render()
{
    // Outputs position data
    NEX_DEMO_LOG("Avatar station ID: %d\n", m_dsInfo.stationID);
    NEX_DEMO_LOG("Avatar Info (                     Pos) >> x:%f y:%f\n", m_dsPosition.x, m_dsPosition.y);
    NEX_DEMO_LOG("Avatar Info (Extrapolated         Pos) >> x:%f y:%f\n", m_dsExtrapolatedPosition.x,
                 m_dsExtrapolatedPosition.y);
    NEX_DEMO_LOG("Avatar Info (ExtrapolatedLoopback Pos) >> x:%f y:%f\n", m_dsExtrapolatedLoopbackPosition.x,
                 m_dsExtrapolatedLoopbackPosition.y);
}

//-----------------------------------------------------------------------------
// Updates movement amount
//-----------------------------------------------------------------------------
void Avatar::UpdateMoveDir()
{
    for (int i=0; i<2; ++i)
    {
        int dir = nn::nex::Platform::GetRandomNumber(2);
        mMove[i] = dir == 0 ? 1.0f : -1.0f * MOVE_POWER;
    }

    // Resets time passed
    mTime = nn::nex::Time::GetTime();
}

//-----------------------------------------------------------------------------
// Updates collisions with world
//-----------------------------------------------------------------------------
void Avatar::UpdateCollisionWithWorld()
{
    UpdateCollisionWithWorldSub(m_dsPosition.x, m_dsPosition.y);
    UpdateCollisionWithWorldSub(m_dsExtrapolatedPosition.x, m_dsExtrapolatedPosition.y);
    UpdateCollisionWithWorldSub(m_dsExtrapolatedLoopbackPosition.x, m_dsExtrapolatedLoopbackPosition.y);
}

//-----------------------------------------------------------------------------
// Updates collisions with world (Sub function)
//-----------------------------------------------------------------------------
void Avatar::UpdateCollisionWithWorldSub(qReal& x, qReal& y)
{
    qReal minx, miny, maxx, maxy;

    World::Ref refWorld(nn::nex::g_hWorld);

    // Gets boundaries
    refWorld->GetBoundary(minx, miny, maxx, maxy);

    if (x < minx || x > maxx)
    {
        if (x < minx)
        {
            x = minx;
        }
        else
        {
            x = maxx;
        }
        mMove[0] *= -1.0f;
        mTime = nn::nex::Time::GetTime();
        NEX_DEMO_LOG("Collision X!!\n");
    }
    if (y < miny || y > maxy)
    {
        if (y < miny)
        {
            y = miny;
        }
        else
        {
            y = maxy;
        }
        mMove[1] *= -1.0f;
        mTime = nn::nex::Time::GetTime();
        NEX_DEMO_LOG("Collision Y!!\n");
    }
}


