﻿/*---------------------------------------------------------------------------*
  (C) Copyright 1998-2014, Technologies Quazal Inc. All rights reserved.
  Quazal Net-Z, Quazal Rendez-Vous and Quazal are trademarks of Quazal.

  Copyright (C) 2011-2014 Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo and/or its licensors and are
  protected by national and international copyright laws.
  They may not be disclosed to third parties or copied or duplicated
  in any form, in whole or in part, without the prior written consent
  of Nintendo.
  The content herein is highly confidential and should be handled
  accordingly.
 *---------------------------------------------------------------------------*/
/*
 * NEX Sample
 *
 * DDLBasic
 *
 * Description:
 *
 *  This sample performs transmissions of DuplicatedObjects (hereafter, DOs).
 *  Processes other than transmissions are the same as in the AutoMatch sample. Please refer to it as well.
 *
 * In the detailed definition of a DO in DDLBasic.ddl, primarily the Avatar.cpp and World.cpp subclasses are listed for the classes that use DOs.
 *  Please refer to those files as well.
 *
 * This sample defines an Avatar as a DO created for each station and a World as the only well-known object that exists for all stations.
 *  We can conceive of a world as comprising of two dimensions and of avatars moving randomly within the world.
 *
 *  Stations that create sessions create a World (a well-known object), and publish it. 
 *  As a result of this, a well-known object is definitely recognized by each station.
 * Each station (including the session master) creates an avatar object after joining the session, and, in the same way, publishes it.
 *   As a result, duplicates (duplicas) of the avatar created by the local host
 are created for each other station.
 *
 *  When communicating changes in the master object to duplicas, call the Update function.
 *  This function is called within the Avatar::DoPhysic function.
 *
 However, when performing type definitions based on the updates by the estimation filter in the DDL (extrapolation_filter), you need to call the Refresh function to reflect the changes.
 *  This also occurs within the Avatar::DoPhysic function in this sample.
 *  There are other updates performed from combinations of loopbacks and the estimation filter.
 *
 *  The update method can vary depending on the type definition. For more information, see the documentation.
 *
 * This sample runs in thread unsafe transport buffer thread mode.
 *
 




*/

#include <nex.h>
#include "../common/NexUtil.h"
#include "../common/NexUtilNetZ.h"

// DDL-related
#include "Avatar.h"
#include "World.h"

//-----------------------------------------------------------------------------
// Constant Definitions
//-----------------------------------------------------------------------------
// Matchmaking mode
#define MATCHMAKE_GAME_MODE 1002

// Number of connected persons
#define MATCHMAKE_MAX_PARTICIPANTS 2

// The number of retries (matchmaking is attempted this number of times when it fails)
#define MATCHMAKE_TRY_NUM 10

// Run matchmaking between friends?
//#define MATCHMAKE_FRIEND_MODE

// Attribute for number of participants connected
#define MATCHMAKE_ATTR_NUM_PARTICIPANTS 0

// The time to wait alone after creating a session
#define ORPHAN_WAIT_TIME 60000

// Amount of time to sleep
#define DISPATCH_PERIOD 100

// Time to wait for a dispatch.
#define DISPATCH_TIMEOUT 3

// Time to send and receive data after matchmaking
#define MATCHMAKING_TIME 60000

// Maximum substream ID.
#define MAX_SUBSTREAM_ID 1
//-----------------------------------------------------------------------------
// Function definitions
//-----------------------------------------------------------------------------
void SetNetZSetting();
qBool CreateNetZGame();
qBool TryCreateNetZSessionAndOpenMatchmakeSession(nn::nex::MatchmakeExtensionClient *pMatchMakingClient,
        nn::nex::GatheringID gatheringID);
qBool EndParticipation(nn::nex::MatchmakeExtensionClient *pMatchMakingClient, nn::nex::GatheringID gatheringID);
qBool JoinNetZGame(nn::nex::qList<nn::nex::StationURL>& lstURL);
qBool JoinMatchmakeSession(nn::nex::MatchmakeExtensionClient *pMatchMakingClient);
qBool AutoMatchmake(nn::nex::MatchmakeExtensionClient *pMatchMakingClient, nn::nex::PrincipalID* pOwnerPID);
qBool WaitToMainLoop(nn::nex::RendezVous *pRV, const nn::nex::NetZ* pNetZ, nn::nex::HostMigrationCallback* pHostMigrationCallback);
void RunGameMainLoop(nn::nex::RendezVous *pRV, const nn::nex::NetZ* pNetZ, nn::nex::HostMigrationCallback* pHostMigrationCallback);
void RunMatchmakingLoop(nn::nex::RendezVous *pRV,
                        nn::nex::Credentials *pCredentials,
                        nn::nex::MatchmakeExtensionClient *pMatchMakingClient);
void OutputConnectionInfo();


//-----------------------------------------------------------------------------
// Event handler to process notifications from the NGS
//-----------------------------------------------------------------------------
// Specify the principal ID of the host.
static nn::nex::PrincipalID s_OwnerPID;

class MyNotificationEventHandler : public nn::nex::NotificationEventHandler
{
public:
    void ProcessNotificationEvent(const nn::nex::NotificationEvent &oEvent)
    {
        nn::nex::StringStream ss;
        ss << oEvent;
        NEX_DEMO_LOG("Notification %ls\n\n",ss.CStr());

        // Per-message processing
        switch (oEvent.GetType())
        {
        case nn::nex::NotificationEvents::ParticipationEvent:
            switch (oEvent.GetSubType())
            {
            case nn::nex::ParticipationEvents::Participate:
                NEX_DEMO_LOG("PARTICIPATION_EVENT PARTICIPATE: [PID %u] join [session %u]\n", oEvent.GetParam2(), oEvent.GetParam1());
                break;

            case nn::nex::ParticipationEvents::EndParticipation:
                NEX_DEMO_LOG("PARTICIPATION_EVENT END: [PID %u] end [session %u]\n", oEvent.GetParam2(), oEvent.GetParam1());
                break;

            case nn::nex::ParticipationEvents::Disconnect:
                NEX_DEMO_LOG("PARTICIPATION_EVENT DISCONNECT: [PID %u] disconnect [session %u]\n", oEvent.GetParam2(),
                             oEvent.GetParam1());
                break;

            default:
                NEX_DEMO_LOG("PARTICIPATION_EVENT: unsupported participation event\n");
                break;
            }
            break;

        case nn::nex::NotificationEvents::OwnershipChangeEvent:
            NEX_DEMO_LOG("NOTIFICATION_EVENT OWNERSHIPCHANGE: [session %u] changed owner to [PID %u]\n",
                         oEvent.GetParam1(), oEvent.GetParam2());

            // Record the new host's PID
            s_OwnerPID = oEvent.GetParam2();
            break;

        case nn::nex::NotificationEvents::GatheringUnregistered:
            NEX_DEMO_LOG("NOTIFICATION_EVENT GATHERINGUNREGISTERED: [session %u] unregistered\n",
                         oEvent.GetParam1());
            break;

        default:
            NEX_DEMO_LOG("NOTIFICATION_EVENT: unsupported notification event\n");
            break;
        }
    }
};


//-----------------------------------------------------------------------------
// Variable Definitions
//-----------------------------------------------------------------------------
// Gathering ID to use for matchmaking
static nn::nex::GatheringID s_JoinedGatheringID = INVALID_GATHERINGID;

// Event handler to process notifications
static MyNotificationEventHandler s_MyNotificationEventHandler;


namespace nn
{
namespace nex
{
extern void InitDOClasses();
}
}


qInt NEXSampleMain(qInt intArg, void* ptrArg)
{
    (void)intArg;
    (void)ptrArg;

    // Connect to the Internet
    if (!nex_demo_util::NetInit())
    {
        NEX_DEMO_LOG("net init failed...\n");
        NEX_DEMO_LOG("Leaving application...\n");

        return 0;
    }

    // Prepare to use tools
    nex_demo_util::Initialize();

    // Initialization
    nn::nex::GlobalVariables::AcquireInstance();
    nn::nex::NetZ::SetDOClassInitializationCallback(&nn::nex::InitDOClasses);

    // Set to thread unsafe transport buffer thread mode
    nn::nex::Core::SetThreadMode(nn::nex::Core::ThreadModeUnsafeTransportBuffer);

    // Timeout settings for NAT traversal.
    nn::nex::RootTransport::SetNATTraversalTimeout(MATCHMAKE_MAX_PARTICIPANTS * 5000);

    // The station mismatch detection count checks at 5-second intervals by default, so set this to be 10 seconds longer than the NAT traversal timeout.
    // 
    nn::nex::Session::SetDetectStationInconsitencyThresholdCount(MATCHMAKE_MAX_PARTICIPANTS + 2);

    // Set the network topology.
    nn::nex::NetZ::SetNetworkTopology(nn::nex::NetworkTopology::Peer2Peer);

    nn::nex::EventLog::GetInstance()->SetLogLevel(nn::nex::EventLog::Debug);
    nn::nex::TraceLog::GetInstance()->SetFlag(TRACE_ALWAYS);

    // Display the mode
    NEX_DEMO_LOG("\n");
#ifdef MATCHMAKE_FRIEND_MODE
    NEX_DEMO_LOG("***** Friend match ***** \n");
#else
    NEX_DEMO_LOG("***** Anybody match ***** \n");
#endif
    NEX_DEMO_LOG("***** Max participants: %d ***** \n", MATCHMAKE_MAX_PARTICIPANTS);

    NEX_DEMO_LOG("\n");

    nn::nex::LANSessionDiscovery::Disable();

    nn::nex::Credentials              *pCredentials        = NULL;
    nn::nex::NgsFacade                *pNgsFacade          = new nn::nex::NgsFacade();
    nn::nex::NATTraversalClient       *pNATTraversalClient = new nn::nex::NATTraversalClient();
    nn::nex::MatchmakeExtensionClient *pMatchMakingClient  = new nn::nex::MatchmakeExtensionClient();

    // Log in to the game server.
    pCredentials = nex_demo_util::GameServerLogin(pNgsFacade, NEXDEMO_GAME_SERVER_ID, NEXDEMO_ACCESS_KEY, 60000);

    // Bind service(s)
    if (pCredentials)
    {
        if (!pMatchMakingClient->Bind(pCredentials))
        {
            NEX_DEMO_LOG("Cannot bind to Matchmaking Service\n");
            nex_demo_util::GameServerLogout(pNgsFacade);
            goto END;
        }
        else
        {
            NEX_DEMO_LOG("Bind to Matchmaking service\n");
        }

        if (!pNATTraversalClient->Bind(pCredentials))
        {
            NEX_DEMO_LOG("Cannot bind to NATTraversal Service\n");
        }
    }
    else
    {
        goto END;
    }

    // Register an event handler to process notifications from the NGS.
    pNgsFacade->RegisterNotificationEventHandler(&s_MyNotificationEventHandler);

    // Main loop.
    RunMatchmakingLoop(pNgsFacade, pCredentials, pMatchMakingClient);

    // Delete the event handler.
    pNgsFacade->UnregisterNotificationEventHandler(&s_MyNotificationEventHandler);

    // Unbind service(s)
    //
    if (!pMatchMakingClient->Unbind())
    {
        NEX_DEMO_LOG("Cannot bind to MatchMaking Service\n");
    }

    if (!pNATTraversalClient->Unbind())
    {
        NEX_DEMO_LOG("Cannot bind to NATTraversal Service\n");
    }

END:
    // Delete each object
    delete pMatchMakingClient;
    delete pNATTraversalClient;

    // Log out
    nex_demo_util::GameServerLogout(pNgsFacade);

    // Delete the login client
    delete pNgsFacade;

    nn::nex::GlobalVariables::ReleaseInstance();

    // Shut down utilities
    nex_demo_util::Finalize();

    // Disconnect from the Internet.
    nex_demo_util::NetTerm();

    return 0;
}

//-----------------------------------------------------------------------------
// Main Function
//-----------------------------------------------------------------------------
qInt GenericMain(qInt argc, qChar** argv)
{
    NEX_DEMO_LOG("*******************************\n");
    NEX_DEMO_LOG("          DDL Trans\n");
    NEX_DEMO_LOG("*******************************\n\n");

#if defined(CAFE_TARGET)

#else
    NEXSampleMain(argc, argv);
#endif

    NEX_DEMO_LOG("Leaving application...\n");

    return 0;
}

NEX_MAIN_WRAPPER(GenericMain)

//-----------------------------------------------------------------------------
// Invoke matchmaking and the main loop
//-----------------------------------------------------------------------------
void RunMatchmakingLoop(nn::nex::RendezVous *pRV,
                        nn::nex::Credentials *pCredentials,
                        nn::nex::MatchmakeExtensionClient *pMatchMakingClient)
{
    for (int i=0; i<MATCHMAKE_TRY_NUM; ++i)
    {
        // Enables NAT traversal.
        nn::nex::ConnectivityManager::GetInstance()->StartNATSession();

        // Create a NetZ object
        nn::nex::NetZ *pNetZ = new nn::nex::NetZ;

        SYSTEMCHECK(pNetZ != NULL);
        if (!pNetZ->IsValid())
        {
            NEX_DEMO_LOG("NetZ is invalid!!\n");
            delete pNetZ;
            return;
        }

        // Setup for communication
        SetNetZSetting();

        // HostMigration is necessary to continue processing when a station leaves a session.
        // A NetZ object is automatically unregistered when it is deleted.
        nn::nex::HostMigrationExtension::Register();
        nn::nex::HostMigrationCallback *pHostMigrationCallback = new nn::nex::HostMigrationCallback(pMatchMakingClient,
                nn::nex::MatchMakingClient::StaticUpdateSessionHost);

        // Start automatic matchmaking.
        qBool bReadyToRun = false;
        s_OwnerPID = INVALID_PRINCIPALID;
        nn::nex::PrincipalID OwnerPrincipalID = INVALID_PRINCIPALID;

        // Try to create a session or join another open session.
        bReadyToRun = AutoMatchmake(pMatchMakingClient, &OwnerPrincipalID);

        // Wait until enough people have gathered.
        if (bReadyToRun)
        {
            s_OwnerPID = OwnerPrincipalID;

            // Enables host migration.
            pHostMigrationCallback->Register(s_JoinedGatheringID);

            // Wait in a loop
            bReadyToRun = WaitToMainLoop(pRV, pNetZ, pHostMigrationCallback);

            // Start the main send/receive processing.
            if (bReadyToRun)
            {
                // Create a DO and publish it
                // With this, duplicas are created in other stations
                // The initial position is random
                qReal x = static_cast<qReal>(nn::nex::Platform::GetRandomNumber(100));
                qReal y = static_cast<qReal>(nn::nex::Platform::GetRandomNumber(100));

                Avatar* pAvatar = Avatar::Create();
                if (pAvatar)
                {
                    // By setting the data before publishing it, this data set is maintained for when the duplicated object is published to other stations.
                    pAvatar->SetPosition(x, y);
                    if (pAvatar->Publish())
                    {
                        // Main loop.
                        RunGameMainLoop(pRV, pNetZ, pHostMigrationCallback);
                    }
                    else
                    {
                        NEX_DEMO_LOG("Avatar publish is faild.\n");
                        /*
                           When publishing fails, it is necessary to delete the main duplicated object reference, as follows.
                           In this case, it is also necessary to destroy the NetZ instance and exit from the session.
                         */
                        pAvatar->DeleteMainRef();
                    }
                }
                else
                {
                    /*
                       In the rare case that a child station could not communicate with the parent station prior to when the duplicated object initilization process completes after joining a session, the `DuplicatedObject::Create` function will return `NULL`.
                       
                       In this case, it is also necessary to destroy the NetZ instance and exit from the session.
                    */
                    NEX_DEMO_LOG("Avatar create is faild.\n");
                }
            }


            // Disables host migration.
            pHostMigrationCallback->Unregister(s_JoinedGatheringID);
        }
        else
        {
            NEX_DEMO_LOG("fail to automatch\n");
        }

        // Delete the host migration callback
        delete pHostMigrationCallback;

        // Delete the NetZ object
        {
            /*
               If the NetZ object is suddenly deleted, a block occurs at deletion that continues until termination, so terminate explicitly.
            

*/
            nn::nex::CallContext oContext;
            pNetZ->Terminate(&oContext);
            oContext.Wait();
            if ( oContext.GetOutcome() == false )
            {
                // It is possible that termination failed and memory leak might have occurred.
                NEX_DEMO_LOG("NetZ::Terminate is failed. Memory leak might be happend!\n");
            }
            delete pNetZ;
        }

        // Ends NAT traversal
        nn::nex::ConnectivityManager::GetInstance()->StopNATSession();

        // Leave the session.
        EndParticipation(pMatchMakingClient, s_JoinedGatheringID);

        // Delete the matchmaking session if we are the owner
        if (s_OwnerPID == pCredentials->GetPrincipalID())
        {
            NEX_DEMO_LOG("Unregister gathering ... \n");
            nn::nex::ProtocolCallContext oContext;
            if (pMatchMakingClient->UnregisterGathering(&oContext, s_JoinedGatheringID))
            {
                oContext.Wait();
                if (oContext.GetState() != nn::nex::CallContext::CallSuccess)
                {
                    nex_demo_util::ReportPCallError(oContext, "failed to Unregister gathering!!\n");
                }
                else
                {
                    NEX_DEMO_LOG("Unregister gathering successfly.\n");
                }
            }
            else
            {
                NEX_DEMO_LOG("UnregisterGathering call failed ... \n");
            }
        }

        // Exit if successful
        if(bReadyToRun)
        {
            break;
        }

        //Exit if you are disconnected from the game  server
        if(!pRV->IsConnected())
        {
            NEX_DEMO_LOG("disconnected\n");
            break;
        }
    }
}

//-----------------------------------------------------------------------------
// Wait for enough players to join.
//-----------------------------------------------------------------------------
qBool WaitToMainLoop(nn::nex::RendezVous *pRV, const nn::nex::NetZ* pNetZ, nn::nex::HostMigrationCallback* pHostMigrationCallback)
{
    qBool bReadyToRun = false;

    NEX_DEMO_LOG("Starting NetZ session...\n");

    // Wait until the prescribed number of people have joined.
    nn::nex::Time tiReadyStart = nn::nex::Time::GetTime();
    nn::nex::Time tiWaitStart = tiReadyStart;
    int oldNumStations = 0;
    while (1)
    {
        nn::nex::Time nowTime = nn::nex::Time::GetTime();

        int numStations = 0;
        nn::nex::Station::SelectionIterator itStations;
        numStations = itStations.Count();

        if (oldNumStations != numStations)
        {
            // Reset the wait time when there has been a change in the number of people connected
            tiWaitStart = nowTime;
        }

        // Remember how many people there were before
        oldNumStations = numStations;

        if (numStations == 1 && nowTime - tiWaitStart >= ORPHAN_WAIT_TIME)
        {
            // Retry matchmaking if no one joins the session for a fixed amount of time.
            NEX_DEMO_LOG("*** time passed as one. match again.\n");
            break;
        }

        if (!pRV->IsConnected())
        {
            //Quit when connection to game server is cut off.
            NEX_DEMO_LOG("*** RV is disconnected.\n");
            break;
        }

        if (nowTime - tiReadyStart >= 3000)
        {
            // Display the number of people who are currently connected.
            NEX_DEMO_LOG("waiting ... NumStations: %d/%d\n", numStations, MATCHMAKE_MAX_PARTICIPANTS);

            // Display connection information.
            OutputConnectionInfo();

            tiReadyStart = nowTime;
        }

        /*
           If the session owner disconnects and only one station remains, host migration will not take place and the session will become ownerless, preventing anyone from connecting. 
           To prevent this, the following function is used to determine whether all other stations have left a session.
           If we detect that there are no other stations in the session, we exit the loop here and retry matchmaking.
        
*/
        if (pHostMigrationCallback->SessionHostIsOrphan())
        {
            NEX_DEMO_LOG("*** host is orphan... match again.\n");
            break;
        }

        // Dispatch messages
        nn::nex::Scheduler::GetInstance()->Dispatch(DISPATCH_TIMEOUT);

        // NetZ fatal error check.
        nn::nex::qResult netZError = pNetZ->GetFatalError();
        if ( QFAILED(netZError.GetReturnCode()) )
        {
            // An error from which restoration is not possible occurs
            nex_demo_util::ReportQResultError(netZError, "NetZ fatal error happend!\n");
            break;
        }

        // Check how many people have connected.
        if (numStations == MATCHMAKE_MAX_PARTICIPANTS)
        {
            // The prescribed number of people have connected and preparations for entering the main loop are complete.
            bReadyToRun = true;
            break;
        }

        nn::nex::Platform::Sleep(DISPATCH_PERIOD);
    }

    return bReadyToRun;
}

//-----------------------------------------------------------------------------
// Main loop after matchmaking
//-----------------------------------------------------------------------------
void RunGameMainLoop(nn::nex::RendezVous *pRV, const nn::nex::NetZ* pNetZ, nn::nex::HostMigrationCallback* pHostMigrationCallback)
{
    NEX_DEMO_LOG("Start NetZ session.\n");
    nn::nex::Time tiSessionStart = nn::nex::Time::GetTime();
    nn::nex::Time tiLogStart = nn::nex::Time::GetTime();
    nn::nex::Time tiSendStart = nn::nex::Time::GetTime();

    // Send and receive data
    while (1)
    {
        nn::nex::Time nowTime = nn::nex::Time::GetTime();

        if (nowTime - tiSessionStart >= MATCHMAKING_TIME)
        {
            // Exit because we have sent and received data for the specified amount of time
            break;
        }

        if (pHostMigrationCallback->SessionHostIsOrphan())
        {
            // Also exit when there is no one else in the session
            NEX_DEMO_LOG("*** host is orphan... abort.\n");
            break;
        }

        if (nowTime - tiSendStart >= 1 * 1000)
        {
            // Update
            tiSendStart = nn::nex::Time::GetTime();

            // Update all avatars
            Avatar::SelectionIterator itAvatar;
            itAvatar.GotoStart();
            while (!itAvatar.EndReached())
            {
                itAvatar->DoPhysic();
                ++itAvatar;
            }
        }

        if (!pRV->IsConnected())
        {
            //Quit when connection to game server is cut off.
            NEX_DEMO_LOG("*** RV is disconnected.\n");
            break;
        }

        // Dispatch messages
        nn::nex::Scheduler::GetInstance()->Dispatch(DISPATCH_TIMEOUT);

        // NetZ fatal error check.
        nn::nex::qResult netZError = pNetZ->GetFatalError();
        if ( QFAILED(netZError.GetReturnCode()) )
        {
            // An error from which restoration is not possible occurs
            nex_demo_util::ReportQResultError(netZError, "NetZ fatal error happend!\n");
            break;
        }

        if (nex_demo_util::IsLowMemory())
        {
            NEX_DEMO_LOG("*** LowMemory! abort.\n");
            return;
        }

        if (nowTime - tiLogStart >= 2 * 1000)
        {
            // Display connected stations.
            tiLogStart = nn::nex::Time::GetTime();

            NEX_DEMO_LOG("\n\nStations currently in Session\n");
            NEX_DEMO_LOG("-----------------------------\n");
            OutputConnectionInfo();

            // Display the state of the DOs
            {
                Avatar::SelectionIterator itAvatar;
                itAvatar.GotoStart();
                while (!itAvatar.EndReached())
                {
                    itAvatar->Render();
                    ++itAvatar;
                }

                // Render the world
                World::Ref refWorld(nn::nex::g_hWorld);
                refWorld->Render();
            }
            NEX_DEMO_LOG("-----------------------------\n");
        }

        nn::nex::Platform::Sleep(DISPATCH_PERIOD);
    }

    NEX_DEMO_LOG("Stop NetZ session.\n");
}

//-----------------------------------------------------------------------------
// Run automatic matchmaking.
// Get the necessary gathering and then join or create a session.
//-----------------------------------------------------------------------------
qBool AutoMatchmake(nn::nex::MatchmakeExtensionClient *pMatchMakingClient, nn::nex::PrincipalID* pOwnerPID)
{
    *pOwnerPID = INVALID_PRINCIPALID;

    // Create search criteria.
    nn::nex::MatchmakeSession* matchmakeSession = new nn::nex::MatchmakeSession();
    matchmakeSession->SetOpenParticipation(false);

    // The description string setting of MatchmakeSession is optional.
    matchmakeSession->SetDescription(_T("DDL Trans"));

    // The minimum and maximum number of participants.
    matchmakeSession->SetMinParticipants(1);
    matchmakeSession->SetMaxParticipants(MATCHMAKE_MAX_PARTICIPANTS);

    // Game mode.
    matchmakeSession->SetGameMode(MATCHMAKE_GAME_MODE);

    // Search attributes that the application can freely define.
    // When the attribute is not set, the type condition will match any value.
    // Here the maximum number of matchmaking participants is added to the search conditions.
    // At present, the different number of people connected will be used for matchmaking and that is why.
    // It is possible that this specification will be revised in the future
    matchmakeSession->SetAttribute(MATCHMAKE_ATTR_NUM_PARTICIPANTS, MATCHMAKE_MAX_PARTICIPANTS);

    // Matchmaking system type.
#ifdef MATCHMAKE_FRIEND_MODE
    matchmakeSession->SetMatchmakeSystemType(nn::nex::MATCHMAKE_SYSTEM_TYPE_FRIEND);
#else
    matchmakeSession->SetMatchmakeSystemType(nn::nex::MATCHMAKE_SYSTEM_TYPE_ANYBODY);
#endif

    // Binary data that the application can freely set.
    nn::nex::qVector<nn::nex::qUnsignedInt8> applicationBuffer;
    applicationBuffer.push_back(1);
    applicationBuffer.push_back(2);
    applicationBuffer.push_back(3);
    matchmakeSession->SetApplicationBuffer(applicationBuffer);

    // Transfer ownership when the owner logs out?
#ifndef MATCHMAKE_FRIEND_MODE
    // Owner migration does not occur during friend matchmaking
    matchmakeSession->SetFlag(nn::nex::GatheringFlags::MigrateOwner);
#endif

    nn::nex::GatheringHolder oInGatheringHolder = matchmakeSession;
    nn::nex::GatheringHolder oOutGatheringHolder;   // The gathering that was joined.

    // Search for or create a matchmaking session.
    nn::nex::ProtocolCallContext oContext;
    if (pMatchMakingClient->AutoMatchmake(&oContext, oInGatheringHolder, &oOutGatheringHolder, _T("ddl trans session")))
    {
        oContext.Wait();
        if (oContext.GetState() != nn::nex::CallContext::CallSuccess)
        {
            nex_demo_util::ReportPCallError(oContext, "Failed to AutoMatchmake\n");
            return false;
        }

        s_JoinedGatheringID = oOutGatheringHolder->GetID();

        //Generates a different signature key for each P2P session by setting the key notified by the game server as P2P session signature key.
        nn::nex::Network::GetInstance()->SetP2PDataPacketSessionSignatureKey(
            reinterpret_cast<nn::nex::MatchmakeSession*>(oOutGatheringHolder.Get())->GetSessionKey());

        nn::nex::PrincipalID OwnerPID = oOutGatheringHolder->GetOwnerPID();
        *pOwnerPID = OwnerPID;
        NEX_DEMO_LOG("[PID Info >> JoinedGID:%d OwnerPID:%d HostPID:%d MyPID:%d]\n", s_JoinedGatheringID, OwnerPID, oOutGatheringHolder->GetOwnerPID(), pMatchMakingClient->GetCredentials()->GetPrincipalID());

        if (*pOwnerPID == pMatchMakingClient->GetCredentials()->GetPrincipalID())
        {
            // Local host creates a session
            return TryCreateNetZSessionAndOpenMatchmakeSession(pMatchMakingClient, oOutGatheringHolder->GetID());
        }
        else
        {
            // Join another session
            return JoinMatchmakeSession(pMatchMakingClient);
        }
    }
    else
    {
        NEX_DEMO_LOG("Failed to AutoMatchmake\n");
        return false;
    }
}

//-----------------------------------------------------------------------------
// Set up communication features
//-----------------------------------------------------------------------------
void SetNetZSetting()
{
    // Set up name displayed in station.
    nex_demo_util::SetupLocalPlayerName();

    //Specifies the maximum SubStreamID for reliable communication.
    nn::nex::Stream::GetSettings().SetMaxUserReliableSubStreams(MAX_SUBSTREAM_ID);
}

//-----------------------------------------------------------------------------
// Create a session
//-----------------------------------------------------------------------------
qBool CreateNetZGame()
{
    // You have to create a well-known object and publish it before creating a session.
    // As a result, when other stations join the session, it is guaranteed that a copy of this object will be created.
    // 
    World* pWorld = World::CreateWellKnown(nn::nex::g_hWorld);
    pWorld->SetBoundary(0, 0, 100, 100);
    pWorld->Publish();

    // Simple NetZ session creation.
    if (nn::nex::Session::CreateSession(NULL))
    {
        NEX_DEMO_LOG("Net-Z session created successfly\n");
        return true;
    }
    else
    {
        NEX_DEMO_LOG("Failed to create a Net-Z session\n");
        return false;
    }
}

//-----------------------------------------------------------------------------
// Create a NetZ session and open MatchmakeSession
//-----------------------------------------------------------------------------
qBool TryCreateNetZSessionAndOpenMatchmakeSession(nn::nex::MatchmakeExtensionClient *pMatchMakingClient,
        nn::nex::GatheringID gatheringID)
{
    if (!CreateNetZGame())
    {
        return false;
    }

    // Open a session
    nn::nex::ProtocolCallContext oContext;
    if (pMatchMakingClient->OpenParticipation(&oContext, gatheringID))
    {
        oContext.Wait();
        if (oContext.GetState() != nn::nex::CallContext::CallSuccess)
        {
            // If it has already been deleted, treat this as a normal exit.
            if(oContext.GetOutcome() == QERROR(RendezVous, SessionVoid))
            {
                NEX_DEMO_LOG("matchmake session is already unregistered. don't care\n");
            }
            else
            {
                nex_demo_util::ReportPCallError(oContext, "Failed to end participation\n");
                return false;
            }
        }
    }
    else
    {
        NEX_DEMO_LOG("Failed to end participation\n");
        return false;
    }
    return true;
}

//-----------------------------------------------------------------------------
// Leave MatchmakeSession
//-----------------------------------------------------------------------------
qBool EndParticipation(nn::nex::MatchmakeExtensionClient *pMatchMakingClient, nn::nex::GatheringID gatheringID)
{
    if (pMatchMakingClient->GetCredentials()->GetPrincipalID() == s_OwnerPID)
    {
        // If you are the owner, you cannot leave the session here if doing so would reduce the number of participants to zero because that would cause `UnregisterGathering` to fail.
        // 
        NEX_DEMO_LOG("I am owner ... skip EndParticipation.\n");
        return true;
    }

    NEX_DEMO_LOG("EndParticipation\n");

    nn::nex::ProtocolCallContext     oContext;

    if (pMatchMakingClient->EndParticipation(&oContext, gatheringID))
    {
        oContext.Wait();
        if (oContext.GetState() != nn::nex::CallContext::CallSuccess)
        {
            nex_demo_util::ReportPCallError(oContext, "Failed to end participation\n");
            return false;
        }
    }
    else
    {
        NEX_DEMO_LOG("Failed to end participation\n");
        return false;
    }
    return true;
}

//-----------------------------------------------------------------------------
// Join a NetZ session
//-----------------------------------------------------------------------------
qBool JoinNetZGame(nn::nex::qList<nn::nex::StationURL>& lstURL)
{
    nn::nex::CallContext oContext;
    if (nn::nex::Session::JoinSession(&oContext, lstURL))
    {
        oContext.Wait();
        if(oContext.GetState() == nn::nex::CallContext::CallSuccess)
        {
#if defined(UNICODE_VERSION)
            NEX_DEMO_LOG("Net-Z session on %ls successfly joined\n", lstURL.front().CStr());
#else
            NEX_DEMO_LOG("Net-Z session on %s successfly joined\n", lstURL.front().CStr());
#endif

            return true;
        }
        else
        {
            NEX_DEMO_LOG("Failed to join the Net-Z session %x\n", oContext.GetOutcome().GetReturnCode());
        }
    }

    NEX_DEMO_LOG("Failed to join the Net-Z session\n");
    return false;
}

//-----------------------------------------------------------------------------
// Join MatchmakeSession
//-----------------------------------------------------------------------------
qBool JoinMatchmakeSession(nn::nex::MatchmakeExtensionClient *pMatchMakingClient)
{
    // List the session URLs and try to join them.
    nn::nex::qList<nn::nex::StationURL> lstURLs;
    nn::nex::ProtocolCallContext oContext;
    if (pMatchMakingClient->GetSessionURLs(&oContext, s_JoinedGatheringID, &lstURLs))
    {
        oContext.Wait();
        if (oContext.GetState() != nn::nex::CallContext::CallSuccess)
        {
            // We may have joined a matchmaking session that has been deleted.
            if(oContext.GetOutcome() == QERROR(RendezVous, SessionVoid))
            {
                NEX_DEMO_LOG("matchmake session is already unregistered.\n");
            }
            nex_demo_util::ReportPCallError(oContext, "Failed obtained a list of URLs to connect to the game host\n");
            return false;
        }

        NEX_DEMO_LOG("successfly obtained a list of URLs to connect to the game host\n");
        nn::nex::qList<nn::nex::StationURL>::iterator it = lstURLs.begin();
        while (it!=lstURLs.end())
        {
            NEX_DEMO_LOG("URL = %ls\n", (*it).CStr());
            ++it;
        }
        // Join with this session list
        return JoinNetZGame(lstURLs);
    }

    NEX_DEMO_LOG("Failed Request URLs...\n");

    return false;
}

//-----------------------------------------------------------------------------
// Display connection information
//-----------------------------------------------------------------------------
void OutputConnectionInfo()
{
    nn::nex::Station::SelectionIterator itStations;
    itStations.GotoStart();
    while (!itStations.EndReached())
    {
        NEX_DEMO_LOG("Station %8x: URL = %ls\n",
                     itStations->GetHandle().GetValue(),
                     itStations->GetStationURL(0));

        ++itStations;
    }
    NEX_DEMO_LOG("\n");
}



