/*---------------------------------------------------------------------------*
  Project:  NEX
  File:     DataStoreSample.cpp

  Copyright (C)2011-2014 Nintendo Co., Ltd.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.
 *---------------------------------------------------------------------------*/

#include <nex.h>

#if defined(CTR_TARGET)
#include <nn/http_Initialize.h>
#elif defined(CAFE_TARGET)
#include <sdk_ver.h>
#if CAFE_OS_SDK_VERSION <= 20908
#include <cafe/nssl/nsslclient.h>
#include <cafe/curl/curl.h>
#else
#include <cafe/nssl/nssl_init.h>
#include <cafe/curl/curl_init.h>
#endif //end of CAFE_OS_SDK_VERSION
#endif //end of defined(CAFE_TARGET)

#include "../common/NexUtil.h"


// Display qResult content in the debugging log.
static void PrintResult(const nn::nex::qResult& result)
{
    result.Trace(TRACE_ALL);
}

// Display multiple DataStoreMetaInfo content in the debugging log.
static void PrintGetMetaResult(const nn::nex::qVector<nn::nex::DataStoreMetaInfo>& metaInfos)
{
    NEX_DEMO_LOG("+---------+----------+------------------------+----------+-----------+\n");
    NEX_DEMO_LOG("| data_id | owner_id |         name           |   size   | data_type |\n");
    NEX_DEMO_LOG("+---------+----------+------------------------+----------+-----------+\n");

    nn::nex::qVector<nn::nex::DataStoreMetaInfo>::const_iterator it;
    for (it = metaInfos.begin(); it != metaInfos.end(); ++it)
    {
        if(!it->IsValid())
        {
            // Do not display invalid data.
            continue;
        }
        NEX_DEMO_LOG("|%9llu|%10u|%-24ls|%10u|%11u|\n",
                     it->GetDataId(),
                     it->GetOwnerId(),
                     it->GetName().GetBufferPtr(),
                     it->GetSize(),
                     it->GetDataType()
                    );
    }
    NEX_DEMO_LOG("+---------+----------+------------------------+----------+-----------+\n");
}

// Display DataStoreSearchResult content in the debugging log.
static void PrintSearchResult(const nn::nex::DataStoreSearchResult& searchResult)
{
    NEX_DEMO_LOG("SearchResult::totalCount = %u\n", searchResult.totalCount);
    PrintGetMetaResult(searchResult.GetResult());
}

class DataStoreSample
{
public:
    DataStoreSample(nn::nex::DataStoreClient* pDataStoreClient)
        :m_DataStoreClient(pDataStoreClient)
    {
    }
    virtual ~DataStoreSample() {}
    void CallVariousApi();

private:
    qBool PostObjectBuffered(qUnsignedInt64 *pDataId);
    qBool GetObjectBuffered(qUnsignedInt64 dataId);
    qBool SearchObject();

private:
    nn::nex::DataStoreClient *m_DataStoreClient;
    static const qUnsignedInt32 POST_DATA_SIZE = 200 * 1024;
    static const qUnsignedInt32 METABINARY_SIZE = 1024;
};

qBool DataStoreSample::PostObjectBuffered(qUnsignedInt64 *pDataId)
{
    nn::nex::ProtocolCallContext oContext;
    nn::nex::DataStorePostParam postParam;
    void* data = std::malloc(POST_DATA_SIZE);
    nn::nex::qResult result;

    // If you set this to false, the buffer must be maintained by the application until all asynchronous processes are complete because the data will not be copied to the internal NEX buffer.
    nn::nex::qBool useInternalBuffer = false;

    postParam.SetSize(POST_DATA_SIZE);
    memset(data, 0xaa, POST_DATA_SIZE);

    m_DataStoreClient->PostObject(&oContext, postParam, pDataId, data, useInternalBuffer);
    oContext.Wait();
    std::free(data);
    result = oContext.GetOutcome();
    if (!result)
    {
        NEX_DEMO_LOG("PostObject failed.\n");
        PrintResult(result);
        return false;
    }
    NEX_DEMO_LOG("PostObjectBuffered succeeded. DataID: %llu\n", *pDataId);

    return true;
}

qBool DataStoreSample::GetObjectBuffered(qUnsignedInt64 dataId)
{
    nn::nex::ProtocolCallContext oContext;
    nn::nex::DataStoreGetParam getParam;
    void* buf = std::malloc(POST_DATA_SIZE);
    nn::nex::qResult result;
    qUnsignedInt32 writtenSize = 0;

    getParam.SetDataId(dataId);
    m_DataStoreClient->GetObject(&oContext, getParam, buf, POST_DATA_SIZE, &writtenSize);
    oContext.Wait();
    std::free(buf);
    result = oContext.GetOutcome();
    if (!result)
    {
        NEX_DEMO_LOG("GetObject failed.\n");
        PrintResult(result);
        return false;
    }
    NEX_DEMO_LOG("GetObjectBuffered succeeded.\n");

    return true;
}

qBool DataStoreSample::SearchObject()
{
    nn::nex::ProtocolCallContext oContext;
    nn::nex::DataStoreSearchParam searchParam;
    nn::nex::DataStoreSearchResult searchResult;
    nn::nex::qResult result;

    searchParam.SetSearchType(nn::nex::DataStoreConstants::SEARCH_TYPE_OWN);

    m_DataStoreClient->SearchObject(&oContext, searchParam, &searchResult);
    oContext.Wait();
    result = oContext.GetOutcome();
    if (!result)
    {
        NEX_DEMO_LOG("SearchObject failed.\n");
        PrintResult(result);
        return false;
    }
    NEX_DEMO_LOG("SearchObject succeeded.\n");

    PrintSearchResult(searchResult);

    return true;
}

void DataStoreSample::CallVariousApi()
{
    qUnsignedInt64 dataId = nn::nex::DataStoreConstants::INVALID_DATAID;

    // Upload data. (use buffer)
    if (!PostObjectBuffered(&dataId))
    {
        return;
    }

    // Get the data. (use buffer)
    if (!GetObjectBuffered(dataId))
    {
        return;
    }

    // Search data.
    if (!SearchObject())
    {
        return;
    }

    NEX_DEMO_LOG("CallVariousApi succeeded.\n");
}


qInt NEXSampleMain(qInt intArg, void* ptrArg)
{
    (void)intArg;
    (void)ptrArg;

    // Connects to the Internet.
    if (!nex_demo_util::NetInit())
    {
        NEX_DEMO_LOG("nex_demo_util::NetInit() failed.\n");
        NEX_DEMO_LOG("Leaving application...\n");
        return 1;
    }

    // Prepare to use utilities
    nex_demo_util::Initialize();

    // Initialize the HTTP library.
#if defined(CTR_TARGET)
    if(nn::http::Initialize().IsFailure())
    {
        NEX_DEMO_LOG("nn::http::initialize() failed.\n");
        NEX_DEMO_LOG("Leaving application...\n");
        return 1;
    }
#elif defined(CAFE_TARGET)
    // libCurl resolves addresses, so set an allocator for that purpose if necessary.
    SOResolverRegisterAllocator(nex_demo_util::AllocFunction, nex_demo_util::FreeFunction);

    if(NSSLInit() < 0)
    {
        NEX_DEMO_LOG("NSSLInit() failed.\n");
        NEX_DEMO_LOG("Leaving application...\n");
        return 1;
    }

    if(curl_global_init(CURL_GLOBAL_ALL) != 0)
    {
        NEX_DEMO_LOG("curl_global_init() failed.\n");
        NEX_DEMO_LOG("Leaving application...\n");
        return 1;
    }
#endif

    // NEX initialization
    nn::nex::GlobalVariables::AcquireInstance();

    // Set to thread unsafe transport buffer thread mode
    nn::nex::Core::SetThreadMode(nn::nex::Core::ThreadModeUnsafeTransportBuffer);

    // Log level settings
    nn::nex::EventLog::GetInstance()->SetLogLevel(nn::nex::EventLog::Debug);
    nn::nex::TraceLog::GetInstance()->SetFlag(TRACE_ALWAYS);

    // You can configure the log format as follows if necessary.
    // For details, see the Reference Manual.
    nn::nex::EventLog::GetInstance()->GetOutputFormat()->ShowDateTime(true); // Add date to EventLog
    nn::nex::TraceLog::GetInstance()->GetOutputFormat()->ShowDateTime(true); // Add date to TraceLog

    nn::nex::NgsFacade  *pNgsFacade  = new nn::nex::NgsFacade();
    nn::nex::Credentials  *pCredentials  = NULL;
    nn::nex::DataStoreClient *pDataStoreClient = new nn::nex::DataStoreClient();

    // Logs into the game server.
    pCredentials = nex_demo_util::GameServerLogin(pNgsFacade, NEXDEMO_GAME_SERVER_ID, NEXDEMO_ACCESS_KEY, 60000);
    if (!pCredentials)
    {
        NEX_DEMO_LOG("Login failed.\n");
        goto END;
    }

    // Bind credentials to service. This sample only binds credentials to DataStore.
    if (!pDataStoreClient->Bind(pCredentials))
    {
        NEX_DEMO_LOG("Cannot bind to DataStore Service\n");
        goto ERROR_BIND;
    }

    // Increase priority of HTTP communication thread. We recommend setting the communication thread's priority higher than the main thread's priority.
    pDataStoreClient->SetHttpThreadPriority(15);

    // Calls various DataStore APIs.
    {
        DataStoreSample sample(pDataStoreClient);
        sample.CallVariousApi();
    }

    // Unbind before logging out from game server.
    pDataStoreClient->Unbind();

ERROR_BIND:
    // Logging out from the game server
    nex_demo_util::GameServerLogout(pNgsFacade);

END:
    delete pDataStoreClient;
    delete pNgsFacade;

    // Finalize the HTTP library.
#if defined(CTR_TARGET)
    nn::http::Finalize();
#elif defined(CAFE_TARGET)
    curl_global_cleanup();
    NSSLFinish();
#endif

    // NEX finalization
    nn::nex::GlobalVariables::ReleaseInstance();

    // Shut down utilities
    nex_demo_util::Finalize();

    // Disconnects from the Internet.
    nex_demo_util::NetTerm();

    return 0;
}

//-----------------------------------------------------------------------------
// Main Function
//-----------------------------------------------------------------------------
qInt GenericMain(qInt argc, qChar** argv)
{
    NEX_DEMO_LOG("*******************************\n");
    NEX_DEMO_LOG("     DataStoreSimple sample\n");
    NEX_DEMO_LOG("*******************************\n\n");

#if defined(CAFE_TARGET)

#else
    NEXSampleMain(argc, argv);
#endif

    NEX_DEMO_LOG("Leaving application...\n");

    return 0;
}

NEX_MAIN_WRAPPER(GenericMain)
