﻿/*---------------------------------------------------------------------------*
  (C) Copyright 1998-2014, Technologies Quazal Inc. All rights reserved.
  Quazal Net-Z, Quazal Rendez-Vous and Quazal are trademarks of Quazal.

  Copyright (C)2011-2014 Nintendo  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo and/or its licensors and are
  protected by national and international copyright laws.
  They may not be disclosed to third parties or copied or duplicated
  in any form, in whole or in part, without the prior written consent
  of Nintendo.
  The content herein is highly confidential and should be handled
  accordingly.
 *---------------------------------------------------------------------------*/

/*
 * NEX Sample
 *
 * Debug
 *
 * Description :
 *
 * This is a sample of the debugging helper features.
 *
 * It uses the DebugClient object to call the game server's API in order to enable the activity log, and then periodically get and output the log.
 * This sample ensures that the same log entry is not retrieved more than once. It specifies the scope of the log to retrieve to the DebugClient::GetAPICalls function as a time.
 *
 * Calls to the API to enable or disable the activity log set the enabled status for the entire game server.
 * Take note of this when calling the DebugClient::EnableApiRecorder or DebugClient::DisableApiRecorder function.
 * In particular, do not call the DebugClient::DisableApiRecorder function if you are using the sample demo.
 *
 */

#include <nex.h>
#include <RendezVous/Services/Jugem/Debug/src/Client/DebugClient.h> // Not included from "nex.h"
#include "../common/NexUtil.h"

// Time to continue displaying the log.
#define API_LOG_TIME 10 * 60 * 1000

// Interval at which to retrieve the log.
#define API_LOG_INTERVAL 30 * 1000

// Main processing
void DebugMain(nn::nex::NgsFacade *pNgsFacade, nn::nex::DebugClient* pClient);

// Call the API to get and display the log for duration API_LOG_TIME, at interval API_LOG_INTERVAL.
void PrintApiRecordLoop(nn::nex::NgsFacade *pNgsFacade, nn::nex::DebugClient* pClient);


qInt NEXSampleMain(qInt intArg, void* ptrArg)
{
    (void)intArg;
    (void)ptrArg;

    // Connects to the Internet.
    if (!nex_demo_util::NetInit())
    {
        NEX_DEMO_LOG("net init failed...\n");
        NEX_DEMO_LOG("Leaving application...\n");

        return 0;
    }

    // Prepare to use demo utilities
    nex_demo_util::Initialize();

    // Initialization
    nn::nex::GlobalVariables::AcquireInstance();

    // Set to thread unsafe transport buffer thread mode
    nn::nex::Core::SetThreadMode(nn::nex::Core::ThreadModeUnsafeTransportBuffer);

    nn::nex::EventLog::GetInstance()->SetLogLevel(nn::nex::EventLog::Info);
    nn::nex::TraceLog::GetInstance()->SetFlag(TRACE_ALWAYS);

    nn::nex::NgsFacade                *pNgsFacade          = new nn::nex::NgsFacade();
    nn::nex::DebugClient              *pDebugClient        = new nn::nex::DebugClient();
    nn::nex::Credentials              *pCredentials        = NULL;

    // Logs into the game server.
    pCredentials = nex_demo_util::GameServerLogin(pNgsFacade, NEXDEMO_GAME_SERVER_ID, NEXDEMO_ACCESS_KEY, 60000);

    if(pCredentials != NULL)
    {
        pDebugClient->Bind(pCredentials);

        DebugMain(pNgsFacade, pDebugClient);

        pDebugClient->Unbind();
    }

    nex_demo_util::GameServerLogout(pNgsFacade);

    // Delete DebugClient object
    delete pDebugClient;

    // Delete the login client
    delete pNgsFacade;

    // End processing
    nn::nex::GlobalVariables::ReleaseInstance();

    // Shut down utilities
    nex_demo_util::Finalize();

    // Disconnects from the Internet.
    nex_demo_util::NetTerm();

    return 0;
}

//-----------------------------------------------------------------------------
// Main Function
//-----------------------------------------------------------------------------
qInt GenericMain(qInt argc, qChar** argv)
{
    NEX_DEMO_LOG("*******************************\n");
    NEX_DEMO_LOG("     Debug sample\n");
    NEX_DEMO_LOG("*******************************\n\n");

#if defined(CAFE_TARGET)

#else
    NEXSampleMain(argc, argv);
#endif

    NEX_DEMO_LOG("Leaving application...\n");

    return 0;
}

// Main processing
void DebugMain(nn::nex::NgsFacade *pNgsFacade, nn::nex::DebugClient* pClient)
{
    nn::nex::ProtocolCallContext oContext;

    // Enable API log
    if(pClient->EnableApiRecorder(&oContext))
    {
        oContext.Wait();
        if (oContext.GetState() != nn::nex::CallContext::CallSuccess)
        {
            nex_demo_util::ReportPCallError(oContext, "Failed to EnableApiRecorder.\n");
            return;
        }
    }
    else
    {
        NEX_DEMO_LOG("cannot call DebugClient::EnableApiRecorder()\n");
        return;
    }

    // Check the status, because other clients may have disabled it.
    nn::nex::qBool isApiRecorderEnabled;
    if(pClient->IsApiRecorderEnabled(&oContext, &isApiRecorderEnabled))
    {
        oContext.Wait();
        if (oContext.GetState() != nn::nex::CallContext::CallSuccess)
        {
            nex_demo_util::ReportPCallError(oContext, "Failed to IsApiRecorderEnabled.\n");
            return;
        }
    }
    else
    {
        NEX_DEMO_LOG("cannot call DebugClient::IsApiRecorderEnabled()\n");
        return;
    }

    NEX_DEMO_LOG("isApiRecorderEnabled:%d\n", isApiRecorderEnabled);
    if(!isApiRecorderEnabled)
    {
        return;
    }

    // Get and display
    PrintApiRecordLoop(pNgsFacade, pClient);

    // Commented out, because if we are using the sample demo server, other developers may be using it.
    /*
    if(pClient->DisableApiRecorder(&oContext))
    {
        oContext.Wait();
        if (oContext.GetState() != nn::nex::CallContext::CallSuccess)
        {
            nex_demo_util::ReportPCallError(oContext, "Failed to DisableApiRecorder.\n");
            return ;
        }
    }
    else
    {
        NEX_DEMO_LOG("cannot call DebugClient::DisableApiRecorder()\n");
        return ;
    }
    */
}

// Call the API to get and display the log for duration API_LOG_TIME, at interval API_LOG_INTERVAL.
void PrintApiRecordLoop(nn::nex::NgsFacade *pNgsFacade, nn::nex::DebugClient* pClient)
{
    // Suppress warning log so that calling the CallContext::Wait function does not output a warning.
    nn::nex::EventLog::GetInstance()->SetLogLevel(nn::nex::EventLog::Error);

    nn::nex::ProtocolCallContext oContext;

    NEX_DEMO_LOG("--------API CALL LOG START----------\n");

    // Get from current time. Specify DateTime::Never if you want to get all log entries.
    nn::nex::DateTime since;
    pNgsFacade->GetServerTime(&since);

    nn::nex::Time startTime = nn::nex::Time::GetTime();
    nn::nex::Time prevCallTime = nn::nex::Time::GetTime();
    nn::nex::qSet<nn::nex::PrincipalID> callerSet; // Set of clients that called an API at least once

    while (true)
    {
        nn::nex::Time nowTime = nn::nex::Time::GetTime();

        if (nowTime - startTime >= API_LOG_TIME)
        {
            break;
        }

        if (nowTime - prevCallTime >= API_LOG_INTERVAL)
        {
            nn::nex::qList<nn::nex::ApiCallData> apiCallDataList;
            nn::nex::qList<nn::nex::PrincipalID> principalIdList; // When not specified, gets call history for everyone.

            // Get.
            if(pClient->GetApiCalls(&oContext, &apiCallDataList, principalIdList, since))
            {
                oContext.Wait();
                if (oContext.GetState() != nn::nex::CallContext::CallSuccess)
                {
                    nex_demo_util::ReportPCallError(oContext, "Failed to GetApiCalls.\n");
                    return;
                }
            }
            else
            {
                nex_demo_util::ReportPCallError(oContext, "cannot call GetApiCalls.\n");
                return;
            }

            // List is guaranteed to be sorted in calling order.
            for(nn::nex::qList<nn::nex::ApiCallData>::iterator it = apiCallDataList.begin(); it != apiCallDataList.end(); ++it)
            {
                NEX_DEMO_LOG("%02d:%02d:%02d %u %ls\n", it->GetCalledTime().GetHour(),
                             it->GetCalledTime().GetMinute(),
                             it->GetCalledTime().GetSecond(),
                             it->GetPrincipalId(), it->GetMethodName().CStr());

                // Remember the PIDs of the clients that called APIs
                callerSet.insert(it->GetPrincipalId());
            }

            // We can ensure that the entries we get from the log do not overlap by specifying the last time in the entries we retrieved as the `since` parameter.
            if(apiCallDataList.size() > 0)
            {
                since = apiCallDataList.back().GetCalledTime();
            }

            prevCallTime = nowTime;
        }

        nn::nex::Scheduler::GetInstance()->DispatchAll();
        nn::nex::Platform::Sleep(30);
    }

    NEX_DEMO_LOG("--------API CALL LOG END----------\n");

    NEX_DEMO_LOG("--------PRINT API CALL SUMMARY START----------\n");

    // Total up the call history of clients remaining in the history that called an API at least once and display them.
    for(nn::nex::qSet<nn::nex::PrincipalID>::iterator itPid = callerSet.begin(); itPid != callerSet.end(); ++itPid)
    {
        nn::nex::qList<nn::nex::ApiCallSummary> summary;

        // Only get clients that exceeded the guideline call frequency (over the whole period)
        if(pClient->GetApiCallSummary(&oContext, &summary, *itPid, nn::nex::DateTime::Never, nn::nex::DateTime::Never, true))
        {
            oContext.Wait();
            if (oContext.GetState() != nn::nex::CallContext::CallSuccess)
            {
                nex_demo_util::ReportPCallError(oContext, "Failed to GetApiCalls.\n");
                return;
            }
        }
        else
        {
            nex_demo_util::ReportPCallError(oContext, "cannot call GetApiCalls.\n");
            return;
        }

        NEX_DEMO_LOG("-- PrincipalID %d Summary -- \n", *itPid);
        for(nn::nex::qList<nn::nex::ApiCallSummary>::iterator it = summary.begin(); it != summary.end(); ++it)
        {
            NEX_DEMO_LOG("%ls, %u, %u, %u, %u, %u\n", it->GetMethodName().CStr(),
                         it->GetPeriod(),
                         it->GetLimit(),
                         it->GetResult(),
                         it->GetCount(),
                         it->GetTotalCount());
        }
        NEX_DEMO_LOG("\n");
    }

    NEX_DEMO_LOG("--------PRINT API CALL SUMMARY END----------\n");
}

NEX_MAIN_WRAPPER(GenericMain)
