﻿/*---------------------------------------------------------------------------*
  (C) Copyright 1998-2014, Technologies Quazal Inc.  All rights reserved.
  Quazal Net-Z, Quazal Rendez-Vous and Quazal are trademarks of Quazal.

  Copyright (C)2011-2014 Nintendo  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo and/or its licensors and are
  protected by national and international copyright laws.
  They may not be disclosed to third parties or copied or duplicated
  in any form, in whole or in part, without the prior written consent
  of Nintendo.
  The content herein is highly confidential and should be handled
  accordingly.
 *---------------------------------------------------------------------------*/
/*
 * NEX Sample
 *
 * IndependentServer
 *
 * Description:
 *
 *  This sample gets an independent server's authentication token.
 *
 * This sample runs in thread unsafe transport buffer thread mode.
 *
 *
 * Note:
 *
 *  When obtaining an authentication token and logging in to the game server, you can take care of both processes at once by using NgsFacade::LoginAndRequestAuthenticationToken.
 *  
 *  
 *
 *  Getting authentication tokens puts a load on the server, so avoid operations such as re-obtaining authentication tokens while they are still valid.
 * 
 *
 */

#include <nex.h>
#include <nn/friends.h>
#include "../../common/NexUtil.h"

// When also performing a login, enable USE_NGS_FACADE.
// #define USE_NGS_FACADE

//-----------------------------------------------------------------------------
// Function definitions
//-----------------------------------------------------------------------------
void DemoInitialize();
void DemoFinalize();

//-----------------------------------------------------------------------------
// Main Function
//-----------------------------------------------------------------------------
void nnMain()
{
    NEX_DEMO_LOG("*******************************\n");
    NEX_DEMO_LOG("     IndependentServer sample\n");
    NEX_DEMO_LOG("*******************************\n\n");

    DemoInitialize();

    // Initialize the friend library
    // Even if you will not need to log in to the friend server, the friend library must always be initialized because nn::friends functions are called in this sample.
    // 
    nn::friends::Initialize();

    NEX_DEMO_LOG("My PrincipalId %d\n", nn::friends::GetMyPrincipalId());

    // Authentication token
    qChar8 authToken[nn::nex::IndependentServer::AUTH_TOKEN_SIZE] = {0};

#ifndef USE_NGS_FACADE

    {
        // Logging in to the friend server is not needed (from CTR-SDK 3.2 on)

        NEX_DEMO_LOG("Request authentication token...\n");

        nn::nex::IndependentServer oIndependentServer;
        // The call context
        nn::nex::ProtocolCallContext oContext;

        if (oIndependentServer.RequestAuthenticationToken(&oContext, NEXDEMO_GAME_SERVER_ID, NEXDEMO_KEY_HASH, authToken, true))
        {
            oContext.Wait();

            if (oContext.GetOutcome())
            {
                // At execution, even when the PrincipalID is 0, PrincipalID can be acquired as long as asynchronous processing succeeds
                NEX_DEMO_LOG("My PrincipalId %d\n", nn::friends::GetMyPrincipalId());

                NEX_DEMO_LOG("authentication token = %s\n", authToken);
            }
            else
            {
                NEX_DEMO_LOG("IndependentServer::RequestAuthenticationToken() failed. Error code : %d\n", oIndependentServer.GetLastErrorCode());
            }
        }

        NN_ASSERT(nn::friends::HasLoggedIn() == false);
    }

#else   // USE_NGS_FACADE

    // You must be logged into the friend server to log into the game server and get an authentication token at the same time.
    // 
    NEX_DEMO_LOG("Logging in to Friend Server...\n");
    {
        nn::Result result;
        nn::os::Event event(true);

        // Log into the friend server
        result = nn::friends::Login(&event);
        if (result.IsSuccess())
        {
            event.Wait();

            result = nn::friends::GetLastResponseResult();
            if (result.IsSuccess())
            {
                NEX_DEMO_LOG("My PrincipalId %d\n", nn::friends::GetMyPrincipalId());
                NEX_DEMO_LOG("nn::friends::Login() succeeded.\n");
            }
            else
            {
                NEX_DEMO_LOG("nn::friends::Login() failed. Error code: %d\n", nn::friends::ResultToErrorCode(result));
                NN_DBG_PRINT_RESULT(result);
            }
        }
        else
        {
            NEX_DEMO_LOG("nn::friends::Login() call failed!!!\n");
            NN_DBG_PRINT_RESULT(result);
        }
    }

    if (nn::friends::HasLoggedIn())
    {
        NEX_DEMO_LOG("Logging in to Nintendo Game Server as well...\n");

        // The call context
        nn::nex::ProtocolCallContext oContext;

        nn::nex::NgsFacade oNgsFacade;

        if (oNgsFacade.LoginAndRequestAuthenticationToken(&oContext, NEXDEMO_GAME_SERVER_ID, NEXDEMO_ACCESS_KEY, NEXDEMO_KEY_HASH, authToken))
        {
            oContext.Wait();

            if (oContext.GetOutcome())
            {
                NEX_DEMO_LOG("authentication token = %s\n", authToken);
            }
            else
            {
                NEX_DEMO_LOG("Ngsfacade::LoginAndRequestAuthenticationToken() failed. Error code : %d\n", oNgsFacade.GetLastLoginErrorCode());
            }
        }

        NEX_DEMO_LOG("Logging out from Game Server...\n");
        {
            // Logging out from the game server
            nn::nex::ProtocolCallContext ccLogout;
            oNgsFacade.Logout(&ccLogout);

            ccLogout.Wait();
        }

        NEX_DEMO_LOG("Logging out from Friend Server...\n");
        {
            // Logging out from the friend server
            nn::friends::Logout();
        }

        NEX_DEMO_LOG("Logged out.\n");
    }

#endif  // USE_NGS_FACADE

    nn::friends::Finalize();

    DemoFinalize();
}

void DemoInitialize()
{
    // Connect to the Internet
    if (!nex_demo_util::NetInit())
    {
        NEX_DEMO_LOG("nn::nex::CTRNetInit() failed!!!\n");

        NEX_DEMO_LOG("Leaving application...\n");
    }

    // Prepare to use utilities
    nex_demo_util::Initialize();

    // Initialize NEX's global variables
    nn::nex::GlobalVariables::AcquireInstance();

    // Set to thread unsafe transport buffer thread mode
    nn::nex::Core::SetThreadMode(nn::nex::Core::ThreadModeUnsafeTransportBuffer);

    // Log level settings
    nn::nex::EventLog::GetInstance()->SetLogLevel(nn::nex::EventLog::Debug);
    nn::nex::TraceLog::GetInstance()->SetFlag(TRACE_ALWAYS);
}

void DemoFinalize()
{
    // Release NEX's global variables
    nn::nex::GlobalVariables::ReleaseInstance();

    // Shut down utilities
    nex_demo_util::Finalize();

    // Disconnect from the Internet
    nex_demo_util::NetTerm();

    NEX_DEMO_LOG("Leaving application...\n");
}
