#!/usr/bin/perl
#
# CTR-NEX
#   Independent Server Authentication Token Generation Program Sample Version 1.1
#
# * Usage
#   tokenenc_ctr.pl key pid friendcode servertype [timestamp]
#      key: Authentication token key (character notation)
#      pid: PID (up to 10 digits, not left-padded with zeros)
#      friendcode: Friend code (fixed at 12 digits, left-padded with zeros)
#      servertype: Two single-byte characters ("D1" for development environments)
#      timestamp: Unix timestamp value
#                   Current date-time used if not specified
#

use strict;
use warnings;
use Crypt::CBC;
use Digest::SHA1 qw(sha1_hex);
use MIME::Base64 qw(encode_base64);

# Set the embedded value
my ($key, $pid, $friendcode, $servertype, $timestamp);
if ($#ARGV < 3) {
  die "usage: tokenenc_ctr.pl key pid friendcode servertype [timestamp]\n";
}
if ($ARGV[0] !~ m/^\w{16}$/xms) {
  die "invalid parameter (key)\n";
}
if ($ARGV[1] !~ m/^\d{1,10}$/xms) {
  die "invalid parameter (pid)\n";
}
if ($ARGV[2] !~ m/^\d{12}$/xms) {
  die "invalid parameter (friendcode)\n";
}
if ($ARGV[3] !~ m/^\w{2}$/xms) {
  die "invalid parameter (servertype)\n";
}
if (defined $ARGV[4] && $ARGV[4] !~ m/^\d+$/xms) {
  die "invalid parameter (timestamp)\n";
}
$key        = $ARGV[0];
$pid        = $ARGV[1];
$friendcode = $ARGV[2];
$servertype = $ARGV[3];
if (defined $ARGV[4]) {
  $timestamp = $ARGV[4];
} else {
  $timestamp = time;
}

# Display values
print "Token Key   : $key\n";
print "PID         : $pid\n";
print "Friend Code : $friendcode\n";
print "Server Type : $servertype\n";
print "Timestamp   : $timestamp\n";


# Concatenation after getting the hash value
my $token = "\\u\\$pid\\f\\$friendcode\\s\\$servertype\\t\\$timestamp";
my $hash = substr sha1_hex($token), 0, 8;
print "SHA-1 Hash  : $hash\n";
$token .= "\\h\\$hash";
print "Plain Token : $token\n";


# AES encryption
# CBC mode, key=128bit, IV=all\0, padding=PKCS#5
my $iv  = pack 'H32', '0' x 32; # IV
my $cipher = Crypt::CBC->new(
  -key         => $key,
  -keysize     => 16,
  -literal_key => 1,
  -cipher      => 'Crypt::Rijndael',
  -iv          => $iv,
  -padding     => 'standard',
  -prepend_iv  => 0,
);
my $enctoken = $cipher->encrypt($token);

# BASE64+encoding
my $tmptoken = encode_base64($enctoken, q{});
$tmptoken =~ tr/\/\+\=/\-\.\*/;
print "Token       : $tmptoken\n";

exit;

