/*--------------------------------------------------------------------------
  Copyright (C)2011-2014 Nintendo  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo and/or its licensors and are
  protected by national and international copyright laws.
  They may not be disclosed to third parties or copied or duplicated
  in any form, in whole or in part, without the prior written consent
  of Nintendo.
  The content herein is highly confidential and should be handled
  accordingly.
 *---------------------------------------------------------------------------*/


#include "DemoFramework.h"

#include <cstdio>

#ifdef DEMO_LIB_AVAILABLE
#include <nn/demo.h>
#else
#include "demo.h"
#endif


#include "../DemoColor.h"


namespace
{


}  // End of anonymous namespace


namespace CTR
{


DemoFramework *DemoFramework::s_pInstance = NULL;

void DemoFramework::CreateInstance(void)
{
    if(s_pInstance)
    {
        // Instance is already created.  Do nothing.
        return;
    }
    else
    {
        s_pInstance = new DemoFramework();
    }
}


void DemoFramework::DestroyInstance(void)
{
    if(s_pInstance)
    {
        delete s_pInstance;
        s_pInstance = NULL;
    }
    else
    {
        // Instance is not available.  Do nothing.
    }
}


DemoFramework* DemoFramework::GetInstance(void)
{
    NN_NULL_ASSERT(s_pInstance);

    return s_pInstance;
}


bool DemoFramework::IsCreated(void)
{
    return s_pInstance!=NULL;
}


DemoFramework::DemoFramework(void)
    :m_accValid(false),
     m_latestX(INVALID_TOUCH_POS),
     m_latestY(INVALID_TOUCH_POS),
     m_renderingTarget(INVALID_DISPLAY),
     m_vsyncInterval(1),
     m_beginFrameVsync(0)
{
    // Device memory-related settings. Set a value for the capacity in the same manner as previous SDKs.
    uptr pDevMem = nn::os::GetDeviceMemoryAddress();
    NN_POINTER_ASSERT(pDevMem);

    const size_t DEVICE_MEMORY_SIZE = 32 * 1024 * 1024;
    nn::Result result = nn::os::SetDeviceMemorySize( DEVICE_MEMORY_SIZE );
    NN_UTIL_PANIC_IF_FAILED( result );

    m_appHeap.Initialize(pDevMem, DEVICE_MEMORY_SIZE, nn::os::ALLOCATE_OPTION_LINEAR);
    m_heapForGx = reinterpret_cast<uptr>(m_appHeap.Allocate(GX_HEAP_SIZE));
    NN_NULL_ASSERT(m_heapForGx);

#ifdef DEMO_LIB_AVAILABLE
    m_pFramework = new demo::DrawFrameworkEx;
#else
    m_pFramework = new demo::RenderSystemDrawing;
#endif
    NN_NULL_ASSERT(m_pFramework);

    m_pFramework->Initialize(m_heapForGx, GX_HEAP_SIZE);
    m_pFramework->SetColor(1.0f, 1.0f, 1.0f);
    m_pFramework->SetClearColor(NN_GX_DISPLAY_BOTH, 0.0f, 0.0f, 0.5f, 0.0f);  // display, r, g, b, a

    // Pass invalid values just in case.
    m_accStatus.x = INVALID_ACCELEROMETER_VALUE;
    m_accStatus.y = INVALID_ACCELEROMETER_VALUE;
    m_accStatus.z = INVALID_ACCELEROMETER_VALUE;
}


DemoFramework::~DemoFramework(void)
{
    m_pFramework->Finalize();
    delete m_pFramework;

    m_appHeap.Free(reinterpret_cast<void*>(m_heapForGx));
}


void DemoFramework::ClearBothDisplays(void)
{
#ifdef DEMO_LIB_AVAILABLE
    m_pFramework->ClearDisplay(NN_GX_DISPLAY_BOTH);
#else
    NN_PANIC("Don't call this function in CTR-SDK 0.10.2!");
#endif
}


void DemoFramework::SetClearColor(const struct DemoColor &clr)
{
    m_pFramework->SetClearColor(NN_GX_DISPLAY_BOTH, clr.r, clr.g, clr.b, clr.a);  // display, r, g, b, a
}


void DemoFramework::DrawText(s32 x, s32 y, const char *fmt, ...)
{
    const size_t BUFSIZE = 1024;
    static char textBuf[BUFSIZE];

    va_list va;
    va_start(va, fmt);
    vsnprintf(textBuf, BUFSIZE, fmt, va);
    va_end(va);
    m_pFramework->SetColor(1.0f, 1.0f, 1.0f, 1.0f);  // White
#ifdef DEMO_LIB_AVAILABLE
    m_pFramework->DrawText2d(static_cast<f32>(x), static_cast<f32>(y), textBuf);
#else
    m_pFramework->SetFontSize(8.0f);
    m_pFramework->DrawText(static_cast<f32>(x), static_cast<f32>(y), textBuf);
#endif
}


void DemoFramework::DrawRect(s32 x, s32 y, s32 w, s32 h, const struct DemoColor &clr)
{
    m_pFramework->SetColor(clr.r, clr.g, clr.b, clr.a);

#ifdef DEMO_LIB_AVAILABLE
    m_pFramework->DrawRectangle2d(x, y, w, h);
#else
    m_pFramework->FillRectangle(x, y, w, h);
#endif
}


void DemoFramework::BeginFrame(void)
{
    // Get VSYNC value at top. Compare with this value using the EndFrame function.
    m_beginFrameVsync = nngxCheckVSync(NN_GX_DISPLAY_BOTH);

    // Restore upper screen as render target every frame
    // (Not cached, so do not call the UseUpperDisplay function)
#ifdef DEMO_LIB_AVAILABLE
    // Clear both screens
    ClearBothDisplays();

    m_pFramework->SetRenderingTarget(NN_GX_DISPLAY0);
#else
    m_pFramework->SetRenderTarget(NN_GX_DISPLAY0);
    m_pFramework->Clear();
#endif
    m_renderingTarget = NN_GX_DISPLAY0;

    // Get key input
    m_padReader.ReadLatest(&m_padStatus);

    // Save past touch panel input information
    m_tpStatusPrev = m_tpStatus;

    // Get latest touch panel input
    bool ret = m_tpReader.ReadLatest(&m_tpStatus);
    if(ret)
    {
        if(m_tpStatus.touch==1)
        {
            m_latestX = m_tpStatus.x;
            m_latestY = m_tpStatus.y;
        }
    }
    else
    {
        NN_LOG("TouchPanelReader::ReadLatest() failed.\n");
    }

    // Get latest acceleration data
    m_accValid = m_accReader.ReadLatest(&m_accStatus);
}


void DemoFramework::EndFrame(void)
{
    while(nngxCheckVSync(NN_GX_DISPLAY_BOTH) - m_beginFrameVsync < m_vsyncInterval - 1)
    {
        nngxWaitVSync(NN_GX_DISPLAY_BOTH);
    }

#ifdef DEMO_LIB_AVAILABLE
    m_pFramework->SwapBuffer(NN_GX_DISPLAY_BOTH, true);  // Waits for VSync so true
#else
    m_pFramework->SwapBuffers();
    m_pFramework->WaitVsync(NN_GX_DISPLAY_BOTH);
#endif
}


bool DemoFramework::SetFramePerSecond(s32 fps)
{
    switch(fps)
    {
    case 60:
        m_vsyncInterval = 1;
        return true;
    case 30:
        m_vsyncInterval = 2;
        return true;
    case 20:
        m_vsyncInterval = 3;
        return true;
    case 15:
        m_vsyncInterval = 4;
        return true;
    case 12:
        m_vsyncInterval = 5;
        return true;
    case 10:
        m_vsyncInterval = 6;
        return true;
    case 6:
        m_vsyncInterval = 10;
        return true;
    case 5:
        m_vsyncInterval = 12;
        return true;
    case 4:
        m_vsyncInterval = 15;
        return true;
    case 3:
        m_vsyncInterval = 20;
        return true;
    case 2:
        m_vsyncInterval = 30;
        return true;
    case 1:
        m_vsyncInterval = 60;
        return true;
    default:
        NN_LOG("Non-supported fps(%d)\n", fps);
        return false;
    }
}


void DemoFramework::UseUpperDisplay(void)
{
    if(m_renderingTarget!=NN_GX_DISPLAY0)
    {
#ifdef DEMO_LIB_AVAILABLE
        m_pFramework->SetRenderingTarget(NN_GX_DISPLAY0);
#else
        // Ignore on purpose.
        //        m_pFramework->SetRenderTarget(NN_GX_DISPLAY0);
#endif
        m_renderingTarget = NN_GX_DISPLAY0;
    }
}


void DemoFramework::UseLowerDisplay(void)
{
    if(m_renderingTarget!=NN_GX_DISPLAY1)
    {
#ifdef DEMO_LIB_AVAILABLE
        m_pFramework->SetRenderingTarget(NN_GX_DISPLAY1);
#else
        // Ignore on purpose.
        //        m_pFramework->SetRenderTarget(NN_GX_DISPLAY1);
#endif
        m_renderingTarget = NN_GX_DISPLAY1;
    }
}


bool DemoFramework::IsHold(bit32 button) const
{
    return m_padStatus.hold & button;
}


bool DemoFramework::IsTrigger(bit32 button) const
{
    return m_padStatus.trigger & button;
}


bool DemoFramework::IsRelease(bit32 button) const
{
    return m_padStatus.release & button;
}


// Determine whether the touch panel was momentarily touched by the stylus.
bool DemoFramework::IsPenTriggered(void) const
{
    return (m_tpStatusPrev.touch==0) && (m_tpStatus.touch==1);
}


bool DemoFramework::IsPenTouched(void) const
{
    return m_tpStatus.touch==1;
}


// Determine whether the stylus was momentarily released from the touch panel.
bool DemoFramework::IsPenReleased(void) const
{
    return (m_tpStatusPrev.touch==1) && (m_tpStatus.touch==0);
}


// Get the coordinates where the stylus last touched the touch panel.
// Return false if the touch panel has not been touched at all.
bool DemoFramework::GetLatestPenPosition(nn::nex::qUnsignedInt16 *x, nn::nex::qUnsignedInt16 *y) const
{
    if((m_latestX==INVALID_TOUCH_POS) || (m_latestY==INVALID_TOUCH_POS))
    {
        return false;
    }
    else
    {
        if(x)
        {
            *x = m_latestX;
        }
        if(y)
        {
            *y = m_latestY;
        }
        return true;
    }
}


f32 DemoFramework::GetAnalogX(void)
{
    return m_padReader.NormalizeStick(m_padStatus.stick.x);
}


f32 DemoFramework::GetAnalogY(void)
{
    return m_padReader.NormalizeStick(m_padStatus.stick.y);
}


void DemoFramework::GetAnalog(f32 *x, f32 *y)
{
    m_padReader.NormalizeStickWithScale(x, y, m_padStatus.stick.x, m_padStatus.stick.y);
}


bool DemoFramework::GetAccelerationData(s16 *x, s16 *y, s16 *z) const
{
    NN_NULL_ASSERT(x);
    NN_NULL_ASSERT(y);
    NN_NULL_ASSERT(z);

    if(m_accValid)
    {
        *x = m_accStatus.x;
        *y = m_accStatus.y;
        *z = m_accStatus.z;
        return true;
    }
    else
    {
        return false;
    }
}


void DemoFramework::PrintCurrentThreadInfo(void) const
{
    using namespace nn::os;
    NN_LOG("Thread ID: %x, Prio: %d, \n", Thread::GetCurrentId(), Thread::GetCurrentPriority());
}


void DemoFramework::SleepCurrentThreadMilliseconds(s32 msec)
{
    nn::os::Thread::Sleep(nn::fnd::TimeSpan::FromMilliSeconds(msec));
}


void* DemoFramework::AllocateDeviceMemory(size_t size)
{
    return m_appHeap.Allocate(size);
}


void DemoFramework::FreeDeviceMemory(void *ptr)
{
    NN_NULL_ASSERT(ptr);

    m_appHeap.Free(ptr);
}


bool DemoFramework::IsInMainThread(void) const
{
    return nn::os::Thread::GetMainThread().GetId()==nn::os::Thread::GetCurrentId();
}



// Get character string representing the elapsed time since the program began execution.
// 00:00:00 (HH::MM::SS) format.
// Allocate a buffer of sufficient size.
void DemoFramework::GetPassedTimeString(char str[])
{
    if(str==NULL)
    {
        return;
    }

    using namespace nn::os;
    s64 t = static_cast<s64>(Tick::GetSystemCurrent());

    s64 sec = t / Tick::TICKS_PER_SECOND;
    s64 ss = sec % 60;
    s64 mm = sec / 60 % 60;
    s64 hh = sec / 60 / 60;

    std::sprintf(str, "%02lld:%02lld:%02lld", hh, mm, ss);
}


void DemoFramework::Checker(void)
{
    NN_LOG("sizeof(m_hid)             = %2u, addr = %08x\n", sizeof(m_hid), &m_hid);
    NN_LOG("sizeof(m_padding)         = %2u, addr = %08x\n", sizeof(m_padding), m_padding);
    NN_LOG("sizeof(m_heapForGx)       = %2u, addr = %08x\n", sizeof(m_heapForGx), &m_heapForGx);
    NN_LOG("sizeof(m_padStatus)       = %2u, addr = %08x\n", sizeof(m_padStatus), &m_padStatus);
    NN_LOG("sizeof(m_padReader)       = %2u, addr = %08x\n", sizeof(m_padReader), &m_padReader);
    NN_LOG("sizeof(m_tpStatus)        = %2u, addr = %08x\n", sizeof(m_tpStatus), &m_tpStatus);
    NN_LOG("sizeof(m_tpStatusPrev)    = %2u, addr = %08x\n", sizeof(m_tpStatusPrev), &m_tpStatusPrev);
    NN_LOG("sizeof(m_tpReader)        = %2u, addr = %08x\n", sizeof(m_tpReader), &m_tpReader);
    NN_LOG("sizeof(m_latestX)         = %2u, addr = %08x\n", sizeof(m_latestX), &m_latestX);
    NN_LOG("sizeof(m_latestY)         = %2u, addr = %08x\n", sizeof(m_latestY), &m_latestY);
    NN_LOG("sizeof(m_pFramework)      = %2u, addr = %08x\n", sizeof(m_pFramework), &m_pFramework);
    NN_LOG("sizeof(m_appHeap)         = %2u, addr = %08x\n", sizeof(m_appHeap), &m_appHeap);
    NN_LOG("sizeof(m_renderingTarget) = %2u, addr = %08x\n", sizeof(m_renderingTarget), &m_renderingTarget);
    NN_LOG("sizeof(m_vsyncInterval)   = %2u, addr = %08x\n", sizeof(m_vsyncInterval), &m_vsyncInterval);
    NN_LOG("sizeof(m_beginFrameVsync) = %2u, addr = %08x\n", sizeof(m_beginFrameVsync), &m_beginFrameVsync);
}


}  // end of namespace CTR


