/*--------------------------------------------------------------------------
  Copyright (C)2011-2014 Nintendo  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo and/or its licensors and are
  protected by national and international copyright laws.
  They may not be disclosed to third parties or copied or duplicated
  in any form, in whole or in part, without the prior written consent
  of Nintendo.
  The content herein is highly confidential and should be handled
  accordingly.
 *---------------------------------------------------------------------------*/


#include "HeapManager.h"

#include <nn/fnd.h>
#include <nn/dbg.h>

#include <OnlineCore/src/Platform/Core/PlatformDecl.h>

#include <OnlineCore/src/Platform/Core/ConsoleIO.h>
#include "../UtilTrace.h"
#include <OnlineCore/src/Platform/Core/EventLog.h>


//n1769 This heap manager uses the CTR-SDK ExpHeap.



namespace
{
const size_t MEMSIZE = 10 * 1024 * 1024;  // 10MB
u32 s_memoryBlock[MEMSIZE/sizeof(u32)] = {};

nn::fnd::ThreadSafeExpHeap s_heap;
bool s_bExpHeapInitialized = false;
size_t s_totalFreeSize = 0;
u32 s_ageCounter = 0;  // Number of times Age function is called.


void* expAlloc(size_t sz)
{
    //n1769
    // Add a region to the start to store the "Age Counter" when the program is executing.
    u32 *p = static_cast<u32*>(s_heap.Allocate(sz + sizeof(u32)));

    *p = s_ageCounter;

    return p+1;
}


void expFree(void *p)
{
    s_heap.Free(static_cast<u32*>(p)-1);
}


void expVisitor(void *pBlock, const nn::fnd::ExpHeapBase *heap, uptr param)
{
    (void)param;
#ifdef NN_SWITCH_DISABLE_ASSERT_WARNING
    (void)heap;
#else
    NN_ASSERT(heap==&s_heap);
#endif
    if(!s_heap.CheckBlock(pBlock))
    {
        NN_PANIC("CheckBlock() failed!\n");
    }

#if 0
    //n1769 Anyway, a method to report all.
    u32 past = *static_cast<u32*>(pBlock);
    QLOG(nn::nex::EventLog::Always, pBlock << _T(" counter: ") << past << _T(" size: ") << s_heap.GetSizeOf(pBlock));
#else
    //n1769 Only the most recent memory block is printed.
    u32 past = *static_cast<u32*>(pBlock);
    if(s_ageCounter - past < 3)
    {
        QLOG(nn::nex::EventLog::Always, pBlock << _T(" counter: ") << past << _T(" size: ") << s_heap.GetSizeOf(pBlock));
    }
#endif
}

#ifdef LOG_ENABLED //n2788: Do not use except with QLOG
// Subcontracting function for expCountMemoryBlocks function.
void expCountVisitor(void *pBlock, const nn::fnd::ExpHeapBase *heap, uptr param)
{
#ifdef NN_SWITCH_DISABLE_ASSERT_WARNING
    (void)heap;
#else
    NN_ASSERT(heap==&s_heap);
#endif
    if(!s_heap.CheckBlock(pBlock))
    {
        NN_PANIC("CheckBlock() failed!\n");
    }

    ++(*reinterpret_cast<u32*>(param));
}


//Count up number of memory blocks existing in the heap.
u32 expCountMemoryBlocks(void)
{
    u32 counter = 0;
    s_heap.VisitAllBlocks(expCountVisitor, reinterpret_cast<uptr>(&counter));
    return counter;
}
#endif //LOG_ENABLED

}  // End of anonymous namespace


namespace CTR
{

bool HeapManager::IsInitialized(void)
{
    return s_bExpHeapInitialized;
}


void HeapManager::Initialize(void)
{
    //n1769 Should lock be done originally with critical sessions?
    s_heap.Initialize(reinterpret_cast<uptr>(s_memoryBlock), MEMSIZE);
    s_totalFreeSize = s_heap.GetTotalFreeSize();
    NN_LOG("totalFreeSize = %u bytes\n", s_totalFreeSize);
    nn::nex::MemoryManager::SetBasicMemoryFunctions(expAlloc, expFree);

    s_bExpHeapInitialized = true;
}


void HeapManager::Finalize(void)
{
    if(IsInitialized())
    {
        s_bExpHeapInitialized = false;

        nn::nex::MemoryManager::SetBasicMemoryFunctions(NULL, NULL);
        s_totalFreeSize = 0;
        s_heap.Finalize();
    }
}


void HeapManager::PrintDebugInfo(bool bDetail)
{
    if (bDetail)
    {
        s_heap.Dump();
    }

#ifdef LOG_ENABLED
    QLOG(nn::nex::EventLog::Always, _T("-------- ExpHeap report begin --------"));
    QLOG(nn::nex::EventLog::Always, _T("ExpHeap free: ") << s_heap.GetTotalFreeSize());
    QLOG(nn::nex::EventLog::Always, _T("ExpHeap memory blocks: ") << expCountMemoryBlocks());
#endif

    if (bDetail)
    {
        s_heap.VisitAllBlocks(expVisitor, 0);
    }

    QLOG(nn::nex::EventLog::Always, _T("-------- ExpHeap report end   --------"));
}


void HeapManager::Age(void)
{
    ++s_ageCounter;
}


bool HeapManager::IsClean(void)
{
    if(IsInitialized())
    {
        return s_heap.GetTotalFreeSize()==s_totalFreeSize;
    }
    else
    {
        return false;
    }
}


bool HeapManager::IsStarved(void)
{
    NN_LOG("%s is not implemented.\n", __PRETTY_FUNCTION__);

    return false;
}


// Returns total free memory.
std::size_t HeapManager::GetTotalFreeSize(void)
{
    return s_heap.GetTotalFreeSize();
}


void HeapManager::ClearLog(void)
{
    NN_LOG("%s is not implemented.\n", __PRETTY_FUNCTION__);
}


void HeapManager::Analyze(void)
{
    NN_LOG("%s is not implemented.\n", __PRETTY_FUNCTION__);
}


void HeapManager::EnableLogging(void)
{
    NN_LOG("%s is not implemented.\n", __PRETTY_FUNCTION__);
}


void HeapManager::DisableLogging(void)
{
    NN_LOG("%s is not implemented.\n", __PRETTY_FUNCTION__);
}

} // end of namespace CTR
