﻿/*--------------------------------------------------------------------------
  Copyright (C)2011-2014 Nintendo  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo and/or its licensors and are
  protected by national and international copyright laws.
  They may not be disclosed to third parties or copied or duplicated
  in any form, in whole or in part, without the prior written consent
  of Nintendo.
  The content herein is highly confidential and should be handled
  accordingly.
 *---------------------------------------------------------------------------*/


#include "ErrorHandling.h"

#include <OnlineCore/src/Platform/Core/UserContext.h>
#include <OnlineCore/src/Platform/Core/SystemError.h>
#include <OnlineCore/src/Platform/Core/SystemErrorResult.h>
#include <OnlineCore/src/Platform/Core/Result.h>
#include <OnlineCore/src/Transport/Interface/TransportError.h>
#include <OnlineCore/src/Platform/Core/ConsoleIO.h>
#include <OnlineCore/src/Core/Scheduler.h>

#include <cstdlib>

#include "NexUtil.h"

namespace
{
// Non-platform dependent (shared) error code related.
nn::nex::qResult        s_commonResult(0);
nn::nex::qInt32         s_commonExtra                = 0;

}  // End of anonymous namespace

bool ErrorHandling::s_bInitialized = false;

// Initialization.
void ErrorHandling::Initialize(void)
{
    if(!s_bInitialized)
    {
        s_bInitialized = true;
    }
}


// End processing.
void ErrorHandling::Finalize(void)
{
    if(s_bInitialized)
    {
        s_bInitialized = false;
    }
}


// Was an error detected on the NEX side?
bool ErrorHandling::IsErrorDetected(void)
{
    NEX_DEMO_ASSERT(s_bInitialized);

    bool bFatal = false;
    nn::nex::NetZ *pNetZ = nn::nex::NetZ::GetInstance();
    if ( pNetZ )
    {
        nn::nex::qResult fatal = pNetZ->GetFatalError();
        if (!fatal)
        {
            nn::nex::_QLOG(nn::nex::EventLog::Error, _T("NetZ fatal error is happend."));
            nn::nex::_QLOG(nn::nex::EventLog::Error, _T("      Error code: 0x%x / %ls"), fatal.GetReturnCode(), fatal.GetErrorName());
            bFatal = true;
        }
    }
    return bFatal;

}


const nn::nex::qResult& ErrorHandling::GetCommonResultCode(void)
{
    return s_commonResult;
}

// Called when NEX API returns false. Gets the error code that caused `false` to be returned and prints its content.
// 
void ErrorHandling::ProcessError(void)
{
    nn::nex::qUnsignedInt err = nn::nex::SystemError::GetLast();
    const size_t SZ = 256;
    nn::nex::qChar buf[SZ];
    nn::nex::SystemError::GetErrorString(err, buf, SZ);
    nn::nex::ConsoleIO::Print(_T("NEX error: %s\n"), buf);
}

// Prints the string representation of the result code to debug output.
void ErrorHandling::Print(void)
{
    PrintCommon(GetCommonResultCode());
}


void ErrorHandling::PrintCommon(const nn::nex::qResult &result)
{
#ifndef BUILD_VERBOSE
    (void)result;
#else
    QLOG(nn::nex::EventLog::Always, _T(" Error:       ") << result.GetErrorName());
    QLOG(nn::nex::EventLog::Always, _T(" Filename:    ") << result.GetFileName());
    QLOG(nn::nex::EventLog::Always, _T(" Line:        ") << result.GetLineNumber());
    QLOG(nn::nex::EventLog::Always, _T(" qResult:     ") << result.GetReturnCode());
#endif
}


void ErrorHandling::PrintCommon(void)
{
    PrintCommon(s_commonResult);
    nn::nex::_QLOG(nn::nex::EventLog::Always, _T(" extra:   %d"), s_commonExtra);
}


// Clears both the nn::Result type and qResult type values being maintained in this class.
void ErrorHandling::Clear(void)
{
    s_commonResult = nn::nex::qResult(0);
    s_commonExtra  = 0;
}


// n1769
// Forcibly sets the result code. Intentionally causes an error to check whether the error processing sequence works correctly.
// 
void ErrorHandling::SetCommonResultCode(nn::nex::qResult result)
{
    s_commonResult = result;
}
