/*---------------------------------------------------------------------------*
  (C) Copyright 1998-2014, Technologies Quazal Inc. All rights reserved.
  Quazal Net-Z, Quazal Rendez-Vous and Quazal are trademarks of Quazal.

  Copyright (C)2011-2014 Nintendo  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo and/or its licensors and are
  protected by national and international copyright laws.
  They may not be disclosed to third parties or copied or duplicated
  in any form, in whole or in part, without the prior written consent
  of Nintendo.
  The content herein is highly confidential and should be handled
  accordingly.
 *---------------------------------------------------------------------------*/

#include <stdarg.h>
#include <string.h>
#include <ctype.h>
#include "Console.h"

using namespace nexdemo;

// Constructor.
Console::Console()
{
    m_ppBuffer      = NULL;
    m_pRenderBuffer = NULL;

    m_isInitialized = false;

    // Initialize variables
    Finalize();
}

// Destructor.
Console::~Console()
{
    Finalize();
}

// Performs initialization.
void Console::Initialize(s16 x, s16 y, u16 windowCharsPerLine, u16 windowLines, u32 bufferCharsPerLine, u32 bufferLines, u8 lineSpace)
{
    if (windowCharsPerLine == 0 || windowLines == 0 || bufferCharsPerLine == 0 || bufferLines == 0)
    {
        return;
    }
    if (m_isInitialized)
    {
        Finalize();
    }

    m_x                      = x;
    m_y                      = y;
    m_windowCharsPerLine     = windowCharsPerLine;
    m_windowLines            = windowLines;
    m_bufferCharsPerLine     = bufferCharsPerLine;
    m_bufferLines            = bufferLines;
    m_usedBufferCharsPerLine = 0;
    m_usedBufferLines        = 0;
    m_lineSpace              = lineSpace;
    m_cursorX                = 0;
    m_cursorY                = 0;
    m_scrollX                = 0;
    m_scrollY                = 0;
    m_isOver                 = false;
    m_isFull                 = false;

    m_ppBuffer = new char*[m_bufferLines];

    for (u32 i = 0; i < m_bufferLines; i++)
    {
        m_ppBuffer[i] = new char[m_bufferCharsPerLine + 1];
        m_ppBuffer[i][m_bufferCharsPerLine] = '\0';
    }

    m_pRenderBuffer = new char[m_windowCharsPerLine + 1];
    m_pRenderBuffer[m_windowCharsPerLine] = '\0';

    // Clear the content of the console
    Clear();

    m_isInitialized = true;
}

// Performs shutdown.
void Console::Finalize()
{
    m_isInitialized = false;

    if (m_ppBuffer)
    {
        for (u32 i = 0; i < m_bufferLines; i++)
        {
            delete[] m_ppBuffer[i];
        }

        delete[] m_ppBuffer;
    }

    if (m_pRenderBuffer)
    {
        delete[] m_pRenderBuffer;
    }

    m_ppBuffer           = NULL;
    m_pRenderBuffer      = NULL;
    m_x                  = 0;
    m_y                  = 0;
    m_windowCharsPerLine = 0;
    m_windowLines        = 0;
    m_bufferCharsPerLine = 0;
    m_bufferLines        = 0;
    m_lineSpace          = 0;
    m_cursorX            = 0;
    m_cursorY            = 0;
    m_scrollX            = 0;
    m_scrollY            = 0;
    m_isOver             = false;
    m_isFull             = false;
}

// Clear the content of the console.
void Console::Clear()
{
    if (!m_isInitialized)
    {
        return;
    }

    if (m_ppBuffer)
    {
        for (u32 i = 0; i < m_bufferLines; i++)
        {
            if (m_ppBuffer[i])
            {
                memset(m_ppBuffer[i], 0, m_bufferCharsPerLine);
            }
        }
    }
}

// Output the string.
void Console::Printf(const char* pFormat, ...)
{
    if (!m_isInitialized || !pFormat)
    {
        return;
    }

    // Buffer
    char buffer[NN_DBG_PRINTF_BUFFER_LENGTH + 1] = {0};
    // Argument list
    va_list pArgs;

    va_start(pArgs, pFormat);
    {
        vsnprintf(buffer, NN_DBG_PRINTF_BUFFER_LENGTH, pFormat, pArgs);
    }
    va_end(pArgs);

    PutString(buffer);
}

// Output the string.
void Console::PutString(const char* pString)
{
    if (!m_isInitialized || !pString)
    {
        return;
    }

    u32 i = 0;

    while (pString[i] != '\0')
    {
        switch (pString[i])
        {
        case '\n':
        {
            PutChar('\0');

            if (m_cursorX > m_usedBufferCharsPerLine)
            {
                m_usedBufferCharsPerLine = m_cursorX;
            }

            m_cursorX = 0;
            m_cursorY++;

            if (m_cursorY >= m_windowLines)
            {
                m_isOver = true;
            }
            if (m_cursorY == m_bufferLines)
            {
                m_isFull  = true;
                m_cursorY = 0;
            }

            if (!m_isFull)
            {
                m_usedBufferLines++;

                if (m_isOver)
                {
                    Scroll(0, 1);
                }
            }

            PutChar('\0');
        }
        break;
        case '\t':
        {
            // Number of spaces (calculate with a tab-width of 4)
            u32 space = 4 - (m_cursorX % 4);

            for (u32 t = 0; t < space; t++)
            {
                PutChar(' ');
            }
        }
        break;
        default:
        {
            // Display only the displayable characters
            if (isprint(pString[i]))
            {
                PutChar(pString[i]);
            }
        }
        }

        i++;
    }
}

// Render the content of the console.
void Console::Render(demo::RenderSystemDrawing* pRenderSystem)
{
    if (!m_isInitialized || !pRenderSystem)
    {
        return;
    }

    // Margin(s)
    u8 margin = GetMargin();
    // Height of one line
    u16 lineHeight = GetLineHeight();

    for (s32 i = 0; i < m_windowLines; i++)
    {
        s32 current = i;

        // When more data has accumulated than the number of lines that can be displayed in a window
        if (m_isOver)
        {
            // When a buffer has been fully used
            if (m_isFull)
            {
                // Enable scrolling, and consider the cursor position
                current = (i + m_scrollY + m_cursorY + 1 + m_bufferLines) % m_bufferLines;
            }
            else
            {
                // Enable scrolling
                current = i + m_scrollY;
            }
        }

        // When the number of characters in the screen buffer is greater than the number of characters in the window, cut off just the amount to be displayed on screen
        if (m_windowCharsPerLine < m_bufferCharsPerLine)
        {
            // Copy the string displayed on screen
            strncpy(m_pRenderBuffer, &m_ppBuffer[current][m_scrollX], m_windowCharsPerLine);
            // Render text
            pRenderSystem->DrawText(m_x, m_y + i * lineHeight + margin, m_pRenderBuffer);
        }
        else
        {
            // Render text
            pRenderSystem->DrawText(m_x, m_y + i * lineHeight + margin, &m_ppBuffer[current][m_scrollX]);
        }
    }
}

// Scroll the content of the console.
void Console::Scroll(s16 scrollX, s16 scrollY)
{
    if (!m_isInitialized)
    {
        return;
    }

    s32 newScrollX = m_scrollX + scrollX;
    s32 newScrollY = m_scrollY + scrollY;

    // Right edge
    s32 rightEnd = m_usedBufferCharsPerLine - m_windowCharsPerLine;
    // Bottom edge
    s32 bottomEnd = m_usedBufferLines - m_windowLines;

    if (newScrollX < 0)
    {
        newScrollX = 0;
    }
    if (newScrollY < 0)
    {
        newScrollY = 0;
    }
    if (newScrollX > rightEnd && rightEnd > 0)
    {
        newScrollX = rightEnd;
    }
    if (newScrollY > bottomEnd && bottomEnd > 0)
    {
        newScrollY = bottomEnd;
    }

    m_scrollX = newScrollX;
    m_scrollY = newScrollY;
}

// Set the scroll position of the x-coordinate.
void Console::SetScrollX(s32 scrollX)
{
    if (!m_isInitialized)
    {
        return;
    }

    // Right edge
    s32 rightEnd = m_usedBufferCharsPerLine - m_windowCharsPerLine;

    if (scrollX < 0)
    {
        scrollX = 0;
    }
    if (scrollX > rightEnd && rightEnd > 0)
    {
        scrollX = rightEnd;
    }

    m_scrollX = scrollX;
}

// Set the scroll position of the y-coordinate.
void Console::SetScrollY(s32 scrollY)
{
    if (!m_isInitialized)
    {
        return;
    }

    // Bottom edge
    s32 bottomEnd = m_usedBufferLines - m_windowLines;

    if (scrollY < 0)
    {
        scrollY = 0;
    }
    if (scrollY > bottomEnd && bottomEnd > 0)
    {
        scrollY = bottomEnd;
    }

    m_scrollY = scrollY;
}

// Add the character(s) to the buffer.
void Console::PutChar(char c)
{
    m_ppBuffer[m_cursorY][m_cursorX] = c;

    if (c != '\0')
    {
        bool isNewLine = false;

        m_cursorX++;

        if (m_cursorX == m_bufferCharsPerLine)
        {
            m_cursorX                = 0;
            m_usedBufferCharsPerLine = m_bufferCharsPerLine;
            m_cursorY++;

            isNewLine = true;
        }

        if (m_cursorY >= m_windowLines)
        {
            m_isOver = true;
        }
        if (m_cursorY == m_bufferLines)
        {
            m_isFull  = true;
            m_cursorY = 0;
        }

        if (isNewLine && !m_isFull)
        {
            m_usedBufferLines++;

            if (m_isOver)
            {
                Scroll(0, 1);
            }
        }
    }
}
