﻿/*---------------------------------------------------------------------------*
  (C) Copyright 1998-2014, Technologies Quazal Inc. All rights reserved.
  Quazal Net-Z, Quazal Rendez-Vous and Quazal are trademarks of Quazal.

  Copyright (C)2011-2014 Nintendo  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo and/or its licensors and are
  protected by national and international copyright laws.
  They may not be disclosed to third parties or copied or duplicated
  in any form, in whole or in part, without the prior written consent
  of Nintendo.
  The content herein is highly confidential and should be handled
  accordingly.
 *---------------------------------------------------------------------------*/
/*
 * NEX Sample
 *
 * JoinFromFriendList
 *
 * Description:
 *
 *  This sample is for joining from the friend list.
 *
 *  In a normal launch, the matchmaking session is registered, the information for joining the gathering is presented to friends, and the system then waits for the stipulated number of people to join the session.
 *  
 *
 * When an application is launched from the friend list, the local device gets the presence information for the user to join-in with, and then joins the P2P session after joining the matchmaking session.
 *  
 *  If, after joining, the stipulated number of participants has not been reached, the information for joining the gathering is presented to friends, and the system waits for the stipulated number of people to gather.
 *  
 *
 *  When the stipulated number of people has been reached, whether they are owner or participant, the information is presented to the local stations' friends indicating that the matchmaking session can no longer be joined.
 *  
 *
 * Note:
 *  Even when updating presence with the nn::frinends::UpdateGameMode function, it can take roughly 10 seconds from the previous update before friends are notified.
 *  Because it is possible for friend presence-related change notifications to be delayed due to network conditions, be aware that even if the friend presence library determines that it is possible to join, that is not always necessarily the case.
 *  
 *
 *  Building this sample requires CTR-SDK 3.1 or later.
 */

#include <nn.h>
#include <nn/applet.h>
#include <nn/friends.h>
#include <nex.h>
#include <RendezVous/Services/NATTraversal.h>

#include "../common/NexUtil.h"
#include "../common/NexUtilNetZ.h"


#include "Graphics.h"
#include "Console.h"

//-----------------------------------------------------------------------------
// Constant Definitions
//-----------------------------------------------------------------------------
// Matchmaking mode
#define MATCHMAKE_GAME_MODE 0

// The mask that limits the joinable game modes.
#define MATCHMAKE_GAME_MODE_MASK 0x0000000000000001

// Number of connected persons.
#define MATCHMAKE_MAX_PARTICIPANTS 2

// Run matchmaking between friends?
//#define MATCHMAKE_FRIEND_MODE

// Index of the search attribute to use.
#define MATCHMAKE_ATTR_INDEX 0

// The maximum search attribute value to use.
#define MATCHMAKE_ATTR_MAX 2

// Specifies the time to wait alone after creating a session.
#define ORPHAN_WAIT_TIME 60000

// Amount of time to sleep
#define DISPATCH_PERIOD 100

// Time to wait for a dispatch.
#define DISPATCH_TIMEOUT 3

// Time to send and receive data after matchmaking
#define MATCHMAKING_TIME 60000

//-----------------------------------------------------------------------------
// Function definitions
//-----------------------------------------------------------------------------
void RunMainLoop(nn::nex::MatchmakeExtensionClient *pMatchMakingClient,
                 bool isFromFriendList);

bool OwnerFunction(nn::nex::MatchmakeSession* pMatchmakeSession,
                   nn::nex::MatchmakeExtensionClient* pMatchMakingClient);

bool ParticipantFunction(nn::nex::MatchmakeSession* pMatchmakeSession,
                         nn::nex::MatchmakeExtensionClient* pMatchMakingClient);

qBool WaitForParticipantsLoop(const nn::nex::NetZ* pNetZ);

void RunGameLoop(const nn::nex::NetZ* pNetZ);

void OutputConnectionInfo();

void UpdateGameMode(const nn::nex::MatchmakeSession& matchmakeSession, bool isJoinable, qBool isOwner);

void RenderLog();

//-----------------------------------------------------------------------------
// Variable Definitions
//-----------------------------------------------------------------------------

// Console
nexdemo::Console s_console;

// Critical section for mutual exclusion
nn::os::CriticalSection s_criticalSection;

//-----------------------------------------------------------------------------
// Main Function
//-----------------------------------------------------------------------------
void nnMain()
{
    // Prepare to use utilities
    nex_demo_util::Initialize();

    // Support for launching of the HOME Menu
    nn::friends::Initialize();

    nn::os::Thread::Sleep(nn::fnd::TimeSpan::FromMicroSeconds(10));

    // Process when Close button was pressed at launch
    if (nn::applet::IsExpectedToCloseApplication())
    {
        nn::applet::PrepareToCloseApplication();
        nn::applet::CloseApplication();
    }

    // HID initialization
    nn::hid::Initialize();
    nn::hid::PadReader oPadReader;

    // Initialization of the log console
    {
        nexdemo::Graphics::Initialize();
        s_criticalSection.Initialize();

        // Space between lines
        u8 lineSpace = 2;
        // Number of characters per line
        u16 charsPerLine = NN_GX_DISPLAY0_HEIGHT / nexdemo::Console::FONT_SIZE;
        // Number of lines
        u16 lines = NN_GX_DISPLAY0_WIDTH / (nexdemo::Console::FONT_SIZE + lineSpace);

        s_console.Initialize(0, 0, charsPerLine, lines, lineSpace);
    }

    NEX_DEMO_LOG("*******************************\n");
    NEX_DEMO_LOG("   JoinFromFriendList sample\n");
    NEX_DEMO_LOG("*******************************\n\n");

    // Connect to the Internet.
    if (!nex_demo_util::NetInit())
    {
        NEX_DEMO_LOG("net init failed...\n");
        NEX_DEMO_LOG("Leaving application...\n");

        return;
    }

    // Initialization
    nn::nex::GlobalVariables::AcquireInstance();

    // Set the thread mode
    nn::nex::Core::SetThreadMode(nn::nex::Core::ThreadModeUnsafeTransportBuffer);

    // Timeout settings for NAT traversal.
    nn::nex::RootTransport::SetNATTraversalTimeout(MATCHMAKE_MAX_PARTICIPANTS * 5000);

    // Default frequency for checking for station inconsistencies is every 5 seconds, so set it to be 10 seconds longer than NAT traversal timeout.
    // 
    nn::nex::Session::SetDetectStationInconsitencyThresholdCount(MATCHMAKE_MAX_PARTICIPANTS + 2);

    // Set the network topology.
    nn::nex::NetZ::SetNetworkTopology(nn::nex::NetworkTopology::Peer2Peer);

    nn::nex::EventLog::GetInstance()->SetLogLevel(nn::nex::EventLog::Debug);
    nn::nex::TraceLog::GetInstance()->SetFlag(TRACE_ALWAYS);

    // Determine whether launched from the friend list
    bool isFromFriendList = false;
    nn::friends::FriendKey friendKey;
    if(nn::friends::IsFromFriendList(&friendKey))
    {
        NEX_DEMO_LOG("this program run from friendlist (pid:%u)\n", friendKey.principalId);

        // Get the presence information, and confirm whether joinable
        nn::friends::FriendPresence presence;
        nn::friends::GetFriendPresence(&presence, &friendKey);
        NEX_DEMO_LOG("  GameId                : %08x\n", presence.gameMode.joinGameId);
        NEX_DEMO_LOG("  GameMode              : %u\n", presence.gameMode.joinGameMode);
        NEX_DEMO_LOG("  GroupId(GatheringId)  : %u\n", presence.gameMode.joinGroupId);
        NEX_DEMO_LOG("  MatchmakeSystemType   : %u\n", presence.gameMode.matchmakeSystemType);
        NEX_DEMO_LOG("  Owner                 : %u\n", presence.gameMode.ownerPrincipalId);
        NEX_DEMO_LOG("  JoinAvailabilityFlag  : %08x\n", presence.gameMode.joinAvailabilityFlag);

        if(presence.IsJoinable(MATCHMAKE_GAME_MODE_MASK))
        {
            isFromFriendList = true;
        }
        else
        {
            // Even when launched from the friend list, not joinable depending on the timing
            NEX_DEMO_LOG("this program run from friendlist but is not joinable.\n");
        }
    }
    else
    {
        NEX_DEMO_LOG("this program run from Home (or Test) Menu\n", friendKey.principalId);
    }

    // Display the mode
    NEX_DEMO_LOG("\n");
#ifdef MATCHMAKE_FRIEND_MODE
    NEX_DEMO_LOG("***** Friend match ***** \n");
#else
    NEX_DEMO_LOG("***** Anybody match ***** \n");
#endif
    NEX_DEMO_LOG("***** Max participants: %d ***** \n", MATCHMAKE_MAX_PARTICIPANTS);
    NEX_DEMO_LOG("***** Send mode: Unreliable ***** \n");
    NEX_DEMO_LOG("\n");

    nn::nex::Credentials              *pCredentials        = NULL;
    nn::nex::NgsFacade                *pNgsFacade          = new nn::nex::NgsFacade();
    nn::nex::NATTraversalClient       *pNATTraversalClient = new nn::nex::NATTraversalClient();
    nn::nex::MatchmakeExtensionClient *pMatchMakingClient  = new nn::nex::MatchmakeExtensionClient();

    // Log in to the game server.
    pCredentials = nex_demo_util::GameServerLogin(pNgsFacade);

    // Bind service(s)
    if (pCredentials)
    {
        if (!pMatchMakingClient->Bind(pCredentials))
        {
            NEX_DEMO_LOG("Cannot bind to Matchmaking Service\n");
            nex_demo_util::GameServerLogout(pNgsFacade);
            goto END;
        }
        else
        {
            NEX_DEMO_LOG("Bind to Matchmaking service\n");
        }

        if (!pNATTraversalClient->Bind(pCredentials))
        {
            NEX_DEMO_LOG("Cannot bind to NATTraversal Service\n");
            nex_demo_util::GameServerLogout(pNgsFacade);
            goto END;
        }
    }
    else
    {
        goto END;
    }

    // Main loop.
    RunMainLoop(pMatchMakingClient, isFromFriendList);

END:
    // Delete each object
    pMatchMakingClient->Unbind();
    delete pMatchMakingClient;

    pNATTraversalClient->Unbind();
    delete pNATTraversalClient;

    // Log out
    nex_demo_util::GameServerLogout(pNgsFacade);

    // Delete the login client
    delete pNgsFacade;

    nn::nex::GlobalVariables::ReleaseInstance();

    // End when a button is pressed
    NEX_DEMO_LOG("Press B to Quit Sample.\n");
    while (1)
    {
        nn::hid::PadStatus oPadStatus;
        oPadReader.ReadLatest(&oPadStatus);

        if (oPadStatus.trigger & nn::hid::BUTTON_B)
        {
            break;
        }

        nn::nex::Platform::Sleep(30);
    }

    NEX_DEMO_LOG("Leaving application...\n");

    // Shut down utilities
    nex_demo_util::Finalize();

    // Disconnect from the Internet.
    nex_demo_util::NetTerm();

    s_console.Finalize();
    nexdemo::Graphics::Finalize();

    nn::os::SetDeviceMemorySize(0);
    nn::os::SetHeapSize(0);

    nn::applet::PrepareToCloseApplication();
    nn::applet::CloseApplication();
}

//-----------------------------------------------------------------------------
// Invoke matchmaking and the main loop
//-----------------------------------------------------------------------------
void RunMainLoop(nn::nex::MatchmakeExtensionClient *pMatchMakingClient,
                 bool isFromFriendList)
{
    qBool isOwner = false;
    // Enables NAT traversal.
    nn::nex::ConnectivityManager::GetInstance()->StartNATSession();

    // Create a NetZ object
    nn::nex::NetZ *pNetZ = new nn::nex::NetZ;

    if (!pNetZ->IsValid())
    {
        NEX_DEMO_LOG("NetZ is invalid!!\n");
        delete pNetZ;
        nn::nex::ConnectivityManager::GetInstance()->StopNATSession();
        return;
    }

    // Set up name displayed in station.
    nex_demo_util::SetupLocalPlayerName();

    // Start matchmaking
    bool bReadyToRun = false;
    nn::nex::MatchmakeSession matchmakeSession;

    // Try to create a session or join another open session.
    if(!isFromFriendList)
    {
        isOwner = true;
        bReadyToRun = OwnerFunction(&matchmakeSession, pMatchMakingClient);
    }
    else
    {
        isOwner = false;
        bReadyToRun = ParticipantFunction(&matchmakeSession, pMatchMakingClient);
    }

    if(!bReadyToRun)
    {
        delete pNetZ;
        nn::nex::ConnectivityManager::GetInstance()->StopNATSession();
        return;
    }

    // Perform presence functions
    UpdateGameMode(matchmakeSession, true, isOwner);

    // Wait
    if (WaitForParticipantsLoop(pNetZ))
    {
        // Close off when owner
        if(isOwner)
        {
            nn::nex::ProtocolCallContext oContext;
            if(pMatchMakingClient->CloseParticipation(&oContext, matchmakeSession.GetID()))
            {
                oContext.Wait();
            }
        }

        // Make unjoinable and set presence information
        UpdateGameMode(matchmakeSession, false, isOwner);
        RunGameLoop(pNetZ);
    }

    // Delete the NetZ object
    {
        nn::nex::CallContext oContext;
        pNetZ->Terminate(&oContext);
        oContext.Wait();
        if ( oContext.GetOutcome() == false )
        {
            // It is possible that termination failed and memory leak might have occurred.
            NEX_DEMO_LOG("NetZ::Terminate is failed. Memory leak might be happend!\n");
        }
        delete pNetZ;
    }

    // Ends NAT traversal
    nn::nex::ConnectivityManager::GetInstance()->StopNATSession();

    // Leave the session.
    {
        nn::nex::ProtocolCallContext oContext;
        if(isOwner)
        {
            if(pMatchMakingClient->UnregisterGathering(&oContext, matchmakeSession.GetID()))
            {
                oContext.Wait();
            }
        }
        else
        {
            if(pMatchMakingClient->EndParticipation(&oContext, matchmakeSession.GetID()))
            {
                oContext.Wait();
            }
        }
    }

    // Make the GameMode, etc., blank and set presence information
    matchmakeSession.Reset();
    UpdateGameMode(matchmakeSession, false, false);
}

//-----------------------------------------------------------------------------
// Wait for enough players to join.
//-----------------------------------------------------------------------------
qBool WaitForParticipantsLoop(const nn::nex::NetZ* pNetZ)
{
    qBool bReadyToRun = false;

    NEX_DEMO_LOG("Starting NetZ session...\n");

    // Wait until the prescribed number of people have joined.
    nn::nex::Time tiReadyStart = nn::nex::Time::GetTime();
    nn::nex::Time tiWaitStart = tiReadyStart;

    int oldNumStations = 0;
    while (1)
    {
        nn::nex::Time nowTime = nn::nex::Time::GetTime();

        int numStations = 0;
        nn::nex::Station::SelectionIterator itStations;
        numStations = itStations.Count();

        if (oldNumStations != numStations)
        {
            // Reset the wait time when there has been a change in the number of people connected
            tiWaitStart = nowTime;
        }

        // Remember how many people there were before
        oldNumStations = numStations;

        // End when no one joins the session for a set period of time
        if (numStations == 1 && nowTime - tiWaitStart >= ORPHAN_WAIT_TIME)
        {
            NEX_DEMO_LOG("*** time passed as. abort.\n");
            break;
        }

        if (nowTime - tiReadyStart >= 3000)
        {
            // Display the number of people who are currently connected.
            NEX_DEMO_LOG("waiting ... NumStations: %d/%d\n", numStations, MATCHMAKE_MAX_PARTICIPANTS);

            // Display connection information.
            OutputConnectionInfo();

            tiReadyStart = nowTime;
        }

        // Dispatch messages
        nn::nex::Scheduler::GetInstance()->Dispatch(DISPATCH_TIMEOUT);

        // NetZ fatal error check.
        nn::nex::qResult netZError = pNetZ->GetFatalError();
        if ( QFAILED(netZError.GetReturnCode()) )
        {
            // An error from which restoration is not possible occurs
            nex_demo_util::ReportQResultError(netZError, "NetZ fatal error happend!\n");
            break;
        }

        // Check how many people have connected.
        if (numStations == MATCHMAKE_MAX_PARTICIPANTS)
        {
            // The prescribed number of people have connected and preparations for entering the main loop are complete.
            bReadyToRun = true;
            break;
        }

        nn::nex::Platform::Sleep(DISPATCH_PERIOD);
    }

    return bReadyToRun;
}

//-----------------------------------------------------------------------------
// Main loop after matchmaking
//-----------------------------------------------------------------------------
void RunGameLoop(const nn::nex::NetZ* pNetZ)
{
    NEX_DEMO_LOG("Start NetZ session.\n");

    nn::nex::Time tiSessionStart = nn::nex::Time::GetTime();
    nn::nex::Time tiLogStart     = nn::nex::Time::GetTime();

    // Send and receive data
    while (1)
    {
        nn::nex::Time nowTime = nn::nex::Time::GetTime();

        if (nowTime - tiSessionStart >= MATCHMAKING_TIME)
        {
            // Exit because we have sent and received data for the specified amount of time
            break;
        }

        // Dispatch messages
        nn::nex::Scheduler::GetInstance()->Dispatch(DISPATCH_TIMEOUT);

        // NetZ fatal error check.
        nn::nex::qResult netZError = pNetZ->GetFatalError();
        if ( QFAILED(netZError.GetReturnCode()) )
        {
            // An error from which restoration is not possible occurs
            nex_demo_util::ReportQResultError(netZError, "NetZ fatal error happend!\n");
            break;
        }

        if (nowTime - tiLogStart >= 3 * 1000)
        {
            // Display connected stations.
            tiLogStart = nn::nex::Time::GetTime();

            NEX_DEMO_LOG("Stations currently in Session\n");
            NEX_DEMO_LOG("-----------------------------\n");
            OutputConnectionInfo();
        }

        nn::nex::Platform::Sleep(DISPATCH_PERIOD);
    }

    NEX_DEMO_LOG("Stop NetZ session.\n");
}

//-----------------------------------------------------------------------------
// Create a NetZ session and open MatchmakeSession
//-----------------------------------------------------------------------------
qBool OwnerFunction(nn::nex::MatchmakeSession* pMatchmakeSession,
                    nn::nex::MatchmakeExtensionClient *pMatchMakingClient)
{
    nn::nex::ProtocolCallContext oContext;

    /////////////////////////////////////////////////////////////////
    // Set the conditions for the matchmaking session
    {
        pMatchmakeSession->Reset();

        // Game mode.
        pMatchmakeSession->SetGameMode(MATCHMAKE_GAME_MODE);

        // The minimum and maximum number of participants.
        pMatchmakeSession->SetMinParticipants(1);
        pMatchmakeSession->SetMaxParticipants(MATCHMAKE_MAX_PARTICIPANTS);

        // Matchmaking system type.
#ifdef MATCHMAKE_FRIEND_MODE
        pMatchmakeSession->SetMatchmakeSystemType(nn::nex::MATCHMAKE_SYSTEM_TYPE_FRIEND);
#else
        pMatchmakeSession->SetMatchmakeSystemType(nn::nex::MATCHMAKE_SYSTEM_TYPE_ANYBODY);
#endif
    }

    /////////////////////////////////////////////////////////////////
    // Matchmaking session registration
    {
        nn::nex::MatchmakeSession* pTempMatchmakeSession = new nn::nex::MatchmakeSession(*pMatchmakeSession);
        nn::nex::GatheringHolder gatheringHolder = pTempMatchmakeSession;
        nn::nex::GatheringID gatheringId = INVALID_GATHERINGID;
        nn::nex::qVector<nn::nex::qByte> sessionKey;

        if(pMatchMakingClient->CreateMatchmakeSession(&oContext, gatheringHolder, &gatheringId, &sessionKey))
        {
            oContext.Wait();
            if(oContext.GetState() == nn::nex::CallContext::CallSuccess)
            {
                NEX_DEMO_LOG("create matchmakesession ID:%u\n", gatheringId);
                pMatchmakeSession->SetID(gatheringId);
                nn::nex::Network::GetInstance()->SetP2PDataPacketSessionSignatureKey(sessionKey);
            }
            else
            {
                nex_demo_util::ReportPCallError(oContext, "Failed to create matchmakesession\n");
                return false;
            }
        }
        else
        {
            NEX_DEMO_LOG("Failed to call CreateMatchmakeSession\n");
            return false;
        }
    }

    /////////////////////////////////////////////////////////////////
    // Open a session
    if (nn::nex::Session::CreateSession(NULL))
    {
        NEX_DEMO_LOG("Net-Z session created successfly\n");
    }
    else
    {
        NEX_DEMO_LOG("Failed to create a Net-Z session\n");
        return false;
    }

    /////////////////////////////////////////////////////////////////
    // Make the matchmaking session open to participants
    if (pMatchMakingClient->OpenParticipation(&oContext, pMatchmakeSession->GetID()))
    {
        oContext.Wait();
        if (oContext.GetState() != nn::nex::CallContext::CallSuccess)
        {
            nex_demo_util::ReportPCallError(oContext, "Failed to open participation.\n");
            return false;
        }
    }
    else
    {
        NEX_DEMO_LOG("Failed to open participation.\n");
        return false;
    }

    return true;
}

//-----------------------------------------------------------------------------
// Join MatchmakeSession
//-----------------------------------------------------------------------------
qBool ParticipantFunction(nn::nex::MatchmakeSession* pMatchmakeSession,
                          nn::nex::MatchmakeExtensionClient *pMatchMakingClient)
{
    nn::nex::ProtocolCallContext oContext;
    pMatchmakeSession->Reset();

    /////////////////////////////////////////////////////////////////
    // Get matchmaking session information from friend presence (it should exist)
    nn::friends::FriendKey friendKey;
    nn::friends::IsFromFriendList(&friendKey);

    // Get presence information
    nn::friends::FriendPresence presence;
    nn::friends::GetFriendPresence(&presence, &friendKey);

    // Just in case confirm once more
    if(!presence.IsJoinable(MATCHMAKE_GAME_MODE_MASK))
    {
        NEX_DEMO_LOG("may be not joinable session\n");
        return false;
    }

    /////////////////////////////////////////////////////////////////
    // Get matchmaking session from NGS (it should exist)
    nn::nex::GatheringHolder gatheringHolder;
    if(pMatchMakingClient->FindByID(&oContext, presence.gameMode.joinGroupId, &gatheringHolder))
    {
        oContext.Wait();
        if (oContext.GetState() == nn::nex::CallContext::CallSuccess)
        {
            *pMatchmakeSession = *(static_cast<nn::nex::MatchmakeSession*>(gatheringHolder.Get()));
        }
        else
        {
            nex_demo_util::ReportPCallError(oContext, "Failed to obtain a matchmakesession\n");
            return false;
        }
    }
    else
    {
        NEX_DEMO_LOG("Failed to call FindByID()\n");
        return false;
    }

    /////////////////////////////////////////////////////////////////
    // Join matchmaking session
    nn::nex::qVector<nn::nex::qByte> sessionKey;
    if(pMatchMakingClient->JoinMatchmakeSession(&oContext, pMatchmakeSession->GetID(), &sessionKey))
    {
        oContext.Wait();
        if(oContext.GetState() != nn::nex::CallContext::CallSuccess)
        {
            nex_demo_util::ReportPCallError(oContext, "Failed to join matchmakesession\n");
            return false;
        }

        nn::nex::Network::GetInstance()->SetP2PDataPacketSessionSignatureKey(sessionKey);
    }
    else
    {
        pMatchmakeSession->Reset();

        NEX_DEMO_LOG("Failed to call JoinMatchmakeSession()\n");
        return false;
    }

    /////////////////////////////////////////////////////////////////
    // Get URL of P2P session host
    nn::nex::qList<nn::nex::StationURL> lstURLs;
    if (pMatchMakingClient->GetSessionURLs(&oContext, pMatchmakeSession->GetID(), &lstURLs))
    {
        oContext.Wait();
        if (oContext.GetState() != nn::nex::CallContext::CallSuccess)
        {
            // The matchmaking session that was just joined may have been deleted.
            if(oContext.GetOutcome() == QERROR(RendezVous, SessionVoid))
            {
                NEX_DEMO_LOG("matchmake session is already unregistered.\n");
            }
            nex_demo_util::ReportPCallError(oContext, "Failed to obtain a list of URLs to connect to the game host\n");
            return false;
        }
    }
    else
    {
        NEX_DEMO_LOG("Failed to call GetSessionURLs()\n");
        return false;
    }

    /////////////////////////////////////////////////////////////////
    // Attempt to join PSP session
    if (nn::nex::Session::JoinSession(&oContext, lstURLs))
    {
        oContext.Wait();
        if(oContext.GetState() != nn::nex::CallContext::CallSuccess)
        {
            nex_demo_util::ReportPCallError(oContext, "Failed to  the Net-Z session\n");
            return false;
        }
    }
    else
    {
        NEX_DEMO_LOG("Failed to JoinSession\n");
        return false;
    }

    return true;
}

//-----------------------------------------------------------------------------
// Display Connection Information
//-----------------------------------------------------------------------------
void OutputConnectionInfo()
{
    nn::nex::Station::SelectionIterator itStations;
    itStations.GotoStart();
    while (!itStations.EndReached())
    {
        NEX_DEMO_LOG("Station %8x: URL = %ls, RTT = %d\n",
                     itStations->GetHandle().GetValue(),
                     itStations->GetStationURL(0),
                     itStations->GetRTTAverage());

        ++itStations;
    }
    NEX_DEMO_LOG("\n");
}

//-----------------------------------------------------------------------------
// Update friend presence based on matchmaking session information
//-----------------------------------------------------------------------------
void UpdateGameMode(const nn::nex::MatchmakeSession& session, bool isJoinable, qBool isOwner)
{
    nn::friends::GameMode gameMode;
    char16* gameDescription = L"Join From FriendList";

    gameMode.joinAvailabilityFlag = isJoinable ? nn::friends::JOIN_AVAILABILITY_JOINABLE : nn::friends::JOIN_AVAILABILITY_NOT_JOINABLE;
    gameMode.joinGameMode = session.GetGameMode();
    gameMode.joinGroupId = session.GetID();
    // If the local device has created the matchmaking session, the GetOwnerPID function cannot get the PrincipalId, so set the owner's PrincipalId.
    if(isOwner)
    {
        gameMode.ownerPrincipalId = nn::friends::GetMyPrincipalId();
    }
    else
    {
        gameMode.ownerPrincipalId = session.GetOwnerPID();
    }
    switch (session.GetMatchmakeSystemType())
    {
    case nn::nex::MATCHMAKE_SYSTEM_TYPE_ANYBODY:
        gameMode.matchmakeSystemType = nn::friends::MATCHMAKE_SYSTEM_TYPE_ANYBODY;
        break;
    case nn::nex::MATCHMAKE_SYSTEM_TYPE_FRIEND:
        gameMode.matchmakeSystemType = nn::friends::MATCHMAKE_SYSTEM_TYPE_FRIEND;
        break;
    default:
        // Handle community matching as ANYBODY
        gameMode.matchmakeSystemType = nn::friends::MATCHMAKE_SYSTEM_TYPE_ANYBODY;
        break;
    }

    nn::friends::UpdateGameMode(gameMode, gameDescription);
}


//-----------------------------------------------------------------------------
// Render the log
//-----------------------------------------------------------------------------
void RenderLog()
{
    // Renderer
    demo::RenderSystemDrawing* pRenderSystem = nexdemo::Graphics::GetRenderSystem();

    // Clear the screen
    pRenderSystem->SetRenderTarget(NN_GX_DISPLAY0);
    pRenderSystem->Clear();

    // Render the background
    pRenderSystem->SetColor(NEXDEMO_RGB(0, 0, 0));
    pRenderSystem->FillRectangle(0, 0, NN_GX_DISPLAY0_HEIGHT, NN_GX_DISPLAY0_WIDTH);

    // Render
    pRenderSystem->SetColor(NEXDEMO_RGB(255, 255, 255));
    s_console.Render(pRenderSystem);

    pRenderSystem->SwapBuffers();
    pRenderSystem->WaitVsync(NN_GX_DISPLAY_BOTH);
}

namespace nn
{
namespace dbg
{
namespace detail
{
// Redirect the debugger output
void PutString(const char* text, s32 length)
{
    // Mutex so there will be no PutString during rendering
    nn::os::CriticalSection::ScopedLock s_scopedLock(s_criticalSection);

    (void)length;

    if (s_console.IsInitialized())
    {
        // Output to console
        s_console.PutString(text);
        // Re-render
        RenderLog();
    }
}
}
}
}
