﻿/*---------------------------------------------------------------------------*
  (C) Copyright 1998-2014, Technologies Quazal Inc. All rights reserved.
  Quazal Net-Z, Quazal Rendez-Vous and Quazal are trademarks of Quazal.

  Copyright (C)2011-2014 Nintendo  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo and/or its licensors and are
  protected by national and international copyright laws.
  They may not be disclosed to third parties or copied or duplicated
  in any form, in whole or in part, without the prior written consent
  of Nintendo.
  The content herein is highly confidential and should be handled
  accordingly.
 *---------------------------------------------------------------------------*/
/*
 * NEX Sample
 *
 * Messaging
 *
 * Description :
 *
 *  This sample performs messaging processing.
 *
 *  Message transmissions can only be used when each sender and receiver is logged in to the same game server.
 *  With this sample, after logging in, messages are sent to clients participating in the same matchmaking session.
 * 
 *
 *  After login, pressing the A Button sends a message to participants in the matchmaking session.
 Prepare two types of messages, text and binary.
 *  If there are other friends who are running this sample, the received message is displayed in the log.
 *  Pressing the B Button quits the sample.
 *
 * This sample runs in thread unsafe transport buffer thread mode.
 *
 *
 * Note:
 *  The messaging feature should be used only when data exchanges are required before the transition to P2P.
 *
 */

#include <nex.h>
#include "../common/NexUtil.h"
#include "../common/DemoPad.h"

//-----------------------------------------------------------------------------
// Constant definitions
//-----------------------------------------------------------------------------
static const qUnsignedInt32 MESSAGING_MAX_PARTICIPANTS = 5;
static const qUnsignedInt32 MESSAGING_GAME_MODE = 1002;

//-----------------------------------------------------------------------------
// Function definitions
//-----------------------------------------------------------------------------
static void SendInstantMessage(nn::nex::MessagingClient *pMessagingClient,
                               nn::nex::PrincipalID target,
                               const nn::nex::String& message, const nn::nex::qVector<qByte>& binary);


//-----------------------------------------------------------------------------
// The event handler that processes the send message
//-----------------------------------------------------------------------------
class MyMessagingNotificationHandler: public nn::nex::MessagingNotificationHandler
{
public:
    void DeliverInstantMessage(const nn::nex::DataHolder& hMsg)
    {
        // Distinguishes between TextMessage and BinaryMessage.

        // If TextMessage::DynamicCast is successful, the receive message is TextMessage.
        nn::nex::TextMessage* pTextMsg = nn::nex::TextMessage::DynamicCast(hMsg.Get());
        if (pTextMsg)
        {
            NEX_DEMO_LOG("*** Recieved message from:[%u] subject[%ls] text body:[%ls]\n",
                         pTextMsg->GetSenderID(),
                         pTextMsg->GetSubject().CStr(),
                         pTextMsg->GetTextBody().CStr());
            return;
        }

        // If TextMessage::DynamicCast is successful, the receive message is TextMessage.
        nn::nex::BinaryMessage* pBinaryMsg = nn::nex::BinaryMessage::DynamicCast(hMsg.Get());
        if (pBinaryMsg)
        {
            NEX_DEMO_LOG("*** Recieved message from:[%u] subject[%ls] binary size:[%u]\n",
                         pBinaryMsg->GetSenderID(),
                         pBinaryMsg->GetSubject().CStr(),
                         pBinaryMsg->GetBufferSize());
            return;
        }
    }
};

//-----------------------------------------------------------------------------
// Join an existing matchmaking session or create a new one
//-----------------------------------------------------------------------------
qBool AutoMatch(nn::nex::MatchmakeExtensionClient *pMatchMakingClient,
                nn::nex::GatheringID* pGatheringId,
                nn::nex::PrincipalID* pOwnerPrincipalId)
{
    *pGatheringId = INVALID_GATHERINGID;
    *pOwnerPrincipalId = INVALID_PRINCIPALID;

    // Create search criteria
    nn::nex::MatchmakeSession* matchmakeSession = new nn::nex::MatchmakeSession();

    // Constant waiting state
    matchmakeSession->SetOpenParticipation(true);

    // The minimum and maximum number of participants
    matchmakeSession->SetMaxParticipants(MESSAGING_MAX_PARTICIPANTS);

    // Game mode
    matchmakeSession->SetGameMode(MESSAGING_GAME_MODE);

    // System type
    matchmakeSession->SetMatchmakeSystemType(nn::nex::MATCHMAKE_SYSTEM_TYPE_ANYBODY);

    nn::nex::GatheringHolder oInGatheringHolder = matchmakeSession;
    nn::nex::GatheringHolder oOutGatheringHolder;   // The gathering that we have joined

    // Search for or create a matchmaking session
    nn::nex::ProtocolCallContext oContext;
    if (pMatchMakingClient->AutoMatchmake(&oContext, oInGatheringHolder, &oOutGatheringHolder))
    {
        oContext.Wait();
        if (oContext.GetState() != nn::nex::CallContext::CallSuccess)
        {
            nex_demo_util::ReportPCallError(oContext, "Failed to AutoMatchmake\n");
            return false;
        }
    }
    else
    {
        NEX_DEMO_LOG("Failed to AutoMatchmake\n");
        return false;
    }

    *pGatheringId = oOutGatheringHolder->GetID();
    *pOwnerPrincipalId = oOutGatheringHolder->GetOwnerPID();

    return true;
}

//-----------------------------------------------------------------------------
// Perform matchmaking session finalization
// When the local host is the owner, delete the matchmaking session, when a participant, leave.
//-----------------------------------------------------------------------------
qBool EndMatch(nn::nex::MatchmakeExtensionClient *pMatchMakingClient,
               nn::nex::GatheringID gatheringId,
               nn::nex::PrincipalID ownerPrincipalId)
{
    nn::nex::ProtocolCallContext     oContext;

    // When the local host is the owner, UnregisterGathering, otherwise, EndParticipation
    if(pMatchMakingClient->GetCredentials()->GetPrincipalID() == ownerPrincipalId)
    {
        if(pMatchMakingClient->UnregisterGathering(&oContext, gatheringId))
        {
            oContext.Wait();
            if (oContext.GetState() != nn::nex::CallContext::CallSuccess)
            {
                nex_demo_util::ReportPCallError(oContext, "Failed to UnregisterGathering\n");
                return false;
            }
            else
            {
                return true;
            }
        }
    }
    else
    {
        if(pMatchMakingClient->EndParticipation(&oContext, gatheringId))
        {
            oContext.Wait();
            if (oContext.GetState() != nn::nex::CallContext::CallSuccess)
            {
                if(oContext.GetOutcome() == QERROR(RendezVous, SessionVoid))
                {
                    NEX_DEMO_LOG("Already unregisterd. Don't care.");
                    return true;
                }

                nex_demo_util::ReportPCallError(oContext, "Failed to EndParticipation\n");
                return false;
            }
            else
            {
                return true;
            }
        }
    }

    NEX_DEMO_LOG("Failed to EndMatch\n");
    return false;
}

//-----------------------------------------------------------------------------
// Send a message to the specified GatheringID of the participants (including the local host)
//-----------------------------------------------------------------------------
void SendInstantMessage(nn::nex::MessagingClient *pMessagingClient,
                        nn::nex::GatheringID target, const nn::nex::String& message, const nn::nex::qVector<qByte>& binary)
{
    nn::nex::TextMessage *pTextMessage = new nn::nex::TextMessage();
    nn::nex::UserMessageDataHolder oDataHolder = pTextMessage;

    // Set a message
    pTextMessage->SetSubject(_T("This is a TextMessage"));
    pTextMessage->SetTextBody(message);

    // Send message
    nn::nex::ProtocolCallContext oContext;
    NEX_DEMO_LOG("Send message to session %d participants ... \n", target);

    // Even when the client or matchmaking session that is the send target of SendInstant/SendChat does not exist
    // Returns that asynchronous processing succeeded.
    if (pMessagingClient->SendChat(&oContext, target, oDataHolder))
    {
        oContext.Wait();
        if (oContext.GetState() != nn::nex::CallContext::CallSuccess)
        {
            nex_demo_util::ReportPCallError(oContext, "MessagingClient::SendChat TextMessage failed!!\n");
        }
        else
        {
            NEX_DEMO_LOG("MessagingClient::SendChat TextMessage successfly.\n");
        }
    }
    else
    {
        NEX_DEMO_LOG("MessagingClient::SendChat TextMessage failed.\n");
    }


    nn::nex::BinaryMessage *pBinaryMessage = new nn::nex::BinaryMessage();
    oDataHolder = pBinaryMessage;

    // Set a message
    pBinaryMessage->SetSubject(_T("This is a BinaryMessage"));
    pBinaryMessage->SetBuffer(static_cast<const void*>(&binary[0]), binary.size());

    // Send message
    NEX_DEMO_LOG("Send message to session %d participants ... \n", target);

    // Even when the client or matchmaking session that is the send target of SendInstant/SendChat does not exist
    // Returns that asynchronous processing succeeded.
    if (pMessagingClient->SendChat(&oContext, target, oDataHolder))
    {
        oContext.Wait();
        if (oContext.GetState() != nn::nex::CallContext::CallSuccess)
        {
            nex_demo_util::ReportPCallError(oContext, "MessagingClient::SendChat BinaryMessage failed!!\n");
        }
        else
        {
            NEX_DEMO_LOG("MessagingClient::SendChat BinaryMessage successfly.\n");
        }
    }
    else
    {
        NEX_DEMO_LOG("MessagingClient::SendChat BinaryMessage failed.\n");
    }
}

//-----------------------------------------------------------------------------
// Variable Definitions
//-----------------------------------------------------------------------------
static MyMessagingNotificationHandler s_MyMessagingNotificationHandler;


qInt NEXSampleMain(qInt intArg, void* ptrArg)
{
    (void)intArg;
    (void)ptrArg;

    // Connects to the Internet.
    if (!nex_demo_util::NetInit())
    {
        NEX_DEMO_LOG("net init failed...\n");
        NEX_DEMO_LOG("Leaving application...\n");

        return 0;
    }

    // Prepare to use utilities
    nex_demo_util::Initialize();

    // Initialization
    nn::nex::GlobalVariables::AcquireInstance();

    // Set to thread unsafe transport buffer thread mode
    nn::nex::Core::SetThreadMode(nn::nex::Core::ThreadModeUnsafeTransportBuffer);

    nn::nex::EventLog::GetInstance()->SetLogLevel(nn::nex::EventLog::Debug);
    nn::nex::TraceLog::GetInstance()->SetFlag(TRACE_ALWAYS);

    nn::nex::Credentials        *pCredentials        = NULL;
    nn::nex::NgsFacade          *pNgsFacade          = new nn::nex::NgsFacade();
    nn::nex::MessagingClient    *pMessageClient      = new nn::nex::MessagingClient();
    nn::nex::MatchmakeExtensionClient *pMatchMakingClient  = new nn::nex::MatchmakeExtensionClient();

    // Logs into the game server.
    pCredentials = nex_demo_util::GameServerLogin(pNgsFacade, NEXDEMO_GAME_SERVER_ID, NEXDEMO_ACCESS_KEY, 60000);

    // Bind service(s)
    if (pCredentials)
    {
        if (!pMatchMakingClient->Bind(pCredentials))
        {
            NEX_DEMO_LOG("Cannot bind to Matchmaking Service\n");
            goto END;
        }
        else
        {
            NEX_DEMO_LOG("Bind to Matchmaking service\n");
        }

        if (!pMessageClient->Bind(pCredentials))
        {
            NEX_DEMO_LOG("Cannot bind to Messaging Service\n");
            goto END;
        }
        else
        {
            NEX_DEMO_LOG("Bind to Messaging service\n");
        }
    }
    else
    {
        goto END;
    }

    // Register an event handler to process message notifications
    pMessageClient->RegisterMessagingNotificationHandler(&s_MyMessagingNotificationHandler);

    // Join a matchmaking session using auto-matchmaking.
    nn::nex::GatheringID gatheringId;
    nn::nex::PrincipalID ownerPrincipalId;

    if(!AutoMatch(pMatchMakingClient, &gatheringId, &ownerPrincipalId))
    {
        NEX_DEMO_LOG("Fail to AutoMatch\n");
        nex_demo_util::GameServerLogout(pNgsFacade);
        goto END;
    }

    // HID initialization
    nex_demo_util::DemoPad::Initialize();
    nex_demo_util::PadStatus oPadStatus;

    // Send a message with a press of the button.
    {
        NEX_DEMO_LOG("\nPress A to Send Message.\n");
        NEX_DEMO_LOG("Press B to Quit Sample.\n");

        while (1)
        {
            nex_demo_util::DemoPad::Update(&oPadStatus);

            if (oPadStatus.trigger & DEMO_PAD_BUTTON_A)
            {
                nn::nex::String message;
                message.Format(_T("Hello! I am %d"), pCredentials->GetPrincipalID());

                nn::nex::qVector<qByte> binary(123);

                SendInstantMessage(pMessageClient, gatheringId, message, binary);
            }
            else if (oPadStatus.trigger & DEMO_PAD_BUTTON_B)
            {
                break;
            }

            nn::nex::Platform::Sleep(30);
            nn::nex::Scheduler::GetInstance()->DispatchAll();
        }
    }

    if(!EndMatch(pMatchMakingClient, gatheringId, ownerPrincipalId))
    {
        NEX_DEMO_LOG("Fail to EndMatch\n");
        nex_demo_util::GameServerLogout(pNgsFacade);
        goto END;
    }

END:
    // Finalize HID
    nex_demo_util::DemoPad::Finalize();

    // Delete each object
    pMessageClient->UnregisterMessagingNotificationHandler();
    pMessageClient->Unbind();
    delete pMessageClient;

    pMatchMakingClient->Unbind();
    delete pMatchMakingClient;

    // Log out
    nex_demo_util::GameServerLogout(pNgsFacade);

    // Delete the login client
    delete pNgsFacade;

    // End processing
    nn::nex::GlobalVariables::ReleaseInstance();

    // Shut down utilities
    nex_demo_util::Finalize();

    // Disconnects from the Internet.
    nex_demo_util::NetTerm();

    return 0;
}

//-----------------------------------------------------------------------------
// Main Function
//-----------------------------------------------------------------------------
qInt GenericMain(qInt argc, qChar** argv)
{
    NEX_DEMO_LOG("*******************************\n");
    NEX_DEMO_LOG("       Messaging sample\n");
    NEX_DEMO_LOG("*******************************\n\n");

#if defined(CAFE_TARGET)

#else
    NEXSampleMain(argc, argv);
#endif

    NEX_DEMO_LOG("Leaving application...\n");

    return 0;
}

NEX_MAIN_WRAPPER(GenericMain)
