﻿/*---------------------------------------------------------------------------*
  (C) Copyright 1998-2014, Technologies Quazal Inc. All rights reserved.
  Quazal Net-Z, Quazal Rendez-Vous and Quazal are trademarks of Quazal.

  Copyright (C)2011-2014 Nintendo  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo and/or its licensors and are
  protected by national and international copyright laws.
  They may not be disclosed to third parties or copied or duplicated
  in any form, in whole or in part, without the prior written consent
  of Nintendo.
  The content herein is highly confidential and should be handled
  accordingly.
 *---------------------------------------------------------------------------*/

/*
 * NEX Sample
 *
 * PersistentGathering
 *
 * Description :
 *
 *  This is a sample of how to use the persistent gathering.
 *  It performs the basic processing, such as the searching, participation and creation of a persistent gathering and the execution of a persistent gathering matching.
 *
 *  This sample comprises the following processes.
 *
 *  1. Logs in to the game server
 *  2. Searches for the persistent gathering you participate in and the one an user with specified PID participates in. 
 *  3. Selects one persistent gathering from the search results and participates in it if not already participating in it.
 *  4. If no persistent gathering is found in step 2, you create one.
 *  5. Perform matching of the persistent gathering that was selected in step 3 or 4 (no actual P2P communications).
 *  6. Logs out.
 *
 *
 *
 * This sample runs in thread unsafe transport buffer thread mode.
 *
 */

#include <nex.h>
#include "../common/NexUtil.h"

//-----------------------------------------------------------------------------
// Constant definitions
//-----------------------------------------------------------------------------
// When matching persistent gathering with a particular user, set PID here.
#define TARGET_PID 1001

//-----------------------------------------------------------------------------
// Function definitions
//-----------------------------------------------------------------------------

// Sample of the persistent gathering features.
void RunPersistentGatheringFunction(nn::nex::MatchmakeExtensionClient* pMatchmakeClient);

// Searches for persistent gatherings by participants.
qBool FindPersistentGatheringByParticipant(nn::nex::MatchmakeExtensionClient* pMatchmakeClient,
                                 nn::nex::PrincipalID principalId,
                                 nn::nex::qList<nn::nex::PersistentGathering>* pPersistentGatheringList);

// Participates in the persistent gathering.
qBool JoinPersistentGathering(nn::nex::MatchmakeExtensionClient* pMatchmakeClient, nn::nex::GatheringID persistentGatheringId);

// Participates in the persistent gathering.
qBool CreatePersistentGathering(nn::nex::MatchmakeExtensionClient* pMatchmakeClient, nn::nex::GatheringID* pPersistentGatheringGatheringId);

// Displays a list of persistent gatherings.
void PrintPersistentGatheringList(const nn::nex::qList<nn::nex::PersistentGathering>& persistentGatheringList);

// Runs the persistent gathering matching of the persistent gathering specified by an argument.
void RunPersistentGatheringMatch(nn::nex::MatchmakeExtensionClient* pMatchmakeClient, nn::nex::GatheringID persistentGatheringId);

//-----------------------------------------------------------------------------
// Event handler to process notifications from the NGS
//-----------------------------------------------------------------------------
class MyNotificationEventHandler : public nn::nex::NotificationEventHandler
{
public:
    void ProcessNotificationEvent(const nn::nex::NotificationEvent &oEvent)
    {
        nn::nex::StringStream ss;
        ss << oEvent;
        NEX_DEMO_LOG("Notification %ls\n\n",ss.CStr());

        // Per-message processing
        switch (oEvent.GetType())
        {
        case nn::nex::NotificationEvents::ParticipationEvent:
            switch (oEvent.GetSubType())
            {
            case nn::nex::ParticipationEvents::Participate:
                NEX_DEMO_LOG("PARTICIPATION_EVENT PARTICIPATE: [PID %u] join [session %u]\n", oEvent.GetParam2(), oEvent.GetParam1());
                break;

            case nn::nex::ParticipationEvents::EndParticipation:
                NEX_DEMO_LOG("PARTICIPATION_EVENT END: [PID %u] end [session %u]\n", oEvent.GetParam2(), oEvent.GetParam1());
                break;

            case nn::nex::ParticipationEvents::Disconnect:
                NEX_DEMO_LOG("PARTICIPATION_EVENT DISCONNECT: [PID %u] disconnect [session %u]\n", oEvent.GetParam2(),
                             oEvent.GetParam1());
                break;

            default:
                NEX_DEMO_LOG("PARTICIPATION_EVENT: unsupported participation event\n");
                break;
            }
            break;

        case nn::nex::NotificationEvents::OwnershipChangeEvent:
            NEX_DEMO_LOG("NOTIFICATION_EVENT OWNERSHIPCHANGE: [session %u] changed owner to [PID %u]\n",
                         oEvent.GetParam1(), oEvent.GetParam2());
            break;

        case nn::nex::NotificationEvents::GatheringUnregistered:
            NEX_DEMO_LOG("NOTIFICATION_EVENT GATHERINGUNREGISTERED: [session %u] unregistered\n", oEvent.GetParam1());

            break;

        default:
            NEX_DEMO_LOG("NOTIFICATION_EVENT: unsupported notification event\n");
            break;
        }
    }
};


qInt NEXSampleMain(qInt intArg, void* ptrArg)
{
    (void)intArg;
    (void)ptrArg;

    // Connects to the Internet.
    if (!nex_demo_util::NetInit())
    {
        NEX_DEMO_LOG("net init failed...\n");
        NEX_DEMO_LOG("Leaving application...\n");

        return 0;
    }

    // Prepare to use utilities
    nex_demo_util::Initialize();

    // Initialization
    nn::nex::GlobalVariables::AcquireInstance();

    // Set to thread unsafe transport buffer thread mode
    nn::nex::Core::SetThreadMode(nn::nex::Core::ThreadModeUnsafeTransportBuffer);

    nn::nex::EventLog::GetInstance()->SetLogLevel(nn::nex::EventLog::Debug);
    nn::nex::TraceLog::GetInstance()->SetFlag(TRACE_ALWAYS);

    // Specify the access key for your own title to the SetPersistentGatheringCodeKey function.
    nn::nex::MatchmakeExtensionClient::SetPersistentGatheringCodeKey(NEXDEMO_ACCESS_KEY);

    // The persistent gathering code can be converted without executing the MatchmakeExtensionClient::Bind function.
    {
        nn::nex::GatheringID gatheringId = 12345;
        qUnsignedInt64 persistentGatheringCode = nn::nex::MatchmakeExtensionClient::GatheringIdToPersistentGatheringCode(12345);

        NEX_DEMO_LOG("GateringID:%u PersistentGatheringCode:%llu\n", gatheringId, persistentGatheringCode);

        SYSTEMCHECK(nn::nex::MatchmakeExtensionClient::IsValidPersistentGatheringCode(persistentGatheringCode));

        SYSTEMCHECK(gatheringId == nn::nex::MatchmakeExtensionClient::PersistentGatheringCodeToGatheringId(persistentGatheringCode));

        // Workaround the warnings in the release build
        (void)gatheringId;
        (void)persistentGatheringCode;
    }

    nn::nex::Credentials              *pCredentials        = NULL;
    nn::nex::NgsFacade                *pNgsFacade          = new nn::nex::NgsFacade();
    nn::nex::MatchmakeExtensionClient *pMatchmakeClient    = new nn::nex::MatchmakeExtensionClient();
    MyNotificationEventHandler myNotificationEventHandler;

    // Logs into the game server.
    pCredentials = nex_demo_util::GameServerLogin(pNgsFacade, NEXDEMO_GAME_SERVER_ID, NEXDEMO_ACCESS_KEY, 60000);

    // Bind service(s)
    if (pCredentials)
    {
        if (!pMatchmakeClient->Bind(pCredentials))
        {
            NEX_DEMO_LOG("Cannot bind to Matchmaking Service\n");
            nex_demo_util::GameServerLogout(pNgsFacade);
            goto END;
        }
        else
        {
            NEX_DEMO_LOG("Bind to Matchmaking service\n");
        }
    }
    else
    {
        goto END;
    }

    // Register an event handler to process notifications from the NGS
    pNgsFacade->RegisterNotificationEventHandler(&myNotificationEventHandler);

    // Main loop
    RunPersistentGatheringFunction(pMatchmakeClient);

    // Delete the event handler
    pNgsFacade->UnregisterNotificationEventHandler(&myNotificationEventHandler);

END:
    // Delete each object
    pMatchmakeClient->Unbind();
    delete pMatchmakeClient;

    // Log out
    nex_demo_util::GameServerLogout(pNgsFacade);

    // Delete the login client
    delete pNgsFacade;

    // Calls the ClearPersistentGatheringCodeKey function to clear it when it is no longer necessary, for instance, when the library ends.
    nn::nex::MatchmakeExtensionClient::ClearPersistentGatheringCodeKey();

    nn::nex::GlobalVariables::ReleaseInstance();

    // Shut down utilities
    nex_demo_util::Finalize();

    // Disconnects from the Internet.
    nex_demo_util::NetTerm();

    return 0;
}

//-----------------------------------------------------------------------------
// Main Function
//-----------------------------------------------------------------------------
qInt GenericMain(qInt argc, qChar** argv)
{
    NEX_DEMO_LOG("*******************************\n");
    NEX_DEMO_LOG("     PersistentGathering sample\n");
    NEX_DEMO_LOG("*******************************\n\n");

#if defined(CAFE_TARGET)

#else
    NEXSampleMain(argc, argv);
#endif

    NEX_DEMO_LOG("Leaving application...\n");

    return 0;
}

NEX_MAIN_WRAPPER(GenericMain)

void RunPersistentGatheringFunction(nn::nex::MatchmakeExtensionClient* pMatchmakeClient)
{
    nn::nex::ProtocolCallContext oContext;
    nn::nex::qList<nn::nex::PersistentGathering> myPersistentGatheringList;
    nn::nex::qList<nn::nex::PersistentGathering> targetPersistentGatheringList;

    nn::nex::GatheringID persistentGatheringId = INVALID_GATHERINGID;
    qUnsignedInt32 maxMatchmakeSessionCount = 0;

    // Selects the persistent gathering for which the persistent gathering matching most frequently runs from among the persistent gatherings you participate in and the ones the users you search for participate in.
    // 

    // Gets the persistent gatherings you participate in.
    if(!FindPersistentGatheringByParticipant(pMatchmakeClient,
                                   pMatchmakeClient->GetCredentials()->GetPrincipalID(),
                                   &myPersistentGatheringList))
    {
        return;
    }

    if(!myPersistentGatheringList.empty())
    {
        persistentGatheringId = myPersistentGatheringList.front().GetID();
        maxMatchmakeSessionCount = myPersistentGatheringList.front().GetMatchmakeSessionCount();

        nn::nex::qList<nn::nex::PersistentGathering>::const_iterator it = myPersistentGatheringList.begin();
        while(it != myPersistentGatheringList.end())
        {
            if(it->GetMatchmakeSessionCount() > maxMatchmakeSessionCount)
            {
                maxMatchmakeSessionCount = it->GetMatchmakeSessionCount();
                persistentGatheringId = it->GetID();
            }
            ++it;
        }
    }

    // Searches for the persistent gatherings that the specified users participate in.
    if(!FindPersistentGatheringByParticipant(pMatchmakeClient,
                                   TARGET_PID,
                                   &targetPersistentGatheringList))
    {
        return;
    }

    if(!targetPersistentGatheringList.empty())
    {
        // If you are not participating in a persistent gathering:
        if(persistentGatheringId == INVALID_GATHERINGID)
        {
            persistentGatheringId = targetPersistentGatheringList.front().GetID();
            maxMatchmakeSessionCount = targetPersistentGatheringList.front().GetMatchmakeSessionCount();
        }

        nn::nex::qList<nn::nex::PersistentGathering>::const_iterator it = targetPersistentGatheringList.begin();
        while(it != targetPersistentGatheringList.end())
        {
            if(it->GetMatchmakeSessionCount() > maxMatchmakeSessionCount)
            {
                maxMatchmakeSessionCount = it->GetMatchmakeSessionCount();
                persistentGatheringId = it->GetID();
            }
            ++it;
        }
    }

    // Participate if the persistent gathering is one you are not participating in.
    if(persistentGatheringId != INVALID_GATHERINGID)
    {
        qBool isJoined = false;
        nn::nex::qList<nn::nex::PersistentGathering>::const_iterator it = myPersistentGatheringList.begin();
        while(it != myPersistentGatheringList.end())
        {
            if(it->GetID() == persistentGatheringId)
            {
                isJoined = true;
            }
            ++it;
        }
        if(!isJoined)
        {
            if(!JoinPersistentGathering(pMatchmakeClient, persistentGatheringId))
            {
                return;
            }
        }
    }


    // Creates a persistent gathering if there is not one you can participate in.
    if(persistentGatheringId == INVALID_GATHERINGID)
    {
        if(!CreatePersistentGathering(pMatchmakeClient, &persistentGatheringId))
        {
            return;
        }
    }

    // Runs a persistent gathering matching.
    RunPersistentGatheringMatch(pMatchmakeClient, persistentGatheringId);
}

qBool FindPersistentGatheringByParticipant(nn::nex::MatchmakeExtensionClient* pMatchmakeClient,
                                 nn::nex::PrincipalID principalId,
                                 nn::nex::qList<nn::nex::PersistentGathering>* pPersistentGatheringList)
{
    nn::nex::ProtocolCallContext oContext;

    NEX_DEMO_LOG("FindPersistentGatheringByParticipant pid = %u\n", principalId);
    if(pMatchmakeClient->FindPersistentGatheringByParticipant(&oContext,
            principalId,
            nn::nex::ResultRange(),
            pPersistentGatheringList))
    {
        oContext.Wait();
        if(oContext.GetOutcome())
        {
            PrintPersistentGatheringList(*pPersistentGatheringList);
        }
        else
        {
            nex_demo_util::ReportPCallError(oContext, "Failed to FindPersistentGatheringByParticipant().\n");
            return false;
        }
    }
    else
    {
        NEX_DEMO_LOG("Fail to call FindPersistentGatheringByParticipant()\n");
        return false;
    }
    return true;
}

qBool JoinPersistentGathering(nn::nex::MatchmakeExtensionClient* pMatchmakeClient, nn::nex::GatheringID persistentGatheringId)
{
    nn::nex::ProtocolCallContext oContext;

    NEX_DEMO_LOG("Join PersistentGathering %u\n", persistentGatheringId);
    if(pMatchmakeClient->JoinPersistentGathering(&oContext, persistentGatheringId))
    {
        oContext.Wait();
        if(oContext.GetOutcome())
        {
            NEX_DEMO_LOG("Success JoinPersistentGathering %u\n", persistentGatheringId);
        }
        else
        {
            // Individually support errors that should be handled by the application (such as displaying application-defined error messages)
            // End if a join fails even once in this sample
            if(oContext.GetOutcome() == QERROR(RendezVous, SessionVoid))
            {
                NEX_DEMO_LOG("PersistentGathering %u is already unregistered.\n", persistentGatheringId);
            }
            else if(oContext.GetOutcome() == QERROR(RendezVous, SessionFull))
            {
                NEX_DEMO_LOG("PersistentGathering %u is already full.\n", persistentGatheringId);
            }
            else if(oContext.GetOutcome() == QERROR(RendezVous, PersistentGatheringParticipationMax))
            {
                NEX_DEMO_LOG("You exceed persistentGathering participation max\n");
            }
            else if(oContext.GetOutcome() == QERROR(RendezVous, InvalidPassword))
            {
                NEX_DEMO_LOG("Bad persistentGathering password\n");
            }
            else if(oContext.GetOutcome() == QERROR(RendezVous, WithoutParticipationPeriod))
            {
                NEX_DEMO_LOG("This persistentGathering is without participation period.\n");
            }
            else
            {
                nex_demo_util::ReportPCallError(oContext, "Failed to JoinPersistentGathering().\n");
            }
            return false;
        }
    }
    else
    {
        NEX_DEMO_LOG("Fail to call JoinPersistentGathering()\n");
        return false;
    }
    return true;
}

qBool CreatePersistentGathering(nn::nex::MatchmakeExtensionClient* pMatchmakeClient, nn::nex::GatheringID* pPersistentGatheringGatheringId)
{
    nn::nex::ProtocolCallContext oContext;

    // Persistent gathering parameter settings.
    nn::nex::PersistentGathering persistentGathering;

    // An open persistent gathering.
    persistentGathering.SetPersistentGatheringType(nn::nex::PersistentGatheringType::Open);

    // Setting the maximum number of participants to 0 removes the limitations
    persistentGathering.SetMaxParticipants(0);

    // Description string.
    persistentGathering.SetDescription(_T("this is test persistentGathering."));

    // Attribute value. As a test, set the local host's PrincipalID
    persistentGathering.SetAttribute(0, pMatchmakeClient->GetCredentials()->GetPrincipalID());

    // Specifying nn::nex::DateTime::Never as the joinable date and time (in UTC time) removes the limitations
    persistentGathering.SetParticipationStartDate(nn::nex::DateTime::Never);

    // Specifying nn::nex::DateTime::Never as the non-joinable date and time (in UTC time) removes the limitations
    persistentGathering.SetParticipationEndDate(nn::nex::DateTime::Never);

    if(pMatchmakeClient->CreatePersistentGathering(&oContext, persistentGathering, pPersistentGatheringGatheringId))
    {
        oContext.Wait();
        if(oContext.GetOutcome())
        {
            NEX_DEMO_LOG("Success CreatePersistentGathering %u\n", *pPersistentGatheringGatheringId);
        }
        else
        {
            // Individually support errors that should be handled by the application (such as displaying application-defined error messages)
            // End if a join fails even once in this sample
            if(oContext.GetOutcome() == QERROR(RendezVous, PersistentGatheringParticipationMax))
            {
                NEX_DEMO_LOG("You exceed persistentGathering participation max\n");
            }
            else if(oContext.GetOutcome() == QERROR(RendezVous, PersistentGatheringCreationMax))
            {
                NEX_DEMO_LOG("You exceed persistentGathering creation max\n");
            }
            else
            {
                nex_demo_util::ReportPCallError(oContext, "Failed to CreatePersistentGathering().\n");
            }
            return false;
        }
    }
    else
    {
        NEX_DEMO_LOG("Fail to call CreatePersistentGathering()\n");
        return false;
    }
    return true;
}

void RunPersistentGatheringMatch(nn::nex::MatchmakeExtensionClient* pMatchmakeClient, nn::nex::GatheringID persistentGatheringId)
{
    NEX_DEMO_LOG("Run PersistentGatheringMatch for PersistentGathering %u\n", persistentGatheringId);

    nn::nex::ProtocolCallContext oContext;

    nn::nex::MatchmakeSession* pMatchmakeSession = new nn::nex::MatchmakeSession();

    // Specifies MATCHMAKE_SYSTEM_TYPE_PERSISTENT_GATHERING because it is a persistent gathering matching.
    pMatchmakeSession->SetMatchmakeSystemType(nn::nex::MATCHMAKE_SYSTEM_TYPE_PERSISTENT_GATHERING);

    // Specifies gathering ID of persistent gathering for the 0th attribute value.
    pMatchmakeSession->SetAttribute(0, persistentGatheringId);

    nn::nex::GatheringHolder inputGatheringHolder = pMatchmakeSession;
    nn::nex::GatheringHolder outputGatheringHolder;

    if(pMatchmakeClient->AutoMatchmake(&oContext, inputGatheringHolder, &outputGatheringHolder))
    {
        oContext.Wait();
        if(oContext.GetOutcome())
        {
            NEX_DEMO_LOG("Success AutoMatchmake id:%u owner:%u\n", outputGatheringHolder->GetID(), outputGatheringHolder->GetOwnerPID());
        }
        else
        {
            nex_demo_util::ReportPCallError(oContext, "Failed to AutoMatchmake().\n");
            return;
        }
    }
    else
    {
        NEX_DEMO_LOG("Fail to call AutoMatchmake()\n");
        return;
    }

    // Wait for roughly 60 seconds. See the AutoMatch sample and others for more on the transition to P2P.
    NEX_DEMO_LOG("Wait 60 seconds.\n");

    nn::nex::Time tiWaitStart = nn::nex::Time::GetTime();
    while(true)
    {
        nn::nex::Scheduler::GetInstance()->DispatchAll();

        nn::nex::Time nowTime = nn::nex::Time::GetTime();
        if(nowTime - tiWaitStart > 60 * 1000)
        {
            break;
        }
        else
        {
            nn::nex::Platform::Sleep(30);
        }
    }

    // If the owner, delete the matchmaking session. If not, leave.
    if(outputGatheringHolder->GetOwnerPID() == pMatchmakeClient->GetCredentials()->GetPrincipalID())
    {
        if(pMatchmakeClient->UnregisterGathering(&oContext, outputGatheringHolder->GetID()))
        {
            oContext.Wait();
            if(oContext.GetOutcome())
            {
                NEX_DEMO_LOG("Success UnregisterGathering %u\n", outputGatheringHolder->GetID());
            }
            else
            {
                // Since the local host is the owner, there should be no failure
                nex_demo_util::ReportPCallError(oContext, "Failed to UnregisterGathering().\n");
                return;
            }
        }
        else
        {
            NEX_DEMO_LOG("Fail to call UnregisterGathering()\n");
            return;
        }
    }
    else
    {
        if(pMatchmakeClient->EndParticipation(&oContext, outputGatheringHolder->GetID()))
        {
            oContext.Wait();
            if(oContext.GetOutcome())
            {
                NEX_DEMO_LOG("Success EndParticipation %u\n", outputGatheringHolder->GetID());
            }
            else
            {
                if(oContext.GetOutcome() == QERROR(RendezVous, SessionVoid))
                {
                    NEX_DEMO_LOG("MatchmakeSession %u is already unregistered.\n", outputGatheringHolder->GetID());
                }
                else
                {
                    nex_demo_util::ReportPCallError(oContext, "Failed to EndParticipation().\n");
                }
                return;
            }
        }
        else
        {
            NEX_DEMO_LOG("Fail to call EndParticipation()\n");
            return;
        }
    }
}

void PrintPersistentGatheringList(const nn::nex::qList<nn::nex::PersistentGathering>& persistentGatheringList)
{
    if(persistentGatheringList.empty())
    {
        NEX_DEMO_LOG("PersistentGatheringList is empty.\n");
    }

    nn::nex::qList<nn::nex::PersistentGathering>::const_iterator it = persistentGatheringList.begin();
    while(it != persistentGatheringList.end())
    {
        NEX_DEMO_LOG("-------------------------------------------\n");
        NEX_DEMO_LOG("GatheringID : %u\n", it->GetID());
        NEX_DEMO_LOG("  PersistentGatheringCode : %llu\n", nn::nex::MatchmakeExtensionClient::GatheringIdToPersistentGatheringCode(it->GetID()));
        NEX_DEMO_LOG("  PersistentGatheringType : %u\n", it->GetPersistentGatheringType());
        NEX_DEMO_LOG("  Owner : %u\n", it->GetOwnerPID());
        NEX_DEMO_LOG("  MaxParticipants : %u\n", it->GetMaxParticipants());
        NEX_DEMO_LOG("  Attribute_0 : %u\n", it->GetAttribute(0));
        NEX_DEMO_LOG("  Attribute_1 : %u\n", it->GetAttribute(1));
        NEX_DEMO_LOG("  Attribute_2 : %u\n", it->GetAttribute(2));
        NEX_DEMO_LOG("  Attribute_3 : %u\n", it->GetAttribute(3));
        NEX_DEMO_LOG("  Attribute_4 : %u\n", it->GetAttribute(4));
        NEX_DEMO_LOG("  Attribute_5 : %u\n", it->GetAttribute(5));
        NEX_DEMO_LOG("  Participation Count : %u\n", it->GetParticipationCount());
        NEX_DEMO_LOG("  MatchmakeSession Count : %u\n", it->GetMatchmakeSessionCount());
        NEX_DEMO_LOG("-------------------------------------------\n");

        ++it;
    }
}
