﻿/*---------------------------------------------------------------------------*
  (C) Copyright 1998-2014, Technologies Quazal Inc. All rights reserved.
  Quazal Net-Z, Quazal Rendez-Vous and Quazal are trademarks of Quazal.

  Copyright (C) 2011-2014 Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo and/or its licensors and are
  protected by national and international copyright laws.
  They may not be disclosed to third parties or copied or duplicated
  in any form, in whole or in part, without the prior written consent
  of Nintendo.
  The content herein is highly confidential and should be handled
  accordingly.
 *---------------------------------------------------------------------------*/

/*
 *  RankingSimple
 *
 *  This sample shows the basic use of ranking.
 *  It does the following:
 *
 *  1. Logs in to game server, and makes rankings available.
 *  2. Registers common ranking data.
 *  3. Registers scores.
 *  4. Gets the ranking list (top 10).
 *  5. Estimates ranking order from scores.
 *  6. Gets ranking statistics.
 *  7. Changes score attributes.
 *  8. Deletes scores.
 *  9. Re-registers scores.
 * 10. Ends use of rankings and logs out from game server.
 */

#include <nex.h>
#include "../common/NexUtil.h"

/*
   To use NEX unique ID's to register scores on a more detailed basis than by using principal ID's, enable this macro to get the NEX unique ID.
   
*/
// #define RANKING_DEMO_USE_NEX_UNIQUE_ID

// Macro for banner display
#define BANNER(...) do                                                               \
                    {                                                                \
                        NEX_DEMO_LOG("\n***************************************\n"); \
                        NEX_DEMO_LOG(__VA_ARGS__);                                   \
                        NEX_DEMO_LOG("\n");                                          \
                    }                                                                \
                    while (0)

namespace
{
// NgsFacade
nn::nex::NgsFacade* s_pNgsFacade = NULL;
// NEX unique ID (to register scores on a principal ID basis, specify INVALID_UNIQUEID)
qUnsignedInt64 s_nexUniqueId = nn::nex::INVALID_UNIQUEID;
// Category
qUnsignedInt32 s_useCategory = 0;
// Groups
qByte s_useGroup0 = 0;
qByte s_useGroup1 = 0;
}

#ifdef RANKING_DEMO_USE_NEX_UNIQUE_ID
// Get NEX unique ID (only when using NEX unique IDs)
qBool AcquireNexUniqueId();
#endif /* RANKING_DEMO_USE_NEX_UNIQUE_ID */

// Upload common ranking data.
qBool UploadCommonData(nn::nex::RankingClient *pRankingClient);
// Upload the scores.
qBool UploadScore(nn::nex::RankingClient *pRankingClient);
// Get the ranking list (top 10).
qBool GetRanking(nn::nex::RankingClient *pRankingClient);
// Estimate ranking order from scores.
qBool GetApproxOrder(nn::nex::RankingClient *pRankingClient);
// Get ranking statistics.
qBool GetStats(nn::nex::RankingClient *pRankingClient);
// Change score attributes.
qBool ChangeAttributes(nn::nex::RankingClient *pRankingClient);
// Delete the scores.
qBool DeleteScore(nn::nex::RankingClient *pRankingClient);
// Display list of rankings.
void PrintRankDataList(const nn::nex::qVector<nn::nex::RankingRankData>& rankDataList);


qInt NEXSampleMain(qInt intArg, void* ptrArg)
{
    (void)intArg;
    (void)ptrArg;

    // Connect to the Internet
    if (!nex_demo_util::NetInit())
    {
        NEX_DEMO_LOG("net init failed...\n");
        NEX_DEMO_LOG("Leaving application...\n");

        return 0;
    }

    // Prepare to use tools
    nex_demo_util::Initialize();

    // Initialization
    nn::nex::GlobalVariables::AcquireInstance();

    // Set to thread unsafe transport buffer thread mode
    nn::nex::Core::SetThreadMode(nn::nex::Core::ThreadModeUnsafeTransportBuffer);

    s_pNgsFacade = new nn::nex::NgsFacade();
    nn::nex::RankingClient* pRankingClient = new nn::nex::RankingClient();

    // Log in to game server.
    BANNER("Game Server Login");

    if(!nex_demo_util::GameServerLogin(s_pNgsFacade, NEXDEMO_GAME_SERVER_ID, NEXDEMO_ACCESS_KEY, 60000))
    {
        NEX_DEMO_LOG("Login failed.\n");
        goto END;
    }

    // Bind the Ranking service.
    BANNER("Bind to Ranking service");
    if (!pRankingClient->Bind(s_pNgsFacade->GetCredentials()))
    {
        NEX_DEMO_LOG("Cannot bind to Ranking Service\n");
        goto ERROR_RANKING_BIND;
    }

#ifdef RANKING_DEMO_USE_NEX_UNIQUE_ID
    // Get NEX unique ID.
    if (!AcquireNexUniqueId())
    {
        goto ERR;
    }
#endif /* RANKING_DEMO_USE_NEX_UNIQUE_ID */

    // Upload common ranking data.
    if (!UploadCommonData(pRankingClient))
    {
        goto ERR;
    }
    // Upload the scores.
    if (!UploadScore(pRankingClient))
    {
        goto ERR;
    }
    // Get the ranking list (top 10).
    if (!GetRanking(pRankingClient))
    {
        goto ERR;
    }
    // Estimate ranking order from scores.
    if (!GetApproxOrder(pRankingClient))
    {
        goto ERR;
    }
    // Get ranking statistics.
    if (!GetStats(pRankingClient))
    {
        goto ERR;
    }
    // Change score attributes.
    if (!ChangeAttributes(pRankingClient))
    {
        goto ERR;
    }
    // Delete the scores.
    if (!DeleteScore(pRankingClient))
    {
        goto ERR;
    }
    // Re-register the scores.
    if (!UploadScore(pRankingClient))
    {
        goto ERR;
    }


ERR:
    // End use of service.
    // Ranking features cannot be used subsequent to here.
    // Call this before logging out.
    BANNER("Unbind from Ranking service");
    pRankingClient->Unbind();

ERROR_RANKING_BIND:
    // Log off from game server.
    BANNER("Game Server Logout");
    nex_demo_util::GameServerLogout(s_pNgsFacade);

END:
    NEX_DEMO_LOG("End.\n");
    delete pRankingClient;
    delete s_pNgsFacade;

    nn::nex::GlobalVariables::ReleaseInstance();

    // Shut down utilities
    nex_demo_util::Finalize();

    // Disconnect from the Internet.
    nex_demo_util::NetTerm();

    return 0;
}

//-----------------------------------------------------------------------------
// Main Function
//-----------------------------------------------------------------------------
qInt GenericMain(qInt argc, qChar** argv)
{
    NEX_DEMO_LOG("*******************************\n");
    NEX_DEMO_LOG("       Ranking sample\n");
    NEX_DEMO_LOG("*******************************\n\n");

#if defined(CAFE_TARGET)

#else
    NEXSampleMain(argc, argv);
#endif

    NEX_DEMO_LOG("Leaving application...\n");

    return 0;
}

NEX_MAIN_WRAPPER(GenericMain)


#ifdef RANKING_DEMO_USE_NEX_UNIQUE_ID
//-----------------------------------------------------------------------------
// Get NEX unique ID (only when using NEX unique IDs)
//-----------------------------------------------------------------------------
qBool AcquireNexUniqueId()
{
    BANNER("AcquireNexUniqueId()");
    /*
     *  Use the Utility service to get the NEX unique ID.
     *  The NEX unique ID is a unique ID in the game server that is newly issued every time the UtilityClient::AcquireNexUniqueId function is called.
     *  NEX unique IDs are used primarily when you would like to divide and register scores on a more detailed basis than you could using principal IDs.
     *
     *  Once you get the NEX unique ID, save it in a save region.
     *  If the NEX unique ID has been saved successfully, get it from the save region on all subsequent occasions.
     */

    // Attempt to get the NEX unique ID from the save region.
    {
        /*
        if (LoadNexUniqueId(&s_nexUniqueId))
        {
            // Check whether the NEX unique ID is valid.
            if(nn::nex::UtilityClient::IsValidNexUniqueId(s_nexUniqueId)) return;
        }
        */
    }

    // If getting it failed, get it from the server.

    // Bind the Utility service.
    NEX_DEMO_LOG("Bind to Utility service\n");
    nn::nex::UtilityClient* pUtilityClient = new nn::nex::UtilityClient();
    if (!pUtilityClient->Bind(s_pNgsFacade->GetCredentials()))
    {
        NEX_DEMO_LOG("Cannot bind to Utility Service\n");
        delete pUtilityClient;
        return false;
    }

    // Get NEX unique ID.
    nn::nex::ProtocolCallContext oContext;
    if (pUtilityClient->AcquireNexUniqueId(&oContext, &s_nexUniqueId))
    {
        oContext.Wait();
        if (oContext.GetState() != nn::nex::CallContext::CallSuccess)
        {
            nex_demo_util::ReportPCallError(oContext, "Failed to AcquireNexUniqueId()\n");
        }
        else
        {
            NEX_DEMO_LOG("AcquireNexUniqueId() Success. (NexUniqueId=%llu)  \n", s_nexUniqueId);
        }
    }

    // End use of the Utility service.
    NEX_DEMO_LOG("Unbind from Utility service\n");
    pUtilityClient->Unbind();
    delete pUtilityClient;

    if (s_nexUniqueId == nn::nex::INVALID_UNIQUEID)
    {
        return false;
    }
    return true;
}
#endif /* RANKING_DEMO_USE_NEX_UNIQUE_ID */

//-----------------------------------------------------------------------------
// Upload common ranking data.
//-----------------------------------------------------------------------------
qBool UploadCommonData(nn::nex::RankingClient *pRankingClient)
{
    BANNER("UploadCommonData()");
    /*
     *  Registers common ranking data.
     *
     *  Common ranking data is data common to a category (for example, a player name).
     *  This data is obtained with the Ranking::GetList function, and is published to all people (meaning anyone can get it).
     */

    // Common ranking data
    const char* pCommonData = "PLAYER";
    // Common ranking data size (player name + termination character)
    qByte commonDataSize = static_cast<qByte>(strlen(pCommonData) + 1);

    // Upload common ranking data.
    nn::nex::ProtocolCallContext oContext;
    if (pRankingClient->UploadCommonData(&oContext, pCommonData, commonDataSize, s_nexUniqueId))
    {
        oContext.Wait();
        if (oContext.GetState() != nn::nex::CallContext::CallSuccess)
        {
            nex_demo_util::ReportPCallError(oContext, "Failed to UploadCommonData()\n");
            return false;
        }
        NEX_DEMO_LOG("UploadCommonData() Success.\n");
    }
    return true;
}

//-----------------------------------------------------------------------------
// Upload the scores.
//-----------------------------------------------------------------------------
qBool UploadScore(nn::nex::RankingClient *pRankingClient)
{
    BANNER("UploadScore()");
    /*
     *  Registers scores.
     *
     *  Determines where to register the score to, specifying the category and group.
     *  Any 64-bit value (for example, a character number) can be specified as a parameter.
     */

    // Create score data.
    nn::nex::RankingScoreData scoreData;
    {
        // Specify the category.
        scoreData.SetCategory(s_useCategory);
        // Set the score.
        scoreData.SetScore(nn::nex::Platform::GetRandomNumber(10000));
        // Specify the sort order for scores.
        scoreData.SetOrderBy(nn::nex::RankingConstants::ORDER_BY_DESC); // Descending order.
        // Specify the score update mode.
        scoreData.SetUpdateMode(nn::nex::RankingConstants::UPDATE_MODE_NORMAL);
        // Set the value for group 0.
        scoreData.SetGroup0(s_useGroup0);
        // Set the value for group 1.
        scoreData.SetGroup1(s_useGroup1);
        // Set the parameter value.
        scoreData.SetParam((qUnsignedInt64)nn::nex::Platform::GetRandomNumber(16));
    }

    // Upload the scores.
    nn::nex::ProtocolCallContext oContext;
    if (pRankingClient->UploadScore(&oContext, scoreData, s_nexUniqueId))
    {
        oContext.Wait();
        if (oContext.GetState() != nn::nex::CallContext::CallSuccess)
        {
            nex_demo_util::ReportPCallError(oContext, "Failed to UploadScore()\n");
            return false;
        }
        NEX_DEMO_LOG("UploadScore() Success. (score=%u)\n", scoreData.GetScore());
    }
    return true;
}

//-----------------------------------------------------------------------------
// Get the ranking list (top 10).
//-----------------------------------------------------------------------------
qBool GetRanking(nn::nex::RankingClient *pRankingClient)
{
    BANNER("GetRanking()");
    /*
     *  Get the ranking list.
     *
     *  - Ranking for scope
     *  - Nearby rankings
     *  - Ranking among friends
     *  - Rankings nearby friends.
     *  All of the above can be obtained.
     *
     *  You can specify another person's principal ID and unique ID, and get that person's nearby rankings.
     */

    // Set the parameter for sort calculation.
    nn::nex::RankingOrderParam orderParam;
    {
        // Specify the method for calculating the order if the scores are identical.
        //  - Rank the third entry as third place if the top two names have the same score.
        orderParam.SetOrderCalculation(nn::nex::RankingConstants::ORDER_CALCULATION_113);
        //  - Rank according to update time if the top two names have the same score.
        //s_orderParam.SetOrderCalculation(nn::nex::RankingConstants::ORDER_CALCULATION_123);

        // Specify the group type to filter on.
        //  - Do not filter by group.
        orderParam.SetFilterGroupIndex(nn::nex::RankingConstants::FILTER_GROUP_INDEX_NONE);
        //  - Filter by group 0.
        //orderParam.SetFilterGroupIndex(nn::nex::RankingConstants::FILTER_GROUP_INDEX_0);
        //  - Filter by group 1.
        //orderParam.SetFilterGroupIndex(nn::nex::RankingConstants::FILTER_GROUP_INDEX_1);

        // Specify the group value to filter on.
        orderParam.SetFilterGroupNum(0);

        // Specify the time period (time scope) for which to calculate the ranking.
        //  - Get rankings for all times.
        orderParam.SetTimeScope(nn::nex::RankingConstants::TIME_SCOPE_ALL);
        //  - The default period (1) set in RMAS is "from 12:00AM on the last Tuesday until now."
        //orderParam.SetTimeScope(nn::nex::RankingConstants::TIME_SCOPE_CUSTOM_0);
        //  - The default period (2) set in RMAS is "from 12:00AM on the first day of this month until now."
        //orderParam.SetTimeScope(nn::nex::RankingConstants::TIME_SCOPE_CUSTOM_1);

        // Specify the maximum length and the offset of the list to get.
        orderParam.SetLength(10);
        orderParam.SetOffset(0);
    }

    // Specify the ranking retrieval mode.
    // - Range ranking (when offset=0 it is top ranking)
    nn::nex::RankingConstants::RankingMode rankingMode = nn::nex::RankingConstants::RANKING_MODE_RANGE;
    // - Nearby ranking (get the rankings nearby the specified user)
    //nn::nex::RankingConstants::RankingMode rankingMode = nn::nex::RankingConstants::RANKING_MODE_NEAR;
    // - Ranking among friends (get the ranking among the user's friends)
    //nn::nex::RankingConstants::RankingMode rankingMode = nn::nex::RankingConstants::RANKING_MODE_FRIEND_RANGE;
    // - Ranking near friends (get ranking near own position in friend ranking)
    //nn::nex::RankingConstants::RankingMode rankingMode = nn::nex::RankingConstants::RANKING_MODE_FRIEND_NEAR;
    // - Get ranking of specified user
    //nn::nex::RankingConstants::RankingMode rankingMode = nn::nex::RankingConstants::RANKING_MODE_USER;

    // Results of getting ranking.
    nn::nex::RankingResult rankingResult;

    // Get the ranking list
    nn::nex::ProtocolCallContext oContext;
    if (pRankingClient->GetRanking(&oContext, rankingMode, s_useCategory, orderParam, &rankingResult, s_nexUniqueId))
    {
        oContext.Wait();
        if (oContext.GetState() != nn::nex::CallContext::CallSuccess)
        {
            nex_demo_util::ReportPCallError(oContext, "Failed to GetRanking()\n");
            return false;
        }
        // Output results.
        NEX_DEMO_LOG("GetRanking(RANKING_MODE_RANGE) Result:\n");
        PrintRankDataList(rankingResult.GetRankDataList());
    }
    return true;
}

//-----------------------------------------------------------------------------
// Estimate ranking order from scores.
//-----------------------------------------------------------------------------
qBool GetApproxOrder(nn::nex::RankingClient *pRankingClient)
{
    BANNER("GetApproxOrder()");
    /*
     *  Gets the estimated ranking order that would result if the scores were uploaded.
     *
     *  This operation does not actually register scores.
     */

    // Score
    qUnsignedInt32 score = nn::nex::Platform::GetRandomNumber(10000);
    NEX_DEMO_LOG("Score: %d\n", score);

    // Set sort calculation parameter.
    nn::nex::RankingOrderParam orderParam;
    {
        orderParam.SetOrderCalculation(nn::nex::RankingConstants::ORDER_CALCULATION_113);
        orderParam.SetFilterGroupIndex(nn::nex::RankingConstants::FILTER_GROUP_INDEX_NONE);
        orderParam.SetTimeScope(nn::nex::RankingConstants::TIME_SCOPE_ALL);
    }

    // Get the estimation result.
    qUnsignedInt32 order = 0;
    nn::nex::ProtocolCallContext oContext;
    if (pRankingClient->GetApproxOrder(&oContext, s_useCategory, orderParam, score, &order, s_nexUniqueId))
    {
        oContext.Wait();
        if (oContext.GetState() != nn::nex::CallContext::CallSuccess)
        {
            nex_demo_util::ReportPCallError(oContext, "Failed to GetApproxOrder()\n");
            return false;
        }
        // Output results.
        NEX_DEMO_LOG("GetApproxOrder() Success. (order=%d)\n", order);
    }
    return true;
}

//-----------------------------------------------------------------------------
// Get ranking statistics.
//-----------------------------------------------------------------------------
qBool GetStats(nn::nex::RankingClient *pRankingClient)
{
    BANNER("GetStats()");
    /*
     *  Get ranking statistical values (Score total, sum, average, minimum and maximum values).
     */

    // Prepare flags specifying which statistical values to get.
    qUnsignedInt32 flags = (
                               nn::nex::RankingConstants::STATS_FLAG_TOTAL   |
                               nn::nex::RankingConstants::STATS_FLAG_SUM     |
                               nn::nex::RankingConstants::STATS_FLAG_MIN     |
                               nn::nex::RankingConstants::STATS_FLAG_MAX     |
                               nn::nex::RankingConstants::STATS_FLAG_AVERAGE
                           );

    // Set sort calculation parameter.
    nn::nex::RankingOrderParam orderParam;
    {
        orderParam.SetFilterGroupIndex(nn::nex::RankingConstants::FILTER_GROUP_INDEX_NONE);
        orderParam.SetTimeScope(nn::nex::RankingConstants::TIME_SCOPE_ALL);
    }

    // Get the statistical results.
    nn::nex::RankingStats stats;
    nn::nex::ProtocolCallContext oContext;
    if (pRankingClient->GetStats(&oContext, s_useCategory, orderParam, flags, &stats))
    {
        oContext.Wait();
        if (oContext.GetState() != nn::nex::CallContext::CallSuccess)
        {
            nex_demo_util::ReportPCallError(oContext, "Failed to GetStats()\n");
            return false;
        }
        // Output results.
        NEX_DEMO_LOG("GetStats Result:\n");
        NEX_DEMO_LOG("** TOTAL: %f\n", stats.GetValue(nn::nex::RankingConstants::STATS_FLAG_TOTAL));
        NEX_DEMO_LOG("** SUM: %f\n", stats.GetValue(nn::nex::RankingConstants::STATS_FLAG_SUM));
        NEX_DEMO_LOG("** MIN: %f\n", stats.GetValue(nn::nex::RankingConstants::STATS_FLAG_MIN));
        NEX_DEMO_LOG("** MAX: %f\n", stats.GetValue(nn::nex::RankingConstants::STATS_FLAG_MAX));
        NEX_DEMO_LOG("** AVERAGE: %f\n", stats.GetValue(nn::nex::RankingConstants::STATS_FLAG_AVERAGE));
    }
    return true;
}

//-----------------------------------------------------------------------------
// Change score attributes.
//-----------------------------------------------------------------------------
qBool ChangeAttributes(nn::nex::RankingClient *pRankingClient)
{
    BANNER("ChangeAttributes()");
    /*
     *  Change score attributes (group 0, group 1, parameter).
     */

    // Create modification parameter.
    nn::nex::RankingChangeAttributesParam changeParam;
    {
        // Specify the attribute to change.
        // You can specify multiple values using the bitwise OR of RankingConstants::ModificationFlag.
        changeParam.SetModificationFlag(nn::nex::RankingConstants::MODIFICATION_FLAG_GROUP0 |
                                        nn::nex::RankingConstants::MODIFICATION_FLAG_GROUP1 |
                                        nn::nex::RankingConstants::MODIFICATION_FLAG_PARAM);
        // Set the value for group 0.
        changeParam.SetGroup0(static_cast<qByte>(nn::nex::Platform::GetRandomNumber(100)));
        // Set the value for group 1.
        changeParam.SetGroup1(static_cast<qByte>(nn::nex::Platform::GetRandomNumber(100)));
        // Set the parameter value.
        changeParam.SetParam(static_cast<qUnsignedInt64>(nn::nex::Platform::GetRandomNumber(16)));

        NEX_DEMO_LOG("Change group0=%d, group1=%d, param=%llu\n",
                     changeParam.GetGroup0(), changeParam.GetGroup1(), changeParam.GetParam());
    }

    // Change attribute value.
    nn::nex::ProtocolCallContext oContext;
    if (pRankingClient->ChangeAttributes(&oContext, s_useCategory, changeParam, s_nexUniqueId))
    {
        oContext.Wait();
        if (oContext.GetState() != nn::nex::CallContext::CallSuccess)
        {
            nex_demo_util::ReportPCallError(oContext, "Failed to ChangeAttributes()\n");
            return false;
        }
        NEX_DEMO_LOG("ChangeAttributes() Success.\n");
    }

    // Get own score after change.
    {
        // Set sort calculation parameter.
        nn::nex::RankingOrderParam orderParam;
        orderParam.SetOrderCalculation(nn::nex::RankingConstants::ORDER_CALCULATION_113);
        orderParam.SetFilterGroupIndex(nn::nex::RankingConstants::FILTER_GROUP_INDEX_NONE);
        orderParam.SetTimeScope(nn::nex::RankingConstants::TIME_SCOPE_ALL);

        // Results of getting ranking.
        nn::nex::RankingResult rankingResult;
        // Specify ranking mode in own ranking data.
        nn::nex::RankingConstants::RankingMode rankingMode = nn::nex::RankingConstants::RANKING_MODE_USER;

        // Get the ranking list
        if (pRankingClient->GetRanking(&oContext, rankingMode, s_useCategory, orderParam, &rankingResult, s_nexUniqueId))
        {
            oContext.Wait();
            if (oContext.GetState() != nn::nex::CallContext::CallSuccess)
            {
                nex_demo_util::ReportPCallError(oContext, "Failed to GetRanking()\n");
                return false;
            }
            // Output results.
            NEX_DEMO_LOG("GetRanking(RANKING_MODE_USER) Result:\n");
            PrintRankDataList(rankingResult.GetRankDataList());
        }
    }
    return true;
}

//-----------------------------------------------------------------------------
// Delete the scores.
//-----------------------------------------------------------------------------
qBool DeleteScore(nn::nex::RankingClient *pRankingClient)
{
    BANNER("DeleteScore()");
    /*
     *  Delete scores.
     *  If a category is specified, only delete the scores for that category.
     *  If a category is not specified, delete all data for all categories at once.
     */

    // Delete the scores.
    nn::nex::ProtocolCallContext oContext;
    if (pRankingClient->DeleteScore(&oContext, s_useCategory, s_nexUniqueId))
    {
        oContext.Wait();
        if (oContext.GetState() != nn::nex::CallContext::CallSuccess)
        {
            nex_demo_util::ReportPCallError(oContext, "Failed to DeleteScore()\n");
            return false;
        }
        NEX_DEMO_LOG("DeleteScore() Success\n");
    }
    return true;
}

//-----------------------------------------------------------------------------
// Display list of rankings.
//-----------------------------------------------------------------------------
void PrintRankDataList(const nn::nex::qVector<nn::nex::RankingRankData>& rankDataList)
{
    if(rankDataList.empty())
    {
        NEX_DEMO_LOG("RankDataList is empty.\n");
    }

    nn::nex::qVector<nn::nex::RankingRankData>::const_iterator it = rankDataList.begin();
    while(it != rankDataList.end())
    {
        NEX_DEMO_LOG("** [%2d] -------------------------------\n", it - rankDataList.begin());
        NEX_DEMO_LOG("** principalId    = %d\n", it->GetPrincipalId());
        NEX_DEMO_LOG("** nexUniqueId    = %llu\n", it->GetUniqueId());
        NEX_DEMO_LOG("** order          = %d\n", it->GetOrder());
        NEX_DEMO_LOG("** category       = %d\n", it->GetCategory());
        const nn::nex::DateTime& updateTime = it->GetUpdateTime();
        (void)updateTime;
        NEX_DEMO_LOG("** updateTime     = %04d-%02d-%02d %02d:%02d:%02d\n",
                updateTime.GetYear(), updateTime.GetMonth(), updateTime.GetDay(),
                updateTime.GetHour(), updateTime.GetMinute(), updateTime.GetSecond());
        NEX_DEMO_LOG("** score          = %d\n", it->GetScore());
        NEX_DEMO_LOG("** group0         = %d\n", it->GetGroup0());
        NEX_DEMO_LOG("** group1         = %d\n", it->GetGroup1());
        NEX_DEMO_LOG("** param          = %llu\n", it->GetParam());
        NEX_DEMO_LOG("** commonDataSize = %d\n", it->GetCommonData().size());

        ++it;
    }
}
