﻿/*---------------------------------------------------------------------------*
  (C) Copyright 1998-2014, Technologies Quazal Inc. All rights reserved.
  Quazal Net-Z, Quazal Rendez-Vous and Quazal are trademarks of Quazal.

  Copyright (C) 2011-2014 Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo and/or its licensors and are
  protected by national and international copyright laws.
  They may not be disclosed to third parties or copied or duplicated
  in any form, in whole or in part, without the prior written consent
  of Nintendo.
  The content herein is highly confidential and should be handled
  accordingly.
 *---------------------------------------------------------------------------*/
/*
 * NEX Sample
 *
 * Description:
 *
 * This sample defines the classes that perform messaging-based sending and receiving.
 * In this feature, the TextMessage object is used for sending and receiving. Use a BinaryMessage object to send and receive binary data.
 * 
 * This feature can be used in applications as-is or revised as necessary for use in the application.
 */

#include <nex.h>

#include "../common/NexUtil.h"


/*
 
 */
class MessageSender
{
public:
    /*
     

     
     */
    MessageSender(nn::nex::NgsFacade* pNgsFacade)
        : m_pNgsFacade(pNgsFacade)
        , m_pMessagingClient(NULL)
        , m_pMessagingNotificationHandler(NULL)
        , m_callContextStore()
        , m_firstError(true)
    {
        Bind();
    };
    virtual ~MessageSender()
    {
        // Cancel all asynchronous processes that are not finished.
        m_callContextStore.CancelAll();
        Unbind();
    };

    /*
     

     
     
     */
    void SendChat(nn::nex::GatheringID target, const nn::nex::String& message)
    {
        NEX_DEMO_LOG("\nCall SendChat().\n");

        nn::nex::TextMessage* pTextMessage = new nn::nex::TextMessage();
        nn::nex::UserMessageDataHolder oDataHolder = pTextMessage;

        // Set a message
        pTextMessage->SetSubject(_T("This is a TextMessage"));
        pTextMessage->SetTextBody(message);
        NEX_DEMO_LOG("*** Sending message to gid:[%u] subject[" STRING_FORMAT "] text body:[" STRING_FORMAT "]\n",
                     target,
                     pTextMessage->GetSubject().CStr(),
                     pTextMessage->GetTextBody().CStr());

        nn::nex::ProtocolCallContext* pCallContext = new nn::nex::ProtocolCallContext();
        nex_demo_util::RegisterCallContextCallback(this, &MessageSender::OnCompletionSend, pCallContext);

        // Only save the generated CallContext for asynchronous processes that are currently running in m_callContextStore so that they can be canceled later if necessary.
        nex_demo_util::CallContextStoreAssociator::Register(&m_callContextStore, pCallContext);
        m_pMessagingClient->SendChat(pCallContext, target, oDataHolder);
    }

    /*
     

     
     
     */
    void SendInstant(nn::nex::PrincipalID target, const nn::nex::String& message)
    {
        NEX_DEMO_LOG("\nCall SendInstant().\n");

        nn::nex::TextMessage* pTextMessage = new nn::nex::TextMessage();
        nn::nex::UserMessageDataHolder oDataHolder = pTextMessage;

        // Set a message
        pTextMessage->SetSubject(_T("This is a TextMessage"));
        pTextMessage->SetTextBody(message);
        NEX_DEMO_LOG("*** Sending message to pid:[%u] subject[" STRING_FORMAT "] text body:[" STRING_FORMAT "]\n",
                     target,
                     pTextMessage->GetSubject().CStr(),
                     pTextMessage->GetTextBody().CStr());

        nn::nex::ProtocolCallContext* pCallContext = new nn::nex::ProtocolCallContext();
        nex_demo_util::RegisterCallContextCallback(this, &MessageSender::OnCompletionSend, pCallContext);

        // Only save the generated CallContext for asynchronous processes that are currently running in m_callContextStore so that they can be canceled later if necessary.
        nex_demo_util::CallContextStoreAssociator::Register(&m_callContextStore, pCallContext);
        m_pMessagingClient->SendInstant(pCallContext, target, oDataHolder);
    }

    /*
     

     
     */
    const nn::nex::qResult& GetFirstError() const { return m_firstError; }

    /*
     

     
     */
    qBool HasReceivedMessage() const { return m_pMessagingNotificationHandler->HasReceivedMessage(); }

    /*
     

     
     */
    nn::nex::TextMessage DequeueReceivedMessage() { return m_pMessagingNotificationHandler->DequeueReceivedMessage(); }

private:
    // Received message event handler.
    class MyMessagingNotificationHandler: public nn::nex::MessagingNotificationHandler
    {
    public:
        MyMessagingNotificationHandler(MessageSender* sender) : m_pSender(sender) {}
        void DeliverInstantMessage(const nn::nex::DataHolder& hMsg)
        {
            // Distinguishes between TextMessage and BinaryMessage.
            // If TextMessage::DynamicCast is successful, the receive message is TextMessage.
            nn::nex::TextMessage* pTextMsg = nn::nex::TextMessage::DynamicCast(hMsg.Get());
            NEX_DEMO_LOG("*** Recieved message from:[%u] subject[" STRING_FORMAT "] text body:[" STRING_FORMAT "]\n",
                         pTextMsg->GetSenderID(),
                         pTextMsg->GetSubject().CStr(),
                         pTextMsg->GetTextBody().CStr());
            if (pTextMsg)
            {
                EnqueueReceivedMessage(*pTextMsg);
                return;
            }
        }
        qBool HasReceivedMessage() const { return !m_notificationQueue.empty(); }
        nn::nex::TextMessage DequeueReceivedMessage() { nn::nex::TextMessage message = m_notificationQueue.front(); m_notificationQueue.pop_front(); return message; }
    private:
        void EnqueueReceivedMessage(const nn::nex::TextMessage& message) { m_notificationQueue.push_back(message); }
        MessageSender* m_pSender;
        nn::nex::qList<nn::nex::TextMessage> m_notificationQueue; // Notification queue.
    };

    void Bind()
    {
        nn::nex::Credentials* pCred = m_pNgsFacade->GetCredentials();
        NEX_DEMO_ASSERT(pCred != NULL);
        m_pMessagingClient = new nn::nex::MessagingClient();
        NEX_DEMO_VERIFY(m_pMessagingClient->Bind(pCred));
        m_pMessagingNotificationHandler = new MyMessagingNotificationHandler(this);
        m_pMessagingClient->RegisterMessagingNotificationHandler(m_pMessagingNotificationHandler);
    }

    void Unbind()
    {
        m_pMessagingClient->UnregisterMessagingNotificationHandler();
        delete m_pMessagingNotificationHandler;
        m_pMessagingNotificationHandler = NULL;
        m_pMessagingClient->Unbind();
        delete m_pMessagingClient;
        m_pMessagingClient = NULL;
    }

    /*
     

     
     */
    void OnCompletionSend(nn::nex::CallContext* pCallContext)
    {
        switch (pCallContext->GetState())
        {
        case nn::nex::CallContext::CallSuccess:
            NEX_DEMO_LOG("     Send*() succeeded.\n");
            break;
        case nn::nex::CallContext::CallCancelled:
            // Do not do anything here because the processes that will be canceled, will be canceled from the ~MessageSender destructor.
            NEX_DEMO_LOG("     Send*() cancelled.\n");
            break;
        case nn::nex::CallContext::CallFailure:
            NEX_DEMO_LOG("     Send*() failed.\n");
            pCallContext->GetOutcome().Trace(TRACE_ALL);
            ErrorOccurred(pCallContext->GetOutcome());
            break;
        default:
            NEX_DEMO_ASSERT(false);
            break;
        }
    }

    void ErrorOccurred(const nn::nex::qResult& result) { if (m_firstError) { m_firstError = result; } }

    nn::nex::NgsFacade* m_pNgsFacade;                           // Pointer to NgsFacade object currently in a logged in state.
    nn::nex::MessagingClient* m_pMessagingClient;               // Pointer to object used to call messaging functions.
    MyMessagingNotificationHandler* m_pMessagingNotificationHandler;    // Pointer to object that handles messaging-related notifications.
    nex_demo_util::CallContextStore m_callContextStore;         // Container that saves CallContext while an asynchronous process is running. Only stores processes that may need to be canceled.
    nn::nex::qResult m_firstError;                              // The first error to occur.
};
