﻿/*---------------------------------------------------------------------------*
  (C) Copyright 1998-2014, Technologies Quazal Inc. All rights reserved.
  Quazal Net-Z, Quazal Rendez-Vous and Quazal are trademarks of Quazal.

  Copyright (C) 2011-2014 Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo and/or its licensors and are
  protected by national and international copyright laws.
  They may not be disclosed to third parties or copied or duplicated
  in any form, in whole or in part, without the prior written consent
  of Nintendo.
  The content herein is highly confidential and should be handled
  accordingly.
 *---------------------------------------------------------------------------*/
/*
 * NEX Sample
 *
 * TeamMatchmake
 *
 * Description:
 *
 *  This sample demonstrates how to form a team and then run matchmaking.
 *  Each team has two players and two teams are matched for a total of four players. Four clients are required for the sample to run to completion.
 *  The number of players can be modified by changing the values of the constants.
 *  
 *  In this sample, a single player will join two matchmaking sessions. One session is for the team and the other is for the match.
 *  After joining the team matchmaking session and gathering the team members with the MatchmakeExtenstionClient::AutoMatchmake function, the owner of the team matchmaking session then uses the MatchmakeExtenstionClient::AutoMatchmake function again to have all players in the team matchmaking session join the match matchmaking session at once.
 *  
 *  Once there are four players in the matchmaking session for the match, the match is determined to have started and the clients begin communicating by messages using MessagingClient.
 *  It is also possible to use P2P communication as necessary.
 *  
 *  If a team member leaves before the match begins, it is handled as an error.
 *  All other required error handling is lumped into the TeamMatchmakeSample::CheckCondition function.
 *  Whenever any kind of error occurs, all incomplete asynchronous processes at that moment are canceled.
 *  
 *  The following sequences are implemented in the sample.
 * 
 * 1. Log in to server (in nex_util).
 * 2. Start auto-matchmaking and then join a team matchmaking session.
 * 3. Wait until the team matchmaking session is full.
 * 4-1. The owner starts auto-matchmaking and has all team members join the matchmaking session for the match.
 * 4-2. Non-owners wait to be added to the matchmaking session for the match.
 * 5. Wait until the matchmaking session for the match is full.
 * 6. Communicate via messaging.
 * 7. After a set amount of time, the owner of the matchmaking session for the match notifies the other clients of the end of communication (game over).
 * 8. Leave the matchmaking session for the match.
 * 9. Leave the team matchmaking session.
 * 10. Log out from the server (in nex_util).
 * 
 * Example of how to embed this in an application:
 * 
 *  The TeamMatchmakeGameLoop function mimics the way the game loop works in an actual game. Implement this process in the game loop.
 *  We recommend calling the nn::nex::Scheduler::DispatchAll function and each of the calculation processes in the order they are called in this sample.
 *  Call functions in the TeamMatchmakeSample class from the main game loop to advance the sequential processes.
 *  Sequential processes are handled separately from the sending and receiving processes. Editing the Communicate function allows you to use any kind of sending and receiving process desired.
 *  
 *  This sample uses messaging-based communication. See MessageSender.h for specific sending and receiving processes.
 *  
 *  This sample uses the MatchmakeExtenstionClient::AutoMatchmake function to gather team members, but teams can also be formed using other methods such as by gathering friends and associates, and by directly joining the matchmaking session for the match without creating a gathering.
 *  When joining the matchmaking session for the match, The MatchmakeExtenstionClient::CreateMatchmakeSession and MatchmakeExtenstionClient::JoinMatchmakeSession functions can also be used instead of MatchmakeExtenstionClient::AutoMatchmake.
 *  These API functions implement the same interface that allows multiple players to join a matchmaking session.
 *  
 *  In this sample, all errors that occur during sequential processes or communication are handled appropriately within the range of the functions and sequences that are implemented herein.
 *  If you decide to add other sequences or modify which API functions are used, however, the errors that can occur in the NEX library may change and you may need additional user-defined result codes.
 *  In this case, you can use the MessageSender::GetFirstError and TeamMatchmakeSample::GetLastResult functions to get the last error that occurred and display it on screen in the application to notify the players or you can handle the error appropriately in the application.
 






*/

#include <nex.h>

#include "../common/NexUtil.h"
#include "MessageSender.h"

//-----------------------------------------------------------------------------
// Constant Definitions
//-----------------------------------------------------------------------------
// Game mode.
static const nn::nex::qUnsignedInt32 MATCHMAKE_GAME_MODE = 1003;

// Number of players per team.
static const nn::nex::qUnsignedInt16 MATCHMAKE_MAX_TEAM_PARTICIPANTS = 2;

// Number of teams to gather.
static const nn::nex::qUnsignedInt16 MATCHMAKE_MAX_NUM_TEAM = 2;

// Number of players in match.
static const nn::nex::qUnsignedInt16 MATCHMAKE_MAX_MATCH_PARTICIPANTS = MATCHMAKE_MAX_TEAM_PARTICIPANTS * MATCHMAKE_MAX_NUM_TEAM;   // 4

// Index of attribute that indicates recruiting mode.
static const nn::nex::qUnsignedInt16 MATCHMAKESESSION_ATTRIBUTE_INDEX_RECRUIT_MODE = 1;

// Duration (in frames) of messaging-based communication.
static const nn::nex::qUnsignedInt32 COMMUNICATION_DURATION = 60 * 60;  // 60 seconds

// User-defined result codes.
enum UserDefinedResultCode
{
    USER_DEFINED_RESULT_CODE_SUCCESS                         =  0,  // Process completed successfully, no errors.
    USER_DEFINED_RESULT_CODE_NETWORK_DISCONNECTED            = -1,  // Detected AC network disconnection.
    USER_DEFINED_RESULT_CODE_MATCHMAKESESSION_UNREGISTERED   = -2,  // Matchmaking session unintentionally deleted.
    USER_DEFINED_RESULT_CODE_UNEXPECTED_RMC_FAILURE          = -3,  // Unexpected RMC failure.
    USER_DEFINED_RESULT_CODE_UNEXPECTED_TEAM_MEMBER_DROP     = -4   // Team member unintentionally leaves.
};

// Recruiting modes.
enum RecruitMode
{
    RECRUIT_MODE_TEAM   = 0,
    RECRUIT_MODE_MATCH  = 1
};

//-----------------------------------------------------------------------------
// State machine that handles auto-matchmaking setup
//-----------------------------------------------------------------------------
class TeamMatchmakeSample
{
public:
    TeamMatchmakeSample(nn::nex::NgsFacade* pNgsFacade)
        : m_state(STATE_AUTOMATCHMAKE_TEAM)
        , m_finished(false)
        , m_errorOccurred(false)
        , m_unexpectedRmcFailureOccurred(false)
        , m_unexpectedGatheringUnregistered(false)
        , m_unexpectedTeamMemberDrop(false)
        , m_lastResult()
        , m_myPid(INVALID_PRINCIPALID)
        , m_teamOwnerPid(INVALID_PRINCIPALID)
        , m_matchOwnerPid(INVALID_PRINCIPALID)
        , m_addedGatheringId(INVALID_GATHERINGID)
        , m_pNgsFacade(pNgsFacade)
        , m_pMatchMakingClient(NULL)
        , m_pNotificationHandler(NULL)
        , m_joinedTeamMatchmakeSession()
        , m_joinedMatchMatchmakeSession()
    {
        Bind();
    };
    virtual ~TeamMatchmakeSample()
    {
        // Cancel all asynchronous processes that are not finished.
        m_callContextStore.CancelAll();
        Unbind();
    };
private:
    enum State
    {
        STATE_AUTOMATCHMAKE_TEAM,
        STATE_WAIT_TEAM_GATHERED,
        STATE_AUTOMATCHMAKE_MATCH,
        STATE_WAIT_MATCH_ADDED,
        STATE_WAIT_MATCH_GATHERED,
        STATE_COMMUNICATE,
        STATE_ENDPARTICIPATION_MATCH,
        STATE_ENDPARTICIPATION_TEAM,
        STATE_RMC_RESULT_WAITING,
        STATE_FINISH
    };

    // Event handler to process notifications from the NGS
    class MyNotificationEventHandler : public nn::nex::NotificationEventHandler
    {
    public:
        MyNotificationEventHandler(TeamMatchmakeSample* sample):m_pSample(sample){};
        virtual ~MyNotificationEventHandler(){};

        // Callback function that is invoked when a notification is received from NGS.
        void ProcessNotificationEvent(const nn::nex::NotificationEvent& oEvent)
        {
            nn::nex::StringStream ss;
            ss << oEvent;
            NEX_DEMO_LOG("[Notified] " STRING_FORMAT "\n", ss.CStr());

            // Temporarily store in a queue and process after the RMC because sometimes a nn::nex::ParticipationEvents::Participate or similar notification is received in advance while calling MatchmakeExtensionClient::AutoMatchmake or other RMC.
            // 
            EnqueueNotificationEvent(oEvent);
        }

        // Returns whether there are any unprocessed notifications.
        qBool HasNotificationEvent() const { return !m_notificationQueue.empty(); }

        // Dequeue and process one notification.
        void DequeueAndProcessNotificationEvent() { ProcessNotificationEventImpl(m_notificationQueue.front()); m_notificationQueue.pop_front(); }
    private:
        void EnqueueNotificationEvent(const nn::nex::NotificationEvent& oEvent) { m_notificationQueue.push_back(oEvent); }
        void ProcessNotificationEventImpl(const nn::nex::NotificationEvent& oEvent)
        {
            // Process for each notification.
            switch (oEvent.GetType())
            {
            case nn::nex::NotificationEvents::ParticipationEvent:
                switch (oEvent.GetSubType())
                {
                case nn::nex::ParticipationEvents::Participate:
                    NEX_DEMO_LOG("  ParticipationEvents::Participate: [PID %u] join [gathering ID %u]\n", oEvent.GetParam2(), oEvent.GetParam1());
                    m_pSample->AddMember(oEvent.GetParam1(), oEvent.GetParam2());
                    break;
                case nn::nex::ParticipationEvents::EndParticipation:
                    NEX_DEMO_LOG("  ParticipationEvents::EndParticipation: [PID %u] end [gathering ID %u]\n", oEvent.GetParam2(), oEvent.GetParam1());
                    m_pSample->RemoveMember(oEvent.GetParam1(), oEvent.GetParam2());
                    break;
                case nn::nex::ParticipationEvents::Disconnect:
                    NEX_DEMO_LOG("  ParticipationEvents::Disconnect: [PID %u] disconnect [gathering ID %u]\n", oEvent.GetParam2(), oEvent.GetParam1());
                    m_pSample->RemoveMember(oEvent.GetParam1(), oEvent.GetParam2());
                    break;
                default:
                    NEX_DEMO_LOG("  ParticipationEvents: unsupported participation event\n");
                    break;
                }
                break;
            case nn::nex::NotificationEvents::OwnershipChangeEvent:
                NEX_DEMO_LOG("  NotificationEvents::OwnershipChangeEvent: [gathering ID %u] changed owner to [PID %u]\n", oEvent.GetParam1(), oEvent.GetParam2());
                m_pSample->GatheringOwnerChanged(oEvent.GetParam1(), oEvent.GetParam2());
                break;
            case nn::nex::NotificationEvents::GatheringUnregistered:
                NEX_DEMO_LOG("  NotificationEvents::GatheringUnregistered: [gathering ID %u] unregistered\n", oEvent.GetParam1());
                m_pSample->GatheringUnregistered(oEvent.GetParam1());
                break;
            case nn::nex::NotificationEvents::AddedToGathering:
                NEX_DEMO_LOG("  NotificationEvents::AddedToGathering: added to [gathering ID %u] by [PID %u]\n", oEvent.GetParam1(), oEvent.GetSource());
                m_pSample->AddedToGahtering(oEvent.GetParam1(), oEvent.GetSource());
                break;
            default:
                NEX_DEMO_LOG("  NotificationEvents: unsupported notification event\n");
                break;
            }
        }
        TeamMatchmakeSample* m_pSample;
        nn::nex::qList<nn::nex::NotificationEvent> m_notificationQueue; // Notification queue.
    };

public:
    qBool ProcessState();
    qBool CanCommunicate() const { return m_state == STATE_COMMUNICATE; }
    void FinishCommunication() { SetState(STATE_ENDPARTICIPATION_MATCH); }
    nex_demo_util::ResultContainer GetLastResult() const { return m_lastResult; }
    const nn::nex::qList<nn::nex::PrincipalID>& GetTeamMembers() const { return m_teamMembers; }
    const nn::nex::qList<nn::nex::PrincipalID>& GetMatchMembers() const { return m_matchMembers; }
    nn::nex::PrincipalID GetTeamOwnerPid() const {return m_teamOwnerPid;}
    nn::nex::PrincipalID GetMatchOwnerPid() const {return m_matchOwnerPid;}
    nn::nex::PrincipalID GetMyPid() const {return m_myPid;}
    nn::nex::GatheringID GetJoinedTeamMatchmakeSessionId() const {return m_joinedTeamMatchmakeSession.GetID();}
    nn::nex::GatheringID GetJoinedMatchMatchmakeSessionId() const {return m_joinedMatchMatchmakeSession.GetID();}
private:
    void UnexpectedRmcFailureOccurred() { m_unexpectedRmcFailureOccurred = true; }
    void UnexpectedGatheringUnregistered() { m_unexpectedGatheringUnregistered = true; }
    void UnexpectedTeamMemberDrop() { m_unexpectedTeamMemberDrop = true; }
    nex_demo_util::ResultContainer CheckCondition();
    void ErrorOccurred(const nex_demo_util::ResultContainer& result);
    void SetState(State state){m_state = state;}
    State GetState() const {return m_state;}
    qBool CanProcessNotificationEvent() { return m_state != STATE_RMC_RESULT_WAITING; }
    void SetTeamOwnerPid(nn::nex::PrincipalID pid){m_teamOwnerPid = pid;}
    void SetMatchOwnerPid(nn::nex::PrincipalID pid){m_matchOwnerPid = pid;}

    void AddMember(nn::nex::GatheringID gid, nn::nex::PrincipalID pid)
    {
        if (gid == m_joinedTeamMatchmakeSession.GetID())
        {
            m_teamMembers.push_back(pid);
        }
        else if (gid == m_joinedMatchMatchmakeSession.GetID() || gid == m_addedGatheringId)
        {
            // Sometimes a Participate notification arrives immediately after a AddedToGathering notification and detailed information is not yet available in m_joinedMatchMatchmakeSession, so make sure it matches m_addedGatheringId.
            // 
            m_matchMembers.push_back(pid);
        }
        else
        {
            NEX_DEMO_LOG("[Ignored] Can not add a member. Unknown gathering id: %u, principal id: %u] \n", gid, pid);
        }
    }
    void RemoveMember(nn::nex::GatheringID gid, nn::nex::PrincipalID pid)
    {
        if (gid == m_joinedTeamMatchmakeSession.GetID())
        {
            m_teamMembers.erase(std::remove(m_teamMembers.begin(), m_teamMembers.end(), pid), m_teamMembers.end());
            if (m_state < STATE_COMMUNICATE)
            {
                // If a team member leaves before the match begins, handle as an error.
                UnexpectedTeamMemberDrop();
            }
        }
        else if (gid == m_joinedMatchMatchmakeSession.GetID())
        {
            m_matchMembers.erase(std::remove(m_matchMembers.begin(), m_matchMembers.end(), pid), m_matchMembers.end());
        }
        else
        {
            NEX_DEMO_LOG("[Ignored] Can not remove a member. Unknown gathering id: %u, principal id: %u] \n", gid, pid);
        }
    }
    void GatheringOwnerChanged(nn::nex::GatheringID gid, nn::nex::PrincipalID pid)
    {
        if (gid == m_joinedTeamMatchmakeSession.GetID())
        {
            SetTeamOwnerPid(pid);
        }
        else if (gid == m_joinedMatchMatchmakeSession.GetID())
        {
            SetMatchOwnerPid(pid);
        }
        else
        {
            NEX_DEMO_LOG("[Ignored] Unknown gathering owner changed. id: %u\n", gid);
        }
    }
    void GatheringUnregistered(nn::nex::GatheringID gid)
    {
        // In this sample, the matchmaking session will not be deleted as long as the local user is still participating in it.
        // Throw an error if the matchmaking session is deleted unexpectedly.
        if (gid == m_joinedTeamMatchmakeSession.GetID())
        {
            UnexpectedGatheringUnregistered();
        }
        else if (gid == m_joinedMatchMatchmakeSession.GetID())
        {
            UnexpectedGatheringUnregistered();
        }
        else
        {
            // Ignore because sometimes the notification is sent even if the player has already left.
            NEX_DEMO_LOG("[Ignored] Unknown gathering unregistered. id: %u\n", gid);
        }
    }
    void AddedToGahtering(nn::nex::GatheringID gid, nn::nex::PrincipalID pid)
    {
        nn::nex::qList<nn::nex::PrincipalID>::iterator it = std::find(m_teamMembers.begin(), m_teamMembers.end(), pid);
        if (it == m_teamMembers.end())
        {
            NEX_DEMO_LOG("Added from unknown pid: %u\n", pid);
            return;
        }
        m_addedGatheringId = gid;
    }
    void Bind();
    void Unbind();
    void MakeTeamByAutoMatchmake();
    void OnCompletionMakeTeamByAutoMatchmake(nn::nex::CallContext* pCallContext);
    void MakeMatchByAutoMatchmake();
    void OnCompletionMakeMatchByAutoMatchmake(nn::nex::CallContext* pCallContext);
    void FindMatchmakeSessionByGatheringIdDetail();
    void OnCompletionFindMatchmakeSessionByGatheringIdDetail(nn::nex::CallContext* pCallContext);
    void EndParticipationMatch();
    void OnCompletionEndParticipationMatch(nn::nex::CallContext* pCallContext);
    void EndParticipationTeam();
    void OnCompletionEndParticipationTeam(nn::nex::CallContext* pCallContext);
    void OnCompletionCloseParticipation(nn::nex::CallContext* pCallContext);
    void CloseParticipation(nn::nex::GatheringID gid);

    State m_state;                                      // The state of sequential processes.
    qBool m_finished;                                   // Whether sequential processes are complete.
    qBool m_errorOccurred;                              // Whether in a state in which an error occurred.
    qBool m_unexpectedRmcFailureOccurred;               // Whether there was an unexpected RMC failure.
    qBool m_unexpectedGatheringUnregistered;            // Whether the matchmaking session was unexpectedly deleted.
    qBool m_unexpectedTeamMemberDrop;                   // Whether any team members were unexpectedly dropped.
    nex_demo_util::ResultContainer m_lastResult;        // Final result of sequential process.
    nex_demo_util::CallContextStore m_callContextStore; // Container that saves CallContext while an asynchronous process is running. Only stores processes that may need to be canceled.
    nn::nex::PrincipalID m_myPid;                       // Local user's principal ID.
    nn::nex::PrincipalID m_teamOwnerPid;                // Owner of team matchmaking session.
    nn::nex::PrincipalID m_matchOwnerPid;               // Owner of matchmaking session for match.
    nn::nex::GatheringID m_addedGatheringId;            // The ID of the added matchmaking session.
    nn::nex::NgsFacade* m_pNgsFacade;                   // Pointer to NgsFacade object currently in a logged in state.
    nn::nex::MatchmakeExtensionClient* m_pMatchMakingClient;    // Pointer to object used to call matchmaking functions.
    MyNotificationEventHandler* m_pNotificationHandler;         // Pointer to object that handles notifications from NGS.
    nn::nex::MatchmakeSession m_joinedTeamMatchmakeSession;     // The joined team matchmaking session.
    nn::nex::qList<nn::nex::PrincipalID> m_teamMembers;         // Principal IDs of team members (including local user).
    nn::nex::MatchmakeSession m_joinedMatchMatchmakeSession;    // The joined matchmaking session for the match.
    nn::nex::qList<nn::nex::PrincipalID> m_matchMembers;        // Principal IDs of all members that have joined the match (including local user).
};

// Initialize each server feature.
void TeamMatchmakeSample::Bind()
{
    nn::nex::Credentials* pCred = m_pNgsFacade->GetCredentials();
    NEX_DEMO_ASSERT(pCred != NULL);
    m_myPid = pCred->GetPrincipalID();
    m_pMatchMakingClient = new nn::nex::MatchmakeExtensionClient();
    // Each Bind process always succeeds as long as pCred is not NULL.
    NEX_DEMO_VERIFY(m_pMatchMakingClient->Bind(pCred));
    m_pNotificationHandler = new MyNotificationEventHandler(this);
    m_pNgsFacade->RegisterNotificationEventHandler(m_pNotificationHandler);
}

// Destroy each server feature.
void TeamMatchmakeSample::Unbind()
{
    m_pNgsFacade->UnregisterNotificationEventHandler(m_pNotificationHandler);
    delete m_pNotificationHandler;
    m_pNotificationHandler = NULL;
    m_pMatchMakingClient->Unbind();
    delete m_pMatchMakingClient;
    m_pMatchMakingClient = NULL;
}

// Check connection state. Returns a user-defined error or a qResult.
nex_demo_util::ResultContainer TeamMatchmakeSample::CheckCondition()
{
    // Whether disconnected from the game server.
    nn::nex::qResult ngsResult = m_pNgsFacade->IsConnected();
    if(!ngsResult)
    {
        NEX_DEMO_LOG("Disconnected from NGS.\n");
        return nex_demo_util::ResultContainer(ngsResult);
    }

    // Whether the matchmaking session has been deleted.
    // In this sample matchmaking sessions are never actively deleted, so if this happens we need to throw an unexpected deletion error.
    if (m_unexpectedGatheringUnregistered)
    {
        NEX_DEMO_LOG("MatchmakeSession is unregistered.\n");
        return nex_demo_util::ResultContainer(USER_DEFINED_RESULT_CODE_MATCHMAKESESSION_UNREGISTERED);
    }

    // Check for unexpected failure of any asynchronous processes.
    if (m_unexpectedRmcFailureOccurred)
    {
        NEX_DEMO_LOG("Detected unexpected failure.\n");
        return nex_demo_util::ResultContainer(USER_DEFINED_RESULT_CODE_UNEXPECTED_RMC_FAILURE);
    }

    // Check for unintentional drop of any team members.
    if (m_unexpectedTeamMemberDrop)
    {
        NEX_DEMO_LOG("Detected unexpected team member drop.\n");
        return nex_demo_util::ResultContainer(USER_DEFINED_RESULT_CODE_UNEXPECTED_TEAM_MEMBER_DROP);
    }

    // Check the network connectivity.
    if(!nex_demo_util::IsNetworkConnected())
    {
        NEX_DEMO_LOG("Network connection unavailable.\n");
        return nex_demo_util::ResultContainer(USER_DEFINED_RESULT_CODE_NETWORK_DISCONNECTED);
    }
    return nex_demo_util::ResultContainer(USER_DEFINED_RESULT_CODE_SUCCESS);
}

// Create or join a team matchmaking session.
void TeamMatchmakeSample::MakeTeamByAutoMatchmake()
{
    NEX_DEMO_LOG("\nCall AutoMatchmake().\n");

    // Set up parameters for matchmaking session if creating one.
    nn::nex::MatchmakeSession matchmakeSession;

    // Set the condition for closing.
    matchmakeSession.SetOpenParticipation(true);

    // The minimum and maximum number of participants.
    matchmakeSession.SetMinParticipants(1);
    matchmakeSession.SetMaxParticipants(MATCHMAKE_MAX_TEAM_PARTICIPANTS);

    // Game mode.
    matchmakeSession.SetGameMode(MATCHMAKE_GAME_MODE);

    // Set attributes for recruiting team members.
    matchmakeSession.SetAttribute(MATCHMAKESESSION_ATTRIBUTE_INDEX_RECRUIT_MODE, RECRUIT_MODE_TEAM);

    // Matchmaking system type.
    matchmakeSession.SetMatchmakeSystemType(nn::nex::MATCHMAKE_SYSTEM_TYPE_ANYBODY);

    // Set to migrate ownership of the matchmaking session when the current owner leaves.
    matchmakeSession.SetFlag(nn::nex::GatheringFlags::MigrateOwner);

    // Use NotifyParticipationEventsToAllParticipantsReproducibly to manage participants in the matchmaking session.
    // All players, including those already joined, receive a nn::nex::ParticipationEvents::Participate notification.
    matchmakeSession.SetFlag(nn::nex::GatheringFlags::NotifyParticipationEventsToAllParticipantsReproducibly);

    // Set up auto-matchmaking parameters.
    // Because we are not using the AutoMatchmakeParam::SetSearchCriteria or AutoMatchmakeParam::SetTargetGatheringId function, the search conditions are created automatically from the parameters specified with the AutoMatchmakeParam::SetSourceMatchmakeSession function.
    // 
    nn::nex::AutoMatchmakeParam autoMatchmakeParam;
    autoMatchmakeParam.SetSourceMatchmakeSession(matchmakeSession);
    autoMatchmakeParam.SetJoinMessage(_T("team session"));

    // Change state ahead of time because the completion callback may be called when an RMC is called.
    SetState(STATE_RMC_RESULT_WAITING);

    // Run automatic matchmaking.
    nn::nex::ProtocolCallContext* pCallContext = new nn::nex::ProtocolCallContext();
    nex_demo_util::RegisterCallContextCallback(this, &TeamMatchmakeSample::OnCompletionMakeTeamByAutoMatchmake, pCallContext);

    // Only save the generated CallContext for asynchronous processes that are currently running in m_callContextStore so that they can be canceled later if necessary.
    nex_demo_util::CallContextStoreAssociator::Register(&m_callContextStore, pCallContext);
    m_pMatchMakingClient->AutoMatchmake(pCallContext, autoMatchmakeParam, &m_joinedTeamMatchmakeSession);
}

// Team auto-matchmaking completion callback.
void TeamMatchmakeSample::OnCompletionMakeTeamByAutoMatchmake(nn::nex::CallContext* pCallContext)
{
    nn::nex::qResult result = pCallContext->GetOutcome();
    if (!result)
    {
        NEX_DEMO_LOG("     AutoMatchmake() failed.\n");
        result.Trace(TRACE_ALL);
        ErrorOccurred(nex_demo_util::ResultContainer(result));
        return;
    }

    SetTeamOwnerPid(m_joinedTeamMatchmakeSession.GetOwnerPID());
    NEX_DEMO_LOG("     AutoMatchmake() succeeded.\n");
    NEX_DEMO_LOG("[PID Info >> JoinedGID:%d OwnerPID:%d MyPID:%d]\n",
                m_joinedTeamMatchmakeSession.GetID(), 
                GetTeamOwnerPid(),
                GetMyPid());

    SetState(STATE_WAIT_TEAM_GATHERED);
}

// Create or join a matchmaking session for a match.
void TeamMatchmakeSample::MakeMatchByAutoMatchmake()
{
    NEX_DEMO_LOG("\nCall AutoMatchmake().\n");

    // Set up parameters for matchmaking session if creating one.
    nn::nex::MatchmakeSession matchmakeSession;

    // Set the condition for closing.
    matchmakeSession.SetOpenParticipation(true);

    // The minimum and maximum number of participants.
    matchmakeSession.SetMinParticipants(1);
    matchmakeSession.SetMaxParticipants(MATCHMAKE_MAX_MATCH_PARTICIPANTS);

    // Game mode.
    matchmakeSession.SetGameMode(MATCHMAKE_GAME_MODE);

    // Set attributes for recruiting members for the match.
    matchmakeSession.SetAttribute(MATCHMAKESESSION_ATTRIBUTE_INDEX_RECRUIT_MODE, RECRUIT_MODE_MATCH);

    // Matchmaking system type.
    matchmakeSession.SetMatchmakeSystemType(nn::nex::MATCHMAKE_SYSTEM_TYPE_ANYBODY);

    // Set to migrate ownership of the matchmaking session when the current owner leaves.
    matchmakeSession.SetFlag(nn::nex::GatheringFlags::MigrateOwner);

    // Use NotifyParticipationEventsToAllParticipantsReproducibly to manage participants in the matchmaking session.
    // All players, including those already joined, receive a nn::nex::ParticipationEvents::Participate notification.
    matchmakeSession.SetFlag(nn::nex::GatheringFlags::NotifyParticipationEventsToAllParticipantsReproducibly);

    // Set up auto-matchmaking parameters.
    // Because we are not using the AutoMatchmakeParam::SetSearchCriteria or AutoMatchmakeParam::SetTargetGatheringId function, the search conditions are created automatically from the parameters specified with the AutoMatchmakeParam::SetSourceMatchmakeSession function.
    // 
    nn::nex::AutoMatchmakeParam autoMatchmakeParam;
    autoMatchmakeParam.SetSourceMatchmakeSession(matchmakeSession);
    autoMatchmakeParam.SetJoinMessage(_T("match session"));

    // Specify the principal IDs of the players that will join at the same time.
    // The caller is always joined, so the local principal ID does not need to be included. Even if the local principal ID is included, it is automatically excluded.
    autoMatchmakeParam.SetAdditionalParticipant(m_teamMembers);
    autoMatchmakeParam.SetGatheringIdForParticipationCheck(m_joinedTeamMatchmakeSession.GetID());   // Assumes that a team matchmaking session has been joined.

    // Change state ahead of time because the completion callback may be called when an RMC is called.
    SetState(STATE_RMC_RESULT_WAITING);

    // Run automatic matchmaking.
    nn::nex::ProtocolCallContext* pCallContext = new nn::nex::ProtocolCallContext();
    nex_demo_util::RegisterCallContextCallback(this, &TeamMatchmakeSample::OnCompletionMakeMatchByAutoMatchmake, pCallContext);

    // Only save the generated CallContext for asynchronous processes that are currently running in m_callContextStore so that they can be canceled later if necessary.
    nex_demo_util::CallContextStoreAssociator::Register(&m_callContextStore, pCallContext);
    m_pMatchMakingClient->AutoMatchmake(pCallContext, autoMatchmakeParam, &m_joinedMatchMatchmakeSession);
}

// Callback called when auto-matchmaking for the match is complete.
void TeamMatchmakeSample::OnCompletionMakeMatchByAutoMatchmake(nn::nex::CallContext* pCallContext)
{
    nn::nex::qResult result = pCallContext->GetOutcome();
    if (!result)
    {
        NEX_DEMO_LOG("     AutoMatchmake() failed.\n");
        result.Trace(TRACE_ALL);
        ErrorOccurred(nex_demo_util::ResultContainer(result));
        return;
    }

    SetMatchOwnerPid(m_joinedMatchMatchmakeSession.GetOwnerPID());
    NEX_DEMO_LOG("     AutoMatchmake() succeeded.\n");
    NEX_DEMO_LOG("[PID Info >> JoinedGID:%d OwnerPID:%d MyPID:%d]\n",
                m_joinedMatchMatchmakeSession.GetID(), 
                GetMatchOwnerPid(),
                GetMyPid());

    SetState(STATE_WAIT_MATCH_GATHERED);
}

// Get information on matchmaking session for the match.
void TeamMatchmakeSample::FindMatchmakeSessionByGatheringIdDetail()
{
    NEX_DEMO_LOG("\nCall FindMatchmakeSessionByGatheringIdDetail().\n");

    // Change state ahead of time because the completion callback may be called when an RMC is called.
    SetState(STATE_RMC_RESULT_WAITING);
    nn::nex::ProtocolCallContext* pCallContext = new nn::nex::ProtocolCallContext();
    nex_demo_util::RegisterCallContextCallback(this, &TeamMatchmakeSample::OnCompletionFindMatchmakeSessionByGatheringIdDetail, pCallContext);

    // Only save the generated CallContext for asynchronous processes that are currently running in m_callContextStore so that they can be canceled later if necessary.
    nex_demo_util::CallContextStoreAssociator::Register(&m_callContextStore, pCallContext);
    m_pMatchMakingClient->FindMatchmakeSessionByGatheringIdDetail(pCallContext, m_addedGatheringId, &m_joinedMatchMatchmakeSession);
}

// Callback invoked when information on the matchmaking session for the match has been retrieved.
void TeamMatchmakeSample::OnCompletionFindMatchmakeSessionByGatheringIdDetail(nn::nex::CallContext* pCallContext)
{
    nn::nex::qResult result = pCallContext->GetOutcome();
    if (!result)
    {
        // Fails if the user is already disconnected from the server.
        NEX_DEMO_LOG("     FindMatchmakeSessionByGatheringIdDetail() failed.\n");
        result.Trace(TRACE_ALL);
        ErrorOccurred(nex_demo_util::ResultContainer(result));
        return;
    }

    SetMatchOwnerPid(m_joinedMatchMatchmakeSession.GetOwnerPID());
    NEX_DEMO_LOG("     FindMatchmakeSessionByGatheringIdDetail() succeeded.\n");
    NEX_DEMO_LOG("[PID Info >> JoinedGID:%d OwnerPID:%d MyPID:%d]\n",
                m_joinedMatchMatchmakeSession.GetID(), 
                GetMatchOwnerPid(),
                GetMyPid());

    SetState(STATE_WAIT_MATCH_GATHERED);
}

// Leave the matchmaking session for the match.
void TeamMatchmakeSample::EndParticipationMatch()
{
    NEX_DEMO_LOG("\nCall EndParticipation().\n");

    // Change state ahead of time because the completion callback may be called when an RMC is called.
    SetState(STATE_RMC_RESULT_WAITING);
    nn::nex::ProtocolCallContext* pCallContext = new nn::nex::ProtocolCallContext();
    nex_demo_util::RegisterCallContextCallback(this, &TeamMatchmakeSample::OnCompletionEndParticipationMatch, pCallContext);

    // Only save the generated CallContext for asynchronous processes that are currently running in m_callContextStore so that they can be canceled later if necessary.
    nex_demo_util::CallContextStoreAssociator::Register(&m_callContextStore, pCallContext);
    m_pMatchMakingClient->EndParticipation(pCallContext, m_joinedMatchMatchmakeSession.GetID());
    m_joinedMatchMatchmakeSession.SetID(INVALID_GATHERINGID);
}

// Callback invoked when a player leaves the matchmaking session for the match.
void TeamMatchmakeSample::OnCompletionEndParticipationMatch(nn::nex::CallContext* pCallContext)
{
    nn::nex::qResult result = pCallContext->GetOutcome();
    if (!result && !(result == QERROR(RendezVous, SessionVoid)))
    {
        // Fails if the user is already disconnected from the server.
        NEX_DEMO_LOG("     EndParticipation() failed.\n");
        result.Trace(TRACE_ALL);
    }
    else
    {
        NEX_DEMO_LOG("     EndParticipation() succeeded.\n");
    }
    SetState(STATE_ENDPARTICIPATION_TEAM);
}

// Leave the team matchmaking session.
void TeamMatchmakeSample::EndParticipationTeam()
{
    NEX_DEMO_LOG("\nCall EndParticipation().\n");

    // Change state ahead of time because the completion callback may be called when an RMC is called.
    SetState(STATE_RMC_RESULT_WAITING);
    nn::nex::ProtocolCallContext* pCallContext = new nn::nex::ProtocolCallContext();
    nex_demo_util::RegisterCallContextCallback(this, &TeamMatchmakeSample::OnCompletionEndParticipationTeam, pCallContext);

    // Only save the generated CallContext for asynchronous processes that are currently running in m_callContextStore so that they can be canceled later if necessary.
    nex_demo_util::CallContextStoreAssociator::Register(&m_callContextStore, pCallContext);
    m_pMatchMakingClient->EndParticipation(pCallContext, m_joinedTeamMatchmakeSession.GetID());
    m_joinedTeamMatchmakeSession.SetID(INVALID_GATHERINGID);
}

// Callback invoked when a player leaves the team matchmaking session.
void TeamMatchmakeSample::OnCompletionEndParticipationTeam(nn::nex::CallContext* pCallContext)
{
    nn::nex::qResult result = pCallContext->GetOutcome();
    if (!result && !(result == QERROR(RendezVous, SessionVoid)))
    {
        // Fails if the user is already disconnected from the server.
        NEX_DEMO_LOG("     EndParticipation() failed.\n");
        result.Trace(TRACE_ALL);
    }
    else
    {
        NEX_DEMO_LOG("     EndParticipation() succeeded.\n");
    }
    SetState(STATE_FINISH);
}

// Close matchmaking session.
void TeamMatchmakeSample::CloseParticipation(nn::nex::GatheringID gid)
{
    NEX_DEMO_LOG("\nCall CloseParticipation().\n");

    nn::nex::ProtocolCallContext* pCallContext = new nn::nex::ProtocolCallContext();
    nex_demo_util::RegisterCallContextCallback(this, &TeamMatchmakeSample::OnCompletionCloseParticipation, pCallContext);

    // Only save the generated CallContext for asynchronous processes that are currently running in m_callContextStore so that they can be canceled later if necessary.
    nex_demo_util::CallContextStoreAssociator::Register(&m_callContextStore, pCallContext);
    m_pMatchMakingClient->CloseParticipation(pCallContext, gid);
}

// Callback invoked when the matchmaking session has been closed to new participants.
void TeamMatchmakeSample::OnCompletionCloseParticipation(nn::nex::CallContext* pCallContext)
{
    switch (pCallContext->GetState())
    {
    case nn::nex::CallContext::CallSuccess:
        NEX_DEMO_LOG("     CloseParticipation() succeeded.\n");
        break;
    case nn::nex::CallContext::CallCancelled:
        // Don't do anything here because the processes that will be canceled will be canceled from the ~TeamMatchmakeSample destructor.
        NEX_DEMO_LOG("     CloseParticipation() cancelled.\n");
        break;
    case nn::nex::CallContext::CallFailure:
        NEX_DEMO_LOG("     CloseParticipation() failed.\n");
        pCallContext->GetOutcome().Trace(TRACE_ALL);
        UnexpectedRmcFailureOccurred();
        break;
    default:
        NEX_DEMO_ASSERT(false);
        UnexpectedRmcFailureOccurred();
        break;
    }
}

// Transition to error state the first time an error occurs.
void TeamMatchmakeSample::ErrorOccurred(const nex_demo_util::ResultContainer& result)
{
    // Don't do anything if already in the error state.
    if (m_errorOccurred)
    {
        return;
    }

    // Cancel all asynchronous processes that are not finished.
    m_callContextStore.CancelAll();

    m_errorOccurred = true;
    m_lastResult = result;

    // After joining the matchmaking session for the match.
    if (m_joinedMatchMatchmakeSession.GetID() != INVALID_GATHERINGID)
    {
        SetState(STATE_ENDPARTICIPATION_MATCH);
    }
    // After joining the team matchmaking session.
    if (m_joinedTeamMatchmakeSession.GetID() != INVALID_GATHERINGID)
    {
        SetState(STATE_ENDPARTICIPATION_TEAM);
    }
    // Not participating in a matchmaking session.
    else
    {
        SetState(STATE_FINISH);
    }
}

// Transition the state.
qBool TeamMatchmakeSample::ProcessState()
{
    // Process notifications.
    // Do not process while asynchronous processes are running. Notifications are often received while MatchmakeExtensionClient::AutoMatchmake is running in particular, so do not process them during that time.
    if (CanProcessNotificationEvent())
    {
        while (m_pNotificationHandler->HasNotificationEvent())
        {
            m_pNotificationHandler->DequeueAndProcessNotificationEvent();
        }
    }

    // Check the communication status and other conditions.
    nex_demo_util::ResultContainer result = CheckCondition();
    if (!result)
    {
        ErrorOccurred(result);
    }

    switch (GetState())
    {
    // Auto-matchmaking for team.
    case STATE_AUTOMATCHMAKE_TEAM:
        MakeTeamByAutoMatchmake();
        break;
    // Wait for team members to assemble.
    case STATE_WAIT_TEAM_GATHERED:
        if (GetTeamMembers().size() >= MATCHMAKE_MAX_TEAM_PARTICIPANTS)
        {
            // Close the session to further participation once the maximum number of team members join.
            // Once the session is closed, new players cannot join even if a player leaves. Players in excess of the maximum will be unable to join even if the session is not closed. 
            if (GetTeamOwnerPid() == GetMyPid())
            {
                CloseParticipation(m_joinedTeamMatchmakeSession.GetID());
                SetState(STATE_AUTOMATCHMAKE_MATCH);
            }
            else
            {
                SetState(STATE_WAIT_MATCH_ADDED);
            }
        }
        break;
    // Auto-matchmaking for the match.
    case STATE_AUTOMATCHMAKE_MATCH:
        MakeMatchByAutoMatchmake();
        break;
    case STATE_WAIT_MATCH_ADDED:
        if (m_addedGatheringId != INVALID_GATHERINGID)
        {
            FindMatchmakeSessionByGatheringIdDetail();
        }
        break;
    // Wait for match members to assemble.
    case STATE_WAIT_MATCH_GATHERED:
        if (GetMatchMembers().size() >= MATCHMAKE_MAX_MATCH_PARTICIPANTS)
        {
            // Close the session to further participation once the maximum number of match members join.
            // Once the session is closed, new players cannot join even if a player leaves. Players in excess of the maximum will be unable to join even if the session is not closed. 
            if (GetMatchOwnerPid() == GetMyPid())
            {
                CloseParticipation(m_joinedMatchMatchmakeSession.GetID());
            }
            SetState(STATE_COMMUNICATE);
        }
        break;
    // Match in progress.
    case STATE_COMMUNICATE:
        {
        }
        break;
    // Leave match gathering.
    case STATE_ENDPARTICIPATION_MATCH:
        EndParticipationMatch();
        break;
    // Leave team gathering.
    case STATE_ENDPARTICIPATION_TEAM:
        EndParticipationTeam();
        break;
    // Wait for the result of the RMC processes (AutoMatchmake, EndParticipation).
    case STATE_RMC_RESULT_WAITING:
        break;
    // All processes are complete.
    case STATE_FINISH:
        m_finished = true;
        break;
    }

    return m_finished;
}

static qBool Communicate(MessageSender& sender, const TeamMatchmakeSample& sample, nn::nex::qUnsignedInt32 communicationFrame)
{
    const nn::nex::String helloMessage = _T("Hi everyone.");
    const nn::nex::String gameOverMessage = _T("Game over.");
    qBool isGameOver = false;

    // Receive.
    while (sender.HasReceivedMessage())
    {
        nn::nex::TextMessage message = sender.DequeueReceivedMessage();

        // If a ping is received, send a reply to the sender.
        if (message.GetTextBody() == helloMessage)
        {
            nn::nex::qChar8 responseMessage[64];
            snprintf(responseMessage, sizeof(responseMessage) - 1, "Hello %u", message.GetSenderID());
            sender.SendInstant(message.GetSenderID(), nn::nex::String(responseMessage));
        }
        // End game when a game over notification is received.
        else if (message.GetTextBody() == gameOverMessage)
        {
            isGameOver = true;
        }
    }

    // Send.
    // Owner sends a message to all participants, including self.
    qBool isMatchOwner = sample.GetMatchOwnerPid() == sample.GetMyPid();
    if (isMatchOwner && communicationFrame % 180 == 0)
    {
        if (communicationFrame >= COMMUNICATION_DURATION)
        {
            // Announce the game is over after a set amount of time.
            sender.SendChat(sample.GetJoinedMatchMatchmakeSessionId(), gameOverMessage);
        }
        else
        {
            // Send a greeting if a set amount of time has not elapsed.
            sender.SendChat(sample.GetJoinedMatchMatchmakeSessionId(), _T("Hi everyone."));
        }
    }
    return isGameOver;
}

// Auto-matchmaking game loop.
static qBool TeamMatchmakeGameLoop(nn::nex::NgsFacade* pNgsFacade)
{
    TeamMatchmakeSample sample(pNgsFacade);
    MessageSender sender(pNgsFacade);
    nn::nex::qUnsignedInt32 communicationFrame = 0;
    nn::nex::qResult communicationResult(true);

    for(;;)
    {
        // Starting game frame.

        // When DispatchAll is called here, packets are mainly received.
        nn::nex::Scheduler::GetInstance()->DispatchAll();

        // Handle game key input here.
        // Perform game calculations here.

        // State transitions.
        qBool finished = sample.ProcessState();

        // Check whether finished.
        if(finished)
        {
            break;
        }

        // Process for when game is in state where communication with match members is possible.
        if (sample.CanCommunicate())
        {
            // Sending and receiving.
            qBool isGameOver = Communicate(sender, sample, communicationFrame++);
            communicationResult = sender.GetFirstError();
            if (isGameOver || !communicationResult)
            {
                // End communication if the game is over or if an error occurs in the sender.
                sample.FinishCommunication();
            }
        }

        // Ideally, the Scheduler::DispatchAll function should be called two times: before and after calculating each game frame. This reduces the delay in packet processes.
        // When DispatchAll is called here, packets are mainly sent.
        nn::nex::Scheduler::GetInstance()->DispatchAll();

        // Perform game rendering here.

        // Synchronize the game in an appropriate frame.
        nn::nex::Platform::Sleep(nex_demo_util::SLEEP_DURATION);
    }

    nex_demo_util::ResultContainer result = sample.GetLastResult();

    // When an error occurs, display the error message.
    if (!communicationResult)
    {
        // Use the EULA applet/error viewer as necessary to display the error message.
        nex_demo_util::ReportQResultError(communicationResult, "An error occurred while communicating.\n");
    }
    else if (!result)
    {
        // Use the EULA applet/error viewer as necessary to display the error message.
        result.ReportError("An error occurred while sequential control.\n");
    }

    return true;
}

qInt NEXSampleMain(qInt intArg, void* ptrArg)
{
    (void)intArg;
    (void)ptrArg;

    // Connect to the Internet
    if (!nex_demo_util::NetInit())
    {
        NEX_DEMO_LOG("net init failed...\n");
        NEX_DEMO_LOG("Leaving application...\n");
        return 1;
    }

    // Prepare to use tools
    nex_demo_util::Initialize();

#if defined(CTR_TARGET)
    // Initialize the friend library.
    nn::friends::Initialize();
#endif

    // Initialization
    nn::nex::GlobalVariables::AcquireInstance();

    // Set to thread unsafe transport buffer thread mode
    nn::nex::Core::SetThreadMode(nn::nex::Core::ThreadModeUnsafeTransportBuffer);

    // Sets level of log.
    nn::nex::EventLog::GetInstance()->SetLogLevel(nn::nex::EventLog::Debug);
    nn::nex::TraceLog::GetInstance()->SetFlag(TRACE_ALWAYS);

    nn::nex::NgsFacade* pNgsFacade = new nn::nex::NgsFacade();

    // Login game loop.
    if (nex_demo_util::LoginGameLoop(pNgsFacade))
    {
        // Team matchmaking game loop.
        if (!TeamMatchmakeGameLoop(pNgsFacade))
        {
            NEX_DEMO_LOG("TeamMatchmakeGameLoop finished with error.\n");
        }

        // Logout game loop.
        NEX_DEMO_VERIFY(nex_demo_util::LogoutGameLoop(pNgsFacade));
    }

    delete pNgsFacade;

#if defined(CTR_TARGET)
    // Finalize the friend library.
    nn::friends::Finalize();
#endif

    // Finalize.
    nn::nex::GlobalVariables::ReleaseInstance();

    // Shut down utilities
    nex_demo_util::Finalize();

    // Disconnect from the Internet.
    nex_demo_util::NetTerm();

    NEX_DEMO_LOG("Leaving application...\n");

    return 0;
}

//-----------------------------------------------------------------------------
// Main Function
//-----------------------------------------------------------------------------
qInt GenericMain(qInt argc, qChar** argv)
{
    NEX_DEMO_LOG("*******************************\n");
    NEX_DEMO_LOG("       TeamMatchmake sample\n");
    NEX_DEMO_LOG("*******************************\n\n");

#if defined(CAFE_TARGET)

#else
    NEXSampleMain(argc, argv);
#endif

    NEX_DEMO_LOG("Leaving application...\n");

    return 0;
}

NEX_MAIN_WRAPPER(GenericMain)
