﻿/*---------------------------------------------------------------------------*
  (C) Copyright 1998-2014, Technologies Quazal Inc. All rights reserved.
  Quazal Net-Z, Quazal Rendez-Vous and Quazal are trademarks of Quazal.

  Copyright (C) 2011-2014 Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo and/or its licensors and are
  protected by national and international copyright laws.
  They may not be disclosed to third parties or copied or duplicated
  in any form, in whole or in part, without the prior written consent
  of Nintendo.
  The content herein is highly confidential and should be handled
  accordingly.
 *---------------------------------------------------------------------------*/

#ifndef _NEXUTIL_H_
#define _NEXUTIL_H_

#include <nex.h>

#if defined(CTR_TARGET)
#include <nn.h>
#elif defined(CAFE_TARGET)
#include <cafe.h>
#endif

#include "NexUtilDefs.h"
#include "NexUtilCallContext.h"


NEXUTIL_BEGIN

//-----------------------------------------------------------------------------
// Constant Definitions
//-----------------------------------------------------------------------------
// Amount of time to sleep
static const nn::nex::qUnsignedInt32 SLEEP_DURATION = 16;

//-----------------------------------------------------------------------------
// Function definitions
//-----------------------------------------------------------------------------


// Function to initialize utilities
// Must be called at the beginning
void Initialize();

// Function to finalize utilities
// This must be called when finalizing
void Finalize();

// This function starts Internet communication.
qBool NetInit();

// This function terminates the Internet connection.
qBool NetTerm();

// This function checks network connectivity.
qBool IsNetworkConnected();

// Log in to the game server. This process blocks until it is finished.
// See LoginApiChain or LoginSample if you do not want this process to block.
nn::nex::Credentials* GameServerLogin(nn::nex::NgsFacade* pNgsFacade,
                                      qUnsignedInt32 gameServerId = NEXDEMO_GAME_SERVER_ID,
                                      const qChar* pAccessKey = NEXDEMO_ACCESS_KEY,
                                      nn::nex::TimeInterval uiTimeout = 30000);

// Log out from game server. This process blocks until it is finished.
// See LogoutSample if you do not want this process to block.
void GameServerLogout(nn::nex::NgsFacade *pNgsFacade);

// Memory functions related to be used in the sample
void* AllocFunction(size_t size);
void FreeFunction(void *pointer);
void AcquireReservedMemory();
void ReleaseReservedMemory();
void PrintMemoryUsage();
qBool IsLowMemory();

// Displays the qResult error.
void ReportQResultError(const nn::nex::qResult& result, const char* pHeaderText);

// Displays the protocol call context error.
void ReportPCallError(const nn::nex::ProtocolCallContext& context, const char* pHeaderText = NULL);

// Endian conversion related
qUnsignedInt16 ConvertHtoLEs( qUnsignedInt16 data );
qUnsignedInt32 ConvertHtoLEl( qUnsignedInt32 data );
qUnsignedInt16 ConvertLEtoHs( qUnsignedInt16 data );
qUnsignedInt32 ConvertLEtoHl( qUnsignedInt32 data );

#if defined(CTR_TARGET)
/*
    
*/
class FriendLoginJob : public nn::nex::SingleStepJob
{
public:
    FriendLoginJob()
        : m_Event(true)
    {};
    virtual ~FriendLoginJob(){};

    // Start logging in to friend server.
    qBool StartJob(nn::nex::CallContext* pContext, qBool deleteOnCompletion=false)
    {
        // Log in to the friend server.
        m_NnResult = nn::friends::Login(&m_Event);
        if (m_NnResult.IsFailure())
        {
            NEX_DEMO_LOG("nn::friends::Login() call failed!!!\n");
            return false;
        }
        return nn::nex::SingleStepJob::Start(pContext, deleteOnCompletion);
    }

    // Check whether finished logging in to friend server.
    virtual qBool OnStep(nn::nex::qResult* result)
    {
        if (m_Event.Wait(0) == false)
        {
            return false;
        }

        m_NnResult = nn::friends::GetLastResponseResult();
        if (m_NnResult.IsSuccess())
        {
            NEX_DEMO_LOG("My PrincipalId %d\n", nn::friends::GetMyPrincipalId());
            NEX_DEMO_LOG("nn::friends::Login() succeeded.\n");
            *result = QRESULT_SUCCESS(Core, Success);
        }
        else
        {
            NEX_DEMO_LOG("nn::friends::Login() failed. Error code: %d\n", nn::friends::ResultToErrorCode(m_NnResult));
            *result = QRESULT_ERROR(Core, Unknown);
        }
        return true;
    }

    // Get the result of logging in to friend server.
    nn::Result GetNnResult() const
    {
        return m_NnResult;
    }
private:
    nn::os::Event m_Event;
    nn::Result m_NnResult;
};
#endif  // CTR_TARGET


/*
    
*/
class ApiChain
{
public:
    ApiChain()
        : m_Result()
        , m_IsCompleted(false)
    {}
    virtual ~ApiChain(){};
    virtual void Call()=0;
    virtual const nn::nex::qResult& GetResult() const { return m_Result; }
    nn::nex::qBool virtual IsCompleted() const { return m_IsCompleted; }
protected:
    virtual void Complete(const nn::nex::qResult& result) { m_IsCompleted = true; m_Result = result; }
private:
    nn::nex::qResult m_Result;
    nn::nex::qBool m_IsCompleted;
};

/*
    

        
        

*/
class LoginApiChain : public ApiChain
{
public:
    LoginApiChain(nn::nex::NgsFacade* ngsFacade, nn::nex::qUnsignedInt32 gameServerId, const nn::nex::qChar* pAccessKey, nn::nex::TimeInterval timeout)
        : m_pNgsFacade(ngsFacade)
        , m_GameServerId(gameServerId)
        , m_AccessKey(pAccessKey)
        , m_Timeout(timeout)
        , m_pCurrentCallContext(NULL)
#if defined(CTR_TARGET)
        , m_pFriendLoginJob(NULL)
#endif  // CTR_TARGET
        , m_Result()
    {
    }
    virtual ~LoginApiChain()
    {
        if(NULL != m_pCurrentCallContext)
        {
            m_pCurrentCallContext->Cancel();
        }
#if defined(CTR_TARGET)
        delete m_pFriendLoginJob;
#endif
    }

    void Call();
    const nn::nex::qResult& GetResult() const { return m_Result; }
    nn::nex::qBool IsCompleted() const { return m_IsCompleted; }

private:
#if defined(CTR_TARGET)
    void LoginFriendServer();
    void OnCompletionLoginFriendServer(nn::nex::CallContext* pCallContext);
#endif  // CTR_TARGET
    void LoginNgs();
    void OnCompletionLoginNgs(nn::nex::CallContext* pCallContext);

    nn::nex::NgsFacade* m_pNgsFacade;
    nn::nex::qUnsignedInt32 m_GameServerId;
    nn::nex::String m_AccessKey;
    nn::nex::TimeInterval m_Timeout;
    nn::nex::CallContext* m_pCurrentCallContext;
#if defined(CTR_TARGET)
    FriendLoginJob* m_pFriendLoginJob;
#endif  // CTR_TARGET
    nn::nex::qResult m_Result;
    nn::nex::qBool m_IsCompleted;
};

/*
    
        
*/
qBool LoginGameLoop(nn::nex::NgsFacade* pNgsFacade);

/*
    
        
*/
qBool LogoutGameLoop(nn::nex::NgsFacade* pNgsFacade);


NEXUTIL_END   // end of namespace nex_util


#endif  // end of _NEXUTIL_H_
