﻿/*---------------------------------------------------------------------------*
  (C) Copyright 1998-2014, Technologies Quazal Inc.  All rights reserved.
  Quazal Net-Z, Quazal Rendez-Vous and Quazal are trademarks of Quazal.

  Copyright (C)2011-2014 Nintendo  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo and/or its licensors and are
  protected by national and international copyright laws.
  They may not be disclosed to third parties or copied or duplicated
  in any form, in whole or in part, without the prior written consent
  of Nintendo.
  The content herein is highly confidential and should be handled
  accordingly.
 *---------------------------------------------------------------------------*/

#ifndef _AddToStoreOperation_H_
#define _AddToStoreOperation_H_

#include <NetZ/src/Core/DOCore/ObjDupDecl.h>
#include <NetZ/src/Core/DOCore/DOOperation.h>
#include <OnlineCore/src/Core/PseudoGlobalVariable.h>

BEGIN_NAMESPACE_NEX

//Group = System Operations

/*
Description: This DOOperation is performed every time a duplicated object is published or discovered.

Remarks: Each Station maintains its own duplicated object store that contains a list of all the active local
Duplication masters (locally published) and Duplica (locally discovered) objects. Objects are added to and removed
from the store as a result of objects being created and deleted at run time. Only objects in the store are
considered active and therefore, updated on the local Station. A Duplication master is considered active if it
exists on the local station, that is, DuplicatedObject::Publish has been called and DuplicatedObject::DeleteMainRef
has not yet been called. A Duplica is considered active if its Duplication master exists in the system and the object
is relevant on the local station, that is, it is relevant in a Duplication space (DuplicationSpace::CheckRelevance returns true)
or a Hard reference to the object exists.

When an object is added to the duplication master store the following sequence of events occurs:
    o   The user calls DuplicatedObject::Create,
    o   The SafetyExecutive is called to determine if the object may be created. If so, we continue, if not we exit.
    o   $PRODUCTNAME$ calls the object constructor,
    o   A DOHandle and MasterID are assigned to the object. The methods DuplicatedObject::IsADuplica, DuplicatedObject::IsADuplicationMaster and DuplicatedObject::GetMasterID are now valid,
    o   The user calls DuplicatedObject::Publish,
    o   The method DataSet::OperationBegin is called on all datasets,
    o   The method DuplicatedObject::OperationBegin is called,
    o   The method DuplicatedObject::InitDO is called,
    o   The duplicated object is added to the duplication master store,
    o   The method DataSet::OperationEnd is called on all datasets,
    o   If in the DDL the object has an Adapter (deprecated class) associated to it, then the AdapterEvent that corresponds to the creation of the object is triggered,
    o   The method DuplicatedObject::OperationEnd is called.

When an object is added to the duplica store the following sequence of events occurs:
    o   $PRODUCTNAME$ calls the object constructor,
    o   A DOHandle and MasterID are assigned to the object. The methods DuplicatedObject::IsADuplica and DuplicatedObject::IsADuplicationMaster are now valid,
    o   The method DataSet::OperationBegin is called on all datasets,
    o   The method DuplicatedObject::OperationBegin is called,
    o   The datasets of the object are initialized,
    o   The method DuplicatedObject::InitDO is called,
    o   The duplicated object is added to the duplica store,
    o   The method DataSet::OperationEnd is called on all datasets,
    o   If in the DDL the object has an Adapter (deprecated class) associated to it, then the AdapterEvent that corresponds to the creation of the object is triggered,
    o   The method DuplicatedObject::OperationEnd is called.

Note: 1. An AddToStoreOperation will be aborted if the Station is in the process of leaving the Session, that is, the
Station::State is either /Leaving/ or /LeavingOnFault/.

Note: 2. Note that the Adapter extension has been deprecated and will be removed in the next release. Refer to System
Operations for information on how to interface discreet class hierarchies. Events can now be caught by a system callback
or a DO callback for a specific object.

When $PRODUCTNAME$ triggers the AdapterEvent (deprecated class), if no Adapter has previously
been attached to the object then the CreateName and AttachName methods for the Adapter are called. The processing of the event,
which is controlled via the AdapterEventDispatcher class, could be processed directly or queued and processed
when the DispatchQueue::Dispatch method, or a similar method, is called. If an Adapter has already been attached
to the object then the triggering of the AdapterEvent does not result in any calls.

See Also: Net-Z Integration, System Operations, Duplicated Object Overview
*/
/* Please see man pages for details 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 */
class AddToStoreOperation: public DOOperation
{
    //Group=Methods
public:
    //{secret}
    AddToStoreOperation(DOHandle hOperationOrigin, DuplicatedObject* pDO, qBool bDuplicationMasterSelection,
                        Message* pInitializationMessage=NULL);
    //{secret}
    virtual ~AddToStoreOperation();

    /*
    Description:  Returns whether the AddToStoreOperation is an addition to the
    duplication master store.
    Returns:  true if the AddToStoreOperation is a result of a local DuplicatedObject::Publish call, in which
    case the duplicated object is added to the duplication master store, false if not or an error occurred.
    See Also: AddToDuplicaSelection
    */
    /* Please see man pages for details
		 
		 
		 
		 
		 
		 
		 
		 
		 
		 
     */
    qBool AddToDuplicationMasterSelection() const
    {
        return m_bDuplicationMasterSelection;
    }

    /*
    Description:  Returns whether the AddToStoreOperation is an addition to the
    duplica store.
    Returns:  true if the AddToStoreOperation is a result of a local discovery of a duplicated object, in which
    case the duplicated object is added to the duplica store, false if not or an error occurred.
    See Also: AddToDuplicationMasterSelection
    */
    /* Please see man pages for details
		 
		 
		 
		 
		 
		 
		 
		 
		 
		 
     */
    qBool AddToDuplicaSelection() const
    {
        return !m_bDuplicationMasterSelection;
    }

    // {secret}
    Message* GetInitializationMessage() const
    {
        return m_pInitializationMessage;
    }

    IMPL_OPERATION_COMMON_METHODS( AddToStore );

private:
#ifdef TRACE_ENABLED    // n2370
    virtual void TraceImpl(Event eEvent, qUnsignedInt64 uiTraceFlags) const;
#endif  // TRACE_ENABLED

    //{secret}
    virtual qBool CallsBackOnDataSet()
    {
        return true;
    }
    //{secret}
    virtual qBool CallsBackOnDataSet(qByte /*byIndex*/)
    {
        return true;
    }

private:
    //{secret}
    qBool m_bDuplicationMasterSelection;
    //{secret}
    Message* m_pInitializationMessage;
};

END_NAMESPACE_NEX

#endif
