﻿/*---------------------------------------------------------------------------*
  (C) Copyright 1998-2014, Technologies Quazal Inc.  All rights reserved.
  Quazal Net-Z, Quazal Rendez-Vous and Quazal are trademarks of Quazal.

  Copyright (C)2011-2014 Nintendo  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo and/or its licensors and are
  protected by national and international copyright laws.
  They may not be disclosed to third parties or copied or duplicated
  in any form, in whole or in part, without the prior written consent
  of Nintendo.
  The content herein is highly confidential and should be handled
  accordingly.
 *---------------------------------------------------------------------------*/

#ifndef _CallRegister_H_
#define _CallRegister_H_

#include <NetZ/src/Core/DOCore/ObjDupDecl.h>

#include <NetZ/src/Core/DOCore/DOCallContext.h>
#include <OnlineCore/src/Core/PeriodicJob.h>
#include <OnlineCore/src/Core/MethodCallJob.h>
#include <OnlineCore/src/Core/SystemSetting.h>



BEGIN_NAMESPACE_NEX
enum
{
    INVALID_CONTEXT_ID =0
};

template<class T> class ItemRegister: public RootObject
{
public:
    typedef typename T::ID ID;
    ItemRegister() {};
    virtual ~ItemRegister()
    {
        // The derived class is responsible to Clear itself upon
        // destruction.  The method clears calls virtual methods -
        // if we call it here, then the default method would end up
        // beeing called.
        SYSTEMCHECK(m_mapItems.empty());
    };

    void Clear()
    {
        ScopedCS oCS(Scheduler::GetInstance()->SystemLock()); //n1589:todo2:TransportLock
        typename qMap<ID, T*>::iterator it;
        while (!m_mapItems.empty())
        {
            it=m_mapItems.begin();
            Unregister(it->second);
        }
    }

    // We register the item.  We assume the Item has a correct ID
    // at this point.  A particular Register could do a little bit more.
    virtual void Register(T* pItem)
    {
        ScopedCS oCS(Scheduler::GetInstance()->SystemLock());
        ACQUIREREF(pItem);

        SYSTEMCHECK(Find(pItem->GetID())==NULL);
        m_mapItems[pItem->GetID()]=pItem;
        SYSTEMCHECK(Find(pItem->GetID())!=NULL);
        SYSTEMCHECK(Find(pItem->GetID())->GetID()==pItem->GetID());
    }

    // We unregister the item.  We assume the Item has a correct ID
    // at this point.  A particular Register could do a little bit more.
    virtual void Unregister(T* pItem)
    {
        ScopedCS oCS(Scheduler::GetInstance()->SystemLock());
        typename qMap<ID, T*>::iterator it=m_mapItems.find(pItem->GetID());
        SYSTEMCHECK(it!=m_mapItems.end());
        m_mapItems.erase(it);
        SYSTEMCHECK(Find(pItem->GetID())==NULL);

        RELEASEREF(pItem);
    }

    T* Find(const ID idItemToFind)
    {
        SYSTEMCHECK(Scheduler::GetInstance()->SystemLock().HasOwnership());
        typename qMap<ID, T*>::iterator it=m_mapItems.find(idItemToFind);
        if (it!=m_mapItems.end())
        {
            SYSTEMCHECK(it->second->GetID()==idItemToFind);
            return it->second;
        }
        else
        {
            return NULL;
        }
    }

    T* FindFirst()
    {
        // Returns the first object that has an ID equal or higher
        // than idItemToFind.
        SYSTEMCHECK(Scheduler::GetInstance()->SystemLock().HasOwnership());
        typename qMap<ID, T*>::iterator it=m_mapItems.begin();
        if (it!=m_mapItems.end())
        {
            return it->second;
        }
        else
        {
            return NULL;
        }
    }

    T* FindNext(const ID idItemToFind)
    {
        // Returns the first object that has an ID equal or higher
        // than idItemToFind.
        SYSTEMCHECK(Scheduler::GetInstance()->SystemLock().HasOwnership());
        typename qMap<ID, T*>::iterator it=m_mapItems.upper_bound(idItemToFind);
        if (it!=m_mapItems.end())
        {
            SYSTEMCHECK(it->second->GetID()>idItemToFind);
            return it->second;
        }
        else
        {
            return NULL;
        }
    }

    T* AcquireRef(const ID idItem)
    {
        ScopedCS oCS(Scheduler::GetInstance()->SystemLock());
        T* pItem=Find(idItem);
        if (pItem!=NULL)
        {
            ACQUIREREF(pItem);
        }
        return pItem;
    }

#ifdef TRACE_ENABLED    // n2370
    void Trace(qUnsignedInt64 uiTraceFlags=TRACE_ALWAYS)
    {
        typename qMap<ID, T*>::iterator it=m_mapItems.begin();
        while (it!=m_mapItems.end())
        {
            it->Trace(uiTraceFlags);
            ++it;
        }
    }
#endif  // TRACE_ENABLED

    class Iterator: public RootObject
    {
    public:
        Iterator(ItemRegister<T>* pAssociatedItemRegister=NULL)
        {
            m_pAssociatedItemRegister=pAssociatedItemRegister;
            m_pItem=NULL;
            GotoStart();
        }
        virtual ~Iterator()
        {
            if (m_pItem!=NULL)
            {
                RELEASEREF(m_pItem);
            }
        }
        void GotoStart(ItemRegister<T>* pAssociatedItemRegister)
        {
            m_pAssociatedItemRegister=pAssociatedItemRegister;
            GotoStart();
        }

        T* GetItem()
        {
            return m_pItem;
        }
        T* GetItem() const
        {
            return m_pItem;
        }
        void SetCurrentItem(T* pNewItem)
        {
            if (pNewItem!=NULL)
            {
                ACQUIREREF(pNewItem);
            }
            if (m_pItem!=NULL)
            {
                RELEASEREF(m_pItem);
            }
            m_pItem=pNewItem;
        }
        void GotoStart()
        {
            T* pNewItem=NULL;
            if (m_pAssociatedItemRegister!=NULL)
            {
                ScopedCS oCS(Scheduler::GetInstance()->SystemLock());
                pNewItem=m_pAssociatedItemRegister->FindFirst();
                if (pNewItem)
                {
                    ACQUIREREF(pNewItem);
                }
            }
            SetCurrentItem(pNewItem);
            if (pNewItem)
            {
                RELEASEREF(pNewItem);
            }
        }
        void Advance()
        {
            if (m_pItem!=NULL)
            {
                SYSTEMCHECK(m_pAssociatedItemRegister!=NULL);
                T* pNewItem=NULL;
                {
                    ScopedCS oCS(Scheduler::GetInstance()->SystemLock());
                    pNewItem=m_pAssociatedItemRegister->FindNext(m_pItem->GetID());
                    if (pNewItem)
                    {
                        ACQUIREREF(pNewItem);
                    }
                }
                SetCurrentItem(pNewItem);
                if (pNewItem)
                {
                    RELEASEREF(pNewItem);
                }
            }
        }
        qBool EndReached()
        {
            return m_pItem==NULL;
        }
        Iterator& operator++()
        {
            Advance();
            return *this;
        }
        Iterator& operator++(qInt iDummy)
        {
            DUMMY_PARAM(iDummy);
            Advance();
            return *this;
        }
        T* operator->() const
        {
            return GetItem();
        }
        T* operator->()
        {
            return GetItem();
        }


    private:
        ItemRegister<T>* m_pAssociatedItemRegister;
        T *m_pItem;
    };

protected:
    // With appropriate locking, the map can be access directly by subclasses.
    qMap<ID, T*> m_mapItems;
    typedef typename qMap<ID, T*>::iterator iterator;
};

class CallRegister: public ItemRegister<DOCallContext>
{
public:

    CallRegister();
    virtual ~CallRegister();

    static CallRegister& GetInstanceRef();

    void Start();
    void CheckExpiredCalls(qInt iDummy);

    virtual void Register(DOCallContext* pContext);
    virtual void Unregister(DOCallContext* pContext);

#ifdef TRACE_ENABLED    // n2370
    void Trace(qUnsignedInt64 uiFlags=TRACE_ALWAYS);
#endif  // TRACE_ENABLED

    void CancelCallToStation(DOHandle hStation);
    void QueueCancelCallToStation(DOHandle hStation);
    void CancelPendingCalls();

    void CancelExpiredCalls();

    void CancelPeriodicJobs();

    // Accessors for FetchContextes.
    void SignalRelevantFetchContextes(DOHandle hFetchedObject, DOCallContext::Outcome oOutcome);
    qBool FetchInProgress(DOHandle hFetchedObject, DOHandle hTargetStation=INVALID_DOHANDLE);
    FetchContext* GetFetchContext(DOHandle hFetchedObject, DOHandle hTargetStation=INVALID_DOHANDLE);

    // Accessors for MigrationContextes.
    qBool MigrationInProgress(DOHandle hMigratingObject, DOHandle hTargetStation=INVALID_DOHANDLE);
    MigrationContext* GetMigrationContext(DOHandle hMigratingObject, DOHandle hTargetStation=INVALID_DOHANDLE);

private:

    DOCallContext::ID GetNextID();

private:

    DOCallContext::ID m_idNextValid;
    MethodCallJob<CallRegister, qInt, PeriodicJob> *m_pCheckExpiredCallsJob;
    static SystemSetting    s_ssCheckExpiredCallsPeriod;
};

END_NAMESPACE_NEX

#endif
