﻿/*---------------------------------------------------------------------------*
  (C) Copyright 1998-2014, Technologies Quazal Inc. All rights reserved.
  Quazal Net-Z, Quazal Rendez-Vous and Quazal are trademarks of Quazal.

  Copyright (C) 2011-2014 Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo and/or its licensors and are
  protected by national and international copyright laws.
  They may not be disclosed to third parties or copied or duplicated
  in any form, in whole or in part, without the prior written consent
  of Nintendo.
  The content herein is highly confidential and should be handled
  accordingly.
 *---------------------------------------------------------------------------*/

#ifndef _DuplicatedObject_H_
#define _DuplicatedObject_H_

#include <NetZ/src/Core/DOCore/ObjDupDecl.h>
#include <NetZ/src/Core/DOCore/RefTemplate.h>
#include <NetZ/src/Core/DOCore/SelectionIteratorTemplate.h>
#include <NetZ/src/Core/DOCore/MethodID.h>
#include <NetZ/src/Core/DOCore/DupSpace.h>
#include <NetZ/src/Core/DOCore/MemberContainer.h>
#include <NetZ/src/Core/DOCore/DOSelections.h>
#include <NetZ/src/Core/DOCore/WKHandle.h>
#include <OnlineCore/src/Core/LogicalClock.h>
#include <OnlineCore/src/Transport/Interface/EndPoint.h> /* For MSG_RELIABLE */
#include <NetZ/src/Core/DOCore/DORole.h>
#include <NetZ/src/Core/DOCore/Selection.h>
#include <OnlineCore/src/Platform/Core/qList.h>
#include <OnlineCore/src/Core/Operation.h>
#include <OnlineCore/src/Core/StateMachine.h>
#include <NetZ/src/Core/DOCore/MasterStationRef.h>
#include <OnlineCore/src/Transport/Interface/SubStreamID.h>


BEGIN_NAMESPACE_NEX

class DOClass;
class DataSet;
class DOOperation;
class FetchContext;
class MigrationContext;
class OperationManager;

class Variable;

class UpdateDataSetOperation;
class ChangeDupSetOperation;
class AddToStoreOperation;
class RemoveFromStoreOperation;
class ChangeMasterStationOperation;

//group=DDL Output

/*
$ifdef VOICE$
{secret}
$else$
Description: Represents the distance between two objects when the distance is unknown or undefined.
Remarks: This value will be used if, for Distance-based dead reckoning or for Distance-based Migration,
the observer or how to calculate the distance between the observer and an object is not defined.
See Also:  Dead Reckoning Overview, DuplicatedObject::ComputeDistance, Station::RegisterObserver,
Distance-based Migration
$endif$
*/
#define UNKNOWN_DISTANCE QREAL(-1.0)

/*
$ifdef VOICE$
{secret}
$else$
Description: Generates the name of the duplication object's base class.
Remarks:
For example, lets say a duplicated object is declared like this in a DDL file :

      doclass GameObject {
          // DDL declarations
      };

Then the user-defined /GameObject/ class is declared like this :

      class GameObject : public DOCLASS(GameObject)  {
          // User specific variables and methods
      };
$endif$
*/


/* Please see man pages for details 
 
 

*/

/* Please see man pages for details 
 
 
 
      
      
 
 
      
      
 
 
 */
#define DOCLASS(C) _DO_##C


/*  

*/
/*
Description: The root of all user-defined duplicated objects.

Remarks: If a DOClass is defined in the DDL, the corresponding DOClass generated by the
DDL compiler will ultimately inherit from the DuplicatedObject class.
The DuplicatedObject class contains regular methods, inherited by
every user-defined doclass.  It also defines system callbacks.  $PRODUCTNAME$ system
callbacks have the semantics of a virtual method meaning that if they are
overridden by the user doclass, the overriding method will be called whenever applicable.

$ifdef NETZ$
In NetZ, all DuplicatedObjects are global objects unless specified nonglobal in their DDL definitions.
When global objects are created and published, they will be instantiated on the local station and
copies, or Duplica s, of the object published on all the remote stations connected to the session.
When an object is deleted, via DeleteMainRef, all Duplica s of the object are also deleted.
$endif$

$ifdef VOICE$
In NetZ, all DuplicatedObjects are global objects unless specified nonglobal in their DDL definitions.
When global objects are created and published, they will be instantiated on the local station and
copies, or Duplica s, of the object published on all the remote stations connected to the session.
When an object is deleted, via DeleteMainRef, all Duplica s of the object are also deleted.
$endif$

See Also: Actions Overview, Fault Tolerance Overview
*/
/* Please see man pages for details 
 
 
 
 
 
 
 
 
 
 
 */
// {tag:DuplicatedObject class}
class DuplicatedObject: public StateMachine
{
    //group=DO Methods
public:

    //Remarks: Duplicated objects should never be directly instantiated.  DuplicatedObject::Create should
    //be used instead.
    /* Please see man pages for details 
     
     
     
     */
    DuplicatedObject();

    /*
    Remarks: The correct way to destroy a duplicated object is through the method
    DuplicatedObject::DeleteMainRef. When a duplicated object is destroyed, the datasets and
    well-known handles associated with it are also destroyed. The duplicated object destructor
    has the semantics of a virtual method and, therefore, it can be rewritten in a user subclass.

    Note: The duplicated object destructor will not be called on an object until all references, both
    Hard reference s and Soft reference s, to the object have been deleted.

    See Also: DataSet, WKHandle, DeleteMainRef
    */
    /* Please see man pages for details 
     
     
     
     
     
     
     
     
     
     */
    virtual ~DuplicatedObject();

    /*
    Summary: Returns a DOHandle to the duplicated object that refers to the local instance
    of the duplicated object.
    Returns     : The DOHandle to the duplicated object or INVALID_DOHANDLE if the call signalled
    an error. The error SYSTEMERROR_GEN_INVALID_OPERATION is possible.
    See also: GetHandleValue
    */
    /* Please see man pages for details 
     
     
     
     
     */
    DOHandle GetHandle() const
    {
        if (m_dohMyself.GetDOID()==0)
        {
            SIGNAL_SYSTEMERROR(SYSTEMERROR_GEN_INVALID_OPERATION_WARNING,0);
            return INVALID_DOHANDLE;
        }
        else
        {
            return m_dohMyself;
        }
    }

    /*
    Summary: Returns a qUnsignedInt32 that represents the duplicated object instance.
    Returns     : A qUnsignedInt32 that represents the duplicated object or INVALID_DOHANDLE if the
    call signaled an error. The error SYSTEMERROR_GEN_INVALID_OPERATION is possible.
    See also: GetHandle
    */
    /* Please see man pages for details 
     
     
     
     
     */
    qUnsignedInt32 GetHandleValue() const
    {
        return GetHandle();
    }

#ifdef DEBUG_VERSION
    // {secret}
    void CheckStateAccess() const;
#else
    // {secret}
    void CheckStateAccess() const {}
#endif

    // {Secret}
    void SetDOClassID(DOClassID uiClassID)
    {
        CheckStateAccess();
        m_dohMyself.SetDOClassID(uiClassID);
    }
    // {Secret}
    void SetDOID(DOID uiObjectID)
    {
        CheckStateAccess();
        m_dohMyself.SetDOID(uiObjectID);
    }
    // {Secret}
    void SetHandle(DOHandle hNewHandle)
    {
        CheckStateAccess();
        m_dohMyself=hNewHandle;
    }

    /*
    Description: Returns whether or not a DuplicatedObject instance is a Global duplicated object.
    Returns     : true if this duplicated object is a Global duplicated object, false if not or an error occurred.
    Remarks: A Global duplicated object is either a Well-known duplicated object or an object that has the
    DDL property /globaldo/ (refer to DDL Properties for details). These objects are known globally across
    the Session and are guaranteed to be duplicated to all stations participating in the session.

    $ifdef NETZ$
    Note: All NetZ duplicated objects are automatically Global duplicated object s unless otherwise specified using
    the nonglobal property in the DDL declaration.
    $endif$

    $ifdef VOICE$
    Note: All NetZ duplicated objects are automatically Global duplicated object s. unless otherwise specified using
    the nonglobal property in the DDL declaration.
    $endif$

    See Also: IsAWellKnownDO, IsACoreDO, IsAUserDO, DDL Properties
    */
    /* Please see man pages for details 
     
     
     
     
     
     
     
     
     
     
     
     */
    qBool IsGlobal() const;

    // {secret}
    virtual qBool HasGlobalDOProperty() const
    {
        return false;
    }

    // {secret}
    virtual qBool HasForcedNonGlobalProperty() const
    {
        return false;
    }

    /*
    Description: Returns the class name of the DuplicatedObject as a string.
    Returns     : A string that represents the class name of the DuplicatedObject.
    Remarks: Gets the class name of the duplicated object instance as a string in your traces.
    Example:
        qReal DuplicatedObject::ComputeDistance(DuplicatedObject * pObservedDuplicatedObject){
            _tprintf(_T("Computing distance for Duplicated Object of class %s"),
                    pObservedDuplicatedObject->GetClassNameString());
            [...]
        }
    */
    /* Please see man pages for details 
     
     
     
     
     
     */
    const qChar* GetClassNameString() const;

    /*
    Description: Returns the class name of the specified dataset associated to this DuplicatedObject.
    Parameters:
    byDatasetIndex: The index of the dataset.
    Returns     : The class name of the dataset associated to this DuplicatedObject specified by the given index as a string.
    Remarks:
    This method is useful for obtaining the dataset class name in traces. $PRODUCTNAME$ sequentially assigns a
    dataset index to each dataset associated to a given duplicated object as declared in the DDL file. The index is a
    positive integer starting at one.
    */
    /* Please see man pages for details 
     
     
     
     
     
     
     
     
     
     
     
     */
    const qChar* GetDatasetNameString(qByte byDatasetIndex) const;

    // {secret}
    // TODO.  Document.  CR212.
    qBool FormatVariableValue(Variable* pDataset, Variable* pVariable, String *pstrResult) const;

    /*
    Description: Returns the DOClassID of the duplicated object.
    Returns     : The DOClassID of the duplicated object. This ID is generated by the DDL compiler.
    See also: DOCLASSID
    */
    /* Please see man pages for details 
     
     
     
     
     
     
     
     
     */
    DOClassID GetDOClassID() const
    {
        return m_dohMyself.GetDOClassID();
    }

    /*
    Description: Returns whether or not a DuplicatedObject instance is an instance
    of a specified doclass.
    Parameters:
    idDOCID:  The DOClassID of the DuplicatedObject.
    Returns     : true if the DuplicatedObject instance has this DOClassID, false if not or an error occurred.
    Remarks: Use the DOCLASSID macro to retrieve the DOClassID for a given DOClass.
    Example:
        NamedObject::SelectionIterator  itNamedObject;
        itNamedObject.GotoStart();
        while(! (itNamedObject.EndReached())){
            if (ItNamedObject->IsA(DOCLASSID(Avatar))){
                itNamedObject++;
            }
        }

    See Also: IsAKindOf
    */
    /* Please see man pages for details 
     
     
     
     
     
     
     
     
     
     */
    qBool IsA(DOClassID idDOCID) const
    {
        return m_dohMyself.GetDOClassID() == idDOCID;
    }

    /*
    Description: Returns whether or not a DuplicatedObject instance is an instance or
    a descendant of the specified DOCLASS.
    Parameters:
    idDOCID: the DOClassID of the DuplicatedObject.
    Returns     : true if the DuplicatedObject instance is an instance of or a descendant class
    of this DOClassID, false if not or an error occurred.
    Remarks: Use the DOCLASSID macro to get an instance of this class.
    Example:
        NamedObject::SelectionIterator  itNamedObject;
        itNamedObject.GotoStart();
        while(! (itNamedObject.EndReached())){
            if (ItNamedObject->IsAKindOf(DOCLASSID(Avatar))){
                itNamedObject++;
            }
        }

    See Also: IsA
    */
    /* Please see man pages for details 
     
     
     
     
     
     
     
     
     
     */
    qBool IsAKindOf(DOClassID idDOCID) const;

    // {Secret}
    DOID GetDOID() const
    {
        return m_dohMyself.GetDOID();
    }

    /*
    Summary: Propagates the contents of all datasets to an object's duplicas or only the contents
    of the specified dataset.
    Parameters:
    refDataset:  An optional pointer to the specific dataset that is to be updated.
    tUpdateTime: (optional) Specifies the point in (absolute) time at which the
    Update is valid. Otherwise, the current time as given by the SessionClock is
    used. This parameter should be set when the time stamp is changed as done when Lookahead
    or Loopback techniques are applied on the Duplication master.
    Remarks:  This method must be called on a Duplication master.  It has the semantics of a virtual
    method and therefore, it can be called on a base class and all the derived class datasets will also
    be updated.  The call can be made
    with or without the parameter to specify which of the duplicated object's datasets should
    be updated. For example, suppose we have a Position dataset and the m_dsPosition dataset
    member of type Position is associated with the Sphere doclass in the DDL file, then it
    is possible to update just that Position dataset as follows:

          // Assuming that all we have is a pointer to our Sphere duplicated object
          poSphere->Update(poSphere->m_dsPosition);

    Alternatively, if we have an extrapolated dataset and a MainLoop method that is called periodically on every
    game object we could update and refresh each Sphere object's datasets via:

            void Sphere::MainLoop() {
                if (IsADuplicationMaster()) {
                    // compute the physics and then update the objects position.
                    Update();
                } else {
                    // Refresh the position.
                    Refresh(m_Pos);
                }
                // Displays the object at the correct position.
            }

    If a station cannot directly communicate with a station to which it wishes to send a message
    $PRODUCTNAME$ will automatically route the message via another station. This effectively deals with
    the problem of stations being unable to communicate with each other because they
    have no common transport type.

    Note: Depending on the dataset update policy, Update does not necessary
    generate a network message. If some of the datasets are extrapolated, Update must be called
    regularly ( /e.g./ after DoPhysics)  even if the contents of the datasets did not change. This
    is necessary to maintain a coherent prediction model for the
    extrapolated datasets.

    Returns     : true if the object was successfully updated, false if not or an error occurred.
    The error SYSTEMERROR_OBJDUP_INVALID_ON_DUPLICA will be signalled if this method is called on a Duplica,
    SYSTEMERROR_GEN_INVALID_PARAMETER if an invalid dataset is specified and
    SYSTEMERROR_MISSING_DATASET_CALLBACK if no callback is defined for a custom update policy.
    See Also: DataSet::SetLoopbackDelay, Refresh
    */
    /* Please see man pages for details 
     
     
     
     
     
     
     
     
     
     
     
     
     
     
     
     
     
     
     
     
     
     
     
     
     
     
     
     
     
     */
    qResult Update()
    {
        return UpdateImpl(NULL, Time::GetSessionTime());
    }

    /* Please see man pages for details 
     
     
     
     
     
     
     
     
     
     
     
     
     
     
     
     
     
     
     
     
     
     
     
     
     
     
     
     
     
     */
    qResult Update(const Time& tUpdateTime)
    {
        return UpdateImpl(NULL, tUpdateTime);
    }

    /* Please see man pages for details 
     
     
     
     
     
     
     
     
     
     
     
     
     
     
     
     
     
     
     
     
     
     
     
     
     
     
     
     
     */
    qResult Update(DataSet& refDataset)
    {
        return UpdateImpl(&refDataset, Time::GetSessionTime());
    }

    /* Please see man pages for details 
     
     
     
     
     
     
     
     
     
     
     
     
     
     
     
     
     
     
     
     
     
     
     
     
     
     
     
     
     
     
     */
    qResult Update(DataSet& refDataset, const Time& tUpdateTime)
    {
        return UpdateImpl(&refDataset, tUpdateTime);
    }



    /*
    Summary: Uses the content of the buffer to refresh the variables of all or a specified buffered dataset of a Duplica.
    Parameters:
    refDataset: An optional pointer to the dataset that is to be refreshed.
    tRefreshTime: (optional) Specifies the point in (absolute) time at which to refresh the value.
    Otherwise the current time as given by the SessionClock is used.
    Returns     : true if the values of the object were successfully refreshed, false if not or an error occurred.
    The error SYSTEMERROR_GEN_INVALID_OPERATION will be signalled
    if the dataset is not buffered or extrapolated,
    SYSTEMERROR_OBJDUP_INVALID_ON_MASTER if this method is called on a Duplication master, and
    SYSTEMERROR_GEN_INVALID_PARAMETER if refDataset is invalid or does not exist.
    Remarks:  This method must be called on the buffered datasets of a Duplica to
    refresh the value of the dataset variables. If an extrapolation filter
    is used, this method will refresh the dataset variables with an
    extrapolation of their value. This method is only valid for the buffered datasets
    of a Duplica.

    If we have an extrapolated dataset and a MainLoop method that is called periodically on every game object we
    could update and refresh each Sphere object's datasets via:

        void Sphere::MainLoop() {
            if (IsADuplicationMaster()) {
                // compute the physics and then update the objects position.
                Update();
            } else {
                // Refresh the position.
                Refresh(m_Pos);
            }
            // Displays the object at the correct position.
        }

    See Also: Update
    */
    /* Please see man pages for details 
     
     
     
     
     
     
     
     
     
     
     
     
     
     
     
     
     
     
     
     
     
     */
    qBool Refresh()
    {
        return RefreshImpl(NULL, Time::GetSessionTime());
    }

    /* Please see man pages for details 
     
     
     
     
     
     
     
     
     
     
     
     
     
     
     
     
     
     
     
     
     */
    qBool Refresh(const Time& tRefreshTime)
    {
        return RefreshImpl(NULL, tRefreshTime);
    }

    /* Please see man pages for details 
     
     
     
     
     
     
     
     
     
     
     
     
     
     
     
     
     
     
     
     
     
     
     */
    qBool Refresh(DataSet& refDataset)
    {
        return RefreshImpl(&refDataset, Time::GetSessionTime());
    }

    /* Please see man pages for details 
     
     
     
     
     
     
     
     
     
     
     
     
     
     
     
     
     
     
     
     
     
     */
    qBool Refresh(DataSet& refDataset, const Time& tRefreshTime)
    {
        return RefreshImpl(&refDataset, tRefreshTime);
    }


    /**********
     *
     * Callbacks.
     *
     **********/

    //{secret}
    qBool CallApproveFaultRecovery();
    //{secret}
    qBool CallApproveEmigration(qUnsignedInt32 uiReason);

    /**********
     *
     * Default callback implementation...
     *
     **********/

    // group=DO System Callbacks

    /*
    Summary: Called by $PRODUCTNAME$ to determine whether a duplicated object
    accepts to become a Duplication master after the station that controlled
    it has failed.
    Returns     : true if a new host can and does become the new Duplication master and gains control of the duplicated object,
    false if the duplicated object is destroyed by $PRODUCTNAME$ or an error occurs.
    Remarks: When a fault occurs, this callback is called on all instances (Duplication master and Duplica s) of the
    given object, meaning all instances of the object must return the same result. The following code shows implementation
    of the ApproveFaultRecovery system callback, which dictates that Avatar objects are fault-tolerant:

            class Avatar : Public DOCLASS(Avatar) {
                public:
                    Avatar();
                    ~Avatar();
                    qBool ApproveFaultRecovery() {
                        return true;
                    }
                    //other application specific methods and variables
            };

    Note: By default, if the game does not implement this system callback,
    fault recovery is disabled for the given DOClass.
    See Also: Fault Tolerance Overview, ApproveEmigration
    */
    /* Please see man pages for details 
     
     
     
     
     
     
     
     
     
     
     
     
     
     
     
     
     
     
     
     
     
     
     */
    virtual qBool ApproveFaultRecovery();

    /*
    Description: Called by $PRODUCTNAME$ to determine if a duplicated object allows Emigration.

    $ifdef ETERNA$
    Parameters:
    uiReason: The reason for which an emigration is requested, which is either
    MIGRATION_REASON_LEAVING_SESSION or MIGRATION_REASON_LOAD_BALANCING.
    The uiReason parameter is usually set internally. However, if you want to control an object's
    response to the ApproveEmigration call, you can add a member variable to the given object's class and use the
    value of this variable to decide whether this method returns true or false.
    $else$
    Parameters:
    uiReason: The reason for which an emigration is requested, which for NetZ can only be
    MIGRATION_REASON_LEAVING_SESSION. This value is set internally by $PRODUCTNAME$.
    $endif$

    Returns     :  true if the given duplicated object can migrate, meaning a new host will become the Duplication master
    of the object, false if $PRODUCTNAME$ will refuse any attempt to migrate the object or an error occurred.
    Remarks: If this system callback returns true, when the Station leaves the session, /i.e./ the uiReason is
    MIGRATION_REASON_LEAVING_SESSION, the objects will be migrated to another station in the Session. if the
    station quits the session cleanly then it migrates its objects itself to another station in the session,
    which is usually the first station it finds once the termination sequence commences. If you want to control the
    station to which objects are migrated then call DuplicatedObject::AttemptEmigration before the station quits the game.
    On the other hand, if the station leaves the session due to a fault then the station to which objects will be migrated
    is determined by the PromotionReferee.

    The following code shows implementation of the ApproveEmigration system callback, which dictates
    that Avatar objects can migrate:

            class Avatar : Public DOCLASS(Avatar) {
                public:
                    Avatar();
                    ~Avatar();
                    qBool ApproveEmigration(qUnsignedInt32 uiReason) {
                        return true;
                    }
                    //other application specific methods and variables
            };

    Note: By default, if a class does not implement this system callback, Emigration
    is disabled for the given DOClass.

    See Also: Fault Tolerance Overview, Object Migration Overview, RemoveFromStoreOperation, ApproveFaultRecovery
    */
    /* Please see man pages for details 
     
     
     
     
     
     
     
     
     
     
     
     
     
     
     
     
     
     
     
     
     
     
     */
    virtual qBool ApproveEmigration(qUnsignedInt32 uiReason);

    /*
    Summary: Called by $PRODUCTNAME$ when the duplicated object has been initialized.
    Remarks: When InitDO is called, the Duplication role of the
    DuplicatedObject is fully defined.  If it is a Duplica, all its
    datasets have been appropriately initialized.

    This section explains how to use the InitDO system callback.

    Suppose we want to create a game where Avatars appear at random
    locations.  We also want to render the graphical representation of the
    Avatars at the same location in 3D space on all the stations
    immediately when they join the session.

    Lets say the object 'Avatar' contains a position dataset defined in the DDL file as follows:

            dataset Position {
                qReal x;
                qReal y;
                qReal z;
            };

    When creating an Avatar via DuplicatedObject::Create, a Duplication
    master is created, so all we need to do before rendering it is set the
    values of its position dataset. When the Avatar is published (via
    DuplicatedObject::Publish), its Duplica appear on all the stations that require a copy of the object
    which then must be rendered at the correct position. So setting the
    Avatar position from within the InitDO function would be performed as follows:

            void Avatar::InitDO(void) {
                if (IsADuplicationMaster()) {
                    Position.x = SomeRandomValue();
                    Position.y = SomeRandomValue();
                    Position.z = SomeRandomValue();
                }
                RenderAtCurrentPosition();
            }

    Note: For this example, because the datasets have default values when
    the Avatar object is instantiated, setting the Avatar position from
    within the constructor will *not* work.

    See Also: DuplicatedObject::Create, DuplicatedObject::Publish, IsADuplicationMaster, IsADuplica
    */
    /* Please see man pages for details 
     
     
     
     
     
     
     
     */
    virtual void InitDO() {}

    /*
    Summary: Called by $PRODUCTNAME$ to determine the distance between two duplicated objects.
    Parameters:
    pOtherObject: A pointer to the duplicated object which you want to compute the
    distance to.
    Remarks: When distance-based dead reckoning is invoked, the desired precision of the extrapolation
    depends on a distance between the duplicated objects. The distance is a value
    specific to the game computed via this system callback and is typically the distance
    between the position of the objects in the virtual world.

    For distance-based dead reckoning $PRODUCTNAME$ needs to compute the distance
    between the registered observer (register via Station::RegisterObserver
    and get via Station::GetObserver) and the Duplica of the
    object for which it is doing the distance-based dead reckoning.  Thus, $PRODUCTNAME$
    calls this method on the Duplica of the observer and passes to the function a pointer
    to the duplicated object for which it is determining the allowable error in the
    extrapolation. The distance must be a positive value, or the special value
    UNKNOWN_DISTANCE. The default implementation returns UNKNOWN_DISTANCE.

    The following is an example that computes the distance between two
    Spheres in a virtual world.  The context of this example is found in the
    Dead Reckoning Overview.

            qReal Sphere::ComputeDistance(DuplicatedObject * pObservedDuplicatedObject) {
                if (pObservedDuplicatedObject->IsA(DOCLASSID(Sphere))) {
                    return sqrt( sqr(m_Pos.x-((Sphere*)pObservedDuplicatedObject)->m_Pos.x)
                                 +sqr(m_Pos.y-((Sphere*)pObservedDuplicatedObject)->m_Pos.y));
                } else {
                    return UNKNOWN_DISTANCE ;
                }
            }

    See Also: Dead Reckoning, Dead Reckoning Overview, ErrorToleranceFunction, UNKNOWN_DISTANCE,
    Station::RegisterObserver, PHBDRParameters
    */
    /* Please see man pages for details 
     
     
     
     
     
     
     
     
     
     
     
     
     
     
     */
    virtual qReal ComputeDistance(DuplicatedObject* pOtherObject);

    // {secret}
    virtual void CallOperationOnDatasets(DOOperation* /*pOperation*/, Operation::Event /*eEvent*/);


    // {secret}
    virtual qBool ValidOperation(DOOperation* pOperation);

    /*
    Description: Called by $PRODUCTNAME$ each time a DOOperation begins.
    Parameters:
    pOperation: A pointer to the DOOperation that just began.
    Remarks: This method is called at the beginning of a DOOperation with a reference to
    the DOOperation that just began as a parameter.  Switch this parameter to find out what
    type of operation it is and then implement your functionality. See
    DuplicatedObject::OperationEnd for an example of how to use this method.
    See Also:  System Operations, OperationEnd, OperationManager
    */
    /* Please see man pages for details 
     
     
     
     
     
     
     
     
     
     
     
     
     */
    virtual void OperationBegin(DOOperation* pOperation);


    /*
    Description: Called by $PRODUCTNAME$ each time a DOOperation ends.
    Parameters:
    pOperation: The DOOperation that just terminated.
    Remarks: This method is called at the end of a DOOperation with a reference to the DOOperation
    that just terminated as a parameter.  Do a switch/case on this parameter to find out what type of operation
    it is and to implement your functionality. The following is an example to capture the DOOperation
    ChangeMasterStationOperation. Thus, each time an AISphere object is locally promoted to the
    role of Duplication master, due to the Migration of the object, the local station will
    take over the control of its AI.
    Example:

            void AISphere::OperationEnd(DOOperation* pOperation) {
                switch (pOperation->GetType()) {
                    case Operation::ChangeMasterStation:
                        if ( ((ChangeMasterStationOperation*) pOperation)->IsAPromotion()){
                            // Take over control of AI
                        }
                    break;
                }
            }

    Note: To avoid deadlocks, while within an OperationEnd context you should not perform any blocking calls,
    such as CallContext::Wait.

    See Also: System Operations, OperationBegin, OperationManager
    */
    /* Please see man pages for details 
     
     
     
     
     
     
     
     
     
     
     
     
     */
    virtual void OperationEnd(DOOperation* pOperation);


    /*
    $ifdef ETERNA$
    Description: Returns the weight of a given duplicated object for load balancing purposes.
    Returns     : The hypothetical weight assigned to a duplicated object for load balancing purposes.
    Remarks: This method is used to assign a hypothetical weight to a given duplicated object
    that will participate in load balancing. The weight of an object is equivalent to how much
    the duplication master object contributes to a station's local load. An object will only
    be migrated if its weight is less than the total load to be transferred during a load balancing
    episode. The default weight of an object is 0. When this is the case the object will not
    participate in load balancing.

    Note: An object's weight must be equal to or greater than 0.

    See Also: LoadBalancingDomain, Object Migration Overview, LoadBalancingDomain::GetLocalLoad
    $else$
    {secret}
    $endif$
    */
    virtual qReal GetWeight()
    {
        return QREAL(0.0);
    }

    //group=DO Methods

    /*
    Description: Traces the default information for a duplicated object instance.
    Remarks: The default trace information for a duplicated object includes its class name,
    DOHandle, Duplication role ( /i.e./ Duplica or Duplication master), its ref count ( /i.e./ how many
    DORef s are presently created in the system for this object), whether or not
    it is flagged for destruction, as well as the station on which its Duplication master instance resides.
    Redefine the Trace method in the implementation of your DOCLASS to
    trace game specific information to that particular DOCLASS.
    However, to continue to trace the default information, make sure to include a call to
    DOCLASS(YourDOCLASS)::Trace().
    Example:
        class Map:DOCLASS(Map){
            [...]
            void Trace(qUnsignedInt64 uiTraceFlag) {
                DOCLASS(Map)::Trace(uiTraceFlag);
                QTRACE(uiTraceFlag,_T("Map number is ") << dec << (GetMapNumber()));
            }
            [...]
        };

    See Also: DOSelections::Trace, Data Logging, TraceLog
    */
    /* Please see man pages for details 
     
     
     */
    virtual void Trace(qUnsignedInt64 uiTraceFlag=TRACE_ALWAYS) const;

    //{secret}
    static void TraceOperationStack(qUnsignedInt64 uiTraceFlag=TRACE_ALWAYS);


    // Discovery related methods...
    // {Secret}
    void SpecificAddDSToDiscoveryMessage(Message* pMsg, Station* pStation)
    {
        DUMMY_PARAM(pMsg);
        DUMMY_PARAM(pStation);
    };
    // {Secret}
    void SpecificExtractDSFromDiscoveryMessage(Message* pMsg)
    {
        DUMMY_PARAM(pMsg);
    };
    // {Secret}
    qBool SpecificRefresh(DataSet * pDataset, const Time& tRefreshTime);
    // {Secret}
    qBool SpecificExtractADataset(Message* pMsg, qByte byDatasetIndex);
    // {Secret}
    qResult SpecificUpdate(DataSet *pDataset, Time tUpdateTime);
    // {Secret}
    void FillDupSpaceInfo(DupSpace::Role eRole, DupSpace::ID* pArray, qUnsignedInt* puiIndex)
    {
        DUMMY_PARAM(eRole);
        DUMMY_PARAM(pArray);
        DUMMY_PARAM(puiIndex);
    }

    // Methods related to action calls.
    //{secret}
    Message* CreateStubMessage(MethodID* pidMethod);
    //{secret}
    qBool SendStubMessage(qBool bSendToDuplicas, Message* pMsg);
    //{secret}
    void SendToAllDuplicas(Message* pMsg, qUnsignedInt32 uiFlags);
    // {secret}
    void SendToSomeDuplicas(Selection* pTargets, Message* pMsg, qUnsignedInt32 uiFlags);

    qBool IsOverReliableSendBufferLimit(const Selection* pTargets, qUnsignedInt32 uiFlags, SubStreamID subStreamID) const;
    qBool IsInvalidSubStreamID(const Selection* pTargets, SubStreamID subStreamID) const;

    /*
     * Adapter methods
     */

    // {secret}
    // Generated code will override this method.  This method is
    // there to provide support for the adapter extension.
    virtual void CallOperationEndOnAdapters(DOOperation* /*pOperation*/) {}

public:

    /**********
     *
     * Operations entry points.
     *
     **********/

    // {secret}
    // TODO.  Cleanup up the arguments of this method.
    qBool RemoveFromStore(DOHandle hSourceStation, qBool bDeleted, qBool bRemoveDuplicas);

    // {secret}
    qBool AddToStoreAsDuplica(DOHandle hSourceStation, Message* pMessageToInitFrom=NULL);
    // {secret}
    qBool AddToStoreAsMaster();
    //{secret}
    qBool UndeleteMainRef();

    // {secret}
    qBool ChangeMasterStation(DOHandle hSourceStation,
                              DOHandle hOldStation,
                              const MasterStationRef& refNewMasterStation,
                              const qList<DOHandle> *plstDupSet,
                              qUnsignedInt uiContext);

    //{secret}
    static void UpdateDatasets(Message *pMessage, DOHandle hTargetDO, qByte byDatasetIndex);

    /*
    Description: Returns a pointer to the DOOperation that is currently in progress.
    Returns     : A pointer to the DOOperation that is currently in progress, or a NULL pointer
    if no DOOperation is currently in progress or an error occurred.
    Remarks: This method is most commonly used to determine if the CallMethodOperation
    is in progress to control the calling of RMCs.
    See Also: DOOperation
    */
    /* Please see man pages for details 
     
     
     
     
     
     
     
     */
    static DOOperation* GetCurrentOperation();
    // {secret}
    qBool ExecuteOperation(DOOperation& oOperation);

protected:

    // {secret}
    virtual void TestInvariants();

private:
    // {secret}
    qResult UpdateImpl(DataSet *pDataset, const Time& tUpdateTime);
    // {secret}
    qResult UpdateImplCore(DataSet *pDataset, const Time& tUpdateTime);

    // {secret}
    qBool RefreshImpl(DataSet* pDataset, const Time& tRefreshTime);
    // {secret}
    qBool RefreshImplCore(DataSet* pDataset, const Time& tRefreshTime);

    qBool ExecuteOperationImpl(DOOperation& oOperation);

    // {secret}
    qBool FaultRecoveryImpl(DOOperation* pOperation);

    // {secret}
    void InValidOperationLog(const DOOperation* pOperation)const;

    void ExecCallMethod(const CallMethodOperation& refOp);
    void ExecChangeMasterStation(const ChangeMasterStationOperation& refOperation);
    void ExecRemoveFromStore(const RemoveFromStoreOperation& refOperation);
    void ExecAddToStore(const AddToStoreOperation& refOp);
    void ExecUpdateDataSet(const UpdateDataSetOperation& refOperation);
    void ExecChangeDupSet(const ChangeDupSetOperation& refOperation);

    /**********
     *
     * Fetch/CreateDuplica methods
     *
     **********/
public:

    /*
    Description: Creates a Duplica of a given duplicated object.
    Parameters:
    hNewDuplicaStation: The DOHandle of the station on which a Duplica will be created.
    Returns     : true if the Duplica is successfully created on the given station, false if not or an
    error occurred.
    Remarks: This method is one manner in which to create a Duplica of an object on a station.
    The method must be called on a Duplication master to create a Duplica of itself.
    See Also: FetchDuplica
    */
    /* Please see man pages for details 
     
     
     
     
     
     
     
     
     */
    qBool CreateDuplica(DOHandle hNewDuplicaStation);

    /*
    Description: Fetches a Duplica of a given duplicated object.
    Parameters:
    hFetchedObject: The DOHandle of the fetched object.
    pContext: A pointer to the FetchContext.

    Remarks: This method is one manner in which to create a Duplica of an object on a station.
    Example:
            bool PlayerManager::RequestPlayer(DOHandle hPlayerID, Player::Ref* prefPlayer) {
                FetchContext oFetchContext;

                // In this example, we assume that the session master holds a copy
                // of the desired player.
                oFetchContext.SetTargetStation(Session::GetInstance()->GetMasterStation());
                oFetchContext.ClearFlag(FetchContext::SynchronousCall);

                if (!DuplicatedObject::FetchDuplica(&oFetchContext, hPlayerID) {
                    printf("Unable to fetch duplica.\n" );
                    return false;
                }

                // We wait until the call completes.
                oFetchContext.Wait();

                // Once the call has completed, we grab a reference to the newly
                // fetched object.  This step is MANDATORY: if we do not do it,
                // then the object that has just been fetched is not considered
                // relevant (as there are no references to it), and will eventually be
                // removed by the system.
                *prefPlayer = oFetchContext.GetFetchedObject();
                if (!prefPlayer->IsValid()) {
                    printf("Invalid player reference.\n" );
                    return false;
                }
                return true;
            }

    See Also: CreateDuplica, FetchContext, FetchInProgress
    */
    /* Please see man pages for details 
     
     
     
     
     
     
     
     
     
     */
    static qBool FetchDuplica(FetchContext* pContext, DOHandle hFetchedObject);

    /*
    Description: Returns whether or not a given object is in the process of being fetched.
    Parameters:
    hTargetStation: The DOHandle of the target station from which the Duplica will be fetched.
    hFetchedObject: The DOHandle of the fetched object.

    Returns     : true if the specified object is in the process of being fetched from the target
    station (if specified), false if not or an error occurred.
    See Also: FetchDuplica

    */
    static qBool FetchInProgress(DOHandle hFetchedObject);
    static qBool FetchInProgress(DOHandle hFetchedObject, DOHandle hTargetStation);


    /**********
     *
     * FaultRecovery methods.
     *
     **********/

    // {secret}
    qBool PerformFaultRecovery(DOHandle hNewMaster, LogicalClock lcNewMasterVersion);

    /**********
     *
     * Migration methods
     *
     **********/
    //group=DO Migration Methods

    /*
    Description: Emigrates a duplicated object to a specific station.
    Parameters:
    idNewLocation: The StationID for the station to which the duplicated object will emigrate (that is, the destination
    StationID for the duplicated object).
    pContext: A pointer to the MigrationContext of this call.
    hTargetStation: The DOHandle of the target station to which the object will migrate.
    Remarks:  This method must be called on the Duplication master.  The signature of this method that
    does not contain a MigrationContext locks until the emigration completes or the error
    SYSTEMERROR_OBJDUP_CANNOT_EMIGRATE is generated due to the unavailability of station idNewLocation.
    On the other hand, the second signature returns immediately before ensuring successful migration to
    hTargetStation or encountering a system error.

    Note: Alternatively, you may use the method AttemptEmigration, which returns immediately but
    does not guarantee a successful emigration.

    Returns     : true if the object is successfully emigrated, false if not or an error occurred.
    The errors SYSTEMERROR_OBJDUP_CANNOT_EMIGRATE and SYSTEMERROR_OBJDUP_INVALID_ON_DUPLICA may be signalled by this call.
    See Also: AttemptEmigration, MigrationInProgress
    */
    // Obsolete
    qBool Emigrate(StationID idNewLocation);
    /* Please see man pages for details 
     
     
     
     
     
     
     
     
     
     
     
     
     
     
     
     
     
     
     
     */
    qBool Emigrate(MigrationContext* pContext, DOHandle hTargetStation);

    /*
    Description: Returns whether or not an object is in the process of migrating.
    Returns     : true if the object is in the process of migrating, false if not or an error occurred.
    Remarks: An object may migrate due to Load Balancing, due to a fault (if the object is fault-tolerant), or
    due to a call to either AttemptEmigration or Emigrate.
    See Also: Fault Tolerance Overview, Emigrate, AttemptEmigration
    */
    /* Please see man pages for details 
     
     
     
     
     
     
     
     
     
     
     
     
     */
    qBool MigrationInProgress() const;
    void MigrationFault() const;

    /*
    Summary: The method attempts to emigrate a duplicated object to a specific station.
    Parameters:
    idNewLocation: The StationID for the station to which the duplicated object will attempt to emigrate
    ( /i.e./ the destination StationID for the duplicated object).
    hNewLocation: The DOHandle of the station to where the duplicated object will attempt to emigrate.
    Returns     : true if the attempt to migrate an object was successful, false if not or an error occurred.
    The errors SYSTEMERROR_OBJDUP_CANNOT_EMIGRATE and SYSTEMERROR_OBJDUP_INVALID_ON_DUPLICA may be signalled by this call.
    Remarks:  This method must be called on the Duplication master of an object. The method returns immediately
    before ensuring successful migration or encountering a system error. If a fault occurs on the destination
    station then migration will not proceed. If you want to control which stations fault tolerant objects
    are migrated to, for load balancing for example, then you should call this method before the Station quits the Session.

    Note: Alternatively, you may use the method Emigrate, one signature of which locks until the emigration process
    is successful or an error is generated.

    See Also: Emigrate , MigrationInProgress
    */
    // Obsolete
    qBool AttemptEmigration(StationID idNewLocation);
    /* Please see man pages for details 
     
     
     
     
     
     
     
     
     
     
     
     
     
     
     
     */
    qBool AttemptEmigration(DOHandle hNewLocation);

    // {secret}
    void PrepareToLeave();
    //{secret}
    DOHandle SelectNewLocation(qUnsignedInt32 uiReason);

    //group=DO Methods

    // Misc methods...
    //{secret}
    DOClass* GetDOClass() const
    {
        return GetDOClass(GetDOClassID());
    }
    // {secret}
    static DOClass* GetDOClass(DOClassID idDOClass);

    /*
    Description:  Returns the DOHandle of the Station that has this object's Duplication master in its store.
    Returns     :  The DOHandle of the Station on which the Duplication master for this duplicated object resides.
    Remarks: In order to determine the Session master of a given Session, call this method on the Session object.

    Remarks: Refer to FetchDuplica for a code sample.
    */
    /* Please see man pages for details 
     
     
     
     
     
     */
    DOHandle GetMasterStation() const
    {
        return m_refMasterStation.GetStationHandle();
    }
    // {secret}
    const Station* GetMasterStationPtr() const
    {
        return m_refMasterStation.GetStation();
    }
    // {secret}
    // TODO - Still required?
    const LogicalClock& GetMasterVersion() const
    {
        return m_refMasterStation.GetVersion();
    }
    // {secret}
    void SetMasterStation(const MasterStationRef& refNewMaster);

    //Description:  Returns the StationID of the station that has this object's Duplication master in its store.
    //Returns     :  The StationID of the station on which the Duplication master for this duplicated object resides.
    /* Please see man pages for details 
     
     
     
     */
    StationID GetMasterID() const;

    /*
    Description: Returns whether or not DuplicatedObject is a Duplica.
    Returns     : true if the DuplicatedObject is a Duplica, false if it is a Duplication master or unknown.
    Remarks: If the method is called before the object's Duplication role has been
    initialized, the object is neither a Duplication master nor a Duplica.
    Typically, this occurs if the methods IsADuplicationMaster or
    IsADuplica are called in the object constructor.  When the
    initialization of an object depends on its role, it should be performed
    via the InitDO system callback, where the role has been defined.
    See Also: Duplication role, IsADuplicationMaster
    */
    /* Please see man pages for details 
     
     
     
     
     
     
     
     
     
     
     
     
     
     
     */
    qBool IsADuplica() const;

    /*
    Description: Returns whether or not the DuplicatedObject is the Duplication master.
    Returns     : true if the DuplicatedObject is a Duplication master, false if it is a Duplica or unknown.
    Remarks: If the method is called before the object's Duplication role has been
    initialized, the object is neither a Duplication master nor a Duplica.
    Typically, this occurs if the methods IsADuplicationMaster or
    IsADuplica are called in the object constructor.  When the
    initialization of an object depends on its role, it should be performed
    through the InitDO system callback, where the role has been defined.
    See Also: Duplication role, IsADuplica
    */
    /* Please see man pages for details 
     
     
     
     
     
     
     
     
     
     
     
     
     
     
     
     
     */
    qBool IsADuplicationMaster() const;

    /*
    Description: Returns whether or not the DuplicatedObject has authority.
    Returns     : true if the DuplicatedObject has authority, false if it does not.
    Remarks: A DuplicatedObject can only have authority if it is the Duplication master instance
    of the object. A Duplication master that has authority can safely perform session critical
    operations. Refer to the Duplicated Object Overview for more information.
    See Also: Duplicated Object Overview
    */
    /* Please see man pages for details 
     
     
     
     
     
     
     
     
     
     
     
     
     
     
     
     
     */
    qBool HasAuthority() const
    {
        return IsADuplicationMaster() && !MigrationInProgress();
    }

    /*
    Summary:  Returns whether or not a duplicated object is an instance of
    a well-known duplicated object class.
    Returns     :  true if the DuplicatedObject instance is a member of a class that was declared in the DDL file as
    well-known, false if not or an error occurred.
    Remarks: A Well-known duplicated object is known globally across the Session and is guaranteed to be
    discovered by a Station before the Session commences on that Station.
    See Also:  DDL File Syntax, DOHandle, WKHandle, IsAUserDO, IsACoreDO, IsGlobal
    */
    /* Please see man pages for details 
     
     
     
     
     
     
     
     
     */
    qBool IsAWellKnownDO() const;

    /*
    Description: Returns whether or not the DuplicatedObject is a CoreDO.
    Returns     : true if the DuplicatedObject is a core object created by $PRODUCTNAME$, false if not or an error occurred.

    $ifdef ETERNA$
    Remarks:
    A CoreDO is a DuplicatedObject created by $PRODUCTNAME$ when a Session is created or when
    Station s join or leave the Session. These objects are essential for the correct operation of
    $PRODUCTNAME$ and include the root classes for all global and non-global DuplicatedObject s,
    the Session, SessionClock, Station, and IDGenerator classes, system classes involved
    in fault tolerance, fault recovery, and multicast messaging mechanisms, a default cell class if the
    DuplicationSpace uses Cell s, and the LoadBalancingDomain and other classes involved in load balancing.

    $else$
    Remarks:
    A CoreDO is a DuplicatedObject created by $PRODUCTNAME$ when a Session is created or when
    Station s join or leave the Session. These objects are essential for the correct operation of
    $PRODUCTNAME$ and include the root classes for all global and non-global DuplicatedObject s,
    the Session, SessionClock, Station, and IDGenerator classes, system classes involved
    in fault tolerance, fault recovery and a default cell class.
    $endif$
    See Also: IsAUserDO, IsAWellKnownDO, IsGlobal
    */
    /* Please see man pages for details 
     
     
     
     
     
     
     
     
     
     
     */
    virtual qBool IsACoreDO() const;

    /*
    Description: Returns whether or not the DuplicatedObject is a UserDO.
    Returns     : true if the DuplicatedObject is a user-defined object, false if not or an error occurred.
    Remarks: A UserDO is a user-defined DuplicatedObject declared by the user in the DDL.
    See Also: IsACoreDO, IsAWellKnownDO, IsGlobal
    */
    /* Please see man pages for details 
     
     
     
     
     
     
     
     
     */
    qBool IsAUserDO() const
    {
        return !IsACoreDO();
    }

    // {secret}
    virtual qBool IsABootstrapDO() const
    {
        return false;
    }

    /*
    Description: Returns a pointer to an OperationManager.
    Returns     : A pointer to an OperationManager.
    Remarks: This method returns a pointer to the OperationManager that manages all DOOperation s. The main
    purpose of the OperationManager is to manage the callbacks called at the beginning and end of the various
    $PRODUCTNAME$ System Operations.
    See Also: OperationManager
    */
    /* Please see man pages for details 
     
     
     
     
     
     
     */
    static OperationManager* GetOperationManager();

    /*
    Description: Instructs $PRODUCTNAME$ to delete the main reference of the duplicated object.
    Returns     :  true if the main Ref of the object was successfully deleted, false if not or an error occurred.
    The error SYSTEMERROR_OBJDUP_INVALID_ON_DUPLICA is returned if this method is called on a Duplica.
    Remarks: This method deletes the main (system) reference to the duplicated object held in the
    duplicated object store. After the main reference is deleted, the object's datasets will cease to be updated
    and the object will no longer be "visible" when a selection of objects is iterated over, nor be listed by
    the Object Monitor. If the main reference was the only reference to the object, the destructor will be
    called and the object will be deleted with this call. If other references to the Duplication master or Duplicas
    of the object exist, /e.g./ created via the RefTemplate or Ref, the destructor will not be called until the last
    reference is released. When the destructor is called, all the Duplicas of the object will be deleted.

    This method can only be called on the Duplication master of an object and may be called before an
    object is published.

    Example:
    // This is one way to delete all the Duplication master objects on a Station.
    {
        Avatar::SelectionIterator itAvatarDM(DUPLICATION_MASTER);
        while (!itAvatarDM.EndReached()) {
            itAvatarDM->DeleteMainRef();
            itAvatarDM++;
        }
    }

    See Also: MainRefIsDeleted, RemoveFromStoreOperation, Publish
    */
    /* Please see man pages for details 
     
     
     
     
     
     
     
     
     
     
     
     
     
     
     
     
     
     
     */
    qBool DeleteMainRef();
    // {secret}
    qBool DeleteDuplicaMainRef();

private:
    // {secret}
    qBool DeleteMainRefImpl();

public:

    // {secret}
    qBool FlagIsSet(qUnsignedInt16 ui16Flag) const
    {
        return (m_uiFlags & ui16Flag)==ui16Flag;
    }
    /*
    Description: Returns whether or not DeleteMainRef has been called on this object.
    Remarks: If the method DeleteMainRef has been called and therefore, the main reference to
    the object has been deleted, the destructor will be called on the object when all references
    to the object have been released, /i.e./ there are no Ref s nor iterators that point to the object.
    Returns     : true if the method DeleteMainRef has been called on this object and therefore, the
    main reference to the object has been deleted, false if not or an error occurred.
    See Also: DeleteMainRef
    */
    /* Please see man pages for details 
     
     
     
     
     
     
     
     
     
     */
    qBool MainRefIsDeleted() const
    {
        return !FlagIsSet(MainReferenceInUse);
    }

    // {secret}
    qBool ConnectOrphanDuplica();
    // {Secret}
    static qUnsignedInt32 GetRealDOClassID()
    {
        return DuplicatedObject_ClassID;
    }

#ifdef FOR_DOC_PURPOSE
    /*
    Summary:  Creates a local instance of a specified doclass.
    Parameters:
    idDOClass: The DOClassID of the duplicated object to be created.
    uiTimeout: The maximum time in milliseconds that the system
    will wait. The default uiTimeout is infinite.
    idUserDefinedID: (Optional). When used, assigns the specified DOID to the duplicated object,
    bypassing the automatic DOID generation mechanism. Note that you MUST assign a different
    DOID to each duplicated object within the same class regardless of its location.
    dohHandle: The DOHandle of the duplicated object to be created.
    Returns     : SYSTEMERROR_GEN_INVALID_WAIT if this method is called in a single thread context
    with a timeout value other than 0.
    Remarks:  The DDL generated duplicated object classes inherit from the DuplicatedObject class.
    A specific constructor for each doclass is generated by the DDL compiler into each DOCLASS.
    It is not possible to create generic DuplicatedObject instances and you need to specify the
    doclass of the instance that you want to create.

    For example, given that Avatar is declared as a doclass in the DDL file, an Avatar object can be created
    and published in any of the following ways:

            // 1. Let $PRODUCTNAME$ assign the IDs for the object.
            {
                Avatar* pNewAvatar=Avatar::Create();
                pNewAvatar->Publish();
            }

            // 2. Assign the DOID.
            {
                ID idNewAvatar=0;
                IDGenerator::Ref refIDGenerator(Avatar::GetIDGenerator());
                if (refIDGenerator->GenerateID(&idNewAvatar, 0)) {
                    Avatar* pNewAvatar=Avatar::Create(idNewAvatar);
                    pNewAvatar->Publish();
                } else {
                    return false;
                }
            }

            // 3. Assign the DOClassID and let $PRODUCTNAME$ assign
            //    the DOID.
            {
                DuplicatedObject* pNewAvatar=DuplicatedObject::Create(DOCLASSID(Avatar));
                pNewAvatar->Publish();
            }

            // 4. Assign both the DOClassID and DOID.
            {
                ID idNewAvatar=0;
                IDGenerator::Ref refIDGenerator(Avatar::GetIDGenerator());
                if (refIDGenerator->GenerateID(&idNewAvatar, 0)) {
                    Avatar* pNewAvatar=(Avatar*)DuplicatedObject::Create(DOCLASSID(Avatar), idNewAvatar);
                    pNewAvatar->Publish();
                } else {
                    return false;
                }
            }

            // 5. Assign a DOHandle.

            // Assume there is a database from which we
            // save/retrieve information.

            DOHandle hPreviouslySavedAvatar = ReadHandleFromDatabase();
            Avatar* pNewAvatar=(Avatar*) DuplicatedObject::Create(hPreviouslySavedAvatar);
            // Initializes the attributes of pNewAvatar with the
            // info saved previously in the database, and then publish
            // the new avatar.
            pNewAvatar->Publish();

    It should be noted that as an object's DOClassID and therefore, its DOHandle may change when you recompile
    your project, alter your DDL file, or upgrade to a new version of $PRODUCTNAME$, it is safer to store and
    retrieve DOID s rather than DOHandle s in the database.

    Note:

    o The method DuplicatedObject* Create(DOID idUserDefinedID) is defined by a specific class.
    o To create a Session duplicated object, Session::CreateSession should be called rather than DuplicatedObject::Create.
    o If Create is called in the context of an RMC a non-zero timeout cannot be specified.
    See Also: Publish, CreateWellKnown
    */
    /* Please see man pages for details 
     
     
     
     
     
     
     
     
     
     
     
     
     
     
     
     
     
     
     
     
     
     
     
     
     
     */
    static DuplicatedObject* Create(qUnsignedInt32 uiTimeout == DEFAULT_CREATE_TIMEOUT)
    {
        // This method is provided by specific classes.
        SYSTEMCHECK(false);
        return NULL;
    }

    /* Please see man pages for details 
     
     
     
     
     
     
     
     
     
     
     
     
     
     
     
     
     
     
     */
    static DuplicatedObject* Create(DOID idUserDefinedID)
    {
        // This method is provided by specific classes.
        SYSTEMCHECK(false);
        return NULL;
    }

#endif
    /* Please see man pages for details 
     
     
     
     
     
     
     
     
     
     
     
     
     
     
     */
    static DuplicatedObject* Create(DOClassID idDOClass, qUnsignedInt32 uiTimeout = WAIT_INFINITE_TIMEOUT);

    /* Please see man pages for details 
     
     
     
     
     
     
     
     
     
     
     
     
     
     
     
     
     
     
     
     */
    static DuplicatedObject* Create(DOClassID idDOClass, DOID idUserDefinedID );

    /* Please see man pages for details 
     
     
     
     
     
     
     
     
     
     
     
     */
    static DuplicatedObject* Create(const DOHandle &dohHandle );

    // {secret}
    static DuplicatedObject* CreateDuplica(DOHandle hDO, const MasterStationRef& refMaster);


    /*
    Summary:  Creates a well-known doclass.
    Parameters:
    hWK: The name of the global variable that references the well-known instance.  This variable has to be
    the same as that specified in the DDL file.
    Returns     : SYSTEMERROR_OBJDUP_WKHANDLE_EXISTS is returned if the well-known instance has already been created, and
    SYSTEMERROR_OBJDUP_INVALID_ON_DUPLICA if this method is called on a station that is not the Session master.
    Remarks:  The DDL generated doclasses inherit from the DuplicatedObject class.  A specific constructor for each
    doclass is generated by the DDL compiler into each DOCLASS.  It is not possible to create generic
    DuplicatedObject instances, you need to specify the doclass of the instance that you want to create.
    For example, given that World is declared as a well-known doclass and the global variable
    name g_hTheWorld is supplied in the DDL file the well-known object can be created as follows:

            World* pNewWorld=World::CreateWellKnown(g_hTheWorld);
            pNewWorld->Publish();

    This method must be called on the Session master.

    Note: As $PRODUCTNAME$ guarantees that all well-known objects are discovered before the game Session is created,
    this method must be called before a Session is created to ensure that all well-known objects are
    created at the appropriate time.

    See Also: Session::RegisterWellKnownDOsFactory, Session::CreateSession, Session::CreateOrJoinSession, Session::JoinSession
    */
    /* Please see man pages for details 
     
     
     
     
     
     
     
     
     
     
     
     
     
     
     
     */
    static DuplicatedObject* CreateWellKnown(WKHandle& hWK);

    /*
    Summary:  Publishes the doclass instance to all stations connected to the session
    that require the object.
    Parameters:
    uiTimeout: The maximum time in milliseconds that the system will wait. The default uiTimeout
    is infinite. This should be set to zero when $PRODUCTNAME$ is run in single-thread mode.
    Returns     : true if the object was successfully published, false if not or an error occurred.
    The error SYSTEMERROR_GEN_INVALID_WAIT will be signalled if this method is called in a single thread context
    and the system cannot return immediately, and SYSTEMERROR_GEN_INVALID_OPERATION if this method is
    called after DeleteMainRef has been called on the object.

    $ifdef VOICE$
    When this method is called on a VoiceChannel object, the method will return false if either
    the CodecID or the packet size has not been set before Publish is called.
    $endif$
    Remarks: For example, given that Avatar is declared as a doclass in the DDL file, an
    instance of the Avatar class can be created and then published as follows. Refer to Create
    for other object creation options.

            Avatar* pNewAvatar=Avatar::Create();
            pNewAvatar->Publish();

    When this method is called in a single thread context, uiTimeout should be set to zero. Note
    that if you want the datasets of the published object to be up-to-date when Publish is called,
    you should initialize the object's datasets between the calls to Create and Publish rather
    than initializing them after calling Publish. If Publish returns false, then you must still
    delete the object (via DeleteMainRef) you were attempting to publish.
    See Also: Create, IsPublished, Threading Model
    */
    /* Please see man pages for details 
     
     
     
     
     
     
     
     
     
     
     
     
     
     
     
     */
    qBool Publish( qUnsignedInt32 uiTimeout = DEFAULT_PUBLISH_TIMEOUT );

    /*
    Description: Returns whether or not the given DuplicatedObject has been published.
    Returns     : true if the object has already been published, false if the object is currently unpublished.
    Remarks: After a DuplicatedObject is created, it must then be published by Publish for it to
    be duplicated on all necessary remote stations in the session.
    See Also: Publish
    */
    /* Please see man pages for details 
     
     
     
     
     
     
     
     */
    qBool IsPublished() const
    {
        return FlagIsSet(Published);
    }


    /*
    Description: Returns a selection that contains the stations where the duplicas of this
    duplicated object are located.
    Returns     : A selection of stations that can be used by a Station::SelectionIterator to iterate
    over all the stations where duplicas of this duplicated object are located.
    SYSTEMERROR_OBJDUP_INVALID_ON_DUPLICA will be returned if this method is called on a Duplica.
    Remarks: This method can only be called on a Duplication master. If called on a Duplica
    SYSTEMERROR_OBJDUP_INVALID_ON_DUPLICA will be returned.

    This method would typically be called as in the following example.

    Example:

        Station::SelectionIterator myIt( myDO->GetDuplicasStationLocations() );

        while(!myIt.EndReached()){
            ...
            }

    */
    Selection *GetDuplicasStationLocations()
    {
        return &m_setDuplicationSet;
    }

    //{secret}
    void FillDuplicaStationsList(qList<DOHandle> *plstDuplicaStations);

    //group=DO Templates

    /*
    Summary: Constructs a reference for safe access to the duplicated object.

    Remarks:  Defined by the duplicated object. Safely access the duplicated object
    by using this class to construct a reference to it from a DOHandle.

    A Ref is a typedef to an appropriate version of the RefTemplate
    template.  All duplicated object subclasses redefine this typedef
    and convert the DOHandle to a reference to the specific subclass of
    duplicated object.

    For example, create a Ref to an Avatar with a DOHandle hAvatar as follows:

            Avatar::Ref refAvatar(hAvatar);
            refAvatar->Render();

    Note: All references to a duplicated object must be released before $PRODUCTNAME$ can call the
    object's destructor. If not all references are released then $PRODUCTNAME$ will wait a certain
    amount of time and issue a warning before the object is deleted.
    See also: RefTemplate
    */
    /* Please see man pages for details 
     
     */
    typedef RefTemplate<DuplicatedObject> Ref;

    /*
    Description: Use your duplicated object class name and the SelectionIteratorTemplate
    class to iterate over duplicated objects.
    Remarks: It is possible to create a SelectionIterator that holds all of the duplicated
    objects of a certain class on a host as follows:

            Avatar::SelectionIterator itALLAvatar;

    This is implemented via a template. You can find more details about how to use it
    under SelectionIteratorTemplate.
    See Also: Iterators Overview, SelectionIteratorTemplate
    */
    /* Please see man pages for details 
     
     */
    typedef SelectionIteratorTemplate<DuplicatedObject> SelectionIterator;

    //group=DO Methods

#ifdef FOR_DOC_PURPOSE
    /*
    Description: Casts (performs explicit type conversion) during run time.
    Parameters:
    pDO: A pointer to a DuplicatedObject class.
    Returns     : A valid pointer to the object if the object is of the expected type, a null
    pointer if it is not.
    Remarks: The DynamicCast method functions in the same manner as the standard C++ dynamic_cast
    operator but using $PRODUCTNAME$ run-time information.

    For example, say we have an Avatar class that inherits from a Duplicated3DObject class, which in
    turn inherits from the root DuplicatedObject class. We can return a pointer to an Avatar
    object, if /pObj/ points to an Avatar object, as follows:

        Avatar *pAvatar = Avatar::DynamicCast(pObj)

    Note: This method is defined by each specific DuplicatedObject class.
    */
    /* Please see man pages for details 
     
     
     
     
     
     
     
     
     
     
     
     
     */
    static DuplicatedObject* DynamicCast(DuplicatedObject* pDO)
    {
        // This method is provided by specific classes.
        SYSTEMCHECK(false);
        return NULL;
    }
    static const DuplicatedObject* DynamicCast(const DuplicatedObject* pDO)
    {
        // This method is provided by specific classes.
        SYSTEMCHECK(false);
        return NULL;
    }
#endif

    // Use this macro to secretize the flags.
#ifdef FOR_DOC_PURPOSE

#else
    typedef enum _Flag
    {
        MainReferenceInUse=1<<0, // Corresponds to bMainReferenceInUse, in doprotocol.spin
        ValidConstruction=1<<2,
        ValidDestruction=1<<3,
        OrphanDuplica=1<<4,
#ifdef DEBUG_VERSION
        Published=1<<5,
        InTheStore=1<<15
#else
        Published=1<<5
#endif
    } Flag;

#endif
    // Relevance count manipulation.
    //{secret}
    qUnsignedInt GetRelevanceCount() const
    {
        return m_uiRelevanceCount;
    }
    //{secret}
    void IncreaseRelevanceCount()
    {
        ScopedCS oCS(s_csRefCount);
        ++m_uiRelevanceCount;
    }
    //{secret}
    void DecreaseRelevanceCount()
    {
        ScopedCS oCS(s_csRefCount);
        SYSTEMCHECK(m_uiRelevanceCount>0);
        --m_uiRelevanceCount;
    }

    // Ref count manipulation.

    // {secret}
    qUnsignedInt GetRefCount() const
    {
        return m_uiRefCount;
    }
    // {secret}
    void IncreaseRefCount()
    {
        if (Core::IsThreadSafe())
        {
            ScopedCS oCS(s_csRefCount);
            ++m_uiRefCount;
        }
        else
        {
            ++m_uiRefCount;
        }
    }

    // Increase both counts...
    //{secret}
    void IncreaseRefAndRelevanceCounts()
    {
        if (Core::IsThreadSafe())
        {
            ScopedCS oCS(s_csRefCount);
            ++m_uiRefCount;
            ++m_uiRelevanceCount;
        }
        else
        {
            ++m_uiRefCount;
            ++m_uiRelevanceCount;
        }
    }


    // {secret}
    qBool DecreaseCountsImplCore(qBool bDecreaseRelevanceCount)
    {
        qBool bQuickDecreaseDone=true; // Assume it is done...
        if (bDecreaseRelevanceCount)
        {
            // This one is decreased here, but does not go up
            // on failure, as does the uiRefCount.  That's ok -
            // the second call to decrease the refcount does
            // not care about relevance.
            SYSTEMCHECK(m_uiRelevanceCount>0);
            --m_uiRelevanceCount;
        }
        SYSTEMCHECK(m_uiRefCount>0);
        --m_uiRefCount;
        if (m_uiRefCount==0)
        {
            // The DO must be deleted.  At this point, we can't
            // do it - the store isn't locked.  Call the
            // complete DecreaseRefCount.
            ++m_uiRefCount;
            bQuickDecreaseDone=false;
        }
        return bQuickDecreaseDone;
    }
    // {secret}
    void DecreaseCountsImpl(qBool bDecreaseRelevanceCount)
    {
        // Quick check - select the fast or the slow decrease ref
        // count...  Most of the time, the fast decrease ref count
        // will do the job - and the store will not have to be
        // locked.  If not - then call the real one - where the DO
        // can be removed from the store.

        // This piece of code is rather complex.  But for
        // performance reasons, it is better to do it the way it
        // is.
        qBool bQuickDecreaseDone=true;
        if (Core::IsThreadSafe())
        {
            ScopedCS oCS(s_csRefCount);
            bQuickDecreaseDone = DecreaseCountsImplCore(bDecreaseRelevanceCount);
        }
        else
        {
            bQuickDecreaseDone = DecreaseCountsImplCore(bDecreaseRelevanceCount);
        }
        if (!bQuickDecreaseDone)
        {
            CompleteDecreaseRefCount();
        }
    }

    // {secret}
    void DecreaseRefAndRelevanceCounts()
    {
        DecreaseCountsImpl(true);
    }
    // {secret}
    void DecreaseRefCount()
    {
        DecreaseCountsImpl(false);
    }

    // {secret}
    void CompleteDecreaseRefCount();

    // {secret}
    void ProtectedDecreaseRefCount()
    {
        // Called with store locked.
        SYSTEMCHECK(Scheduler::GetInstance()->SystemLock().HasOwnership()); //n1589:mark:TransportLock:off
        qBool bDestroyDO=false;
        {
            ScopedCS oCS(s_csRefCount);
            SYSTEMCHECK(m_uiRefCount>0);
            --m_uiRefCount;
            if (m_uiRefCount==0)
            {
                bDestroyDO=true;
            }
        }
        if (bDestroyDO)
        {
            DestroyDO();
        }
    }

    // {secret}
    void DestroyDO();

    // Flags manipulation.

    // {secret}
    void SetFlag(qUnsignedInt16 ui16Flag);
    // {secret}
    void ClearFlag(qUnsignedInt16 ui16Flag);

    // {secret}
    void ReleaseReferenceToMaster();
    // {secret}
    void AcquireReferenceToMaster();
    // {secret}
    qBool OwnsReferenceToMaster() const;

    //{secret}
    virtual void UpdateCellStats(qInt iNbPublishers,
                                 qInt iNbSubscribers,
                                 qInt iNbPublishersAndSubscribers)
    {
        DUMMY_PARAM(iNbPublishers);
        DUMMY_PARAM(iNbSubscribers);
        DUMMY_PARAM(iNbPublishersAndSubscribers);
    };

#if NEX_ENABLE_DOREF_LOCK_OPTIMIZATION
    //{secret}
    static CriticalSection* GetRefCountCS()
    {
        return &s_csRefCount;
    }
    //{secret}
    void DecreaseRelevanceCountNoLock()
    {
        SYSTEMCHECK(m_uiRelevanceCount>0);
        --m_uiRelevanceCount;
    }
    //{secret}
    void IncreaseRelevanceCountNoLock()
    {
        ++m_uiRelevanceCount;
    }
    //{secret}
    void DecreaseRefCountNoLock()
    {
        --m_uiRefCount;
    }
    // {secret}
    void IncreaseRefCountNoLock()
    {
        ++m_uiRefCount;
    }

#endif  // NEX_ENABLE_DOREF_LOCK_OPTIMIZATION
    void PrepareToDestroy()
    {
        m_uiRefCount = 1;
        m_uiRelevanceCount = 1;
    }

protected:

    void SetStationSpecialRelevance();

private:
    // {secret}
    static DuplicatedObject *CreateMasterImpl(DOHandle hMaster, DOClassID idDOClass, DOID idUserSuppliedDOID);


    /******
     *
     * Access to DuplicationSets
     *
     ******/
public:

    // {secret}
    qBool IsDuplicatedOn(DOHandle hLocation);

    // {secret}
    qBool IsInCachedDuplicationSet(DOHandle hStation) const;
    // {secret}
    qBool AddToCachedDuplicationSet(const Station* pStation);
    // {secret}
    qBool RemoveFromCachedDuplicationSet(DOHandle hStation);

    // {secret}
    qBool IsInDuplicationSet(DOHandle hStation) const;

private:
    // Always called via the ChangeDupSetOperation operation.
    // {secret}
    qBool AddToDuplicationSet(Station* pStation);
    qBool RemoveFromDuplicationSet(DOHandle hStation);

protected:
    // When a duplica is fetched, even if the duplica is currently
    // in the DupSet, we re-add it to the DupSet.  We hide the
    // removal to the user (no corresponding operation), since
    // logically, there has been no removal.
    // {secret}
    qBool ForceRedundantAddDuplica(DOHandle hStation)
    {
        return RemoveFromDuplicationSet(hStation);
    }

public:
    // This method is called by a leaving station.
    // {secret}
    static void RemoveAllDuplicasOnLeavingStation(DOHandle hStation);


    /******
     *
     * Implementation of DOProtocol basic operations
     *
     * See doprotocol.spin for use in protocol.
     *
     ******/

private:

    // Implementation of DOProtocol basic operations.
    // See doprotocol.spin for use in protocol.
    //{secret}
    void AcquireMainReference()
    {
        SYSTEMCHECK(!FlagIsSet(MainReferenceInUse));
        IncreaseRefCount();
        SetFlag(MainReferenceInUse);
    };

public:
    // TODO - Make private!
    //{secret}
    void ReleaseMainReference()
    {
        SYSTEMCHECK(FlagIsSet(MainReferenceInUse));
        ClearFlag(MainReferenceInUse);
        // The object could be destroyed after this.
        DecreaseRefCount();
    };




protected:

    /******
     *
     * Prevent use of copy constructor and assignment operation on
     * DuplicatedObjects.
     *
     ******/

#ifdef FOR_DOC_PURPOSE

#else
    DuplicatedObject(const DuplicatedObject&); // Not implemented.
    DuplicatedObject& operator=(const DuplicatedObject&); // Not implemented.
#endif

protected:

    /*******
     *
     * States
     *
     *******/

    // For NEX Developer <wiki/Duplicated Object State Machine#DOStates>.

    void SetInitialState(const StateMachine::QEvent& e);

    QSTATE InvalidState(const StateMachine::QEvent& e);

    QSTATE ValidState(const StateMachine::QEvent& e);
    QSTATE InitialState(const StateMachine::QEvent& e);
    QSTATE DuplicationMasterState(const StateMachine::QEvent& e);
    QSTATE UnpublishedMasterState(const StateMachine::QEvent& e);
    QSTATE UnidentifiedMasterState(const StateMachine::QEvent& e);
    QSTATE InStoreMasterState(const StateMachine::QEvent& e);
    QSTATE DeletedMasterState(const StateMachine::QEvent& e);

    QSTATE DuplicaState(const StateMachine::QEvent& e);
    QSTATE InStoreDuplicaState(const StateMachine::QEvent& e);
    QSTATE OrphanDuplicaState(const StateMachine::QEvent& e);
    QSTATE ConnectedDuplicaState(const StateMachine::QEvent& e);
    QSTATE DeletedDuplicaState(const StateMachine::QEvent& e);

private:
    /**********************************************
     **********************************************
     ***
     *** MEMBER VARIABLES
     ***
     **********************************************
     **********************************************/

    /******
     *
     * Member variables used in ref counting
     *
     ******/

    qUnsignedInt16 m_uiRefCount;
    qUnsignedInt16 m_uiRelevanceCount;
    static CriticalSection s_csRefCount;


    /******
     *
     * Member variables that compose the DO state
     *
     * All write access to these values are protected via the
     * system lock.  The method CheckStateAccess is called in all
     * methods that write to it, to system check if it isn't the case.
     *
     ******/

    MasterStationRef m_refMasterStation;

    qUnsignedInt16 m_uiFlags;

    Selection m_setDuplicationSet;

    DOHandle m_dohMyself;

    // Cached info, managed by the duplica.
    //{secret}
    Selection m_setCachedDuplicationSet;

public:
    /* Please see man pages for details 
     
     
     
     */
    static const qUnsignedInt32 DEFAULT_CREATE_TIMEOUT = 500;
    /* Please see man pages for details 
     
     
     
     */
    static const qUnsignedInt32 DEFAULT_PUBLISH_TIMEOUT = 500;
};

END_NAMESPACE_NEX

#endif

#include <NetZ/src/Core/DOCore/DOFilters.h>

