﻿/*---------------------------------------------------------------------------*
  (C) Copyright 1998-2014, Technologies Quazal Inc.  All rights reserved.
  Quazal Net-Z, Quazal Rendez-Vous and Quazal are trademarks of Quazal.

  Copyright (C)2011-2014 Nintendo  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo and/or its licensors and are
  protected by national and international copyright laws.
  They may not be disclosed to third parties or copied or duplicated
  in any form, in whole or in part, without the prior written consent
  of Nintendo.
  The content herein is highly confidential and should be handled
  accordingly.
 *---------------------------------------------------------------------------*/

#ifndef _MEMBERLISTITERATOR_H_
#define _MEMBERLISTITERATOR_H_

#include <NetZ/src/Core/DOCore/ObjDupDecl.h>

BEGIN_NAMESPACE_NEX

//Group = Member Containers

template<class TDDL_Type> class MemberList;

/*
Description:
This class offers standard C++ operators that can be performed on iterators over
a MemberList.

Remarks:
Assuming that the dataset MyContainer is declared in the DDL as follows:

        dataset MyContainer {
            list<real> lst;
        };

        doclass Avatar {
            MyContainer m_dsContainer;
        };

The following syntax creates an iterator for this container:

        MemberListIterator<DDLTYPE(qReal)> it;
*/
template<class TDDL_Type>
class MemberListIterator: public RootObject
{
    //Group =
public:
    // {secret}
    typedef typename TDDL_Type::CPPType Type;

    /*
    Remarks:
    Assuming that the dataset MyContainer is declared in the DDL as follows:

            dataset MyContainer {
                list<real> lst;
            };

            doclass Avatar {
                Mycontainer m_dsContainer;
            };

    The following syntax creates an iterator for this container:

            MemberListIterator<DDLTYPE(qReal)> it;

    Note: 'MemberListIterator' is the required name in order to specify that it
    is the 'lst' dataset member of MyContainer that we want to create the
    iterator for.
    */
    MemberListIterator()
    {
        m_uiPos = s_uiINVALID_POSITION;
    }

    MemberListIterator(const MemberListIterator& it)
    {
        m_pmlList = it.GetMemberList();
        m_pstdlList = m_pmlList->GetStdList();
        m_uiPos = it.m_uiPos;
        m_it = it.m_it;
    }

    //{secret}
    void Init(MemberList<TDDL_Type>& ml)
    {
        m_pmlList = &ml;
        m_pstdlList = m_pmlList->GetStdList();
        Begin();
    }

    //{secret}
    qBool IsValid() const
    {
        return (m_uiPos != s_uiINVALID_POSITION);
    }

    //{secret}
    void Invalidate()
    {
        m_uiPos = s_uiINVALID_POSITION;
    }

    //{secret}
    void Begin()
    {
        m_uiPos = 0;
        m_it = m_pstdlList->begin();
    }

    //{secret}
    qBool Next()
    {
        if (m_it != m_pstdlList->end() && IsValid())
        {
            ++m_it;
            m_uiPos++;
        }
        else
        {
            Invalidate();
        }
        return IsValid();
    }

    //{secret}
    qBool Previous()
    {
        if (m_it != m_pstdlList->begin() && IsValid())
        {
            --m_it;
            m_uiPos--;
        }
        else
        {
            Invalidate();
        }
        return IsValid();
    }

    // Member prefix (++x)
    MemberListIterator& operator++()
    {
        QVERIFY(Next());
        return *this;
    }

    // Member postfix (x++)
    MemberListIterator operator++(qInt iBid)
    {
        DUMMY_PARAM(iBid);
        const MemberListIterator itResult = *this;
        operator++();
        return itResult;
    }

    MemberListIterator& operator+(qInt iInc)
    {
        for(qInt i = 0; i < iInc; i++)
        {
            QVERIFY(Next());
        }
        return *this;
    }

    MemberListIterator& operator+=(qInt iInc)
    {
        for(qInt i = 0; i < iInc; i++)
        {
            QVERIFY(Next());
        }
        return *this;
    }

    // Member prefix (--x)
    MemberListIterator& operator--()
    {
        QVERIFY(Previous());
        return *this;
    }

    // Member postfix (x--)
    MemberListIterator operator--(qInt iBid)
    {
        DUMMY_PARAM(iBid);
        const MemberListIterator itResult = *this;
        operator--();
        return itResult;
    }

    MemberListIterator& operator-(qInt iInc)
    {
        for(qInt i = 0; i < iInc; i++)
        {
            QVERIFY(Previous());
        }
        return *this;
    }

    MemberListIterator& operator-=(qInt iInc)
    {
        for(qInt i = 0; i < iInc; i++)
        {
            QVERIFY(Previous());
        }
        return *this;
    }

    qInt operator-(MemberListIterator& it)
    {
        return  Refresh() - it.Refresh();
    }

    MemberListIterator& operator=(const Type& tElem)
    {
        *m_it = tElem;
        m_pmlList->Assign(*this, tElem);
        return *this;
    }

    MemberListIterator& operator=(const MemberListIterator& it)
    {
        m_pmlList = it.GetMemberList();
        m_pstdlList = m_pmlList->GetStdList();
        m_uiPos = it.m_uiPos;
        m_it = it.m_it;
        return *this;
    }

    qBool operator==(const MemberListIterator& it)
    {
        return (m_it == it.m_it);
    }

    qBool operator!=(const MemberListIterator& it)
    {
        return (m_it != it.m_it);
    }

    const Type& operator*() const
    {
        return *m_it;
    }

    //{secret}
    qInt Refresh()
    {
        // Asserting here through Previous() and Next() calls may imply that
        // operations like "Push_Front" were performed on the container. Such
        // scenario makes the iterators instances invalid.

        qInt iCount1 = 0, iCount2 = 0;
        while(m_it != m_pstdlList->begin())
        {
            QVERIFY(Previous());
            iCount1++;
        }

        while(iCount2 < iCount1)
        {
            QVERIFY(Next());
            iCount2++;
        }

        m_uiPos = iCount1;
        return m_uiPos;
    }

    //{secret}
    qBool SetPos(qUnsignedInt32 iPos)
    {
        Begin();
        for(qUnsignedInt i = 0; i <= iPos; i++)
        {
            QVERIFY(Next());
        }
        return IsValid();
    }

    //{secret}
    qUnsignedInt32 GetPos() const
    {
        return m_uiPos;
    }

    //{secret}
    qBool IncPos()
    {
        if(!IsValid())
        {
            return false;
        }
        m_uiPos++;
        ++m_it;
        return true;
    }

    //{secret}
    qBool IncPos(qInt i)
    {
        SYSTEMCHECK(i > 0);
        qInt n;
        for (n = 0; n < i && IncPos(); n++)
        {
        }
        return (n == i);
    }

    // This method is only used for MemberVector class implementation
    //{secret}
    void SetStdIterator(const typename qList<Type>::iterator& it)
    {
        m_it = it;
    }

    // This method is only used for MemberList class implementation
    //{secret}
    const typename qList<Type>::iterator& GetStdIterator() const
    {
        return m_it;
    }

private:
    //{secret}
    MemberList<TDDL_Type>* GetMemberList() const
    {
        return m_pmlList;
    }

    typename qList<Type>::iterator m_it;
    MemberList<TDDL_Type> *m_pmlList;
    qList<Type> *m_pstdlList;
    qUnsignedInt32 m_uiPos;

    static const qUnsignedInt32 s_uiINVALID_POSITION;
};

template<class TDDL_Type>
const qUnsignedInt32 MemberListIterator<TDDL_Type>::s_uiINVALID_POSITION = static_cast<qUnsignedInt32>(-1);

END_NAMESPACE_NEX

#endif//_MEMBERLISTITERATOR_H_
