﻿/*---------------------------------------------------------------------------*
  (C) Copyright 1998-2014, Technologies Quazal Inc.  All rights reserved.
  Quazal Net-Z, Quazal Rendez-Vous and Quazal are trademarks of Quazal.

  Copyright (C)2011-2014 Nintendo  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo and/or its licensors and are
  protected by national and international copyright laws.
  They may not be disclosed to third parties or copied or duplicated
  in any form, in whole or in part, without the prior written consent
  of Nintendo.
  The content herein is highly confidential and should be handled
  accordingly.
 *---------------------------------------------------------------------------*/

#ifndef _MEMBERVECTOR_H_
#define _MEMBERVECTOR_H_

#include <NetZ/src/Core/DOCore/ObjDupDecl.h>
#include <NetZ/src/Core/DOCore/MemberContainer.h>
#include <NetZ/src/Core/DOCore/MemberVectorIterator.h>
#include <OnlineCore/src/Platform/Core/qVector.h>

BEGIN_NAMESPACE_NEX

//Group = Member Containers

/*
Description:  Contains methods to control the elements of a DDL-defined MemberVector.
Remarks: MemberVector objects are basically just standard C++ vectors optimized for $PRODUCTNAME$.
The MemberVectorIterator keeps track of the operations performed on the MemberVector in order
to transmit updates efficiently, /i.e./ transmit just those elements that change between updates.
On the other hand, the std_vector transmits all its elements each update. In most cases it is
preferable to use MemberVectors however, if you need to do a lot of manipulation on a dataset
that changes infrequently you will probably find the std_vector easier to use given its standard API.

A MemberVector instance is associated with a doclass in the DDL file in the following way:

        dataset VectorDataset{
            vector<double> vec;
        };

        doclass MyDOClass{
            VectorDataset m_dsVectorOfDoubles;
        };

The MyDOClass implementation inherits from the DDL generated DOCLASS (MyDOClass) that
associates the m_dsVectorOfDoubles to your class.  Thus you can access this dataset member
and perform operations on it just like any other dataset member of your doclass.  For example:

        class MyDOClass: public DOCLASS( MyDOClass) {
            public:

            void PushBackAndUpdate(qReal dNewValue) {
                m_dsVectorOfDoubles.vec.Push_Back (dNewValue);
                Update();
            }
            // other user-implemented code
        }

*/
template<class TDDL_Type>
class MemberVector : public MemberContainer<TDDL_Type>
{
    //Group=Methods
public:
    //{secret}
    typedef MemberVectorIterator<TDDL_Type> Iterator;
    // {secret}
    typedef typename TDDL_Type::CPPType Type;
    //{secret}
    typedef typename NAMESPACE_NEX MemberContainer<TDDL_Type>::ContOp ContOp_t;

    //{secret}
    MemberVector()
    {
        m_it.Init(*this);
        m_begin.Init(*this);
        m_end.Init(*this);
    }
    //{secret}
    MemberVector(const MemberVector<TDDL_Type>& mv)
    {
        m_it.Init(*this);
        m_begin.Init(*this);
        m_end.Init(*this);

        // Need because the const version of Iterator return an Iterator const (and not a iterator of const type)
        for(Iterator it = mv.Begin(); it != mv.End(); ++it)
        {
            Push_Back(*it);
        }
    }

    /*
    Description: Pushes an element to the back of the vector.
    Parameters:
    tElem: The element to insert.
    */
    void Push_Back(const Type& tElem)
    {
        MemberContainer<TDDL_Type>::AddOperation(ContOp_t(ContOp_t::PUSH_BACK_OPCODE, tElem));
        m_stdvVector.push_back(tElem);
    }

    /*
    Description: Pops an element from the back of the vector.
    */
    void Pop_Back()
    {
        MemberContainer<TDDL_Type>::AddOperation(ContOp_t(ContOp_t::POP_BACK_OPCODE));
        m_stdvVector.pop_back();
    }

    /*
    Description: Inserts an element into a MemberVector at the current position of the MemberVectorIterator.
    Parameters:
    it:  The MemberVectorIterator to use to insert the element.
    tElem: The element to insert.
    n: Optional parameter that specifies how many copies of the element will be inserted.
    Returns: An iterator that points to the position where the new element was
    inserted into the vector. Or if the Boolean method is used, true if
    the element is successfully inserted into the MemberVector, false if
    not or an error occurred.
    */
    Iterator& Insert(Iterator& it, const Type& tElem)
    {
        SYSTEMCHECK(it.IsValid());
        it.Refresh();
        it.SetStdIterator(m_stdvVector.insert(it.GetStdIterator(), tElem));
        //it.IncPos();
        MemberContainer<TDDL_Type>::AddOperation(ContOp_t(ContOp_t::INSERT_PX_OPCODE, it.GetPos(), tElem));
        return it;
    }

    Iterator Insert(Iterator it, qUnsignedInt n, const Type& tElem)
    {
        SYSTEMCHECK(it.IsValid());
        it.Refresh();
        m_stdvVector.insert(it.GetStdIterator(), n, tElem);
        MemberContainer<TDDL_Type>::AddOperation(ContOp_t(ContOp_t::INSERT_PNX_OPCODE, it.GetPos(), n, tElem));
        return it;
    }

    // This method is only used for MemberVectorIterator::operator= implementation
    // NOTE: This is an internal method not to mix up with Assign(n, tElement)
    //{secret}
    qBool Assign(Iterator& it, const Type& tElem)
    {
        if(!it.IsValid())
        {
            return false;
        }

        it->Refresh();
        MemberContainer<TDDL_Type>::AddOperation(ContOp_t(ContOp_t::ASSIGN_OPCODE, it.GetPos(), tElem));
        return true;
    }

    /*
    Description:  Erases the vector content and copies the specified element n
    times in the empty vector.
    Parameters:
    n: The number of copies of an element being inserted into the vector.
    tElement: The value of the element being inserted into the vector.
    */
    void Assign(typename qVector<Type>::size_type n, const Type& tElement)
    {
        Clear();
        Insert(Begin(), n, tElement);
    }

    /*
    Description:  Returns the element at the specified position.
    Returns:  The element at the specified position.
    Parameters:
    iPos: Index into the vector
    */
    const Type& operator[](const qInt iPos) const
    {
        return m_stdvVector[iPos];
    }

    /*
    Description: Sets the element at the specified position.
    Parameters:
    iPos: Element Index into the vector
    tElement: The element to assigned at the specified index.
    */
    void SetElement(const qInt iPos, const Type& tElement)
    {
        ReplayAssign(iPos, tElement);
        MemberContainer<TDDL_Type>::AddOperation(ContOp_t(ContOp_t::ASSIGN_OPCODE, iPos, tElement));
    }

    /*
    Description:  Returns the element at the specified position.
    Returns:  The element at the specified position.
    Parameters:
    iPos: Index into the vector
    */
    const Type& GetElement(const qInt iPos) const
    {
        return m_stdvVector[iPos];
    }

    /*
    Description: Removes an element from specified positions in the vector.
    Parameters:
    it: The iterator indicating which element to erase.
    Remarks: This operation invalidates the iterator.
    */
    Iterator& Erase(Iterator& it)
    {
        it.Refresh();
        it.SetStdIterator(m_stdvVector.erase(it.GetStdIterator()));
        MemberContainer<TDDL_Type>::AddOperation(ContOp_t(ContOp_t::ERASE_OPCODE, it.GetPos(), 0));
        it.Invalidate();
        return it;
    }

    //{secret}
    void Clear()
    {
        MemberContainer<TDDL_Type>::AddOperation(ContOp_t(ContOp_t::CLEAR_OPCODE));
        m_stdvVector.clear();
    }

    /*
    Description: Returns the element at the front of the MemberVector.
    Returns: The element at the front of the vector.
    */
    const Type& Front() const
    {
        return m_stdvVector.front();
    }

    /*
    Description:  Returns the element at the back of the MemberVector.
    Returns: The element at the back of the vector.
    */
    const Type& Back() const
    {
        return m_stdvVector.back();
    }

    /*
    Description:  Returns whether or not the MemberVector is empty.
    Returns:  true if the MemberVector is empty, false if not or an error occurred.
    */
    qBool Empty() const
    {
        return m_stdvVector.empty();
    }

    /*
    Description:  Returns the number of elements in the MemberVector.
    Returns:  The number of elements in the MemberVector.
    */
    typename qVector<Type>::size_type Size() const
    {
        return m_stdvVector.size();
    }

    /*
    Description:  Sets the MemberVectorIterator to point to the first element in the MemberVector.
    Returns:  The MemberVectorIterator that now points to the beginning of the MemberVector.
    */
    const Iterator& Begin() const
    {
        MemberVector<TDDL_Type>& oThis = *const_cast<MemberVector<TDDL_Type>*>(this);
        oThis.m_begin.SetStdIterator(oThis.m_stdvVector.begin());
        return m_begin;
    }

    /*
    Description:  Sets the MemberVectorIterator to point to the last element in the MemberVector.
    Returns:  The MemberVectorIterator that now points to the last element in the MemberVector.
    */
    Iterator& End()
    {
        m_end.SetStdIterator(m_stdvVector.end());
        SYSTEMCHECK(m_end.GetStdIterator() == m_stdvVector.end());
        return m_end;
    }

    const Iterator& End() const
    {
        MemberVector<TDDL_Type>& oThis = *const_cast<MemberVector<TDDL_Type>*>(this);
        oThis.m_end.SetStdIterator(oThis.m_stdvVector.end());
        SYSTEMCHECK(m_end.GetStdIterator() == m_stdvVector.end());
        return m_end;
    }

private:
    ////////////////////////////
    // Replay operations (BEGIN)
    ////////////////////////////
    //{secret}
    void ReplayInsert(qInt iPos, const Type& tElem)
    {
        m_it.Begin();
        for(qInt i = 0; i < iPos-1; i++)
        {
            ++m_it;
        }
        m_it.SetStdIterator(m_stdvVector.insert(m_it.GetStdIterator(), tElem));
    }

    //{secret}
    void ReplayInsert(qInt iPos, qUnsignedInt n, const Type& tElem)
    {
        m_it.Begin();
        for(qUnsignedInt i = 0; i < (iPos-n); i++)
        {
            ++m_it;
        }
        m_stdvVector.insert(m_it.GetStdIterator(), n, tElem);
    }

    //{secret}
    void ReplayErase(qInt iPos)
    {
        m_it.Begin();
        for(qInt i = 0; i < iPos; i++)
        {
            ++m_it;
        }
        m_stdvVector.erase(m_it.GetStdIterator());
    }

    //{secret}
    void ReplayAssign(qInt iPos, const Type& tElem)
    {
        m_it.Begin();
        typename qVector<Type>::iterator stdIt;
        for(qInt i = 0; i < iPos; i++)
        {
            ++m_it;
        }
        stdIt = m_it.GetStdIterator();
        *stdIt = tElem;
    }
    ////////////////////////////
    // Replay operations (END)
    ////////////////////////////

protected:
    //{secret}
    virtual void AddContentToMsg(Message *pMsg)
    {
        const qUnsignedInt uiSize = m_stdvVector.size();
        *pMsg << uiSize;
        for(qUnsignedInt i = 0; i < m_stdvVector.size(); i++)
        {
            TDDL_Type::Add(pMsg, m_stdvVector[i]);
        }
    }

    //{secret}
    virtual void ExtractContentFromMsg(Message *pMsg)
    {
        qUnsignedInt uiNbElem=0;
        Type tElem;
        m_stdvVector.clear();
        *pMsg >> uiNbElem;
        typename qVector<Type>::iterator it = m_stdvVector.begin();
        for(qUnsignedInt i = 0; i < uiNbElem; i++)
        {
            TDDL_Type::Extract(pMsg, &tElem);
            it = m_stdvVector.insert(it, tElem);
            ++it;
        }
    }

public:
    //{secret}
    virtual qBool ReplayOperations()
    {
        ContOp_t *pcoOpTmp=NULL;
        while(!MemberContainer<TDDL_Type>::m_coOperationList.empty())
        {
            pcoOpTmp = MemberContainer<TDDL_Type>::m_coOperationList.back().GetPtr();
            switch(pcoOpTmp->GetOpCode())
            {
            case ContOp_t::PUSH_BACK_OPCODE:
                m_stdvVector.push_back(pcoOpTmp->GetData());
                break;
            case ContOp_t::POP_BACK_OPCODE:
                m_stdvVector.pop_back();
                break;
            case ContOp_t::INSERT_PX_OPCODE:
                ReplayInsert(pcoOpTmp->GetPos(), pcoOpTmp->GetData());
                break;
            case ContOp_t::INSERT_PNX_OPCODE:
                ReplayInsert(pcoOpTmp->GetPos(), pcoOpTmp->GetRepeat(), pcoOpTmp->GetData());
                break;
            case ContOp_t::INSERT_PFL_OPCODE:
                SYSTEMCHECK(false && _T("INSERT_PFL_OPCODE not implemented."));
                break;
            case ContOp_t::ERASE_OPCODE:
                ReplayErase(pcoOpTmp->GetPos());
                break;
            case ContOp_t::CLEAR_OPCODE:
                m_stdvVector.clear();
                break;
            case ContOp_t::ASSIGN_OPCODE:
                ReplayAssign(pcoOpTmp->GetPos(), pcoOpTmp->GetData());
                break;
            default:
                SYSTEMCHECK(false && _T("Invalid container operation code for MemberVector."));
            }
            MemberContainer<TDDL_Type>::m_coOperationList.pop_back();
        }
        return true;
    }

    //{secret}
    qVector<Type>* GetStdVector()
    {
        return &m_stdvVector;
    }

    MemberVector<TDDL_Type>& operator=(const MemberVector<TDDL_Type>& mvSource)
    {
        MemberVector<TDDL_Type>& mv = const_cast<MemberVector&>(mvSource);
        Clear();

        for(Iterator it = mv.Begin(); it != mv.End(); ++it)
        {
            Push_Back(*it);
        }

        return *this;
    }

    virtual qBool operator==(const MemberVector<TDDL_Type>& mv) const
    {
        return (this->m_stdvVector == mv.m_stdvVector);
    }
    virtual qBool operator!=(const MemberVector<TDDL_Type>& mv) const
    {
        return (this->m_stdvVector != mv.m_stdvVector);
    }


#if defined(TRACE_ENABLED)
    // Warning: TraceContent assumes the template type is qReal
    //{secret}
    qBool TraceContent(qUnsignedInt64 uiTraceFlag) const
    {
        QTRACE(uiTraceFlag, _T("\n\nTracing vector content: "));
        for(qUnsignedInt i = 0; i < m_stdvVector.size(); i++)
        {
            QTRACE(uiTraceFlag, m_stdvVector[i]);
        }
        return true;
    }
#endif

private:
    virtual qUnsignedInt32 SizeOfContent() const
    {
        return sizeof(Type) * m_stdvVector.size();
    }

private:
    qVector<Type> m_stdvVector;
    Iterator m_it;
    Iterator m_begin;
    Iterator m_end;
};

END_NAMESPACE_NEX

#endif//_MEMBERVECTOR_H_
