﻿/*---------------------------------------------------------------------------*
  (C) Copyright 1998-2014, Technologies Quazal Inc.  All rights reserved.
  Quazal Net-Z, Quazal Rendez-Vous and Quazal are trademarks of Quazal.

  Copyright (C)2011-2014 Nintendo  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo and/or its licensors and are
  protected by national and international copyright laws.
  They may not be disclosed to third parties or copied or duplicated
  in any form, in whole or in part, without the prior written consent
  of Nintendo.
  The content herein is highly confidential and should be handled
  accordingly.
 *---------------------------------------------------------------------------*/

#ifndef _MEMBERVECTORITERATOR_H_
#define _MEMBERVECTORITERATOR_H_

#include <NetZ/src/Core/DOCore/ObjDupDecl.h>

BEGIN_NAMESPACE_NEX

//Group = Member Containers

template<class TDDL_Type> class MemberVector;

/*
Description:
Offers standard C++ operators that can be performed on iterators over
a MemberVector.

Remarks:
Assuming that the dataset MyContainer is declared in the DDL as follows:

        dataset MyContainer {
            vector<real> vec;
        };

        doclass Avatar {
            MyContainer m_dsContainer;
        };

The following syntax creates an iterator for this container:

        MemberVectorIterator<DDLTYPE(qReal)> it;
*/
template<class TDDL_Type>
class MemberVectorIterator: public RootObject
{
    //Group = Operations
public:
    // {secret}
    typedef typename TDDL_Type::CPPType Type;

    /*
    Remarks:
    Assuming that the dataset MyContainer is declared in the DDL as follows:

            dataset MyContainer {
                vector<real> vec;
            };

            doclass Avatar {
                MyContainer m_dsContainer;
            };

    The following syntax creates an iterator for this container:

            MemberVectorIterator<DDLTYPE(qReal)> it;

    Note: 'MemberVectorIterator' is the required name in order to specify that
    it is the 'vec' dataset member of MyContainer that we want to create the
    iterator for.
    */
    MemberVectorIterator()
    {
        m_uiPos = s_uiINVALID_POSITION;
    }

    MemberVectorIterator(const MemberVectorIterator& it)
    {
        m_pmvVector = it.GetMemberVector();
        m_pstdvVector = m_pmvVector->GetStdVector();
        m_uiPos = it.m_uiPos;
        m_it = it.m_it;
    }

    //{secret}
    void Init(MemberVector<TDDL_Type>& mv)
    {
        m_pmvVector = &mv;
        m_pstdvVector = m_pmvVector->GetStdVector();
        Begin();
    }

    //{secret}
    qBool IsValid() const
    {
        return (m_uiPos != s_uiINVALID_POSITION);
    }

    //{secret}
    void Invalidate()
    {
        m_uiPos = s_uiINVALID_POSITION;
    }

    //{secret}
    void Begin()
    {
        m_uiPos = 0;
        m_it = m_pstdvVector->begin();
    }

    //{secret}
    qBool Next()
    {
        if (m_it != m_pstdvVector->end() && IsValid())
        {
            ++m_it;
            m_uiPos++;
        }
        else
        {
            Invalidate();
        }
        return IsValid();
    }

    //{secret}
    qBool Previous()
    {
        if (m_it != m_pstdvVector->begin() && IsValid())
        {
            --m_it;
            m_uiPos--;
        }
        else
        {
            Invalidate();
        }
        return IsValid();
    }

    // Member prefix (++x)
    MemberVectorIterator& operator++()
    {
        QVERIFY(Next());
        return *this;
    }

    // Member postfix (x++)
    MemberVectorIterator operator++(qInt iBid)
    {
        DUMMY_PARAM(iBid);
        const MemberVectorIterator itResult = *this;
        operator++();
        return itResult;
    }

    MemberVectorIterator& operator+(qInt iInc)
    {
        for(qInt i = 0; i < iInc; i++)
        {
            QVERIFY(Next());
        }
        return *this;
    }

    MemberVectorIterator& operator+=(qInt iInc)
    {
        for(qInt i = 0; i < iInc; i++)
        {
            QVERIFY(Next());
        }
        return *this;
    }

    // Member prefix (--x)
    MemberVectorIterator& operator--()
    {
        QVERIFY(Previous());
        return *this;
    }

    MemberVectorIterator operator--(qInt iBid)
    {
        DUMMY_PARAM(iBid);
        const MemberVectorIterator itResult = *this;
        operator--();
        return itResult;
    }

    MemberVectorIterator& operator-(qInt iInc)
    {
        for(qInt i = 0; i < iInc; i++)
        {
            QVERIFY(Previous());
        }
        return *this;
    }

    MemberVectorIterator& operator-=(qInt iInc)
    {
        for(qInt i = 0; i < iInc; i++)
        {
            QVERIFY(Previous());
        }
        return *this;
    }

    qInt operator-(MemberVectorIterator& it)
    {
        return  Refresh() - it.Refresh();
    }

    MemberVectorIterator& operator=(const Type& tElem)
    {
        *m_it = tElem;
        m_pmvVector->Assign(*this, tElem);
        return *this;
    }

    MemberVectorIterator& operator=(const MemberVectorIterator& it)
    {
        m_pmvVector = it.GetMemberVector();
        m_pstdvVector = m_pmvVector->GetStdVector();
        m_uiPos = it.m_uiPos;
        m_it = it.m_it;
        return *this;
    }

    qBool operator==(const MemberVectorIterator& it)
    {
        return (m_it == it.m_it);
    }

    qBool operator!=(const MemberVectorIterator& it)
    {
        return (m_it != it.m_it);
    }

    const Type& operator*() const
    {
        return *m_it;
    }

    //{secret}
    qInt Refresh()
    {
        // Asserting here through Previous() and Next() calls may imply that
        // operations like "Push_Front" were performed on the container. Such
        // scenario makes the iterators instances invalid.

        qInt iCount1 = 0, iCount2 = 0;
        while(m_it != m_pstdvVector->begin())
        {
            QVERIFY(Previous());
            iCount1++;
        }

        while(iCount2 < iCount1)
        {
            QVERIFY(Next());
            iCount2++;
        }

        m_uiPos = iCount1;
        return m_uiPos;
    }

    //{secret}
    qBool SetPos(qUnsignedInt32 iPos)
    {
        Begin();
        for(qUnsignedInt i = 0; i <= iPos; i++)
        {
            QVERIFY(Next());
        }
        return IsValid();
    }

    //{secret}
    qUnsignedInt32 GetPos() const
    {
        return m_uiPos;
    }

    //{secret}
    qBool IncPos()
    {
        if(!IsValid())
        {
            return false;
        }
        m_uiPos++;
        ++m_it;
        return true;
    }

    //{secret}
    qBool IncPos(qInt i)
    {
        SYSTEMCHECK(i > 0);
        qInt n;
        for (n = 0; n < i && IncPos(); n++)
        {
        }
        return (n == i);
    }

    // This method is only used for MemberVector class implementation
    //{secret}
    void SetStdIterator(const typename qVector<Type>::iterator& it)
    {
        m_it = it;
    }

    // This method is only used for MemberVector class implementation
    //{secret}
    const typename qVector<Type>::iterator& GetStdIterator() const
    {
        return m_it;
    }

private:
    //{secret}
    MemberVector<TDDL_Type>* GetMemberVector() const
    {
        return m_pmvVector;
    }

    typename qVector<Type>::iterator m_it;
    MemberVector<TDDL_Type> *m_pmvVector;
    qVector<Type> *m_pstdvVector;
    qUnsignedInt32 m_uiPos;

    static const qUnsignedInt32 s_uiINVALID_POSITION;
};

template<class TDDL_Type>
const qUnsignedInt32 MemberVectorIterator<TDDL_Type>::s_uiINVALID_POSITION = static_cast<qUnsignedInt32>(-1);

END_NAMESPACE_NEX

#endif//_MEMBERVECTORITERATOR_H_
