/*---------------------------------------------------------------------------*
  (C) Copyright 1998-2014, Technologies Quazal Inc. All rights reserved.
  Quazal Net-Z, Quazal Rendez-Vous and Quazal are trademarks of Quazal.

  Copyright (C)2011-2014 Nintendo  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo and/or its licensors and are
  protected by national and international copyright laws.
  They may not be disclosed to third parties or copied or duplicated
  in any form, in whole or in part, without the prior written consent
  of Nintendo.
  The content herein is highly confidential and should be handled
  accordingly.
 *---------------------------------------------------------------------------*/

#ifndef _RefTemplate_H_
#define _RefTemplate_H_

#include <NetZ/src/Core/DOCore/ObjDupDecl.h>
#include <NetZ/src/Core/DOCore/SystemError.h>
#include <NetZ/src/Core/DOCore/DORef.h>
#include <NetZ/src/Core/DOCore/DOClass.h>

BEGIN_NAMESPACE_NEX


class DuplicatedObject;

DOClass* GetDOClass(DuplicatedObject* pDO);

// Group=Template Classes

/*
Description: Creates a safe way to access a duplicated object.
Remarks: In $PRODUCTNAME$, the safe way to access a duplicated object is by reference.  The
template RefTemplate provides a safe reference to a duplicated object. You can create a reference
to a duplicated object to call any of your C++ methods.  If you have a reference to an object,
you can perform actions on it just like any other C++ object.

Every duplicated object class defines a typedef to the appropriate RefTemplate.

As long as a reference to a duplicated object exists, the duplicated object pointer
will not be invalidated by $PRODUCTNAME$.  Typically a reference would be created
on the stack and destroyed as soon as it was no longer required. It is important to
ensure that a reference to an object is only kept for a short period of time as even
though an object wil be removed from a given station's object store (refer to RemoveFromStoreOperation),
the object's destructor cannot be called while there is still a vaild reference to the object.

A reference may be created by specifying a DOHandle to an object or by specifying a reference to
an object. The default constructor may also be used and then an object assigned to the reference
via the assignment operator.

For example, if the DOClass Sphere is defined in the DDL and hSphere is a DOHandle to a Sphere,
then the following code provides access to the Sphere object referenced by hSphere.

        {
            Sphere::Ref refSphere(hSphere);
                if (refSphere.IsValid()){
                    // Call the Sphere method SelectSound;
                    refSphere->SelectSound();
                    // Call the action PlaySound on all duplicas of this Sphere
                    refSphere->PlaySound_OnDuplicas();
                }
        }

Note: All references to a duplicated object must be released before $PRODUCTNAME$ can call the
object's destructor.
See also: DuplicatedObject::Ref, DORef
*/
/* Please see man pages for details 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 */
template<class DO> class RefTemplate: public DORef
{
    // Group=Methods
public:
    /*
    Description: Creates a reference to a duplicated object.
    Parameters:
    hDO: DOHandle that refers to the duplicated object to which a
    reference is wanted.
    oRef: A reference to a reference of a duplicated object.
    Remarks: The method DORef::IsValid should be used to check if the reference was correctly
    constructed.
    See Also: Duplicated Object Overview, DORef
    */
    /* Please see man pages for details 
     
     
     
     */
    RefTemplate() {}
    /* Please see man pages for details 
     
     
     
     
     
     
     
     
     
     
     
     
     */
    RefTemplate(DOHandle hDO): DORef(hDO) {}
    /* Please see man pages for details 
     
     
     
     
     
     
     
     
     
     
     */
    RefTemplate(const DORef& oRef): DORef(oRef) {}

    virtual ~RefTemplate() {};

    DO* operator->()
    {
        if (!IsValid())
        {
            return NULL;
        }
        else
        {
            return (DO*) GetDOPtr();
        }
    }

    const DO* operator->() const
    {
        if (!IsValid())
        {
            return NULL;
        }
        else
        {
            return (const DO*) GetDOPtr();
        }
    }

    /*
    Summary: Provides access to the duplicated object.
    Remarks: Returns a pointer to the duplicated object of a specific doclass.  This
    pointer can safely be used to call actions or methods.
    See Also: Duplicated Object Overview, DORef
    */
    /* Please see man pages for details 
     
     
     
     
     
     */
    DO* operator*()
    {
        if (!IsValid())
        {
            return NULL;
        }
        else
        {
            return (DO*) GetDOPtr();
        }
    }

    const DO* operator*() const
    {
        if (!IsValid())
        {
            return NULL;
        }
        else
        {
            return (const DO*) GetDOPtr();
        }
    }

    RefTemplate<DO>& operator=(DO* pDO)
    {
        DORef::operator=(pDO);
        return *this;
    }

    inline qBool IsValid() const
    {
        if (GetDOPtr()==NULL)
        {
            SIGNAL_SYSTEMERROR(SYSTEMERROR_OBJDUP_DO_NOT_FOUND, 0);
            return false;
        }
        else
        {
            if (!GetDOClass(GetDOPtr())->IsAKindOf(DO::GetRealDOClassID()))
            {
                SIGNAL_SYSTEMERROR(SYSTEMERROR_OBJDUP_MISMATCHED_DOCLASS, 0);
                return false;
            }
            return true;
        }
    }

};


END_NAMESPACE_NEX

#endif
