﻿/*---------------------------------------------------------------------------*
  (C) Copyright 1998-2014, Technologies Quazal Inc.  All rights reserved.
  Quazal Net-Z, Quazal Rendez-Vous and Quazal are trademarks of Quazal.

  Copyright (C)2011-2014 Nintendo  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo and/or its licensors and are
  protected by national and international copyright laws.
  They may not be disclosed to third parties or copied or duplicated
  in any form, in whole or in part, without the prior written consent
  of Nintendo.
  The content herein is highly confidential and should be handled
  accordingly.
 *---------------------------------------------------------------------------*/

#ifndef _RemoveFromStoreOperation_H_
#define _RemoveFromStoreOperation_H_

#include <NetZ/src/Core/DOCore/ObjDupDecl.h>
#include <NetZ/src/Core/DOCore/DOOperation.h>
#include <OnlineCore/src/Core/PseudoGlobalVariable.h>

BEGIN_NAMESPACE_NEX


//Group = System Operations

/*
Description: This DOOperation is performed each time a duplicated object is deleted.

Remarks: Each Station maintains its own duplicated object store that contains a list of all the active local
Duplication master (locally published) and Duplica (locally discovered) objects. Objects are added to and removed
from the store as a result of objects being created and deleted at run time. Only objects in the store are
considered active and therefore, updated on the local Station. A Duplication master is considered active if it
exists on the local station, that is, DuplicatedObject::Publish has been called and DuplicatedObject::DeleteMainRef
has not yet been called. A Duplica is considered active if its Duplication master exists and the object is relevant
on the local station, that is, it is relevant in a Duplication space (DuplicationSpace::CheckRelevance returns true)
or a Hard reference to the object exists.

The RemoveFromStoreOperation behaves differently depending on whether it is a Duplication master
or a Duplica that is being deleted.

A Duplication master instance is deleted as a result of one of these two possible sequences of events:
    o   A station that holds a Duplication master instance leaves the session and the DuplicatedObject::ApproveEmigration or DuplicatedObject::ApproveFaultRecovery system callback returns false,
    o   A Duplication master instance of a duplicated object is deleted via the method DuplicatedObject::DeleteMainRef.

When an object is removed from the duplication master store the following sequence of events occurs:
    o   The SafetyExecutive is called to determine if the object may be deleted. If so, we continue, if not we exit.
    o   The method DataSet::OperationBegin is called on all datasets,
    o   The method DuplicatedObject::OperationBegin is called,
    o   The duplicas of the Duplication master object are notified of the deletion via a multicasted message which results in duplica deletion operations (see below),
    o   The reference to the object in the duplication master store is deleted,
    o   The method DataSet::OperationEnd is called on all datasets,
    o   If in the DDL the object has an adapter associated to it, then the AdapterEvent (deprecated class) that corresponds to the deletion of the object is triggered,
    o   The method DuplicatedObject::OperationEnd is called. Iterations can no longer be performed over the duplication master object.
    o   When there are no more references (hard or soft) to the object, created via the RefTemplate, the object destructor is called.

A Duplica instance may also be deleted as a result of the DuplicationSpace::CheckRelevance and DuplicationSpace::PSMatch
virtual functions returning false.

When an object is deleted from the duplica store, the following sequence of events occurs:
    o   The SafetyExecutive is called to determine if the object may be deleted. If so, we continue, if not we exit.
    o   The method DataSet::OperationBegin is called on all datasets,
    o   The method DuplicatedObject::OperationBegin is called,
    o   The reference to the object in the duplica store is deleted,
    o   The method DataSet::OperationEnd is called on all datasets,
    o   If in the DDL the object has an adapter associated to it, then the AdapterEvent (deprecated class) that corresponds to the deletion of the object is triggered,
    o   The method DuplicatedObject::OperationEnd is called. Iterations can no longer be performed over the duplica object.
    o   When there are no more references (hard or soft) to the object, created via the RefTemplate, the object destructor is called.

This operation will not be called on a DuplicatedObject when the system state is locked via Scheduler::SystemLock.

Note: Note that the Adapter extension has been deprecated and will be removed in the next release. Refer to System
Operations for information on how to interface discreet class hierarchies. Events can now be caught by a system callback
or a DO callback for a specific object.

When $PRODUCTNAME$ triggers the AdapterEvent (deprecated class), if the Adapter has not previously been deleted the
DeleteName method for the Adapter is called. The processing of the event, which is controlled via the
AdapterEventDispatcher class, could be processed directly or queued and processed when the
DispatchQueue::Dispatch method, or a similar method, is called.

See Also: DuplicatedObject::DeleteMainRef, System Operations, Duplicated Object Overview
*/
/* Please see man pages for details 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 */
class RemoveFromStoreOperation: public DOOperation
{

public:
    //{secret}
    RemoveFromStoreOperation(DOHandle hOperationOrigin, DuplicatedObject* pDO, qBool bDeleted, qBool bRemoveDuplicas);
    //{secret}
    virtual ~RemoveFromStoreOperation();

    //Group=Methods

    /*
    Description: Returns whether the RemoveFromStoreOperation was triggered by a duplicated object deletion.
    Remarks: Returns true if the RemoveFromStoreOperation is being performed as a result of a call to
    DuplicatedObject::DeleteMainRef or the disconnection from the session of a master
    station for a non-fault-tolerant duplicated object.
    Returns: true if the RemoveFromStoreOperation is being performed as a result of a call to
    DuplicatedObject::DeleteMainRef or the disconnection from the session of a master
    station followed by a return of false for the DuplicatedObject::ApproveEmigration or
    DuplicatedObject::ApproveFaultRecovery system callbacks, false if not or an error occurred.
    */
    /* Please see man pages for details
		 
		 
		 
		 
		 
		 
     */
    qBool ObjectIsDeleted()
    {
        return m_bDeleted;
    }

    /*
    Description: Returns whether the RemoveFromStoreOperation was triggered by a Relevance check in
    the DuplicationSpace.
    Remarks: This method can only return true if the RemoveFromStoreOperation is being performed on a Duplica,
    since only duplicas can become irrelevant.  If this method returns true, it implies that the
    removal from the duplica store that is performed is a result of a DuplicationSpace::CheckRelevance
    call followed by the DuplicationSpace::PSMatch function returning false.
    Returns: true if the RemoveFromStoreOperation is being performed as a result of a
    DuplicationSpace::CheckRelevance call followed by the DuplicationSpace::PSMatch function returning false.
    False if not or an error occurred.
    See Also: DOOperation::GetAssociatedDO
    */
    qBool ObjectIsIrrelevant() const
    {
        return !m_bDeleted;
    }

    // {secret}
    qBool RemoveDuplicas() const
    {
        return m_bRemoveDuplicas;
    }

    IMPL_OPERATION_COMMON_METHODS( RemoveFromStore );

private:

#ifdef TRACE_ENABLED    // n2370
    virtual void TraceImpl(Event eEvent, qUnsignedInt64 uiTraceFlags) const;
#endif  // TRACE_ENABLED

    //{secret}
    virtual qBool CallsBackOnDataSet()
    {
        return true;
    }
    //{secret}
    virtual qBool CallsBackOnDataSet(qByte /*byIndex*/)
    {
        return true;
    }

private:
    //{secret}
    qBool m_bDeleted;
    //{secret}
    qBool m_bRemoveDuplicas;
};

END_NAMESPACE_NEX

#endif
