﻿/*---------------------------------------------------------------------------*
  (C) Copyright 1998-2014, Technologies Quazal Inc. All rights reserved.
  Quazal Net-Z, Quazal Rendez-Vous and Quazal are trademarks of Quazal.

  Copyright (C)2011-2014 Nintendo  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo and/or its licensors and are
  protected by national and international copyright laws.
  They may not be disclosed to third parties or copied or duplicated
  in any form, in whole or in part, without the prior written consent
  of Nintendo.
  The content herein is highly confidential and should be handled
  accordingly.
 *---------------------------------------------------------------------------*/

#ifndef _SafetyExecutive_H_
#define _SafetyExecutive_H_

#include <NetZ/src/Core/DOCore/ObjDupDecl.h>
#include <OnlineCore/src/Core/PseudoSingleton.h>

BEGIN_NAMESPACE_NEX

class Operation;

// Group = System Operations

/*
Description: Approves or denies $PRODUCTNAME$ System Operations.
Remarks: This class is used to approve or deny $PRODUCTNAME$ System Operations before they are called
to prevent a given Operation being called incorrectly. An incorrect call may be due to a bug in
the game code or due to a hack performed by a player. During internal or beta testing of your game,
where cheating is unlikely, incorrectly called System Operations will most probably be due to bugs in
the game code.

The SafetyExecutive class can be used to prevent specific operations, such as object creation, deletion, and migration,
being perfomed on a given Station under specific conditions. For example, you could allow object migration
to occur under normal circumstances but then prevent migration under your deemed abnormal conditions.
Refer to RegisterCustomOperationCheck for details as to how to restrict when System Operations may be called.
*/
/* Please see man pages for details 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 */
class SafetyExecutive: public RootObject
{
    //Group =Type Definitions
public:
    /*
    Description: Defines a CustomOperationCheck.
    Parameters:
    /oOperation/ : The Operation that this CustomOperationCheck will be called on.
    Remarks: A CustomOperationCheck is registered via RegisterCustomOperationCheck. It sets the conditions
    under which a given Operation will be accepted or denied.
    See Also: RegisterCustomOperationCheck
    */
    /* Please see man pages for details 
     
     
     
     */
    typedef qBool (*CustomOperationCheck)(const Operation& oOperation);

    /*
    Description: Defines an InvalidOperationCallback.
    Parameters:
    /oOperation/ : The Operation that this InvalidOperationCallback will be called on.
    Remarks: An InvalidOperationCallback is registered via RegisterInvalidOperationCallback. It defines what
    action will be taken when an Operation is denied.
    See Also: RegisterInvalidOperationCallback
    */
    /* Please see man pages for details 
     
     
     
     */
    typedef void (*InvalidOperationCallback)(const Operation& oOperation);

    //Group =Methods

    // A SafetyExecutive object should never be directly instantiated. Instead, it is created by
    // $PRODUCTNAME$ on all Station s when a session is created.
    /* Please see man pages for details 
     
     
     
     */
    SafetyExecutive();

    // A SafetyExecutive object should never be directly destroyed. It is destroyed by $PRODUCTNAME$
    // when the Session terminates.
    /* Please see man pages for details 
     
     
     
     */
    virtual ~SafetyExecutive();

    /*
    Description: Returns a pointer to the SafetyExecutive instance.
    Returns: A reference to the SafetyExecutive object.
    Remarks: Use the reference returned by this method to control the approval or denial
    of $PRODUCTNAME$ System Operations. For example, register a custom operation check as follows:

            SafetyExecutive::GetInstance()->RegisterCustomOperationCheck(MyOperationCheck);
    */
    /* Please see man pages for details 
     
     
     
     
     
     
     
     
     
     */
    static SafetyExecutive* GetInstance();

    /*
    Description: Sets the conditions under which an Operation will be accepted or denied.
    Parameters:
    pfCustomOperationCheck: A user-defined function that defines the conditions
    under which an Operation is approved. This function must return true if the Operation
    is approved and false otherwise.
    Remarks: This method prevents a given Operation being called incorrectly, which may be due
    to a bug in the game code or due to a hack performed by a player. Currently only the
    AddToStoreOperation and ChangeDupSetOperation can be approved or denied via the SafetyExecutive,
    all other System Operations are automatically accepted. However, in future versions of $PRODUCTNAME$
    all System Operations will pass via the SafetyExecutive before being processed.

    For example, this method could be used to prevent Avatar objects being created on a client station
    as follows. The action taken, if any, when an Operation is denied, that is, when pfCustomOperationCheck returns
    false, is defined by the /pfInvalidOperationCallback/, which is registered via RegisterInvalidOperationCallback.

            // Define the pfCustomOperationCheck
            qBool DenyClientAvatarCreation(const Operation& oOperation) {
                // I will not create duplicas of my Avatar on a client station.
                if (oOperation.GetType()==Operation::ChangeDupSet) {
                    const ChangeDupSetOperation* pOp=ChangeDupSetOperation::DynamicCast(&oOperation);
                    if (pOp->IsADuplicaAddition() && pOp->GetAssociatedDO()->
                                                        IsA(DOCLASSID(Avatar))) {
                        Station::Ref refStation(pOp->GetTargetStation());
                        if (refStation->GetProcessType()==Station::ClientProcess) {
                            return false;
                        }
                    }
                }
                return true;
            }

            // Then register the custom check
            SafetyExecutive::GetInstance()->RegisterCustomOperationCheck(DenyClientAvatarCreation);

    See Also: RegisterInvalidOperationCallback
    */
    /* Please see man pages for details 
     
     
     
     
     
     
     
     
     
     
     
     
     
     
     
     */
    void RegisterCustomOperationCheck(CustomOperationCheck pfCustomOperationCheck);

    /*
    Description: Determines what action will be taken when an Operation
    is denied.
    Parameters:
    pfInvalidOperationCallback: A user-defined function that dictates what action will be taken when
    an Operation is denied.
    Remarks: This method is called by $PRODUCTNAME$ when an Operation is denied by the SafetyExecutive,
    that is, the /pfCustomOperationCheck/ as given by RegisterCustomOperationCheck returns false. When an
    Operation is denied you define what action, if any, should be taken via an pfInvalidOperationCallback.
    For example, you could implement a reporting system as illustrated below. Here, if we are in debug or
    testing mode we assume the invalid operation call is due to a bug and output a trace when it occurs.
    On the other hand, if we are in release mode we assume an invalid operation is possibly due to a cheat
    and report the event to the server where we keep stats for each station. If a station is deemed to be
    cheating, we can then take appropriate action.

            // Define the pfInvalidOperationCallback
            void ReportCheats(const Operation& oOperation) {
                // For example, at this point we could report to the server that
                // the player on station 'oOperation.GetOrigin()' has performed an
                // invalid operation and is potentially cheating. This event would
                // be reported via a RMC on a well-known 'CheatManager' object.
                // The server could then keep stats on the reports and if a large
                // number of people report that the same person is cheating,
                // appropriate action be taken. Note that an invalid operation may
                // also be an indication of a bug in the game and not just a cheat.
            }

            void TraceInvalidOperations(const Operation& oOperation) {
                // During development testing, we know that as no one cheats if an
                // invalid operation is called then it is due to a bug in the code.
                QTRACE(TRACE_ALWAYS,_T("** An invalid operation has been performed **"));
                SYSTEMCHECK(false); // Stop here.
            }

            // Then register the custom callback.
            // This callback will be called in release mode,
            SafetyExecutive::GetInstance()->RegisterInvalidOperationCallback(ReportCheats);
            // and this callback will be called when testing in internal or debug mode.
            SafetyExecutive::GetInstance()->RegisterInvalidOperationCallback(TraceInvalidOperations);

    See Also: RegisterCustomOperationCheck
    */
    /* Please see man pages for details 
     
     
     
     
     
     
     
     
     
     
     
     
     
     
     
     
     
     
     
     
     */
    void RegisterInvalidOperationCallback(InvalidOperationCallback pfInvalidOperationCallback);

    // {secret}
    qBool CheckOperationValidity(const Operation& oOperation);

    /*
    Description: Sets whether the local station is trusted.
    Parameters:
    bFlag: If true (the default), the local station is trusted.
    Remarks: If the local station is trusted then the validity of any System Operations that the
    station triggers will not be checked by the SafetyExecutive before being called. If the local
    station is not trusted then System Operations will be validated by the SafetyExecutive before being
    called.
    See Also: LocalStationIsTrusted
    */
    /* Please see man pages for details 
     
     
     
     
     
     
     
     
     
     
     
     */
    void TrustLocalStation(qBool bFlag)
    {
        m_bTrustLocalStation=bFlag;
    }

    /*
    Description: Returns whether the local station is trusted.
    Returns: true if the local Station is trusted, which is the default behaviour, false if not or an error occurred.
    Remarks: If the local station is trusted then the validility of any System Operations that the
    station triggers will not be checked by the SafetyExecutive before being called. If the local
    station is not trusted then System Operations will be validated by the SafetyExecutive before being
    called.
    See Also: TrustLocalStation
    */
    /* Please see man pages for details 
     
     
     
     
     
     
     
     
     
     
     
     
     */
    qBool LocalStationIsTrusted() const
    {
        return m_bTrustLocalStation;
    }

private:
    qBool ValidOperation(const Operation& oOperation);

private:
    CustomOperationCheck m_pfCustomOperationCheck;
    InvalidOperationCallback m_pfInvalidOperationCallback;
    qBool m_bTrustLocalStation;
};


END_NAMESPACE_NEX

#endif
