/*---------------------------------------------------------------------------*
  (C) Copyright 1998-2014, Technologies Quazal Inc. All rights reserved.
  Quazal Net-Z, Quazal Rendez-Vous and Quazal are trademarks of Quazal.

  Copyright (C)2011-2014 Nintendo  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo and/or its licensors and are
  protected by national and international copyright laws.
  They may not be disclosed to third parties or copied or duplicated
  in any form, in whole or in part, without the prior written consent
  of Nintendo.
  The content herein is highly confidential and should be handled
  accordingly.
 *---------------------------------------------------------------------------*/

#ifndef _SELECTIONITERATORTEMPLATE_H_
#define _SELECTIONITERATORTEMPLATE_H_

#include <NetZ/src/Core/DOCore/ObjDupImpl.h>
#include <NetZ/src/Core/DOCore/IteratorOverDOs.h>
#include <NetZ/src/Core/DOCore/DOFilters.h>
#include <NetZ/src/Core/DOCore/DORole.h>
#include <NetZ/src/Core/DOCore/Selection.h>


BEGIN_NAMESPACE_NEX

// Group=Template Classes

/*
Description: Use this template to create a SelectionIterator for a specific doclass.
Remarks: This iterator template can be used to iterate over a specific doclass such as
user-defined DuplicatedObject classes and Station s.

For example, it is possible to create a SelectionIterator that holds all of the duplicated
objects of a certain class on a station via the following:

         Sphere::SelectionIterator itALLSphere;

It is also possible to create a SelectionIterator that holds all of the Duplica or
Duplication master instances of a given doclass on the local station.
For example, assuming that the Sphere doclass was declared in the DDL file, to create a
Sphere SelectionIterator that holds all of the Duplication master Spheres on a station the
following is used:

        Sphere::SelectionIterator itDMSphere(DUPLICATION_MASTER);
        itDMSphere.GotoStart();
        while(! (itDMSphere.EndReached())) {
            itDMSphere->DoPhysics();
            itDMSphere++;
        }

Similarly, if you wanted to list all the Station s in a Session, you could as follows:

        Station::SelectionIterator   itStations;
        itStations.GotoStart();
        while (!itStations.EndReached()) {
            itStations->Trace();
        }

*/
/* Please see man pages for details 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 */
template<class DO>
class SelectionIteratorTemplate: public IteratorOverDOs
{
    // Group=Methods
public:

    /* Please see man pages for details 
     
     
     
     
     
     
     */
    SelectionIteratorTemplate(): IteratorOverDOs(true,true,false), m_isAKindOfDOFilter(DO::GetRealDOClassID())
    {
        // n1769
        // CTR_PROFILING_UNIT_WITH_NAME(10, "SelectionIteratorTemplate()");
#if defined(LOCK_DEBUG_ENABLED) && defined(NEX_CLIENT_OPTIMIZATION)
        SYSTEMCHECK( Scheduler::GetInstance()->IsLockedByCurrentThread() && "Need Scheduler::GetInstance()->SystemLock() Lock to access DOs. Please see the manual");
#endif

        InitFilter();
        {
            // n1769
            // CTR_PROFILING_UNIT_WITH_NAME(11, "GoToStart()");

            GotoStart();
        }
    }

    /*
    Parameters:
    rSelectionType: The Role of the objects of this SelectionIteratorTemplate. The role may be either
    DUPLICA or DUPLICATION_MASTER.
    */
    /* Please see man pages for details 
     
     
     
     
     
     
     
     
     
     
     
     
     */
    SelectionIteratorTemplate(Role rSelectionType):
        IteratorOverDOs(rSelectionType==DUPLICA? true: false,
                        rSelectionType==DUPLICATION_MASTER? true:false,
                        false
                       ),
        m_isAKindOfDOFilter(DO::GetRealDOClassID())
    {
        // n1769
        // CTR_PROFILING_UNIT(11);
#if defined(LOCK_DEBUG_ENABLED) && defined(NEX_CLIENT_OPTIMIZATION)
        SYSTEMCHECK( Scheduler::GetInstance()->IsLockedByCurrentThread() && "Need Scheduler::GetInstance()->SystemLock() Lock to access DOs. Please see the manual");
#endif

        InitFilter();
        GotoStart();
    }

    /*
    Parameters:
    pSelection: A pointer to the Selection of this SelectionIteratorTemplate.
    */
    SelectionIteratorTemplate(Selection *pSelection): m_isAKindOfDOFilter(DO::GetRealDOClassID())
    {
        // n1769
        // CTR_PROFILING_UNIT(12);
#if defined(LOCK_DEBUG_ENABLED) && defined(NEX_CLIENT_OPTIMIZATION)
        SYSTEMCHECK( Scheduler::GetInstance()->IsLockedByCurrentThread() && "Need Scheduler::GetInstance()->SystemLock() Lock to access DOs. Please see the manual");
#endif

        SetSelection(pSelection); // Duplica.
        InitFilter();
        GotoStart();
    }

    /*
    Description:  Creates a copy of a SelectionIterator over a Selection of duplicated objects.
    Remarks:  If you do not know whether a SelectionIterator holds duplicas, duplication masters, or both
    and you need to iterate to check all of the objects against all others, /e.g./ for collisions, you can
    create a copy of the SelectionIterator in the following way:

            qBool CheckSphereCollisions(SelectionIteratorTemplate &itSphere){
                itSphere.GotoStart();
                //create a copy of the parameter
                Sphere::SelectionIterator itSphereCopy (itSphere);
                while (!itSphere.EndReached()){
                    itSphereCopy.GotoStart();
                    while (!itSphereCopy.EndReached()){

                        //Check for collision...

                        itSphereCopy++;
                    }
                    itSphere++;
                }
            }

    Parameters:
    oIt:  An existing SelectionIterator that in effect creates a copy of the specified SelectionIterator.
    */
private:
    /* Please see man pages for details 
     
     
     
     
     
     
     
     */
    SelectionIteratorTemplate(const SelectionIteratorTemplate& oIt) : IteratorOverDOs(oIt), m_isAKindOfDOFilter(DO::GetRealDOClassID())
    {
        GotoStart();
    }

public:
    virtual ~SelectionIteratorTemplate()
    {
        // Set to NULL while m_isAKindOfDOFilter still exists, and release the IteraterOverDOs::m_pFilter pointer reference.
        SetFilter(NULL);
    }
private:
    //{secret}
    void InitFilter()
    {
        //n1769 Average of 225 usec
        //CTR_PROFILING_UNIT(20);

        // Init filter.
        SetCacheFilter(&m_isAKindOfDOFilter);
        // Release the reference to the newly created filter - the
        // IteratorOverDOs has kept the one it needed.

        //n1769
        //CTR_PROFILING_UNIT_WITH_NAME(22, "RELEASEREF(pFilter)");
    }
public:

    /*
    Description: Returns a pointer to a DuplicatedObject of a specific class.
    Returns: A pointer to a DuplicatedObject of a specific class.
    */
    /* Please see man pages for details 
     
     
     
     
     */
    DO* operator->()
    {
        return (DO*) IteratorOverDOs::GetDOPtr();
    }

    /*
    Description: Returns a pointer to a DuplicatedObject of a specific class.
    Returns: A pointer to a DuplicatedObject of a specific class.
    */
    /* Please see man pages for details 
     
     
     
     
     */
    DO* operator*()
    {
        return (DO*) IteratorOverDOs::GetDOPtr();
    }
    const DO* operator*() const
    {
        return (const DO*) IteratorOverDOs::GetDOPtr();
    }

    /*
    Description: Returns a pointer to a DuplicatedObject of a specific class.
    Returns: A pointer to a DuplicatedObject of a specific class.
    */
    /* Please see man pages for details 
     
     
     
     
     */
    DO* GetDOPtr()
    {
        return (DO*) IteratorOverDOs::GetDOPtr();
    }
private:
    IsAKindOfDOFilter m_isAKindOfDOFilter;
};



END_NAMESPACE_NEX

#endif//_SELECTIONITERATORTEMPLATE_H_
