﻿/*---------------------------------------------------------------------------*
  (C) Copyright 1998-2014, Technologies Quazal Inc. All rights reserved.
  Quazal Net-Z, Quazal Rendez-Vous and Quazal are trademarks of Quazal.

  Copyright (C) 2011-2014 Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo and/or its licensors and are
  protected by national and international copyright laws.
  They may not be disclosed to third parties or copied or duplicated
  in any form, in whole or in part, without the prior written consent
  of Nintendo.
  The content herein is highly confidential and should be handled
  accordingly.
 *---------------------------------------------------------------------------*/

#include <NetZ/src/Core/DOCore/DOClassTemplate.h> // special case
#include <NetZ/src/Core/DOCore/DataSetTemplate.h> // special case

#ifndef _SessionClock_H_
#define _SessionClock_H_

#include <NetZ/src/Core/DOCore/ObjDupDecl.h>
#include <NetZ/src/Core/DOCore/DOClass.h>
#include <NetZ/src/Core/DOCore/DuplicatedObject.h>
#include <NetZ/src/Core/SessionClock/SessionClockDDL.h>

#include <OnlineCore/src/Core/SmoothedAverage.h>
#include <OnlineCore/src/Core/LatestMedian.h>
#include <OnlineCore/src/Platform/Core/Time.h>

//n2967 Option to use median value instead of moving average for session clock.
#define NEX_SESSIONCLOCK_USE_MEDIAN 1

BEGIN_NAMESPACE_NEX

// Group=Time Management

// This macro is never defined.  It is used only to get around a compilation problem with
// DocJet, which does not like the inheritance to the DOCLASS(SessionClock).
#ifdef FOR_DOC_PURPOSE

/*
Description:  The Session's synchronized time provider.
Remarks:  The time provided by the session clock is synchronised across all stations that belong to
the session. The session clock is automatically set to zero on the session master when Session::CreateSession
is called. Only the session master may reset the session clock via the method SetTime.

Use the session clock to synchronize or compare events on different stations. However,
in order to measure the time it takes to perform a set of operations on a single station
use an instance of SystemClock or Chrono. Note that poorly synchronized SessionClock s will affect various
$PRODUCTNAME$ processes, especially dead reckoning.

The accuracy of the synchronization is determined by the latency and jitter of
the network. Latency is the time lag before a message reaches its destination,
while jitter is the variation in latency. The synchronization algorithm
assumes that the latency between two stations is symmetrical. Therefore, the
non-symmetry of latency between two stations will cause an error in the session
clock on the local station. Non-symmetrical latency can be caused by one Station
performing additional tasks, such as loading files (especially upon a join), which delay synchronization
pulses leading to poorly synchronized SessionClock s.
See also: SystemClock, Chrono, Time
*/
/* Please see man pages for details 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 */
class SessionClock : DuplicatedObject
{

#else

class SessionClock : public DOCLASS(SessionClock)
{

#endif
    //group=Methods
public:

    // API for the class.

    /*
    Summary: Returns the current time of the session clock on the local station
    Remarks:  The local session clock is synchronized with the session master's session clock.
    Any connected station can call this method.
    Returns     :  The current time, in milliseconds, of the session clock on the local station.
    */
    /* Please see man pages for details 
     
     
     
     
     
     
     
     
     */
    static Time GetTime();

    /*
    Description: Sets the time, in milliseconds, of the session clock on the
    session master.
    Remarks: Only the session master can set the time of the session clock.
    Returns     : true if the time was successfully set, false if not or an error occurred.
    The error SYSTEMERROR_DC_NOT_CLUSTER_MASTER is signalled if the method is called
    on a station that is not the session master.
    Parameters:
    tNow: The time in milliseconds to which the session clock on the session master will be set.
    */
    /* Please see man pages for details 
     
     
     
     
     
     
     
     
     
     
     
     
     */
    static qBool SetTime(Time tNow);


    //n1769:memo There is no instance of this function, so it is not mentioned in the API Reference.
    /*
    Description: Sends a message to all SessionClock duplicas.
    Parameters:
    pMsg: A pointer to the message that will be broadcast.
    Returns     : true if the message was successfully broadcast, false if not or an error occurred.
    */
    /* Please see man pages for details 
    
     
     
     
     
     
     
     */
    static qBool BroadcastMessage(Message *pMsg);

    // group=Fine Tuning Methods

    /*
    Summary:  Sets the initial minimum number of synchronization pulses sent to synchronize
    the SessionClock before leaving the Session::JoinSession method.
    Remarks: The default number of synchronization pulses is s_uiDefaultMinimumNbSyncPulse.
    This means that at least s_uiDefaultMinimumNbSyncPulse messages will be sent between the
    host and session master to synchronize the clock before we leave the Session::JoinSession method.
    Increase the value of this parameter for a more precise time synchronization, although the
    synchronization will take a longer time to complete and use more bandwidth.
    Parameters:
    uiMinSyncPulse:  The minimum number of synchronization pulses that will be sent
    before we leave Session::JoinSession. uiMinSyncPulse can take any value greater than 0.
    Returns     :  true if the uiMinSyncPulse was successfully set, false if not or an error occurred.
    The error SYSTEMERROR_GEN_OUT_OF_RANGE is signalled if uiMinSyncPulse is not greater than 0.
    */
    /* Please see man pages for details 
     
     
     
     
     
     
     
     
     
     
     
     
     
     */
    static qBool SetMinimumNbSyncPulse(qUnsignedInt uiMinSyncPulse);

    static qUnsignedInt GetMinimumNbSyncPulse()
    {
        return s_uiSyncPulseSampleSize;
    }

    /*
    Summary:  Sets the delay between consecutive pulses sent to keep the session clock
    synchronized after session creation.
    Remarks: The default value is s_tiDefaultSyncPulseDelay. You can set the tiPulseDelay to any value
    greater than zero or to INFINITE. Setting the minimum synchronization pulse delay to INFINITE
    means that the SessionClock will synchronize initially using the minimum number of synchronization
    pulses at Session::JoinSession, but will not send any pulses after that to prevent drift. Decrease
    the value of this parameter to prevent drift, decrease the time taken to synchronize stations, and
    to improve the accuracy of synchronization, at the cost of increased bandwidth usage.

    This method can be called on any station and also at run time. Therefore, it is possible to set a certain
    initial pulse delay and then to reduce it once the initial clock synchronization is complete, for example.
    Parameters:
    tiPulseDelay:  The amount of delay, in milliseconds, between two consecutive pulses
    sent. tiPulseDelay can take any value greater than 0.
    Returns     :  true if the tiPulseDelay was successfully set, false if not or an error occurred.
    The error SYSTEMERROR_GEN_OUT_OF_RANGE is signalled if tiPulseDelay is not greater than 0.
    */
    /* Please see man pages for details 
     
     
     
     
     
     
     
     
     
     
     
     
     
     
     
     
     
     
     
     
     
     
     */
    static qBool SetSyncPulseDelay(TimeInterval tiPulseDelay);

    /*
    Description: Signals whether or not to take station induced latency into account
    when performing time synchronization calculations.
    Parameters:
    bAdjust: If false (the default), the time synchronization calculations will not take station induced
    latency into account.
    Remarks: AdjustForStationLatency affects how this Station synchronizes its SessionClock
    with the session master. When bAdjust is set to true, this Station will take into
    account the values of the input and output latency (set via Station::SetLatencyInfo) of
    this Station and of the session master in order to more accurately calculate the session
    clock time on this Station. For the majority of cases it is not necessary to adjust for station
    latency as the input and output latency of one station's connection will often be similar to
    that of the session master.
    See Also: Station::SetLatencyInfo
    */
    static void AdjustForStationLatency(qBool bAdjust)
    {
        s_bAdjustForStationLatency=bAdjust;
    }

    // {secret}
    static void SendSyncRequest();

    /*
    Summary:  Sets the size (number) of pulses used to synchronize the local session
    clock with that of the session master.
    Remarks: The default is s_uiDefaultSyncPulseSampleSize. The session clock on the local
    station is synchronized by sending a series of pulses to the session master at a set time
    interval. This method sets the size (in number of pulses) of each sample that will be sent.
    Increase the value of this parameter to improve the accuracy of synchronization, at the cost of greater
    bandwidth usage and an increase in the time taken to synchronize stations.
    Parameters:
    uiSyncPulseSampleSize: The number of pulses sent per sample. s_uiDefaultSyncPulseSampleSize can take
    any value greater than 0.
    Returns     :  true if the uiSyncPulseSampleSize was successfully set, false if not or an error occurred.
    The error SYSTEMERROR_GEN_OUT_OF_RANGE is signalled if uiSyncPulseSampleSize is not greater than 0.
    */
    /* Please see man pages for details 
    
    

    
    

    

    
    

    */
    static qBool SetSyncPulseSampleSize(qUnsignedInt uiSyncPulseSampleSize);

    // {secret}
    static void Enable();
    // {secret}
    static void Disable();

    //group=Default Constants

    // Summary: The default value for the minimum number of initial synchronization pulses is 5.
    /* Please see man pages for details 
    
     */
    static const qUnsignedInt s_uiDefaultMinimumNbSyncPulse;

    // Summary: The default value for the pulse delay to keep the SessionClock synchronized is
    // 2000 milliseconds ( = 2 seconds).
    /* Please see man pages for details 
    
     */
    static const TimeInterval s_tiDefaultSyncPulseDelay;

    // Summary: The default value for the sample size of the synchronization pulse is 10 pulses.
    static const qUnsignedInt s_uiDefaultSyncPulseSampleSize;


#ifdef NEX_SESSIONCLOCK_USE_MEDIAN
    /* Please see man pages for details 
     
     
     
     */
    static void SetDefaultMedianSize(qUnsignedInt s)
    {
        SYSTEMCHECK(s > 0);
        s_uiMedianSize = s;
    }

    /* Please see man pages for details 
     
     
     
     */

    static qUnsignedInt GetDefaultMedianSize()
    {
        return s_uiMedianSize;
    }
#endif

public:
    // Interface to the class - as seen by the system

    /*
    Remarks: A SessionClock object should never be directly instantiated. Instead, it is created by
    $PRODUCTNAME$ when a session is created and initialized to 0 ms.
    */
    /* Please see man pages for details 
     
     
     
     
     */
    SessionClock();

    //{Secret}
    static qBool IsInitializedOrDisabled();

    // {secret}
    virtual void OperationEnd(DOOperation* pOperation);


    /*
    Remarks: A SessionClock object should never be directly destroyed. It is destroyed by $PRODUCTNAME$
    when the session terminates.
    */
    /* Please see man pages for details 
     
     
     
     */
    virtual ~SessionClock();

    // {Secret}
    void GetFrequency(Time* pliFrequency);

    // {Secret}
    void StartTheTime();

    // {Secret}
    static void RestoreRealTimeRatio();

    // {Secret}
    void _Sync(Time liCurrentTime);


    // {Secret}
    qBool SetTimeOnInstance(Time liCurrentTime);

    // {Secret}
    Time _GetTime();
    //{secret}
    void AdjustTime(LargeTimeInterval tiDiff);
    //{secret}
    void SyncRequest(Time uiTimeStamp);
    //{secret}
    void SyncResponse(Time uiTimeStamp, Time uiCurrentTime, TimeInterval iServerDispatchTime);

private:
    // {Secret}
    Time _GetTimeImpl();

    class JobSendPulse;
#if !defined(DOT_NET_COMPILER)
    /* The following must be specified */
    friend class JobSendPulse;
#endif
    void StartSyncPulse();
    void StopSyncPulse();


    // {Secret}
    qBool _SetRealTimeRatio(qReal dRatio);

    qBool _SendSyncRequest();
    qBool _Pause();
    qBool _Resume();

public:

    // {secret}
    static void ApplyDefaultsIfRequired();
    // {secret}
    static void ApplyDefaults();
    // TODO: Should use a ref.
    //{secret}
    static SessionClock* GetInstance();

private:
    // {Secret}
#ifdef NEX_SESSIONCLOCK_USE_MEDIAN
    LatestMedian<TimeInterval> m_tiOffsetStatsMedian;
    SmoothedAverage<TimeInterval> m_tiOffsetStats;
    static const qUnsignedInt s_uiDefaultMedianSize;
#else
    SmoothedAverage<TimeInterval> m_tiOffsetStats;
#endif
    TimeInterval m_tiOffset;
    qBool m_bSet;
    qBool m_bPaused;

    qReal m_dRealTimeRatio;
    qReal m_dRealTimeRatioBeforePause;

    static  qBool s_bDefaultsApplied;

private:
    qUnsignedInt m_uiTimeSyncReceived;
    Time m_tLastSyncPulse;

private:
    // Fine tuning values.
    // {secret}
    static TimeInterval s_tiPulseDelay;
    static qUnsignedInt s_uiMinNbSyncPulse;
    static qUnsignedInt s_uiSyncPulseSampleSize;
    static qBool s_bAdjustForStationLatency;
    static qBool s_bEnabled;
#ifdef NEX_SESSIONCLOCK_USE_MEDIAN
    static qUnsignedInt s_uiMedianSize;
#endif

public:
    // {Secret}
    void FakeTheTime();//For tests

public :
    // {secret}
    static qBool s_bSimulateDelayOnReception;

#ifdef TEST_ENABLED
public:
    // Interface for the tests.

    // {Secret}
    void FakeTheTime();//For tests

    // {Secret}
    static qBool Test();
#endif

private :
    qBool m_bInitialized;

    // {secret}
    qBool m_boIsSyncing;

    Time m_tiLastSyncRequest;

    JobSendPulse *m_pJobSendPulse;

#ifdef TEST_ENABLED
public:
    // This method is there to implement non-regression tests for bug 2147.
    qUnsignedInt32 GetInstanceVersion() const
    {
        return m_uiInstanceVersion;
    }

private:
    qUnsignedInt32 m_uiInstanceVersion;
    static qUnsignedInt32 s_uiInstanceVersion;
#endif
};

// For the TraceOutput - to trace the session time.
// {secret}
extern qUnsignedInt32 GetSessionTime();

END_NAMESPACE_NEX

#endif
