﻿/*---------------------------------------------------------------------------*
  (C) Copyright 1998-2014, Technologies Quazal Inc.  All rights reserved.
  Quazal Net-Z, Quazal Rendez-Vous and Quazal are trademarks of Quazal.

  Copyright (C)2011-2014 Nintendo  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo and/or its licensors and are
  protected by national and international copyright laws.
  They may not be disclosed to third parties or copied or duplicated
  in any form, in whole or in part, without the prior written consent
  of Nintendo.
  The content herein is highly confidential and should be handled
  accordingly.
 *---------------------------------------------------------------------------*/

#ifndef _DeadReckoningModel_H_
#define _DeadReckoningModel_H_

#include <NetZ/src/Core/DOCore/ObjDupDecl.h>

BEGIN_NAMESPACE_NEX

//Group=Custom Dead Reckoning

/*
Description: The base class for all custom dead reckoning models.
Remarks: When you implement your custom dead reckoning model your model class must inherit from the
base DeadReckoningModel class and implement the IsAccurateEnough, UpdateModelOnMaster, UpdateModelOnDuplica,
and ComputeValue methods. Refer to the MyOwnModel template class for details.
See Also: MyOwnModel
*/
class DeadReckoningModel: public RootObject
{
public:
    //Group=Methods

    DeadReckoningModel()
        : m_uiLastUpdate(QUNSIGNEDINT32_MAX), m_uiLastUpdateLoop(QUNSIGNEDINT32_MAX) {}
    ~DeadReckoningModel() {}

    // {secret}
    void SetLastUpdate(qUnsignedInt32 uiUpdateLoopNumber)
    {
        m_uiLastUpdate=uiUpdateLoopNumber;
    }
    // {secret}
    qUnsignedInt32 GetLastUpdate() const
    {
        return m_uiLastUpdate;
    }
    // {secret}
    void SetLastUpdateLoop(qUnsignedInt32 uiUpdateLoopNumber)
    {
        m_uiLastUpdateLoop=uiUpdateLoopNumber;
    }
    // {secret}
    qUnsignedInt32 GetLastUpdateLoop() const
    {
        return m_uiLastUpdateLoop;
    }

private:
    qUnsignedInt32 m_uiLastUpdate;
    qUnsignedInt32 m_uiLastUpdateLoop;
};

#ifdef FOR_DOC_PURPOSE
//Group=Custom Dead Reckoning

/*
Description: Sample template class describing the methods that need to be implemented with your custom dead reckoning model(s).
Remarks: When you implement a custom dead reckoning model your model class must inherit from the
base DeadReckoningModel class and implement the IsAccurateEnough, UpdateModelOnMaster, UpdateModelOnDuplica,
and ComputeValue methods. These methods will be called when either DuplicatedObject::Update or DuplicatedObject::Refresh is
called and an update is required. Via these methods you respectively define how large an error is
tolerated between the Duplication master and a Duplica, how new data points are added to the Duplication master's
and Duplica's local dead reckoning models, and finally how these stored points are used to compute the dataset's parameter values.

Note that the purpose of this class is to illustrate the implementation of a custom dead reckoning model and
therefore, the class has no actual functionality.
*/
template <class DS>
class MyOwnModel: public DeadReckoningModel
{
public:
    //Group=Methods

    MyOwnModel() {};
    ~MyOwnModel() {};

    /*
    Description: Returns whether or not a Duplica's datset values are close enough to that of
    its Duplication master.
    Parameters:
    oDataset: The DataSet whose values on the Duplication master and Duplica will be compared.
    tTimestamp: The time for which the oDataset compared values are valid. This is equal to
    the /tUpdateTime/ specified in the DuplicatedObject::Update method.
    pDO: A pointer to the DuplicatedObject whose DataSet values oDataset on the Duplication master
    and a Duplica will be compared.
    pTarget: A pointer to the Station where the Duplica that will be compared to the Duplication master resides.
    Returns: This method should be written so that it returns true if the Duplica's dataset values are
    close enough to that of its Duplication master and false if they are not close enough.
    Remarks: This method is called on the Duplication master station when DuplicatedObject::Update is called.
    If this method returns true then the Duplication master will not update the oDataset DataSet on
    its Duplica. On the other hand, if this method returns false then an update will be sent.

    For example, if we define a Position DataSet in the DDL which has two parameters /x/ and /y/, we can enforce
    a maximum error between the Duplication master and a Duplica via the following. For the complete implementation
    of this and other examples refer to the source code of the Marbles sample provided with the distribution.

            qBool MyOwnModel::IsAccurateEnough(const Position& posReal, Time tTimestamp) {
                Position posExtrapolated;
                // compute extrapolated (Duplica's) value at tTimestamp
                ComputeValue(&posExtrapolated, tTimestamp);
                // compute difference between master and duplica values
                qFloat fError = posExtrapolated.ComputeError(posReal);
                return fError < 1;
            }
    */
    qBool IsAccurateEnough(const DS& oDataset, Time tTimestamp, DuplicatedObject* pDO, Station* pTarget);

    /*
    Description: Adds new dataset value(s) to the Duplication master's local instance of the duplica's
    dead reckoning model.
    Parameters:
    oCurrentValue: The value(s) of the dataset parameters that will be added to the local instance of the
    dead reckoning model.
    tUpdateTime: The time for which the new dataset values are valid. This is equal to the time that is passed
    when DuplicatedObject::Update is called.
    Remarks: When an update is sent to a duplica ( /i.e./ IsAccurateEnough returns false), this method is
    called on the Duplication master to add the new dataset values to the Duplication master's local instance
    of the duplica's dead reckoning model.

    For example, if we define a Position DataSet in the DDL which has two parameters /x/ and /y/, to update the
    values in our model we could simply replace the existing values with the new values as follows. For the complete
    implementation of this and other examples refer to the source code of the Marbles sample provided
    with the distribution.

            void MyOwnModel::UpdateModelOnMaster(const Position& oCurrentValue, Time tUpdateTime) {
                m_fX[0] = m_fX[1];
                m_fY[0] = m_fY[1];
                m_tT[0] = m_tT[1];

                m_fX[1] = oCurrentValue.m_fX;
                m_fY[1] = oCurrentValue.m_fY;
                m_tT[1] = tUpdateTime;
            }
    */
    void UpdateModelOnMaster(const DS& oCurrentValue, Time tUpdateTime);

    /*
    Description: Adds new dataset value(s) to the Duplica's local instance of the dead reckoning model.
    Parameters:
    oCurrentValue: The value(s) of the dataset parameters that will be added to the local instance of the
    dead reckoning model.
    Remarks: This method is called on a Duplica when the Duplica receives an update from its master.

    Net-Z does not send a timestamp in the packet that can be reused. To know what time the update is valid for,
    you can include a timestamp in the update packet that you send from the master (a 4 byte timestamp, for example,
    which would wrap every 49 days) or you can independently send a timestamp and then pass it into the UpdateModelOnDuplica
    call.  Otherwise, when this method is called, Net-Z uses the time at which the update from the Duplication
    master was received as the time stamp.
    */
    void UpdateModelOnDuplica(const DS& oCurrentValue);

    /*
    Description: Defines how the DataSet's values should be calculated.
    Parameters:
    pDS: A pointer to the DataSet whose parameter values will be recomputed according to the specified scheme.
    tTarget: The time for which the parameter values are computed, /i.e./ the time at which they are valid. This is
    equal to the time that is passed when DuplicatedObject::Refresh is called.
    Remarks: This method is called on the Duplica's Station when DuplicatedObject::Refresh is called to recompute
    the dataset values, using the values stored in the local model, according to the implemented dead reckoning model.
    Essentially, this method defines the mathematical model that describes how data extrapolation or interpolation is
    performed, /i.e./ how to use the data stored in the model to calculate the dataset values at the current time.

    For example, if we define a Position DataSet in the DDL which has two parameters /x/ and /y/, we could implement
    a simple dead reckoning model that calculates the current position based on the object's previous position and
    velocity as follows. For the complete implementation of this and other examples refer to the source code of the
    Marbles sample provided with the distribution.

            void MyOwnModel::ComputeValue(Position* pPos, Time tTarget) const {
                qFloat tDelta1 = static_cast<qFloat>(tTarget - m_tT[1]);
                qFloat tDelta0 = static_cast<qFloat>(m_tT[1] - m_tT[0]);

                pPos->m_fX = m_fX[1] + (m_fX[1] - m_fX[0])/tDelta0 * tDelta1;
                pPos->m_fY = m_fY[1] + (m_fY[1] - m_fY[0])/tDelta0 * tDelta1;
            }
    */
    void ComputeValue(DS* pDS, Time tTarget) const;
};

#endif




END_NAMESPACE_NEX

#endif

