﻿/*---------------------------------------------------------------------------*
  (C) Copyright 1998-2014, Technologies Quazal Inc.  All rights reserved.
  Quazal Net-Z, Quazal Rendez-Vous and Quazal are trademarks of Quazal.

  Copyright (C)2011-2014 Nintendo  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo and/or its licensors and are
  protected by national and international copyright laws.
  They may not be disclosed to third parties or copied or duplicated
  in any form, in whole or in part, without the prior written consent
  of Nintendo.
  The content herein is highly confidential and should be handled
  accordingly.
 *---------------------------------------------------------------------------*/

#ifndef _PHBDRParameters_H_
#define _PHBDRParameters_H_

#include <NetZ/src/Core/DOCore/ObjDupDecl.h>
#include <OnlineCore/src/Platform/Core/Time.h>

BEGIN_NAMESPACE_NEX

/* {group: Dead Reckoning}
Summary:  Holds the parameters which fine-tune the Position History-Based
Dead Reckoning algorithm.
Remarks: This class holds the parameters used to fine-tune the dead reckoning algorithm.
A PHBDRParameters object should never be directly instantiated. Rather, DataSet::GetPHBDRParameters is used
to access the object.
*/
/* Please see man pages for details 
 
 
 
 
 
 
 
 
 */
class PHBDRParameters: public RootObject
{
public:
    /* Please see man pages for details
    
     */
    enum ANGLE_TYPE
    {
        //Summary: The angle supplied as a parameter to the function is in radians.
        //See Also: PHBDRParameters::SetTrackingAngleThreshold, PHBDRParameters::SetConvergenceAngleThreshold,
        //ANGLE_IN_DEGREES
        ANGLE_IN_RADIANS = 1, //

        // Summary:  The angle supplied as a parameter to the function is in degrees.
        //See Also:   PHBDRParameters::SetTrackingAngleThreshold,  PHBDRParameters::SetConvergenceAngleThreshold,
        //ANGLE_IN_RADIANS
        ANGLE_IN_DEGREES = 2 //
    };
    //Group=Methods

    /*
    Remarks: A PHBDRParameters object should never be directly instantiated. Instead, it is created by $PRODUCTNAME$
    prior to session creation. To access the PHBDRParameters object from within your game use the
    DataSet::GetPHBDRParameters method.
    See Also: DataSet::GetPHBDRParameters, DDL File Syntax
    */
    PHBDRParameters();

    /*
    Remarks: A PHBDRParameters object should never be directly destroyed. It is destroyed by $PRODUCTNAME$
    when the session terminates.
    */
    virtual ~PHBDRParameters();

public:

    //{secret}
    static void PostdateUpdates(TimeInterval tiDelay)
    {
        SetExtrapolationDelay(tiDelay);
    }

    /*
    Description: Sets the extrapolation delay invoked on a duplica.
    Parameters:
    tiDelay: Value of the delay in milliseconds.
    Remarks: Delays the extrapolation on a duplica to create a delayed
    but more accurate reflection of the Duplication master.
    The extrapolation delay changes the timestamp on an update message when the message
    is received. Therefore, if /t/ is the initial timestamp on the message received from
    the Duplication master, then with an extrapolation delay the value used by the Duplica
    would be t+tiDelay.

    Setting an extrapolation delay enables $PRODUCTNAME$ to more accurately predict the dataset of
    the duplicas as extrapolation is performed for a shorter period of time. Delaying
    extrapolation on the duplicas results in a trade-off between the duplica displaying more accurate,
    but delayed dataset values, and less accurate values that are time synchronized with the duplication
    master. An extrapolation delay is typically set when Loopback is applied.

    See Also: GetExtrapolationDelay
    */
    /* Please see man pages for details
         
		 
		 
         
     */
    static void SetExtrapolationDelay(TimeInterval tiDelay);

    /*
    Description: Returns the extrapolation delay in milliseconds.
    Returns: The current extrapolation delay in milliseconds.
    See Also: SetExtrapolationDelay
    */
    /* Please see man pages for details
         
		 
		 
         
     */
    static TimeInterval GetExtrapolationDelay()
    {
        return s_tiExtrapolationDelay;
    }

    /*
    Summary:  Signals whether a convergence algorithm will be used.
    Parameters:
    bFlag: If true (the default), a convergence algorithm is used. If false,
    only a tracking algorithm is used.
    Remarks: When a convergence algorithm
    is used and an update message is received, rather than the dataset extrapolation value
    directly changing from the previous extrapolated value to the new one, resulting in a
    jerky trajectory, the dataset extrapolation value will smoothly change from the old value
    to the new one.

    The convergence delay is the time required to converge from the old
    value to the new value on the predicted path. $PRODUCTNAME$ uses several
    heuristics to compute an adequate convergence time, but the minimum and
    maximum delay for the convergence can be defined using the method
    SetConvergenceDelay.

    To fine-tune dead reckoning, as convergence will hide bad tracking you should
    first fine-tune the tracking and then the convergence algorithm. To start, turn off
    the convergence algorithm (ApplyConvergence) and then tweak the
    tracking algorithm (SetTrackingAngleThreshold) until the object's trajectory is relatively
    accurate but jerky. Then, tweak the values of the convergence angle threshold (SetConvergenceAngleThreshold)
    and convergence delay (SetConvergenceDelay) until the trajectory of the object is satisfactory.

    See also: SetConvergenceAngleThreshold, SetConvergenceDelay
    */
    /* Please see man pages for details
         
         
		 
     */
    void ApplyConvergence(qBool bFlag);

    /*
    Summary: Sets the minimum and maximum values for the convergence delay.
    Parameters:
    tiMinCPDelay:  Minimum convergence delay in milliseconds.
    tiMaxCPDelay:  Maximum convergence delay in milliseconds.
    Remarks: $PRODUCTNAME$ uses several heuristics to compute an adequate convergence time,
    but the minimum and maximum value for the convergence can be defined
    using the method SetConvergenceDelay. The default values for the convergence
    delay are a minimum of 250 ms and a maximum of 1000 ms.

    The convergence delay should be set so that the variable converges to the predicted
    path before the next update is received. If this is not the case the extrapolation
    will appear irregular.

    Returns: true if the convergence delay was successfully set, false if not or an error occurred.
    The error SYSTEMERROR_GEN_OUT_OF_RANGE is signalled if the values are negative or if the
    maximum is less than or equal to the minimum.
    */
    /* Please see man pages for details
         
         
		 
         
         
         
         
     */
    qBool SetConvergenceDelay(TimeInterval tiMinCPDelay, TimeInterval tiMaxCPDelay);

    //Description: Returns the minimum convergence delay in milliseconds.
    //Returns: The mimimum convergence delay in milliseconds.
    //See also: SetConvergenceDelay, GetMaxCP
    TimeInterval GetMinCP() const;

    //Summary: Returns the maximum convergence delay in milliseconds.
    //Returns: The mamimum convergence delay in milliseconds.
    //See also: SetConvergenceDelay, GetMinCP
    TimeInterval GetMaxCP() const;
    // {Secret}
    TimeInterval ComputeCP(TimeInterval tiLastInterval) const;

    /*
    Description: Returns the maximum local correction delay in milliseconds.
    Returns: The maximum local correction delay in milliseconds.
    See also: SetMaximumLocalCorrectionDelay
    */
    /* Please see man pages for details
	
	
     */
    TimeInterval GetMaximumLocalCorrectionDelay() const
    {
        return m_tiMaximumLocalCorrectionDelay;
    }

    /*
    Description: Sets the maximum local correction delay in milliseconds.
    Parameters:
    tiMaximumLocalCorrectionDelay: The maximum local correction delay in milliseconds. The
    default value is 3500 ms.
    Remarks: The maximum local correction delay is the maximum time that local corrections
    will be applied to a duplica. Once the local correction delay expires, if the duplica has
    not converged to the position given by the extrapolation model, the object will be teleported
    to this position. This ensures that in the rare situations where local corrections, such as
    for hitting a wall, are repeatedly performed on a duplica which prevent the duplica from
    converging to the extrapolation model, then after a short period of time the position of
    the duplica will be corrected. For most games the default value should not need to be changed.
    See also: GetMaximumLocalCorrectionDelay
    */
    /* Please see man pages for details
	
	
     */
    void SetMaximumLocalCorrectionDelay(TimeInterval tiMaximumLocalCorrectionDelay)
    {
        m_tiMaximumLocalCorrectionDelay=tiMaximumLocalCorrectionDelay;
    }

    /*
    n2343
    When predicting with dead-reckoning and the Time difference of two values is less than m_tiMinimumTimeInterval,
    stuff into m_tiMinimumTimeInterval.
    The default is 0, so it is effectively invalid.
    */
    TimeInterval GetMinimumTimeInterval() const
    {
        return m_tiMinimumTimeInterval;
    }
    void SetMinimumTimeInterval(TimeInterval tiMinimumTimeInterval)
    {
        m_tiMinimumTimeInterval=tiMinimumTimeInterval;
    }

    friend class PHBDRVar_qInt32;
    friend class PHBDRVar_qInt16;
    friend class PHBDRVar_qInt8;
    friend class PHBDRVar_qDouble;
    friend class PHBDRVar_qFloat;
    friend class PHBDRModel;


    /* Please see man pages for details
         
         
		 
     */
    qBool ApplyConvergence() const;

    // {Secret}
    static void ApplyExtrapolationDelay(Time* ptTime);

    //Description: Ensures that a linear tracking algorithm is always used.
    //See also: ForceNonLinearTracking, SetTrackingAngleThreshold
    /* Please see man pages for details
	
     */
    void ForceLinearTracking();

    // {secret}
    qBool UseLinearTracking() const
    {
        return m_dTrackingAngleThreshold==4;
    }

    /*
    Description: Ensures that a non-linear tracking algorithm is always used.
    Remarks: For example, if we have a dataset ExtrapolatedPosition we can ensure that non-linear tracking
    is used via the following:

            ExtrapolatedPosition::GetPHBDRParameters()->ForceNonLinearTracking();

    See also: ForceLinearTracking, SetTrackingAngleThreshold
    */
    /* Please see man pages for details
	
     */
    void ForceNonLinearTracking();

    // {secret}
    qBool UseNonLinearTracking() const
    {
        return m_dTrackingAngleThreshold==-1;
    }

    //Description: Ensures that a linear convergence algorithm is always used.
    //See also: ForceNonLinearConvergence, SetConvergenceAngleThreshold
    /* Please see man pages for details
	
     */
    void ForceLinearConvergence();

    // {secret}
    qBool UseLinearConvergence() const
    {
        return m_dConvergenceAngleThreshold==-1;
    }

    //Description: Ensures that a non-linear convergence algorithm is always used.
    //See also: ForceLinearConvergence, SetConvergenceAngleThreshold
    /* Please see man pages for details
	
     */
    void ForceNonLinearConvergence();

    // {secret}
    qBool UseNonLinearConvergence() const
    {
        return m_dConvergenceAngleThreshold==4;
    }

    /*
    Summary:  Sets the tracking angle threshold that is used to determine if a
    linear or non-linear prediction algorithm is used.
    Parameters:
    dAngle:  The threshold angle that decides if a linear or non-linear tracking algorithm is used.  If the
    actual angle of embrace is less than this angle, a linear tracking algorithm is used.  If the actual angle
    of embrace is larger than this angle, a non-linear tracking algorithm is used.
    byAngleUnit:  Specifies if the angle supplied is in radians (ANGLE_IN_RADIANS),
    which is the default, or degrees (ANGLE_IN_DEGREES).
    Remarks: The threshold, given in radians or degrees is set for each dataset using this method.
    If the angle of embrace calculated with the last three dataset values is less than the tracking
    threshold angle, then a linear tracking algorithm is used. If the angle of embrace is larger, then a
    non-linear tracking algorithm is used. To ensure coherence between the datasets of objects on different
    stations, the tracking angle threshold must be set the same on each station across the network.

    To fine-tune dead reckoning, as convergence will hide bad tracking you should
    first fine-tune the tracking and then the convergence algorithm. To start, turn off
    the convergence algorithm (ApplyConvergence) and then tweak the
    tracking algorithm (SetTrackingAngleThreshold) until the object's trajectory is relatively
    accurate but jerky. Then, tweak the values of the convergence angle threshold (SetConvergenceAngleThreshold)
    and convergence delay (SetConvergenceDelay) until the trajectory of the object is satisfactory.

    Note: By default, the tracking angle threshold is set to ForceLinearTracking.

    See also: ForceLinearTracking, ForceNonLinearTracking
    */

    /* Please see man pages for details
         
         
         
		 
		 
         
     */
    void SetTrackingAngleThreshold(qReal dAngle, ANGLE_TYPE byAngleUnit=ANGLE_IN_RADIANS);

    /*
    Summary:  Sets the convergence angle threshold that is used to determine if a
    linear or non-linear convergence algorithm is used.
    Parameters:
    dAngle:  The threshold angle that dictates if a linear or non-linear convergence is used.  If the
    actual angle of embrace is less than this angle, a non-linear convergence algorithm is used.
    If the actual angle of embrace is larger than this angle, a linear convergence algorithm is used.
    byAngleUnit:  Specifies if the angle supplied is in radians (ANGLE_IN_RADIANS),
    which is the default, or degrees (ANGLE_IN_DEGREES).
    Remarks:  This method sets the convergence angle threshold, given in radians or degrees, for each dataset.
    If the angle of embrace is less than the convergence threshold angle, then a non-linear tracking
    algorithm is used. If the angle of embrace is larger, then a linear tracking algorithm is used. As the
    convergence algorithm is only implemented on the local station each station across the network may set
    a different convergence angle threshold.

    To fine-tune dead reckoning, as convergence will hide bad tracking you should
    first fine-tune the tracking and then the convergence algorithm. To start, turn off
    the convergence algorithm (ApplyConvergence) and then tweak the
    tracking algorithm (SetTrackingAngleThreshold) until the object's trajectory is relatively
    accurate but jerky. Then, tweak the values of the convergence angle threshold (SetConvergenceAngleThreshold)
    and convergence delay (SetConvergenceDelay) until the trajectory of the object is satisfactory.

    Note: By default, the convergence angle threshold is set to ForceLinearConvergence.

    See also: ForceLinearConvergence, ForceNonLinearConvergence
    */
    /* Please see man pages for details
         
         
         
		 
		 
         
         
     */
    void SetConvergenceAngleThreshold(qReal dAngle, ANGLE_TYPE byAngleUnit=ANGLE_IN_RADIANS);

private:

    qReal m_dTrackingAngleThreshold;
    qReal m_dConvergenceAngleThreshold;

    TimeInterval m_tiMaxCPDelay;
    TimeInterval m_tiMinCPDelay;
    qBool m_bApplyConvergence;

    TimeInterval m_tiMaximumLocalCorrectionDelay;

    TimeInterval m_tiMinimumTimeInterval;

    static TimeInterval s_tiExtrapolationDelay;
};

END_NAMESPACE_NEX

#endif

