﻿/*---------------------------------------------------------------------------*
  (C) Copyright 1998-2014, Technologies Quazal Inc. All rights reserved.
  Quazal Net-Z, Quazal Rendez-Vous and Quazal are trademarks of Quazal.

  Copyright (C)2011-2014 Nintendo  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo and/or its licensors and are
  protected by national and international copyright laws.
  They may not be disclosed to third parties or copied or duplicated
  in any form, in whole or in part, without the prior written consent
  of Nintendo.
  The content herein is highly confidential and should be handled
  accordingly.
 *---------------------------------------------------------------------------*/

#ifndef _CallContext_H_
#define _CallContext_H_

#include <OnlineCore/src/Platform/Core/PlatformDecl.h>
#include <OnlineCore/src/Platform/Core/UserContext.h>
#include <OnlineCore/src/Platform/Core/RefCountedObject.h>
#include <OnlineCore/src/Platform/Core/qList.h>
#include <OnlineCore/src/Platform/Core/qVector.h>
#include <OnlineCore/src/Platform/Core/ConnectionID.h>

BEGIN_NAMESPACE_NEX

class CallbackRoot;
class CallContextRegister;

// $ifdef NETZ$
// Group=Results Classes
// $else$
// group=Return Results Classes
// $endif$


/*
Description: The root class of all call contexts.
Remarks: This root class defines the context of all call contexts. You may make any number of
non-blocking calls using different call contexts and $PRODUCTNAME$ will guarantee that the
order of execution is maintained on the remote station. Nevertheless, if you create a dependency
chain between calls, that is, a given call depends on the result returned by a previous call,
you must Wait between calls so that the system has time to complete the first request before calling
the second.

The State diagram for calls is as follows, where the State may be returned via GetState.

{image:RMCStateDiagram}

If an Asynchronous call returns false, that is, call initiation failed, then the State of the call's
CallContext will be either Available or CallFailure depending on when $PRODUCTNAME$ detected the error.
*/

/* Please see man pages for details 





*/
class CallContext: public RefCountedObject
{
    //Group=Type Definitions
public:

    /* Please see man pages for details 
    
    */
    CallContext();

    // Remarks: If the DeleteOnCompletion flag is set, the destructor is automatically called upon
    // completion of the call that uses this CallContext.

    /* Please see man pages for details 
    

    

    
    */
    virtual ~CallContext();

    /* Description: A timeout interval which "fails" a call if the call
    exceeds the interval. That is, the call's state changes to
    CallContext::CallFailure the reason being Core::Timeout.
    Parameters: The interval in milliseconds.
    */

    //Do not expose this API publicly as it does not work as advertised in
    // all the places a call context can be used. See CallContext.SetDeadline()
    // {secret}
    /* Please see man pages for details 
     

     
     

     
     */
    void SetTimeout(TimeInterval tiTimeout)
    {
        SetDeadline(Time::ConvertTimeoutToDeadline(tiTimeout));
    };

    /* Description: A timeout deadline which "fails" a call if the call
    exceeds the deadline. That is, the call's state changes to
    CallContext::CallFailure the reason being Core::Timeout.
    Parameters: The deadline which takes the time in the form of the
        number of milliseconds since the session started.
    */

    //Do not expose this API publicly as it does not work as advertised in
    // all the places a call context can be used. In fact in most places
    // this deadline is ignoreded by jobs. Some exceptions to this which do
    // use the call context deadline: SyncSim.JoinSimulation,
    // AuthenticationClient.Login
    // {secret}
    void SetDeadline(Time tDeadline)
    {
        m_tDeadline = tDeadline;
    };

    // {secret}
    Time GetDeadline() const
    {
        return m_tDeadline;
    };

    // {secret}
    qBool HasDeadline() const
    {
        return m_tDeadline != Time(0);
    };


    // Description: Enumerates the various States of a CallContext.
    // See Also: State

    /* Please see man pages for details 
    
    */
    enum State
    {
        // The call can be called. When this is done the State will change to
        // CallInProgress. If Reset is called when the State is one of CallSuccess,
        // CallFailure, or CallCancelled the State will change to Available.
        Available,    //
        /*
        $ifdef RENDEZVOUS$
        The call is in progress.
        $else$
        The call is in progress. The method CallContext::Cancel may be called.
        $endif$
        */
        CallInProgress, //
        // The call has terminated. The method Reset may be called.
        CallSuccess,    //
        // The call failed. The method Reset may be called.
        CallFailure,    //
        /*
        $ifdef RENDEZVOUS$
        The call was canceled by the user. The method Reset may be called.
        $else$
        The call was canceled by the user via the method CallContext::Cancel. The method
        Reset may be called.
        $endif$
        */
        CallCancelled    //
    };

    // Description: Enumerates the various Flags that may be used with a CallContext.
    /* Please see man pages for details 
    
    */
    enum Flags
    {
        /*
        $ifdef RENDEZVOUS$
        {secret}
            $else$
        The call cannot be canceled. This flag can only be set for a MigrationContext.
         $endif$
        */
        CancellationNotAllowed  = 0x01,
        // The CallContext of this call will be deleted upon completion of the call that uses this CallContext.
        // When you dynamically create a CallContext, you need to ensure that it is deleted once it is no longer
        // needed.
        /* Please see man pages for details 
         
         */
        DeleteOnCompletion      = 0x02,
        /*
        $ifdef RENDEZVOUS$
        {secret}
            $else$
           The call will be unidirectional. It will not return the result of the called method. This
           flag can only be set for a RMCContext. If the remote method returns a value, this value will be ignored
           and not returned to the caller (no return message will be sent). However, as always, the call will still
           return a Boolean that indicates whether the message was sent to the target station. If this flag
           is set, the call can be made on multiple targets and the flags SendUnreliableMessage, CallOnDuplicas,
           CallOnMaster, CallOnNeighbouringStations, and CallOnLocalStation may also be set.
           $endif$
        */
        /* Please see man pages for details 
         
         
         
         
         
         
         */
        OneWayCall              = 0x04,
        /*
        $ifdef RENDEZVOUS$
        {secret}
            $else$
         The call will be sent over an unreliable channel. This flag can only be set for a RMCContext and is only
         valid when the OneWayCall flag is set.
          $endif$
        */
        /* Please see man pages for details 
         
         
         
         */
        SendUnreliableMessage   = 0x08,
        /*
        $ifdef RENDEZVOUS$
        {secret}
            $else$
         The call message will not be bundled and will therefore be sent as soon as possible.
         This flag can only be set for a RMCContext.
        $endif$
        */
        SendUnbundledMessage    = 0x10,
        // The first available flag that a derived class may use.
        FirstFlagAvailable      = 0x20
    };

    /*
    Description: Defines a CompletionCallback.
    Parameters:
    o /CallContext/ : A pointer to the CallContext of this CompletionCallback.
    o /UserContext/ : A pointer to the UserContext associated with this CompletionCallback.
    Remarks: A CompletionCallback may be registered via RegisterCompletionCallback to ensure that a
    given user-defined function is called upon completion of the CallContext.
    See Also: RegisterCompletionCallback, RegisterCompletionCallback
    */

    /* Please see man pages for details 
    

     

    
    */
    typedef void (*CompletionCallback)(CallContext*, const UserContext*);

    //group=Methods

    /*
    Description: Returns the current State of the call.
    Returns: The current State of the call which is one of Available,
    CallInProgress, CallSuccess, CallFailure, or CallCancelled.
    See Also: State
    */

    /* Please see man pages for details 
    

    

    
    */
    State GetState() const
    {
        return m_eState;
    }

    /*
    Description: Sets the UserContext associated to this CallContext.
    Parameters:
    iIndex: The user-defined index of the UserContext that is associated to this CallContext.
    oUserContext: A pointer to the UserContext associated with this CallContext.
    Remarks: A UserContext can be associated to a CallContext via SetUserContext and returned via
    GetUserContext.
    See Also: GetUserContext
    */

    /* Please see man pages for details 
    

    
    

    
    */
    void SetUserContext(qUnsignedInt32 iIndex, const UserContext& oUserContext);

    /*
    Description: Returns the UserContext associated to this CallContext.
    Parameters:
    iIndex: The user-defined index of the UserContext that will be returned.
    Returns: A pointer to the UserContext associated with this CallContext.
    Remarks: A UserContext can be associated to a CallContext via SetUserContext and returned via
    GetUserContext.
    See Also: SetUserContext
    */

    /* Please see man pages for details 
    

    

    

    
    */
    const UserContext& GetUserContext(qUnsignedInt32 iIndex) const;

public:
    // {secret}
    qBool InitiateCall();
    // {secret}
    qBool InitiateSpecialCall();

private:
    // {secret}
    qBool SetCallInProgressState();

public:
    // {secret}
    void SignalFailure()
    {
        SetStateImpl(CallContext::CallFailure,QRESULT_ERROR(Core,Unknown),false);   //TODO: Why is that one using "false"?
    }
    // {secret}
    void SignalFailure(qResult rOutcome);
    // {secret}
    void SignalSuccess()
    {
        SetStateImpl(CallContext::CallSuccess,QRESULT_SUCCESS(Core,Success),true);
    }
    // {secret}
    void SignalSuccess(qResult rOutcome);
    // {secret}
    // Calls into SignalSuccess/Failure, based on outcome.
    void SignalCompletion(qResult rOutcome);
    // {secret}
    // Signals special terminations (e.f. Cancellation)
    // This is not the prefered interface to use.
    void SignalSpecialCompletion(State eNewState, qResult rOutcome)
    {
        SetStateImpl(eNewState,rOutcome,true);
    }

public:

    /*
    Description: Sets the flag to be used with the call.
    Parameters:
    uiFlag: The flag with which the call is called.
    See Also: ClearFlag, FlagIsSet, Flags
    */

    /* Please see man pages for details 
    

    

    
    */
    void SetFlag(qUnsignedInt32 uiFlag);

    /*
    Description: Returns whether the combination of flags currently in use is valid.
    Returns: true if the combination of flags currently in use is valid, false if not or an error occurs.
    See Also: Flags
    */

    /* Please see man pages for details 
    

    

    
    */
    virtual qBool FlagsAreValid() const;

    /*
    Description: Clears the flag set via SetFlag.
    Parameters:
    uiFlag: The flag with which the call is called.
    See Also: SetFlag, FlagIsSet, Flags
    */

    /* Please see man pages for details 
    

    

    
    */
    void ClearFlag(qUnsignedInt32 uiFlag);

    /*
    Description: Returns whether a given flag is set.
    Parameters:
    uiFlag: The flag with which the call is called.
    Returns: true if the stated flag is set, false if not or an error occurred.
    See Also: SetFlag, ClearFlag, Flags
    */

    /* Please see man pages for details 
    

    

    

    
    */
    qBool FlagIsSet(qUnsignedInt32 uiFlag) const;
    // {secret}
    qUnsignedInt32 GetFlags() const
    {
        return m_uiFlags;
    }

private:
    // {secret}
    void SetState(State eNewState)
    {
        SetStateImpl(eNewState,QRESULT_ERROR(Core,Unknown),false);
    }

    // {secret}
    void SetStateImpl(State eNewState, qResult rOutcome, qBool bSetOutcome);

public:

    // {secret}
    static const qChar* GetStateString(State eState);

    /*
    Description: Sets the operations to perform when this CallContext completes.
    Parameters:
    pfCompletionCallback: A pointer to the call back function for the CallContext.
    oContext: The UserContext associated with this CallContext.
    pCallback: A pointer to the Callback object to be called on completion.
    bCallOnSuccess: If true (the default), the Callback object will be called only if
    the CallContext succeeds.
    bAddToEnd: If true (the default), the pCallback will be added to the end of the list
    of callbacks. If false, the callback will be added to the beginning of the list.

    Remarks: A CompletionCallback presents an easy and efficient way to ensure a given function
    is called upon completion of the CallContext. When methods are called asynchronously, use of this
    callback avoids having to continually poll the CallContext to determine whether it has
    completed. If multiple callbacks are registered, then by default they are stored and called in the order that
    they were registered, with the first registered callback being called first. If you wish to change the order
    you need to set the bAddToEnd parameter when you register a given callback.

    A complete example of the implementation of a completion callback is given in the Inventory and Item classes
    of the Real-time 3D Framework, part of which is given below. Refer to the framework source code for full details.
    First we declare the ChangeOwner RMC in the DDL:

                // DDL declarations for Item class.
                doclass Item : Duplicated3DObject {
                    ItemOwnership m_dsOwner;
                    bool ChangeOwner(in dohandle hCurrentOwner, in dohandle hNewOwner,
                    dohandle hTargetItemMaster);
                };

    Then in the /Inventory.cpp/ file we implement the PickUp method that uses the ChangeOwner RMC:

    $import CallContextRegister1 TOP/Frameworks\Eterna\RealTime3D\FoundationClasses\Inventory.cpp$

    Then we call the RMC as follows. When the RMC completes, the method Inventory::AddItem will be called.

    $import CallContextRegister2 TOP/Frameworks\Eterna\RealTime3D\FoundationClasses\Inventory.cpp$

    $ifdef RENDEZVOUS$
    $else$
    See Also: Callback
    $endif$
    */

    /* Please see man pages for details 
    

    

    
    

    
    

    

    
    
    

    */
    void RegisterCompletionCallback(CompletionCallback pfCompletionCallback, const UserContext& oContext, qBool bAddToEnd=true);

    /* Please see man pages for details 
    

    

    

    
    
    

    */
    void RegisterCompletionCallback(CallbackRoot* pCallback, qBool bCallOnSuccess=true, qBool bAddToEnd=true);

    // {secret}
    void RegisterCancellationCallback(CallbackRoot* pCallback);



private:

#ifdef TEST_ENABLED
    friend class CallContextRegister;
#endif

    // {secret}
    // Called when the call is initiated.
    void Register();

    // {secret}
    // Called when the CallContext reaches a final state.
    void Unregister();

public:
    // {secret}
    qBool IsRegistered() const;

    /*
    Description: Resets the State of the call to Available.
    Remarks: This method may only be called if the current State of the call is one of
    Sucess, Failure, or Cancelled. The error SYSTEMERROR_GEN_INVALID_OPERATION will
    result if this method is called when the current State is
    CallInProgress.
    If the State is Available, the method returns true, but it behaves as a no-op.
    See Also: State
    */

    /* Please see man pages for details 
    

    

    
    

    
    */
    qBool Reset();

    /*
    Description: Causes the system to wait until the Outcome of the CallContext is known or the timeout is exceeded.
    Parameters:
    uiTimeout: The maximum time in milliseconds that the system will wait. The default
    uiTimeout is infinite.
    Returns: true if the outcome of the CallContext is known, false if not or
    the timeout is exceeded. The possible outcomes of a given call are specific to the call
    that is made.
    Remarks: This method may only be called if the current CallContext::State of the call is
    CallInProgress and the flag (DOCallContext::Flags) SynchronousCall is set to false. The method
    returns true if the Wait has succeeded, that is, the state is no longer CallInProgress. The error
    SYSTEMERROR_GEN_TIMEOUT will be generated if the timeout is exceeded.

    In order to prevent deadlocks, if this method is called on an embeded RMC then the only valid
    uiTimeout is zero. If any other timeout is set then the error SYSTEMERROR_GEN_INVALID_WAIT will
    be signalled. Refer to CallMethodOperation for an example of how to use this method with embeded RMCs.

    $ifdef ETERNA$
    Refer to DuplicatedObject::FetchDuplica for an example of this method in use.
    $endif$

    See Also: State, GetOutcome
    */

    /* Please see man pages for details 
    

    
    
    

    
    
    

    

    

    
                

    
    */
    qBool Wait(qUnsignedInt32 uiTimeout=WAIT_INFINITE_TIMEOUT);

    /*
    Description: Cancels the call that is associated to this CallContext.
    Parameters:
    eNewState: The new State of this CallContext. The default value is CallCancelled.
    Returns: true if the call was successfully canceled, false if not or an error occurred.
    The error SYSTEMERROR_GEN_INVALID_OPERATION will be returned if this method is called when the
    current State of the call is any state other than CallInProgress.
    Remarks: When this method is called the CallContext::State of the call will be set to CallCancelled
    and the place holder for the results deleted. If the call associated with the context has not yet been
    issued, it will be canceled, if it has already been issued it cannot then be canceled but
    the issuing station will ignore the return message, if any. This method may only be called
    if the current state of the call is CallInProgress.

    See Also: State, Reset
    */

    /* Please see man pages for details 
    

    
    
    

    

    

    
               

    

    
    */
    qBool Cancel(State eNewState=CallCancelled);

    // {secret}
    qUnsignedInt32 GetCallID() const
    {
        return m_uiID;
    }

    // {secret}
    void SetDependentConnection(void* pConnection, ConnectionID idConnection);
    // {secret}
    void* GetDependentConnection() const
    {
        return m_pDependentConnection;
    }
    // {secret}
    ConnectionID GetDependentConnectionID() const
    {
        return m_idConnection;
    }

    /*
    Description: Returns the Outcome of the call as a qResult.
    Returns: The Outcome of the call as a qResult.
    Remarks: The possible outcomes of a given call are specific to the call
    that is made. Refer to the documentation for the specific calls that may take a
    CallContext as a parameter for a description of the possible outcomes for that call.
    */

    /* Please see man pages for details 
    

    

    
    

    
    */
    qResult GetOutcome() const
    {
        return m_rCallOutcome;
    }

protected:

    // {secret}
    static qBool TransitionIsValid(State eInitialState, State eNewState);
    // {secret}
    static qBool StateIsTerminal(State eState)
    {
        return (eState==CallSuccess || eState==CallFailure || eState==CallCancelled);
    }
    // Used by the RMCContext - for now.
    //{secret}
    void SetOutcome(qResult rOutcome)
    {
        m_rCallOutcome = rOutcome;
    }

    //{secret}
    qBool CancelImpl(State eNewState);

public:
    // {secret}
    qBool CallIsComplete() const
    {
        return StateIsTerminal(m_eState);
    }

    // {secret}
    qBool HasExpired(Time tNow)
    {
        if  (m_tDeadline == NAMESPACE_NEX Time(0))
        {
            return false;
        }
        else
        {
            return (tNow > m_tDeadline ? true :false);
        }
    }

#ifdef TRACE_ENABLED    // n2370
    /* Please see man pages for details 
    
            
    
    */
    void Trace(qUnsignedInt64 uiTraceFlags = TRACE_ALWAYS);
#else
    void Trace(qUnsignedInt64 uiTraceFlags = TRACE_ALWAYS)
    {
        DUMMY_PARAM(uiTraceFlags);
    }
#endif  // TRACE_ENABLED

private:

    // {secret}
    virtual void BeginTransition(State /*eNewState*/, qResult /*rOutcome*/, qBool /*bSetOutcome*/) {}

    // Deactivate copy-ctor + op=
    // {secret}
    CallContext(const CallContext&);
    // {secret}
    CallContext& operator=(const CallContext&);

protected:
    // {secret}
    virtual void ProcessCallCompletion() {}


public:

#ifdef TEST_ENABLED
    //{secret}
    static qBool Test();
#endif

protected:
    qUnsignedInt32 m_uiFlags;

private:
    State m_eState;

    //{secret}
    struct InternalData
    {
        CompletionCallback m_Callback;
        UserContext        m_Context;

        InternalData( const CompletionCallback &cb,
                      const UserContext &c ) : m_Callback(cb), m_Context(c) {}
    };

    qList<InternalData> m_lstCompletionCallback;
    qVector<UserContext> m_vUserContext;

    qResult m_rCallOutcome;

    // n2370:memo
    /* Please see man pages for details 
     
     
     
     
     */
    qUnsignedInt32 m_uiID;

    void* m_pDependentConnection;
    ConnectionID m_idConnection;
    CallbackRoot* m_pCancellationCallback;
    Time          m_tDeadline;
};

//{secret}
extern void InvokeCallbackOnSuccess(CallContext* pContext, const UserContext* pUserParameter);
//{secret}
extern void InvokeCallbackOnCompletion(CallContext* pContext, const UserContext* pUserParameter);

END_NAMESPACE_NEX

#endif
