/*---------------------------------------------------------------------------*
  (C) Copyright 1998-2014, Technologies Quazal Inc. All rights reserved.
  Quazal Net-Z, Quazal Rendez-Vous and Quazal are trademarks of Quazal.

  Copyright (C)2011-2014 Nintendo  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo and/or its licensors and are
  protected by national and international copyright laws.
  They may not be disclosed to third parties or copied or duplicated
  in any form, in whole or in part, without the prior written consent
  of Nintendo.
  The content herein is highly confidential and should be handled
  accordingly.
 *---------------------------------------------------------------------------*/

#ifndef _Chrono_H_
#define _Chrono_H_

#include <OnlineCore/src/Core/UtilityDecl.h>
#include <OnlineCore/src/Platform/Core/Time.h>
#include <OnlineCore/src/Core/LocalClock.h>

/*
X = The function call is not allowed
O = The function call is allowed

State/Function  | Start | Pause | Resume | Check | Stop | Reset | GetState
---------------------------------------------------------------------------
Paused          |   X   |   X   |   O    |   O   |   O   |  O   |    O
Started         |   X   |   O   |   X    |   O   |   O   |  O   |    O
Stopped         |   O   |   X   |   X    |   O   |   X   |  O   |    O
*/





BEGIN_NAMESPACE_NEX

//Group = Time Management

/*
Description:  Measures the elapsed time from Start to Stop.
Remarks:  An instance of this class measures the elapsed time, in milliseconds, on
the local station. Use an instance of this class to precisely measure time in the
same manner as you would with a stopwatch.
*/

/* Please see man pages for details 
 
 
 
 
 
 */
class Chrono: public RootObject
{

    //Group =Methods

public:
    enum
    {
        CHRONO_IS_STARTED=0,
        CHRONO_IS_PAUSED =1,
        CHRONO_IS_STOPPED=2,
        INVALID_TIME_INTERVAL=-1
    };

    //Remarks:  By default, upon creation the Chrono is stopped. Use Start to start it.

    /* Please see man pages for details 
     
     
     */

    Chrono();

    /* Please see man pages for details 
     
     */
    virtual ~Chrono();

    //Description: Starts the Chrono at an initial time of zero. This method is
    //equivalent to pushing 'start' on a stopwatch.
    //Remarks:  You can only Start a Chrono that is presently stopped.
    //Returns: true if the Chrono was successfully started, false if not or an error occurred.
    // SYSTEMERROR_GEN_INVALID_OPERATION is signalled if the Chrono is not stopped.

    /* Please see man pages for details 
     
     
     
     
     
     
     
     */
    qBool Start();

    //Description: Pauses the Chrono if it was currently running.
    //Remarks: Pauses the Chrono until you resume it via Resume. You can resume the Chrono
    //after pausing it, which you cannot do if you Stop the Chrono.
    //Returns: The current elapsed time of the Chrono. The error SYSTEMERROR_GEN_INVALID_OPERATION is
    //signalled if this method is called and the Chrono is not started.

    /* Please see man pages for details 
     
     
     
     
     
     
     
     
     */
    TimeInterval Pause();

    //Description: Resumes the Chrono if it was temporarily paused.
    //Remarks: Resumes the operation of the Chrono if it was in a paused state.
    //You can Resume the Chrono after pausing it, which you cannot do if you stop the Chrono.
    //Returns: true if the Chrono successfully resumed operation, false if not or an error occurred.
    //The error SYSTEMERROR_GEN_INVALID_OPERATION is signalled if this method is called and the Chrono is not paused.
    /* Please see man pages for details 
     
     
     
     
     
     
     
     
     */
    qBool Resume();

    //Description: Returns the current elapsed time without stopping the Chrono.
    //Returns: The current elapsed time of the Chrono.
    //Remarks:  You can check the Chrono at anytime.

    /* Please see man pages for details 
     
     
     
     
     */
    TimeInterval Check();

    //Description: Stops the chrono and returns the elapsed time since the Chrono was started.
    //Remarks: The time given by the Chrono is the elapsed time since the Chrono
    //was started excluding the time that the Chrono was paused.
    //Returns: The current elapsed time of the Chrono. The error SYSTEMERROR_GEN_INVALID_OPERATION is signalled
    // if this method is called on a Chrono that is not started.
    /* Please see man pages for details 
     
     
     
     
     
     
     
     */
    TimeInterval Stop();

    //Description:  Resets the Chrono to zero.
    //Remarks: Sets the state of the Chrono to stopped and resets the time to zero.
    //You may Reset the Chrono at anytime.

    /* Please see man pages for details 
     
     */
    void Reset();

    // {secret}
    // Description:
    // Return the present state of the Chrono.
    // Remarks:  The state if one of: CHRONO_IS_STARTED, CHRONO_IS_STOPPED, CHRONO_IS_PAUSED.
    qUnsignedInt32 GetState() const;


#ifdef TEST_ENABLED
    //{secret}
    static qBool Test();
#endif

private:
    // {secret}
    TimeInterval Elapsed() const;

    // {secret}
    void UpdateAccumulatedTime();


private:
    LocalClock m_oClock;
    Time m_tInitialTime;
    TimeInterval m_tiAccumulatedTime;

    qUnsignedInt32 m_uiState;
};

END_NAMESPACE_NEX

#endif
