﻿/*---------------------------------------------------------------------------*
  (C) Copyright 1998-2014, Technologies Quazal Inc.  All rights reserved.
  Quazal Net-Z, Quazal Rendez-Vous and Quazal are trademarks of Quazal.

  Copyright (C)2011-2014 Nintendo  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo and/or its licensors and are
  protected by national and international copyright laws.
  They may not be disclosed to third parties or copied or duplicated
  in any form, in whole or in part, without the prior written consent
  of Nintendo.
  The content herein is highly confidential and should be handled
  accordingly.
 *---------------------------------------------------------------------------*/

#ifndef _DispatchQueue_H_
#define _DispatchQueue_H_

#include <OnlineCore/src/Core/UtilityDecl.h>
#include <OnlineCore/src/Platform/Core/CriticalSection.h>
#include <OnlineCore/src/Platform/Core/ScopedCS.h>
#include <OnlineCore/src/Platform/Core/RefCountedObject.h>
#include <OnlineCore/src/Platform/Core/EventHandler.h>
#include <OnlineCore/src/Platform/Core/Event.h>
#include <OnlineCore/src/Platform/Core/Time.h>
#include <OnlineCore/src/Platform/Core/qList.h>
#include <OnlineCore/src/Core/PseudoSingleton.h>

BEGIN_NAMESPACE_NEX

// This class is used by the AdapterEventDispatcher.

// A DispatchQueue is used with an Item that is refcounted, and that
// provides an access to an InstanceContext.

//group=Miscellaneous Classes

/*
Description: Controls how system events are dispatched.
Remarks: This class controls how events that occur in the system are dispatched.
*/
template<class Item> class DispatchQueue: public RootObject
{
    //group=Methods
public:

    /*
    Remarks: A DispatchQueue object should never be directly instantiated. Instead, it is created
    by $PRODUCTNAME$ prior to session creation.
    */
    DispatchQueue(): m_oEventHandler(1)
    {
        m_pWakeUpEvent = m_oEventHandler.CreateEventObject(0,0);
    }

    /*
    Remarks: A DispatchQueue object should never be directly destroyed. It is destroyed by
    $PRODUCTNAME$ when the session terminates.
    */
    virtual ~DispatchQueue()
    {
        // Must be empty when we get here.  It's too late to
        // dispatch - the derived class has been deleted.
        SYSTEMCHECK(IsEmpty());
        m_oEventHandler.DeleteEventObject(m_pWakeUpEvent);
    }

    //{secret}
    void Queue(Item* pItem)
    {
        {
            ScopedCS oCS(m_csQueue);
            ACQUIREREF(pItem);
            m_lstQueue.push_back(pItem);
        }
        m_pWakeUpEvent->Set();
    }

    //{secret}
    Item* PopFirstItem()
    {
        ScopedCS oCS(m_csQueue);
        if (!m_lstQueue.empty())
        {
            Item* pItem=m_lstQueue.front();
            SYSTEMCHECK(pItem!=NULL);
            m_lstQueue.pop_front();
            return pItem;
        }
        else
        {
            return NULL;
        }
    }

    /*
    Description: Dispatches all items in the queue for processing.
    Remarks: This method dispatches all items in the dispatch queue for processing. The method will
    return once all items have been dispatched and therefore, the queue is empty.
    See Also: DispatchFirstItem, Dispatch
    */
    void DispatchAll()
    {
        Dispatch(WAIT_INFINITE_TIMEOUT, true);
    }

    //{secret}
    void PurgeAll()
    {
        Item* pFirstItem=PopFirstItem();
        while (pFirstItem!=NULL)
        {
            DeleteItem(pFirstItem);
            pFirstItem=PopFirstItem();
        }
    }

    /*
    Description: Dispatches the first item in the queue for processing.
    Returns: true if the first item was successfully dispatched, false if not or an error occurred.
    Remarks: This method dispatches the first item in the dispatch queue for processing.
    See Also: DispatchAll, Dispatch
    */
    qBool DispatchFirstItem()
    {
        Item* pFirstItem=PopFirstItem();
        if (pFirstItem!=NULL)
        {
            CallProcessItem(pFirstItem);
            return true;
        }
        else
        {
            return false;
        }
    }

    /*
    Description: Dispatches items in the queue for processing for a given time.
    Remarks: This method dispatches items in the dispatch queue for processing until either the timeout
    is exceeded or the queue is empty. If bStopWhenEmpty is set to true, the method will return when either
    the timeout is exceeded or after all items have been dispatched, whichever occurs first. If bStopWhenEmpty
    is set to false, the method will return when the timeout is exceeded regardless of whether there are still
    items in the queue or not.
    Parameters:
    uiDispatchTimeout: The maximum time in milliseconds that the system will wait. The default
    uiDispatchTimeout is infinite.
    bStopWhenEmpty: If true (the default) and if all items are dispatched before the timeout is exceeded,
    this method will return when the queue is empty.

    See Also: DispatchAll, DispatchFirstItem
    */
    void Dispatch(qUnsignedInt32 uiDispatchTimeout, qBool bStopWhenEmpty=true)
    {
        qBool bDone=false;

        Time tiStart=Time::GetTime();
        while (!bDone)
        {
            qBool bEmptyQueue=true; // Assume the queue is empty...
            {
                Item* pFirstItem=PopFirstItem();
                if (pFirstItem!=NULL)
                {
                    // The queue isn't empty - process the first
                    // item.
                    CallProcessItem(pFirstItem);
                    // As a result of the process, the queue may
                    // become empty - but we don't really care -
                    // the next iteration will catch this case.
                    bEmptyQueue=false;
                }
            }

            if (bEmptyQueue)
            {
                // If the queue is empty - we either wait or stop.
                if (bStopWhenEmpty)
                {
                    // Stop.
                    bDone=true;
                }
                else
                {
                    // Wait.
                    qBool bWaitOutcome;
                    if (uiDispatchTimeout==WAIT_INFINITE_TIMEOUT)
                    {
                        bWaitOutcome=WaitForEvent(WAIT_INFINITE_TIMEOUT);
                        SYSTEMCHECK(bWaitOutcome==true); // Timeout can't happen: timeout is infinite!
                    }
                    else
                    {
                        // Compute new timeout and wait.
                        qUnsignedInt32 uiTimeSpent=qUnsignedInt32(Time::GetTime()-tiStart);
                        if (uiTimeSpent<uiDispatchTimeout)
                        {
                            bWaitOutcome=WaitForEvent(uiDispatchTimeout-uiTimeSpent);
                        }
                        else
                        {
                            bWaitOutcome=false; // Assume a timeout.
                        }
                    }
                    if (bWaitOutcome==false)
                    {
                        // Stop.
                        bDone=true;
                    }
                }
            }

            if (uiDispatchTimeout!=WAIT_INFINITE_TIMEOUT && Time::GetTime()-tiStart>uiDispatchTimeout)
            {
                // Timed out processing...
                bDone=true;
            }
        }
    }

    //{secret}
    qBool IsEmpty()
    {
        ScopedCS oCS(m_csQueue);
        return m_lstQueue.empty();
    }

    //{secret}
    qInt GetQueueLength()
    {
        ScopedCS oCS(m_csQueue);
        return m_lstQueue.size();
    }

private:

    qBool WaitForEvent(qUnsignedInt32 uiTimeout)
    {
        Event *pEvent;
        return m_oEventHandler.WaitForEvent(uiTimeout, &pEvent);
    }

    void CallProcessItem(Item* pItem)
    {
        // Record the current context.  This method will change the
        // instance context to the one stored in the item.  We must
        // be able to set it back after.
        ICContext icInitial=PseudoSingleton::GetCurrentContext();
        PseudoSingleton::SetContextIfRequired(pItem->GetInstanceContext());
        ProcessItem(pItem);
        RELEASEREF(pItem);
        // Set it back.
        PseudoSingleton::SetContextIfRequired(icInitial);
    }

    void DeleteItem(Item* pItem)
    {
        // Record the current context.  This method will change the
        // instance context to the one stored in the item.  We must
        // be able to set it back after.
        ICContext icInitial=PseudoSingleton::GetCurrentContext();
        PseudoSingleton::SetContextIfRequired(pItem->GetInstanceContext());
        RELEASEREF(pItem);
        // Set it back.
        PseudoSingleton::SetContextIfRequired(icInitial);
    }

protected:

    virtual void ProcessItem(Item* pItem)=0;
    CriticalSection& CSRef()
    {
        return m_csQueue;
    }

private:

    CriticalSection m_csQueue;
    qList<Item*> m_lstQueue;
    EventHandler m_oEventHandler;
    Event *m_pWakeUpEvent;
};

#ifdef TEST_ENABLED
//{secret}
qBool DispatchQueueTest();
#endif

END_NAMESPACE_NEX

#endif
