﻿/*---------------------------------------------------------------------------*
  (C) Copyright 1998-2014, Technologies Quazal Inc. All rights reserved.
  Quazal Net-Z, Quazal Rendez-Vous and Quazal are trademarks of Quazal.

  Copyright (C) 2011-2014 Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo and/or its licensors and are
  protected by national and international copyright laws.
  They may not be disclosed to third parties or copied or duplicated
  in any form, in whole or in part, without the prior written consent
  of Nintendo.
  The content herein is highly confidential and should be handled
  accordingly.
 *---------------------------------------------------------------------------*/

// The following code example is adapted from the book
// C++ Templates - The Complete Guide
// by David Vandevoorde and Nicolai M. Josuttis, Addison-Wesley, 2002
// Copyright David Vandevoorde and Nicolai M. Josuttis 2002

#ifndef _Holder_H_
#define _Holder_H_

#include <algorithm>

BEGIN_NAMESPACE_NEX

//==============================================================================
// Holds the reference of a dynamically allocated instance (through "new"). That
// is, if the holder deletes the instance it refers when it is destroyed while
// having the ownership of the pointed object.
//==============================================================================
/* Please see man pages for details 
 

        
 */
template <typename T>
class Holder : public RootObject
{
private:
    T* m_ptr; // Refers to the object it holds (if any)

public:
    // Default constructor
    Holder() : m_ptr(NULL) {}

    // Constructor for a pointer: let the holder refer to where the pointer refers
    explicit Holder (T* p) : m_ptr(p) {}

    // Destructor: Releases the object to which it refers (if any)
    ~Holder()
    {
        if (m_ptr!=NULL)
        {
            delete m_ptr;
        }
    }

    // Assigns of new reference deleting the previous (if any)
    Holder<T>& operator= (T* p)
    {
        if (m_ptr!=NULL)
        {
            delete m_ptr;
        }
        m_ptr = p;
        return *this;
    }

    // Pointer operators
    T& operator* () const
    {
        return *m_ptr;
    }

    // Get reference in the same way as *
    T& GetRef() const
    {
        return *m_ptr;
    }

    T* operator-> () const
    {
        return m_ptr;
    }

    // Gets referenced object (if any)
    T* Get() const
    {
        return m_ptr;
    }

    // Releases ownership of referenced object
    T* Release()
    {
        T* ret=m_ptr;
        m_ptr = NULL;
        return ret;
    }

    // Exchanges ownership with an other holder
    void ExchangeWith (Holder<T>& h)
    {
        std::swap(m_ptr, h.m_ptr);
    }

    // Exchanges ownership with an other pointer
    void ExchangeWith (T*& p)
    {
        std::swap(m_ptr, p);
    }

private:
    // Copying and copy assignment not allowed
    Holder (Holder<T> const&);
    Holder<T>& operator= (Holder<T> const&);
};

END_NAMESPACE_NEX

#endif
