﻿/*---------------------------------------------------------------------------*
  (C) Copyright 1998-2014, Technologies Quazal Inc. All rights reserved.
  Quazal Net-Z, Quazal Rendez-Vous and Quazal are trademarks of Quazal.

  Copyright (C) 2011-2014 Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo and/or its licensors and are
  protected by national and international copyright laws.
  They may not be disclosed to third parties or copied or duplicated
  in any form, in whole or in part, without the prior written consent
  of Nintendo.
  The content herein is highly confidential and should be handled
  accordingly.
 *---------------------------------------------------------------------------*/

#ifndef _Operation_H_
#define _Operation_H_

#include <OnlineCore/src/Core/CoreDecl.h>
#include <OnlineCore/src/Platform/Core/UserContext.h>
#include <OnlineCore/src/Platform/Core/SystemError.h>
#include <OnlineCore/src/Core/StateMachine.h>

BEGIN_NAMESPACE_NEX

//Group = System Operations

/*
Description: This is the root class of all $PRODUCTNAME$ operations.
Remarks:
$ifdef ETERNA$
The operations in $PRODUCTNAME$ fall into one of the categories DOOperation,
DupSpaceOperation, and SessionOperation, which respectively incorporate the operations
relating to duplicated objects, duplication spaces, and the game session.

The Operation class contains methods that enable the extraction of information about the type
of operation that it is and if there is a user-defined context associated to the operation.
$else$
The operations in $PRODUCTNAME$ fall into either of the categories DOOperation
or SessionOperation, which respectively incorporate the operations
relating to duplicated objects and the game session.

The Operation class contains methods that enable the extraction of information about the type
of operation that it is and if there is a user-defined context associated to the operation.
$endif$
*/
/* Please see man pages for details 
 
 
 

 
 
 
 
 
 
 
 */
#ifdef FOR_DOC_PURPOSE
class Operation: public RootObject
#else
class Operation: public StateMachine::QEvent
#endif
{
    //Group=Type Definitions
public:

    // This macro is never defined.  It is used only to get around a compilation problem with
    // DocJet, as Connect and Disconnect cannot be made secret.
#ifdef FOR_DOC_PURPOSE

    //$ifdef NETZ$
    //Description: This enumerates all types of operations (duplicated objects and game session)
    // that can be called by $PRODUCTNAME$.
    //Remarks: The defined operations fall into either the DOOperation or SessionOperation categories:
    //operations for duplicated objects and operations for the game session (JoinSessionOperation
    //and LeaveSessionOperation) respectively.
    //See Also: Type
    /* Please see man pages for details 
     
     
     */
    enum Type
    {
        /* See ChangeMasterStationOperation*/
        /* Please see man pages for details 
         
         */
        ChangeMasterStation,
        /* See ChangeDupSetOperation*/
        /* Please see man pages for details 
         
         */
        ChangeDupSet,
        /* See FaultRecoveryOperation*/
        /* Please see man pages for details 
         
         */
        FaultRecovery,
        /* See AddToStoreOperation*/
        /* Please see man pages for details 
         
         */
        AddToStore,
        /* See RemoveFromStoreOperation*/
        /* Please see man pages for details 
         
         */
        RemoveFromStore,
        /* See CallMethodOperation */
        /* Please see man pages for details 
         
         */
        CallMethod,
        /* See UpdateDataSetOperation */
        /* Please see man pages for details 
         
         */
        UpdateDataSet,
        /* See JoinSessionOperation */
        /* Please see man pages for details 
         
         */
        JoinSession,
        /* See LeaveSessionOperation */
        /* Please see man pages for details 
         
         
         */
        LeaveSession,
        /* Please see man pages for details 
         
         */
        JoinStation,
        /* Please see man pages for details 
         
         */
        LeaveStation
    };

#else

    enum Type
    {
        MasterIDChange = StateMachine::FirstUserSignal,
        FaultRecovery,
        AddToStore,
        CallMethod,
        UpdateDataSet,
        RemoveFromStore,
        Match,
        Connect,
        Disconnect,
        ChangeMasterStation,
        ChangeDupSet,
        JoinSession,
        LeaveSession,
        JoinStation,
        LeaveStation,
        CallProtocolMethod,
        CreateMaster,
        RendezVousLogin,
        RendezVousLogout,
        Test
    };


    enum Event
    {
        Begin, End, Approval, Denial, CustomTrace
    };

#endif

    //Group=Methods

    //{secret}
    Operation(qUnsignedInt32 uiOrigin=0);

    //{secret}
    virtual ~Operation()
    {
    }

    // {secret}
    void SetOrigin(qUnsignedInt32 uiOrigin)
    {
        m_uiOrigin = uiOrigin;
    }
    // {secret}
    qUnsignedInt32 GetOrigin() const
    {
        return m_uiOrigin;
    }

    //n1589: Implementation moved to cpp
#ifdef TRACE_ENABLED
    void Trace(Event oEvent) const;
    void Trace(qUnsignedInt64 uiTraceFlags = TRACE_ALWAYS) const;
#else
    void Trace(Event oEvent) const
    {
        DUMMY_PARAM(oEvent);
    }
    void Trace(qUnsignedInt64 uiTraceFlags = TRACE_ALWAYS) const
    {
        DUMMY_PARAM(uiTraceFlags);
    }
#endif

private:
    virtual StateMachine::Signal GetSignal() const
    {
        return static_cast<StateMachine::Signal>(GetType());
    }

public:

    /*
    Description: Returns the Type of the operation.
    Returns     : The Type of operation that has been called.
    Remarks: Refer to Type for the different types of operations that $PRODUCTNAME$ may call and to
    SetTraceFilter for an example of this method in use.
    See Also: Type
    */
    //n1589:memo:GetType is defined by IMPL_OPERATION_COMMON_METHODS
    /* Please see man pages for details 
     
     
     */
    virtual Type GetType() const =0;

    /*
    Description: Returns the user-defined value that was set via the method SetUserData.
    Returns     : The user-defined value set via the method SetUserData. The value may be either
    a qUnsignedInt32, qReal, or pointer to an object.

    $ifdef ETERNA$
    Remarks: The user-defined data should be set when one of the methods DataSet::OperationBegin,
    DuplicatedObject::OperationBegin, or DuplicationSpace::OperationBegin is called and retrieved
    when one of the methods DataSet::OperationEnd, DuplicatedObject::OperationEnd, or
    DuplicationSpace::OperationEnd is called.
    $endif$

    $ifdef NETZ$
    Remarks: The user-defined data should be set when either of the methods DataSet::OperationBegin
    or DuplicatedObject::OperationBegin is called and retrieved when either of the methods
    DataSet::OperationEnd or DuplicatedObject::OperationEnd is called.
    $endif$

    Note: The user must ensure that the methods SetUserData and GetUserData are called in
    relation to the same operation.

    See Also: SetUserData, UserContext
    */
    /* Please see man pages for details 
     
     
     
     
     
     
     
     
     
     */
    UserContext GetUserData();

    /*
    Description: Sets a user-defined value that can be obtained via the method GetUserData.
    Parameters:
    uUserData: A user-defined value which may be either a qUnsignedInt32, qReal, or pointer to an
    object.

    $ifdef ETERNA$
    Remarks: The user-defined value may be either a qUnsignedInt32, qReal, or pointer to an
    object. It should be set when one of the methods DataSet::OperationBegin,
    DuplicatedObject::OperationBegin, or DuplicationSpace::OperationBegin is called and retrieved
    when one of the methods DataSet::OperationEnd, DuplicatedObject::OperationEnd, or
    DuplicationSpace::OperationEnd is called.
    $endif$

    $ifdef NETZ$
    Remarks: The user-defined value may be either a qUnsignedInt32, qReal, or pointer to an
    object. It should be set when either of the methods DataSet::OperationBegin or
    DuplicatedObject::OperationBegin is called and retrieved
    when either of the methods DataSet::OperationEnd or DuplicatedObject::OperationEnd is called.
    $endif$

    Note: The user must ensure that the methods SetUserData and GetUserData are called in
    relation to the same operation.

    See Also: GetUserData, UserContext
    */
    /* Please see man pages for details 
     
     
     
     
     
     
     
     
     */
    void SetUserData(UserContext uUserData);

    /*
    Description: Returns the class name of this Operation as a string.
    Returns     : A string that represents the class name of the Operation.
    Remarks: Use this method to get the class name of the operation as a string. Refer to Type for the
    different types of operations that $PRODUCTNAME$ may call.
    See Also: Type
    */
    /* Please see man pages for details 
     
     
     */
    virtual const qChar* GetClassNameString() const = 0;

    // {secret}
    virtual void ForceImplOperationCommonMethodsMacro() = 0;


#ifdef FOR_DOC_PURPOSE
    /*
    Description: Use this method for casting (explicit type conversion) during run time.
    Parameters:
    pOperation: A pointer to an Operation class.
    Returns     : A valid pointer to the Operation if the Operation is of the expected type, a null
    pointer if it is not.
    Remarks: The DynamicCast method functions in the same manner as the standard C++ dynamic_cast
    operator but using $PRODUCTNAME$ run-time information.

    For example, we can return a pointer to the AddToStoreOperation object, if /pOperation/ points to a
    AddToStoreOperation object, as follows:

        AddToStoreOperation *pAddToStoreOperation = AddToStoreOperation::DynamicCast(pOperation)

    Note: This method is defined by each specific Operation class.
    */
    /* Please see man pages for details 
     
     
     
     
     
     
     
     
     
     
     
     
     
     
     */
    static SpecificOperation *DynamicCast( Operation *pOperation );
    static const SpecificOperation *DynamicCast( const Operation *pOperation );

#endif

    /*
    Description: Sets which operations will be traced via the TraceLog.
    Parameters:
    pfTraceFilter: A user defined function that takes a pointer to an Operation and returns a Boolean. If this
    is set to NULL, all operations for all UserDO s will be traced.
    Remarks: This method is called by $PRODUCTNAME$ to determine which operations should be traced via the TraceLog.
    To trace operations, you need to call the SetTraceFilter method that will enable the
    registration of a callback, set the TraceLog flag to trace operations, and then register a callback
    via the pfTraceFilter function that defines how the traces are filtered. If this function is not
    defined then by default all operations for all objects except CoreDOs are traced, that is, all UserDOs.

    Below is an example of how to call this method and how to register a callback to trace selected operations
    on a user object class.

        // Register a callback that defines how the traces will be filtered
        qBool FilterAvatarTraces(const Operation* pOperation) {
            DuplicatedObject* pAssociatedDO=NULL;
            // Select the operations to be traced
            switch (pOperation->GetType()) {
                case Operation::UpdateDataSet:
                case Operation::ChangeMasterStation:
                case Operation::FaultRecovery:
                case Operation::AddToStore:
                case Operation::RemoveFromStore:
                case Operation::CallMethod:
                case Operation::ChangeDupSet:
                    pAssociatedDO=((DOOperation*)pOperation)->GetAssociatedDO();
                    break;
                default:
                    break;
            }
            if (pAssociatedDO==NULL) {
                // Keep traces that are not related to a specific DO.
                return true;
            }

            // Trace the operation if the object is of the Avatar class.
            if (pAssociatedDO->IsA(DOCLASSID(Avatar))) {
                return true;
            } else {
                return false;
            }
        }

        // To trace operations, call the SetTraceFilter method and
        // then set the TraceLog flag to trace operations
        Operation::SetTraceFilter(FilterAvatarTraces);
        TraceLog::GetInstance()->SetFlag(TRACE_OPERATION);

    See Also: TraceLog
    */
    //n1589:memo:called from DOCore::DOCore
    /* Please see man pages for details 
     
     
     
     
     
     
     
     
     
     
     
     
     
     
     
     
     */
    static void SetTraceFilter(qBool (*pfTraceFilter)(const Operation*))
    {
        s_pfTraceFilter=pfTraceFilter;
    }

    // {secret}
    void AbortOperation();

    // {secret}
    qBool OperationIsAborted() const
    {
        return m_bOperationAborted;
    }

protected:
    // {secret}
    qChar* GetEventString(Event eEvent) const;

private:
#ifdef TRACE_ENABLED    // n2370
    virtual void TraceImpl(Event oEvent, qUnsignedInt64 uiTraceFlags) const =0;
#endif  // TRACE_ENABLED

private:
    // {secret}
    qBool m_bOperationAborted;

    // {secret}
    UserContext m_uUserData;

    static qBool (*s_pfTraceFilter)(const Operation*);

    //{secret}
    qUnsignedInt32 m_uiOrigin; // TODO - Change this to CID/PID
};

#define IMPL_OPERATION_COMMON_METHODS( TYPE )                               \
    virtual void ForceImplOperationCommonMethodsMacro(){};                                          \
    static TYPE##Operation *DynamicCast( Operation *pOperation ){                       \
        if( pOperation == NULL || pOperation->GetType() != TYPE ){ return NULL; }   \
        return static_cast< TYPE##Operation * >( pOperation );                          \
    }                                                                                   \
    static const TYPE##Operation *DynamicCast( const Operation *pOperation ){                       \
        if( pOperation == NULL || pOperation->GetType() != TYPE ){ return NULL; }   \
        return static_cast< const TYPE##Operation * >( pOperation );                            \
    }                                                                                   \
    virtual Type GetType() const {return TYPE;}                                         \
    virtual const qChar* GetClassNameString() const {return _T(#TYPE);}

// secretizes TestOperation class
#ifdef FOR_DOC_PURPOSE

#else
#ifdef TEST_ENABLED

class TestOperation: public Operation
{
public:
    TestOperation(qUnsignedInt32 uiOrigin = 0): Operation(uiOrigin) {}
    IMPL_OPERATION_COMMON_METHODS(Test);

    virtual void TraceImpl(Event eEvent, qUnsignedInt64 uiTraceFlags) const;
};

#endif
#endif

END_NAMESPACE_NEX

#endif
