/*---------------------------------------------------------------------------*
  (C) Copyright 1998-2014, Technologies Quazal Inc. All rights reserved.
  Quazal Net-Z, Quazal Rendez-Vous and Quazal are trademarks of Quazal.

  Copyright (C)2011-2014 Nintendo  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo and/or its licensors and are
  protected by national and international copyright laws.
  They may not be disclosed to third parties or copied or duplicated
  in any form, in whole or in part, without the prior written consent
  of Nintendo.
  The content herein is highly confidential and should be handled
  accordingly.
 *---------------------------------------------------------------------------*/

#ifndef _Pacer_H_
#define _Pacer_H_

#include <OnlineCore/src/Core/UtilityDecl.h>
#include <OnlineCore/src/Platform/Core/Time.h>
#include <OnlineCore/src/Core/Chrono.h>

BEGIN_NAMESPACE_NEX

class EventHandler;
class Event;

// Group=Time Management

/*
Summary:  Provides an easy way to ensure an activity goes no faster than a specified frequency.
Remarks:  The Pacer class provides an easy way to pace a particular activity
to ensure that it goes no faster than a specified frequency. It is important to note
that the pacer can only sleep to affect the frequency and therefore, can slow down but not
speed up processing time.
*/
/* Please see man pages for details 
 
 
 
 
 */
class Pacer: public RootObject
{

    // Group=Methods

public:

    /*
    Parameters:
    dFrequency: The maximum frequency at which you want the activity to be performed. The default value
    is zero.
    Remarks: This constructor constructs a Pacer object and calls SetFrequency(dFrequency) on it.
    */
    /* Please see man pages for details 
     
     
     
     
     
     */
    Pacer(qReal dFrequency=0);

    /* Please see man pages for details 
     
     
     */
    virtual ~Pacer();

    // Description: Sets the maximum frequency, in Hertz, at which
    // the paced activity will be performed.
    // Parameters:
    // dFrequency: Frequency, in Hertz, at which the Pacer runs. If
    // the value of dFrequency is 0 the Pacer will not delay the calls to Pace.
    // Remarks:  A frequency of 10 Hz (that is, 10 intervals per second) will cause
    // the Pacer to remain in the Pace method long enough to ensure that the calls
    // to perform the paced activity do not occur without at least a 0.1 second
    // delay (that is, 1/10). Therefore, if the processing time between two consecutive
    // calls to Pace is 0.02 seconds, Pace will sleep for 0.08 seconds.
    // See Also: GetFrequency
    /* Please see man pages for details 
     
     
     
     
     
     
     
     
     */
    void SetFrequency(qReal dFrequency);

    // Description: Returns the frequency of the Pacer.
    // Returns: The pacer frequency, in Hertz, that was set via SetFrequency or upon construction.
    // See also: SetFrequency
    /* Please see man pages for details 
     
     
     
     
     
     
     */
    qReal GetFrequency() const
    {
        return m_dFrequency;
    }

    // Description: Causes the pacer to return immediately from the Pace call.
    // Remarks: This method is called from another thread to cause the thread
    // that called Pace to exit from the method. Use this function to exit
    // from Pace while it is sleeping to pace an activity's frequency of occurrence.
    /* Please see man pages for details 
     
     
     */
    void Interrupt();

    /*
    Description: Causes a delay to ensure that the paced activity is called at a rate
    no faster than a set frequency.
    Remarks:  This is the main method of the pacer. It actually does the pacing.
    Assume that the frequency is set via SetFrequency to 10Hz (i.e.,
    10 intervals/second or 0.1 seconds/interval). If the delay between two
    calls to Pace is 0.06 s, the second call will sleep for 0.04 s to ensure
    that the method is called at a rate no faster than the set frequency.

    The Pacer also records statistics about the processing time, which is the
    time between two consecutive calls to the method Pace. For the above
    example the processing time is equal to 0.06 s.

    Note:  When Pace is called, the thread goes to sleep if a delay shorter
    than 1/dFrequency has occured since the last call. If the value of
    dFrequency is 0, it returns immediately.
    */
    /* Please see man pages for details 
     
     
     
     */
    virtual void Pace();

    // Description: Returns the statistics associated with the Pacer.
    // Returns: The statistics associated with the Pacer. The
    // statistics are calculated since the last time the method ResetStats
    // was called, or since the Pacer was created, whichever happened last.
    // Parameters:
    //   ptiTotal: Total elapsed time since the creation of the Pacer or
    //   since the last call to ResetStats.
    //   ptiTotalProcessing: Total time spent between the calls to Pace.
    //   ptiMinProcessing: Minimum time spent between two calls to Pace.
    //   ptiMaxProcessing: Maximum time spent between two calls to Pace.
    //   puiLoops: The number of times Pace was called since the creation
    //   of the Pacer or a call to ResetStats.
    //See Also: ResetStats



    /* Please see man pages for details 
     
     
     
     
     
     
     */
    void GetStats(TimeInterval* ptiTotal,
                  TimeInterval* ptiTotalProcessing,
                  TimeInterval* ptiMinProcessing,
                  TimeInterval* ptiMaxProcessing,
                  qUnsignedInt32 *puiLoops) const;

    /*
    Description: Resets all the statistics for the Pacer to zero.
    Remarks: Refer to GetStats for a description of the Pacer statistics that can be returned.
    See Also: GetStats
    */
    /* Please see man pages for details 
     
     
     */
    void ResetStats();

public:
#ifdef TEST_ENABLED
    /* {Secret} */
    static qBool Test();
#endif


private:
    qReal m_dFrequency;
    Chrono m_oChrono;
    qBool m_bFirstLoop;
    EventHandler *m_pEventHandler;
    Event *m_pInterruptEvent;

protected:
    TimeInterval m_tiDesiredLoopTime;

    TimeInterval m_tiTotal, m_tiTotalProcessing, m_tiMinProcessing, m_tiMaxProcessing;

    qUnsignedInt32 m_uiLoops;
    qBool m_bStatsRecorded;
};

END_NAMESPACE_NEX

#endif

