﻿/*---------------------------------------------------------------------------*
  (C) Copyright 1998-2014, Technologies Quazal Inc. All rights reserved.
  Quazal Net-Z, Quazal Rendez-Vous and Quazal are trademarks of Quazal.

  Copyright (C) 2011-2014 Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo and/or its licensors and are
  protected by national and international copyright laws.
  They may not be disclosed to third parties or copied or duplicated
  in any form, in whole or in part, without the prior written consent
  of Nintendo.
  The content herein is highly confidential and should be handled
  accordingly.
 *---------------------------------------------------------------------------*/

#ifndef _Scheduler_H_
#define _Scheduler_H_

#include <OnlineCore/src/Core/CoreDecl.h>

#include <OnlineCore/src/Core/Job.h>
#include <OnlineCore/src/Platform/Core/qChain.h>
#include <OnlineCore/src/Platform/Core/qReservedQueue.h>
#include <OnlineCore/src/Platform/Core/qMultiMap.h>
#include <OnlineCore/src/Platform/Core/Time.h>
#include <OnlineCore/src/Platform/Core/EventHandler.h>
#include <OnlineCore/src/Platform/Core/ProfilingUnit.h>
#include <OnlineCore/src/Platform/Core/HighResolutionChrono.h>
#include <OnlineCore/src/Core/WorkerThreads.h>
#include <OnlineCore/src/Core/Core.h>
#include <OnlineCore/src/Core/SingleThreadCallPolicy.h>

#ifdef WATERMARK_ENABLED
#include <OnlineCore/src/Platform/Core/WaterMark.h>
#endif


BEGIN_NAMESPACE_NEX

class Core;

//group=Miscellaneous Classes

typedef qBool (*PreventBlockCallCallback)(RootObject* object, qUnsignedInt32 uiTimeout);

/*
Description: Dispatches Jobs when $PRODUCTNAME$ is run in user-thread mode.
Remarks: This class is used to dispatch Jobs when $PRODUCTNAME$ is run in user-thread mode. If
you direct $PRODUCTNAME$ to use internal threads for processing you should not call any of
the methods of this class as Job scheduling is performed automatically.
See Also: Threading Model, Core::UseThreads
*/
/* Please see man pages for details 
 
 
 
 
 
 
 
 */
class Scheduler: public RootObject
{
    //group=Methods
public:
    /* Please see man pages for details 
     
     
     */
    class SchedulerWorkerThread: public WorkerThreads
    {
    public:
        SchedulerWorkerThread()
        {
            m_pScheduler=NULL;
        }
        virtual ~SchedulerWorkerThread() {};

        void AssociateScheduler(Scheduler* pScheduler)
        {
            m_pScheduler=pScheduler;
        }

        virtual void Initialize()
        {
            m_pScheduler->InitializeThread();
        }
        //n1589:todo1:m_pScheduler->Dispatch(100
        virtual void Work()
        {
            m_pScheduler->Dispatch(100, ContinueWhenEmpty);
        }

    protected:
        // This accessor is there in case we want to change the
        // way that the dispatch work is done.
        Scheduler* GetScheduler()
        {
            return m_pScheduler;
        }

    private:
        Scheduler* m_pScheduler;
    };

    // Description: Enumerates the flags you can use to control the Scheduler's behavior.
    // See Also: DispachFlag
    /* Please see man pages for details 
     
     */
    enum DispachFlag
    {
        // This flag keeps the Scheduler running until the uiDispatchTimeout expires, even if
        // all jobs have already been processed. If any new jobs arrive before uiDispatchTimeout
        // expires, they will be processed.
        /* Please see man pages for details 
         
         
         
         */
        ContinueWhenEmpty = 0x01,
        // This flag keeps the Scheduler running to prevent fault detection. With this flag, the Scheduler
        // will not dispatch any jobs for processing; it only provides a heartbeat to keep the
        // client connected.
        /* Please see man pages for details 
         
         
         */
        DispatchKeepAliveOnly = 0x02
    };

private:
    // Not allowed, not implemented.
    //{secret}
    Scheduler(const Scheduler&);
    Scheduler& operator=(const Scheduler&);

public:
    friend class Core;

    /*
    Remarks: A Scheduler object should never be directly instantiated. Instead, it is created
    by $PRODUCTNAME$ prior to session creation.
    */
    /* Please see man pages for details 
     
     
     
     */
    //n1589
#if 0
    //n1589:memo:original
    Scheduler(qByte byExecutionLockGroup = LockMasks::SystemLock, SchedulerWorkerThread* pSchedulerWorkerThread=NULL);
#else
    Scheduler(qUnsignedInt32 uiExecutionLockGroup = LockMasks::SystemLock, SchedulerWorkerThread* pSchedulerWorkerThread = NULL);
#endif

    /*
    Remarks: A Scheduler object should never be directly destroyed. It is destroyed by
    $PRODUCTNAME$ when the session terminates.
    */
    /* Please see man pages for details 
     
     
     
     */
    virtual ~Scheduler();

public:

    // {secret}
    void SetName(const qChar* szName);

    //{secret}
    void Queue(Job* pJob, qBool bFront = false);

    //{secret}
    qBool Cancel(Job *pJob);

    /*
    Description: Dispatches all Jobs in the dispatch queue for processing.
    Returns     : An error will be returned if this method is called when $PRODUCTNAME$ is set to "internal threads" mode.
    Refer to the Threading Model overview for threading details.
    Remarks: This method dispatches all Jobs in the dispatch queue for processing. The method will
    return once all Jobs ready to execute have been dispatched.

    $ifdef NETZ$
    Typically this method should be
    called each frame as given below to ensure all objects receive the most recent information
    before their datasets are updated or refreshed. Within the main loop the recommended sequence
    of calls to minimize latency is given below. It is possible to call DispatchAll or Dispatch
    at other points in the sequence or only once within the main loop however, this will add some
    latency to the system. In addition, dispatching jobs twice within this loop has no negative
    effect on performance.

        1.  Dispatch jobs
        2.  Refresh all duplicas
        3.  Perform game state change (physics, AI, etc.)
        4.  Update all duplication masters
        5.  Dispatch jobs
        6.  Render the game state

    Even if you call DispatchAll in the main loop, depending on how you write your code you may also need to call this
    method at other times. The most common situation is during load time, or any other time consuming sequence for that
    matter, where if messages are not regularly dispatched it will be assumed that a given Station has faulted and left
    the Session as it is not sending messages.

    Refer to the Threading Model overview for an example of how to use this method.
    $endif$
    $ifdef RENDEZVOUS$
    Note: If Dispatch or DispatchAll is not called at least each minute, a fault will be declared by Rendez-Vous.
    $endif$

    See Also: Dispatch
    */
    /* Please see man pages for details 
     
     
     
     
     
     
     
     
     
     
     
     
     
     
     
     
     
     
     
     
     
     
     
     
     
     */
    void DispatchAll()
    {
        Dispatch(WAIT_INFINITE_TIMEOUT, 0);
    }

    /*
    Description: Returns a pointer to the Scheduler instance.
    Returns     : A reference to the Scheduler object.
    Remarks: Use the reference returned by this method to control the Scheduler object.
    For example, all Jobs in the dispatch queue could be dispatched for processing as follows:

            Scheduler::GetInstance()->DispatchAll();
    */
    /* Please see man pages for details 
     
     
     
     */
    static Scheduler* GetInstance()
    {
        Core* pCore = Core::GetInstance();
        if (pCore==NULL)
        {
            return NULL;
        }
        else
        {
            return pCore->GetScheduler();
        }
    }

    /*
    Description: Starts a dispatcher thread on this Scheduler object.
    Returns     : true if a dispatcher thread was successfully started, false if not or an error occurred.
    Remarks: If Core::UseThreads is set to true, then when $PRODUCTNAME$ is created a Scheduler object is automatically
    created and a dispatcher thread started. This thread may then be stopped (StopDispatcherThread) and restarted
    (StartDispatcherThread) at will during run time. For example, if you want a specific section of code to be run in
    user threads but the rest run in internal threads, then you can stop the dispatcher thread while that specific code
    is being executed.
    See Also: StopDispatcherThread, HasDispatcherThread, Core::UseThreads
    */
    qBool StartDispatcherThread();

    /*
    Description: Stops the dispatcher thread on this Scheduler object.
    Returns     : true if the dispatcher thread was successfully stopped, false if not or an error occurred.
    Remarks: If Core::UseThreads is set to true, then when $PRODUCTNAME$ is created a Scheduler object is automatically
    created and a dispatcher thread started. This thread may then be stopped (StopDispatcherThread) and restarted
    (StartDispatcherThread) at will during run time. For example, if you want a specific section of code to be run in
    user threads but the rest run in internal threads, then you can stop the dispatcher thread while that specific code
    is being executed.
    See Also: StartDispatcherThread, HasDispatcherThread, Core::UseThreads
    */
    qBool StopDispatcherThread();

    /*
    Description: Returns whether or not this Scheduler object uses a dispatcher thread.
    Returns     : true if this Scheduler object uses a dispatcher thread, false if not or an error occurred.
    Remarks: The Scheduler will use dispatcher threads if $PRODUCTNAME$ has been instructed to uses threads,
    which is achieved by via setting Core::UseThreads to true.
    See Also: StartDispatcherThread, StopDispatcherThread, Core::UseThreads
    */
    qBool HasDispatcherThread();

    //n1589:memo
    /*
    Description: Dispatches Jobs in the queue for processing for a given time.
    Parameters:
    uiDispatchTimeout: The maximum time in milliseconds that the system will wait. The default
    uiDispatchTimeout is infinite.
    uiFlags: _DispatchFlag provides a method for keeping the Scheduler running even if there are no
    threads to process.
    uiFlags: You can set DispatchFlag s to determine behavior if all Jobs have been dispatched before
    the timeout is exceeded. You can set multiple flags using AND.
    Returns     : An error will be returned if this method is called when $PRODUCTNAME$ is set to "internal threads" mode.
    Refer to the Threading Model overview for threading details.
    Remarks: This method dispatches Jobs in the dispatch queue for processing until either the timeout
    is exceeded or the queue is empty. The ContinueWhenEmpty flag keeps the Scheduler running until the
    uiDispatchTimeout expires, even if all jobs have already been processed. If any new jobs arrive before
    uiDispatchTimeout expires, they will be processed. The DispatchKeepAliveOnly flag keeps the Scheduler
    running to prevent fault detection. With this flag, the Scheduler will not dispatch any jobs for processing;
    it only provides a heartbeat to keep the client connected.

    $ifdef NETZ$
    Typically this method should be called each frame as given below to ensure all objects receive the
    most recent information before their datasets are updated or refreshed. Within the main loop the
    recommended sequence of calls to minimize latency is given below. It is possible to call DispatchAll
    or Dispatch at other points in the sequence or only once within the main loop however, this will add
    some latency to the system. In addition, dispatching jobs twice within this loop has no negative
    effect on performance.

        1.  Dispatch jobs
        2.  Refresh all duplicas
        3.  Perform game state change (physics, AI, etc.)
        4.  Update all duplication masters
        5.  Dispatch jobs
        6.  Render the game state

    Even if you call Dispatch in the main loop, depending on how you write your code you may also need to call this
    method at other times. The most common situation is during load time, or any other time consuming sequence for that
    matter, where if messages are not regularly dispatched it will be assumed that a given Station has faulted and left
    the Session as it is not sending messages.

    Refer to the Threading Model overview for an example of how to use this method.
    $endif$
    $ifdef RENDEZVOUS$
    Note: If Dispatch or DispatchAll is not called at least each minute, a fault will be declared by Rendez-Vous.
    $endif$

    See Also: DispatchAll, DispatchFlag
    */
    /* Please see man pages for details 
     
     
     
     
     
     
     
     
     
     
     
     
     
     
     
     
     
     
     
     
     
     
     
     
     
     
     
     
     
     
     
     
     
     
     
     
     
     
     
     */
    qUnsignedInt Dispatch(qUnsignedInt32 uiDispatchTimeout, qUnsignedInt32 uiFlags = 0);

    //{secret}
    void SpecialDispatch(const Time& tDeadline);
    //{secret}
    void InitializeThread();

    //{secret}
    qBool WorkerThreadIsRunning() const
    {
        return m_pWorkerThreads->IsRunning();
    }


    //{secret}
    WorkerThreads* GetWorkerThreads()
    {
        return m_pWorkerThreads;
    }

    //{secret}
    qUnsignedInt GetNbWorkers() const
    {
        return m_pWorkerThreads->GetNbWorkers();
    }

    //{secret}
    qBool IsAWorker( ThreadID id ) const
    {
        return m_pWorkerThreads->IsAWorker( id );
    }

    //{secret}
    static void GlobalSingleThreadDispatch(qUnsignedInt32 uiTimeout=WAIT_INFINITE_TIMEOUT);
    //{secret}
    static void GlobalSingleThreadSpecialDispatch();
    //{secret}
    static void BypassGlobalSingleThreadDispatch();
    //{secret}
    static void UseGlobalSingleThreadDispatch();
private:
    static void SingleThreadDispatch(qUnsignedInt32 uiTimeout);
    static void SingleThreadSpecialDispatch();

public:

    //{secret}
    void EnableActivityTrace(qBool bState = true);

    //{secret}
    qBool DoesTraceActivity() const
    {
        return m_bTraceActivity;
    }

#ifdef TRACE_ENABLED
    /*
    Description: Traces the contents of this Scheduler object.
    Remarks: This method traces the list of jobs that are currently scheduled.
    */
    void Trace(qUnsignedInt64 uiTraceFlags=TRACE_ALWAYS) const;
    //{secret}
    void TraceJobState(const qChar* szEvent, Job* pJob);
#else
    void Trace(qUnsignedInt64 uiTraceFlags=TRACE_ALWAYS) const
    {
        DUMMY_PARAM(uiTraceFlags);
    };
    void TraceJobState(const qChar* /*szEvent*/, Job* /*pJob*/) {}
#endif

    // Outputs a warning to the log when the time spent executing one JobStep exceeds the specified time. Not output to log when INFINITE_LARGE_TIME_INTERVAL is specified.
    // Argument is in microseconds
    static void SetJobStepExecutionBusyIntervalThreshold(LargeTimeInterval interval);

    void WaitForAllJobsToComplete(qUnsignedInt32 timeout, qBool immidiately=true);
    //{secret}
    void WaitForEmptyQueues() const;
    //{secret}
    qUnsignedInt GetTotalQueueSize() const;
    //{secret}
    qUnsignedInt GetReadyQueueSize() const;
    /* Please see man pages for details */
    qUnsignedInt GetOneTypeJobsCount(Job::JobType jobType) const;
#if NEX_ENABLE_SCHEDULE_BUSY_CHECK
    /*
    Description: Gets whether the scheduler that has a large volume of jobs registered is in a busy state.
    */
    qBool IsJobQueueBusy() const;

    /*
    Description: Gets whether the feature to deny new Station participation in the Session while the Scheduler is in a busy state is enabled.
    Returns     : true when SetCheckingQueueToDenyJoiningEnabled is set to true, and Core::IsUsingThreads function is true.
    */
    qBool IsCheckingQueueToDenyJoiningEnabled() const;

    /*
    Description: Sets the threshold value to determine whether the Scheduler is busy.
    Remarks: The default value is set to 100.
    */
    void SetBusyBorder( qUnsignedInt nJobCount )
    {
        m_nQueueBusyBorder = nJobCount;
    }

    /*
    Description: Sets whether to deny new participation in a Session when the Scheduler is in a busy state.
    Remarks: The default value is set to true.
    */
    void SetCheckingQueueToDenyJoiningEnabled( qBool bEnabled )
    {
        m_bCheckingQueueToDenyJoiningEnabled = bEnabled;
    }
#endif  // NEX_ENABLE_SCHEDULE_BUSY_CHECK

    //{secret}
    void ClearAllQueues();

    //{secret}
    static qBool CurrentThreadCanWaitForJob();

    //{secret}
    qBool CurrentThreadIsDispatchingJobs() const
    {
        return m_oDispatchCallPolicy.CurrentThreadIsDispatchingJobs();
    }

    // {secret}
    void IgnoreWaitTime()
    {
        m_bIgnoreWaitTime=true;
    }
    // {secret}
    void RestoreWaitTime()
    {
        m_bIgnoreWaitTime=false;
    }

    //{secret}
    void RegisterSpecialDispatchJob(Job* pJob, qBool bFront = false);
    //{secret}
    void UnregisterSpecialDispatchJob(Job *pJob);
    //{secret}

    /*
    Description: Returns a reference to the CriticalSection that currently locks the system.
    Returns     : A reference to the CriticalSection that is currently used to lock the system.
    Remarks: Once you have a reference to a CriticalSection, you may then enter and leave it
    as required. Note that once a CriticalSection is entered and therefore, the system locked,
    you cannot call any blocking functions, such as CallContext::Wait or Dispatch, until you
    leave the CriticalSection. Refer to Threading Model for more information on blocking functions.
    Example:
        void MyFunction() {
            // Lock the system. No state change will occur until
            // the block is exited.
            ScopedCS oCS(Scheduler::GetInstance()->SystemLock()); //n1589:mark:TransportLock:off

            // Protected block of code. You cannot call any blocking
            // function while the system is locked.
        }

    See Also: CriticalSection, ScopedCS, Threading Model
    */

    /* Please see man pages for details 
     
     
     
     
     
     
     
     
     */
    CriticalSection& SystemLock()
    {
        // We do not often check the validity of the Scheduler::GetInstance()
        // pointer before we call this method and we are therefore able to
        // return a reference to an invalid CriticalSection which
        // then leads to hard to diagnose problems since the program will
        // only crash once the member of some invalid object is actually
        // referenced. Checking that the 'this' pointer is valid will catch
        // this error.
        SYSTEMCHECK(this != NULL);

        return m_csSystemLock;
    }

#ifdef LOCK_DEBUG_ENABLED
    //Note for bug 2531: if a SYSTEMCHECK happens upon returned value of IsLockedByCurrentThread(),
    //even if no internal threads are used (for example, Core::UseThreads(false); ), this means that
    //another thread, somewhere, has entered the scope of the system lock.
    //In this case, we might recommend any given user to not make calls of RV or NetZ from another
    //thread, unless he is implementing his own locking mechanism.
    //See bug 2531 for details.

    //{secret}
    qBool IsLockedByCurrentThread()
    {
        return m_csSystemLock.HasOwnership();
    }

#endif //LOCK_DEBUG_ENABLED

#if NEX_ENABLE_LOCK_OPTIMIZATION
    CriticalSection& GetTransportLock()
    {
        SYSTEMCHECK(this != NULL);
        return m_csTransportLock;
    }

#ifdef LOCK_DEBUG_ENABLED
    qBool IsTransportLockedByCurrentThread()
    {
        return m_csTransportLock.HasOwnership();
    }
#endif
#endif

    //{secret}
    static CriticalSection& GlobalSystemLock()
    {
        return s_csGlobalSystemLock;
    }

    //n1589
#ifdef LOCK_DEBUG_ENABLED
    qBool IsGlobalLockedByCurrentThread()
    {
        return s_csGlobalSystemLock.HasOwnership();
    }
#endif

    //{secret}
    qBool IsQueued(const Job* pJob) const;

    qUnsignedInt32 GetNbSpecialJobs() const;

    void ProcessJobsInTermination(Job::JobType jobType);
#if NEX_ENABLE_IMPROVE_DO_CORRUPT
    void TerminateDOCoreJobs();
    void DeleteDOCoreJobs(qChain<Job*>* m_lstQueue);
    void DeleteDOCoreJobs(qMultiMap<Time, Job*>* m_mapQueue);
#endif
    void ForceTerminateJobs();

    void RequeueWaitingJob(Job* pWaitingJob, Time waitTime);

private:
    void QueueImpl(Job* pJob, qBool bFront);

    void IncrementNbJobs();
    void DecrementNbJobs();

    void DispatchBegins();
    void DispatchEnds();

    void PopFirstJob();

    void WaitForJobToExecute(qUnsignedInt32 uiTimeout);

    void ProcessJob(Job* pJob, const Time& tDeadline);
    void DeleteJob(Job* pJob);

    // {secret}
    void ProcessCurrentJob(const Time& tDeadline );
    //{secret}
    qBool QueueNewlyReadyJobs();
    // {secret}
    qBool JobIsReady(Job* pJob, const Time& tExecutionTime, const Time& tNow);

    qUnsignedInt32 GetTotalNumberOfJobs() const;
    //{secret}
    void DispatchImpl(Time tDeadline);
    //{secret}
    qUnsignedInt32 TimeToDeadline(Time tDeadline) const;
    //{secret}
    qBool DeadlineExpired(Time tDeadline) const
    {
        return TimeToDeadline(tDeadline)==0;
    }

#if NEX_DISABLE_SCHEDULER_PERFCOUNTER_OPTIMIZATION
#else
    // {secret}
    ProfilingUnit& DispatchProfilingUnit()
    {
        return m_puDispatchInterval;
    }
    // {secret}
    const ProfilingUnit& DispatchProfilingUnit() const
    {
        return m_puDispatchInterval;
    }
    // {secret}
    void SetMaxDispatchInterval(TimeInterval tiInterval)
    {
        s_tiMaxDispatchInterval = tiInterval;
    }
#endif  // !defined(NEX_DISABLE_SCHEDULER_PERFCOUNTER_OPTIMIZATION)

public:

#ifdef TEST_ENABLED
    //{secret}
    static qBool Test();
#endif

private:
    Job*  m_pCurrentJob;
    volatile qBool m_bCurrentJobValid;
    SchedulerWorkerThread* m_pWorkerThreads;
    EventHandler m_oEventHandler;

    Event *m_pWakeUpEvent;

    //{secret}
    mutable CriticalSection m_csSystemLock;

    //n1589: Declared with NEX_ENABLE_LOCK_OPTIMIZATION to simplify the code
#if NEX_ENABLE_LOCK_OPTIMIZATION
    mutable CriticalSection m_csTransportLock;
#endif

    //{secret}
    mutable CriticalSection m_csQueue;

    // For NEX Developer <wiki/Core Threading Model#queues>
    //{secret}
    qChain<Job*> m_lstReadyQueue;
    qChain<Job*> m_lstSuspendedQueue;
    qMultiMap<Time, Job*> m_mapWaitingQueue;

    // Special jobs.  These special dispatch job is used to
    // integrate the transport work, in single threaded operation mode.
    // This is a hack - as no generalized form of such a special job is
    // required.  This mechanism should eventually be generalized.
    // {secret}
    qChain<Job*> m_lstSpecialDispatchJobs;

    qBool m_bIgnoreWaitTime;

#ifdef WATERMARK_ENABLED
    WaterMark m_oSchedulerJobs;
    WaterMark m_oBusyRatio;
#endif
    ProfilingUnit m_oJobProcessing;

    static qBool s_bGlobalDispatch;
    Time m_tiLastDispatchStart;
    Time m_tiLastDispatchEnd;
    // A lock valid across all instances of the scheduler (when using multiple contexts)
    //{secret}
    static CriticalSection s_csGlobalSystemLock;

    // Instance context in which this scheduler must run.
    //{secret}
    ICContext m_icContext;

    //{secret}
    // Following the information found at
    // For NEX Developer <wiki/CallPolicy#declaration>, we declare a particular call
    // policy, that will be used to control that Dispatch is always
    // called by a single thread.
    SingleThreadCallPolicy m_oDispatchCallPolicy;

    Time m_prevDispatchTime;
#if NEX_DISABLE_SCHEDULER_PERFCOUNTER_OPTIMIZATION
#else

    // (Perfcounter) Measures the frequency at which dispatch
    // is called.  This is useful to debug situations where it
    // isn't called often enough.
    ProfilingUnit m_puDispatchInterval;
    HighResolutionChrono m_oChrono; // to report info to the DispatchInverval.
    qBool m_bRecordDispatchInterval;

#endif  // !defined(NEX_DISABLE_SCHEDULER_PERFCOUNTER_OPTIMIZATION)
    static TimeInterval s_tiMaxDispatchInterval;

#if NEX_ENABLE_SCHEDULE_BUSY_CHECK
    //
    qUnsignedInt m_nQueueBusyBorder;

    //
    qBool m_bCheckingQueueToDenyJoiningEnabled;
#endif  // NEX_ENABLE_SCHEDULE_BUSY_CHECK

    qBool m_bTraceActivity;

public:
    static qBool PreventBlockCall(
        qBool bPotentialBlockTest,
        qUnsignedInt32 uiTimeout,
        qBool bErrorValue,
        PreventBlockCallCallback cbCallback,
        RootObject* object,
        qBool* pbReturn
    );

    static qBool PreventRegularBlockCall(
        qBool bPotentialBlockTest,
        qUnsignedInt32 uiTimeout,
        PreventBlockCallCallback cbCallback,
        RootObject* object,
        qResult* prReturn
    );
};

END_NAMESPACE_NEX

#endif
