﻿/*---------------------------------------------------------------------------*
  (C) Copyright 1998-2014, Technologies Quazal Inc. All rights reserved.
  Quazal Net-Z, Quazal Rendez-Vous and Quazal are trademarks of Quazal.

  Copyright (C) 2011-2014 Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo and/or its licensors and are
  protected by national and international copyright laws.
  They may not be disclosed to third parties or copied or duplicated
  in any form, in whole or in part, without the prior written consent
  of Nintendo.
  The content herein is highly confidential and should be handled
  accordingly.
 *---------------------------------------------------------------------------*/

#ifndef _MemoryManager_H_
#define _MemoryManager_H_

// See jwiki:MemoryAllocation for an overview of memory allocation in the system.

#ifdef __cplusplus  // see else below...

#include <OnlineCore/src/Platform/Core/PlatformDecl.h>
#include <OnlineCore/src/Platform/Core/CriticalSection.h>
#include <OnlineCore/src/Platform/Core/MutexPrimitive.h>
#include <OnlineCore/src/Platform/Core/Time.h>

// The amount of 0 padding to add to each block
#ifndef HEAP_TRAILER_PADDING_SIZE
// {secret}
// If not specified, defaults to the same size as the signature contained in the
// trailer since anything less would end up taking that much in order for the
// Trailer to be aligned.
#define HEAP_TRAILER_PADDING_SIZE (sizeof(qUnsignedInt32))
#endif //HEAP_TRAILER_PADDING_SIZE

#ifdef GLOBALNEW_DETECTION_ENABLED
// Global operator.

// We define the standard operator new/delete.  There is a total of 6 versions
// of each - the version with exceptions or not, array and not, plus
// the placement version.

// We do not redefine the placement new/delete.  This is typically
// defined inline by the compiler includes, and we do not need to
// redefine it (see also bug:4675 for a related question).

// {secret}
#ifdef CTR_TARGET
void *operator new(size_t) throw(std::bad_alloc);
#else
void *operator new(size_t);
#endif
// {secret}
#ifdef CTR_TARGET
void *operator new[](size_t size) throw(std::bad_alloc);
#else
void *operator new[](size_t size);
#endif
// {secret}
void* operator new(std::size_t, const std::nothrow_t&) throw();
// {secret}
void* operator new[](std::size_t, const std::nothrow_t&) throw();

// {secret}
void operator delete(void* uptr) throw();
// {secret}
void operator delete[](void *uptr) throw();
// n2605: warning 831-D: support for placement delete is disabled. (see ARM\include\new.)
// {secret}
//void operator delete(void*, const std::nothrow_t&) throw();
// {secret}
//void operator delete[](void*, const std::nothrow_t&) throw();

// We now redefine our own versions of the operator new.  This is the
// one that gets called when qNew is used, in debug mode.  qNew adds
// extra debug info.

//{secret}
void *operator new(size_t size, NAMESPACE_NEX qChar *szFile, NAMESPACE_NEX qUnsignedInt32 uiLine) throw();
//{secret}
void *operator new[](size_t size, NAMESPACE_NEX qChar *szFile, NAMESPACE_NEX qUnsignedInt32 uiLine) throw();




#endif //GLOBALNEW_DETECTION_ENABLED

BEGIN_NAMESPACE_NEX

#ifdef WATERMARK_ENABLED
class WaterMark;
#endif

class MutexPrimitive;

//group=Miscellaneous Classes
//$ifdef WII_COMMON$
//group=Net-Z API
//$endif$

/*
Description: Overrides the default $PRODUCTNAME$ memory allocation functions.
Remarks: This class is used to override the default $PRODUCTNAME$ memory allocation functions.

$ifdef WII_COMMON$
Note: Memory access on the Wii is not thread-safe. So it is assumed that you will create your own
thread-safe memory allocation functions. If you wish to use them with Net-Z objects, you must register
them using the SetBasicMemoryFunctions method.
$endif$

*/
/* Please see man pages for details 
 

 
 
 

 
 
 */
class MemoryManager: public RootObject
{
public:

    typedef enum _InstructionType
    {
        Unknown,
        DirectCall,
        UntrackedDirectCall,
        RootObjectNewDelete,
        RootObjectNewDeleteArray,
        GlobalNewDelete,
        GlobalNewDeleteArray,
        AllocatorClass,
        qSpecialNewDelete,
        qSpecialNewDeleteArray,
        qAllocFree
    } InstructionType;

    //n1769
    const qChar* GetInstructionTypeString(InstructionType t) const;

#ifdef HEAP_DEBUGGING_ENABLED
    typedef struct _Trailer
    {
        qUnsignedInt32 m_uiSignature;
        qByte m_pPadding[HEAP_TRAILER_PADDING_SIZE];
    } Trailer;
#endif


    /* Please see man pages for details 
    

       
       

    
    
    
    */
    typedef void* (* MallocFunction)(size_t stSize);


    /* Please see man pages for details 
    

       

    
    
    */
    typedef void (*FreeFunction)(void *pMemory);



#ifdef DBG_STACKWALK_SUPPORT
    enum { MaxDbgStackSize = 12 };
#endif

    typedef struct _Header
    {
#ifdef HEAP_DEBUGGING_ENABLED
        qUnsignedInt32  m_uiLine;
        const qChar             *m_szFile;
        InstructionType m_eSourceInstruction;
        qByte m_padding[7];
        qUnsignedInt64  m_tiTime;
        MemoryManager   *m_pMemoryManager;
        Trailer         *m_pTrailer;
        struct _Header  *m_pPrevious;
        struct _Header  *m_pNext;
        qUnsignedInt32  m_uiSignature;
#ifdef DBG_STACKWALK_SUPPORT
        qUnsignedInt32 m_pStack[MaxDbgStackSize];
#endif
#endif
#ifdef LEAK_DETECTION_ENABLED
        size_t          m_stRealSize;
#endif
        FreeFunction    m_fcnFree;
        void*           m_pInfo; // Additional information
    } Header;

protected:
    /*
    Parameters:
    szName: The name of this MemoryManager.
    Remarks: A MemoryManager object should never be directly instantiated. It is created
    by $PRODUCTNAME$ when required.
    */
    MemoryManager(qChar* szName);
public:
    /*
    Remarks: A MemoryManager object should never be directly destroyed. It is destroyed
    by $PRODUCTNAME$ when no longer required.
    */
    virtual ~MemoryManager();

#ifdef HEAP_DEBUGGING_ENABLED

    //{secret}
    static void *Allocate(size_t stSize, InstructionType eSourceInstruction=DirectCall)
    {
        return Allocate( stSize, _T("Unknown"), 0, eSourceInstruction );
    }
    //{secret}
    static void *Allocate(size_t stSize, const qChar *szFile, qUnsignedInt32 uiLine, InstructionType eSourceInstruction=DirectCall);
    //{secret}
    static void Free(void *pPtr, InstructionType eSourceInstruction=DirectCall);

#else   // HEAP_DEBUGGING_ENABLED

    //{secret}
    static void *Allocate(size_t stSize, const qChar *szFile, qUnsignedInt32 uiLine, InstructionType eSourceInstruction=DirectCall)
    {
        DUMMY_PARAM( szFile );
        DUMMY_PARAM( uiLine );
        DUMMY_PARAM( eSourceInstruction );
        return Allocate( stSize );
    }
    //{secret}
    static void *Allocate(size_t stSize, InstructionType eSourceInstruction)
    {
        DUMMY_PARAM( eSourceInstruction );
        return Allocate( stSize );
    }
    //{secret}
    static void Free(void *pPtr, InstructionType eSourceInstruction)
    {
        DUMMY_PARAM( eSourceInstruction );
        Free( pPtr );
    }
    //{secret}
    static void *Allocate(size_t stSize);
    //{secret}
    static void Free(void *pPtr);

#endif  // HEAP_DEBUGGING_ENABLED


protected:
    // Methods used by the manager.

    static void* GenericMalloc(size_t stRealSize);
    static void GenericFree(FreeFunction fcnFree, void* pPtr);

#if defined(HEAP_DEBUGGING_ENABLED)
    virtual void BeginProtection();
    virtual void EndProtection();
#else
    virtual void BeginProtection() {}
    virtual void EndProtection() {}
#endif
    void IncreaseMemUsage(size_t stRealSize);
    void DecreaseMemUsage(size_t stRealSize);

public:
#ifndef NEX_DEFAULT_ALIGNMENT
#define NEX_DEFAULT_ALIGNMENT   4
#endif
    // When the user specifies the amount of memory they want to be allocated, calculate how much memory should actually be allocated here.
    static qUnsignedInt GetAlignedSize(qUnsignedInt size)
    {
        return size + ((NEX_DEFAULT_ALIGNMENT - (size % NEX_DEFAULT_ALIGNMENT)) % NEX_DEFAULT_ALIGNMENT);
    }
    //{secret}
    static qUnsignedInt GetHeaderSize()
    {
        return GetAlignedSize(sizeof(Header));
    }
    //{secret}
    static Header* GetHeaderPtr(void* pUserPtr)
    {
        return (Header*)((qByte*)pUserPtr-GetHeaderSize());
    }

    // To cleanup.

#ifdef LEAK_DETECTION_ENABLED
    //{secret}
    qBool MemoryLeakDetected(qUnsignedInt32 uiDesiredMemoryUsage=0)
    {
        return m_uiMemUsage!=uiDesiredMemoryUsage;
    }
#ifdef HEAP_DEBUGGING_ENABLED   //Release build fix; requires m_tiTime member from Header (HEAP_DEBUGGING_ENABLED only)
    //{secret}
    void DumpAllocatedMemory(qUnsignedInt64 tiTime = 0, qBool boShowStack = false);
#endif
    //{secret}
    static size_t GetRealAllocationSize(Header* pHeader);
    //{secret}
    static size_t GetUserAllocationSize(Header* pHeader);
    //{secret}
    qUnsignedInt32 GetMemoryUsage()
    {
        return m_uiMemUsage;
    }
#endif

#ifdef HEAP_DEBUGGING_ENABLED
    //{secret}
    static void CheckPointer(MemoryManager* pManager, void* pUserPtr, InstructionType eSourceInstruction=Unknown)
    {
        CheckBlock(pManager, GetHeaderPtr(pUserPtr), eSourceInstruction);
    }

    //{secret}
    static void CheckPointer(void* pUserPtr, InstructionType eSourceInstruction=Unknown)
    {
        CheckBlock(GetDefaultMemoryManager(), GetHeaderPtr(pUserPtr), eSourceInstruction);
    }

    //{secret}
    static void CheckBlock(MemoryManager* pManager, Header* pHeader, InstructionType eSourceInstruction=Unknown);
    //{secret}
    qBool ValidBlock(Header *pHeader, InstructionType eSourceInstruction=Unknown);
    //{secret}
    qBool ValidPointer(void* pPtr, InstructionType eSourceInstruction=Unknown)
    {
        return ValidBlock(GetHeaderPtr(pPtr), eSourceInstruction);
    }

    //{secret}
    void CheckMemoryState();

    //{secret}
    void SetLog(Log *pLog)
    {
        m_pLog = pLog;
    }

    //{secret}
    void GetMemoryBlockInfo(Header *pHeader, qChar *szInfo, qBool bDumpStack = false);

    //{secret}
    void LogSnapshot(Log *pLog, qUnsignedInt64 tStartTime = 0, qUnsignedInt64 tEndTime = 0, qBool bDumpStack = false);
#endif

    static MemoryManager* GetDefaultMemoryManager();

    //{secret}
    static void ShutdownDefaultMemoryManager();

    //group=Methods

    /*
    Description: Enables the developer to override the default memory allocation functions used
    by $PRODUCTNAME$.
    Parameters:
    fcnMalloc: A user-defined Malloc function.
    fcnFree: A user-defined Free function.
    Remarks: This method enables the developer to override the memory allocation functions used
    by the $PRODUCTNAME$ memory manager. By default $PRODUCTNAME$ calls the /malloc/ and /free/
    functions to allocate memory. Using this method, however, you can change these functions to your
    own memory allocation functions. This method may be called at any time. However, it is normally
    called before the $PRODUCTNAME$ object is created.

    By redefining the MallocFunction and FreeFunction methods, SetBasicMemoryFunctions can be used,
    for example, to allocate memory via your own memory manager or to calculate variables such as
    how much memory $PRODUCTNAME$ uses. If you redefine the fcnMalloc and fcnFree functions, they
    must use the same signatures as given in the example below.

    For example, we could redefine the MallocFunction and FreeFunction to count the number
    of allocated memory blocks as follows:

        qUnsignedInt g_uiAllocatedBlocksCnt=0;

        void* SpecialMalloc(size_t stSize) {
            g_uiAllocatedBlocksCnt++;
            return malloc(stSize);
        }

        void SpecialFree(void* pBlock) {
            g_uiAllocatedBlocksCnt--;
            free(pBlock);
        }

        MemoryManager::SetBasicMemoryFunctions(SpecialMalloc, SpecialFree);

    */


    /* Please see man pages for details 
     

     
     

     

     
     
     

     
     

     
     

     
     
     
     
     
     

     
     
     

     

     

     

     
     
     */
    static void SetBasicMemoryFunctions(MallocFunction fcnMalloc, FreeFunction fcnFree)
    {
        s_fcnMalloc = fcnMalloc;
        s_fcnFree = fcnFree;
    }

#ifdef CAFEIOP_TARGET
    /* Please see man pages for details 
     

         
         

     
     
     */
    static void SetIpcMemoryFunctions(MallocFunction fcnMalloc, FreeFunction fcnFree)
    {
        s_fcnIpcMalloc = fcnMalloc;
        s_fcnIpcFree = fcnFree;
    }
    static void *AllocateIpc(size_t stSize)
    {
        return s_fcnIpcMalloc(stSize);
    }
    static void FreeIpc(void *pPtr)
    {
        s_fcnIpcFree(pPtr);
    }
#endif



#ifdef WII_TARGET
    // {secret}
    static qBool IsMemoryFunctionSet()
    {
        return (s_fcnMalloc != NULL);
    }
#endif
    // {secret}
    void Trace();

private:

#ifdef HEAP_DEBUGGING_ENABLED
    //{secret}
    static void LinkHeader(MemoryManager* pManager, Header* pHeader);
    //{secret}
    static void UnlinkHeader(MemoryManager* pManager, Header* pHeader);
#endif
    const qChar   *m_szName;
    qUnsignedInt32 m_uiMemUsage;
    Header        *m_pLastHeader;
    Log           *m_pLog;
#ifdef WATERMARK_ENABLED
    WaterMark   *m_pMemoryWaterMark;
#endif
    static enum eShutDownState
    {
        ShutDownPending,
        ShutDownInProcess,
        ShutDownDone
    } s_eShutDownState;

    MutexPrimitive* m_pMutex;

    //{secret}
    static MallocFunction s_fcnMalloc;
    //{secret}
    static FreeFunction s_fcnFree;
#ifdef CAFEIOP_TARGET
    //{secret}
    static MallocFunction s_fcnIpcMalloc;
    //{secret}
    static FreeFunction s_fcnIpcFree;
#endif
};

//{secret}
template <class T> T* SpecialNewArray(qUnsignedInt uiNbObjects, qChar* szFile, qUnsignedInt32 uiLine)
{
    // Allocate the memory.
#ifdef HEAP_DEBUGGING_ENABLED
    void *pMemoryBlock=MemoryManager::Allocate(sizeof(T)*uiNbObjects + sizeof(qUnsignedInt32), szFile, uiLine, MemoryManager::qSpecialNewDeleteArray);

#else /* HEAP_DEBUGGING_ENABLED */
    DUMMY_PARAM(szFile);
    DUMMY_PARAM(uiLine);
    void *pMemoryBlock=MemoryManager::Allocate(sizeof(T)*uiNbObjects + sizeof(qUnsignedInt32), MemoryManager::qSpecialNewDeleteArray);
#endif /* HEAP_DEBUGGING_ENABLED */

    qUnsignedInt32 *pNbObjects=(qUnsignedInt32*) pMemoryBlock;
    T* pTypePtr=(T*) (pNbObjects+1);
    // Store the number of objects.
    *pNbObjects=uiNbObjects;
    // Create the objects.
#ifndef SN_COMPILER
    for (qUnsignedInt ui=0; ui<uiNbObjects; ui++)
    {
        new ((qChar*)(pTypePtr+ui)) T;
    }
#endif
    return pTypePtr;
}

//{secret}
template <class T> void SpecialDeleteArray(T *pPtr)
{
    // Retrieve number of objects.
    qUnsignedInt32 *pNbObjects=((qUnsignedInt32*) pPtr)-1;
    void *pMemoryBlock=pNbObjects;
    // Destroy the objects.
    T* pTypePtr=(T*) pPtr;
    // This pointer is actually not used at all if T is a simple type
    // that has no destructor - the compiler can get rid of the loop
    // that follows.
    DUMMY_PARAM(pTypePtr);
    for (qUnsignedInt ui=0; ui<*pNbObjects; ui++)
    {
        pTypePtr[ui].~T();
    }
    // Free the memory.
    MemoryManager::Free(pMemoryBlock, MemoryManager::qSpecialNewDeleteArray);
}

//{secret}
template <class T> T* SpecialNew(qChar* szFile, qUnsignedInt32 uiLine)
{
    // Allocate the memory.
#ifdef HEAP_DEBUGGING_ENABLED
    void *pObjectLocation=MemoryManager::Allocate(sizeof(T), szFile, uiLine, MemoryManager::qSpecialNewDelete);

#else /* HEAP_DEBUGGING_ENABLED */
    DUMMY_PARAM(szFile);
    DUMMY_PARAM(uiLine);
    void *pObjectLocation=MemoryManager::Allocate(sizeof(T), MemoryManager::qSpecialNewDelete);
#endif /* HEAP_DEBUGGING_ENABLED */

#ifdef SN_COMPILER
    return (T*)pObjectLocation;
#else /* SN_COMPILER */
    return new (pObjectLocation) T;
#endif /* SN_COMPILER */
}

//{secret}
template <class T> void SpecialDelete(T *pPtr)
{
    // Call the object destructor.
    pPtr->~T();
    // Free the memory.
    MemoryManager::Free(pPtr, MemoryManager::qSpecialNewDelete);
}


/* Please see man pages for details 
  
  
 */

// New macros
#ifdef HEAP_DEBUGGING_ENABLED

#define qAlloc(TYPE,NUMBER)             (TYPE*) NAMESPACE_NEX MemoryManager::Allocate(sizeof(TYPE)*(NUMBER), _T(__FILE__), __LINE__, NAMESPACE_NEX MemoryManager::qAllocFree)
#define qSpecialNew(TYPE)               NAMESPACE_NEX SpecialNew<TYPE>(_T(__FILE__), __LINE__)
#define qSpecialNewArray(TYPE,NUMBER)   NAMESPACE_NEX SpecialNewArray<TYPE>(NUMBER, _T(__FILE__), __LINE__)
#define qNew                            new(_T(__FILE__), __LINE__)
#define qNewMayFail(pool)               new(pool, _T(__FILE__), __LINE__)

#else /* HEAP_DEBUGGING_ENABLED */

#define qAlloc(TYPE,NUMBER)             (TYPE*) NAMESPACE_NEX MemoryManager::Allocate(sizeof(TYPE)*(NUMBER), NAMESPACE_NEX MemoryManager::qAllocFree)
#define qSpecialNew(TYPE)               NAMESPACE_NEX SpecialNew<TYPE>(_T(""), 0)
#define qSpecialNewArray(TYPE,NUMBER)   NAMESPACE_NEX SpecialNewArray<TYPE>(NUMBER, _T(""), 0)

/* Please see man pages for details 
 
 */
#define qNew                            new
#define qNewMayFail(pool)               new(pool)

#endif /* HEAP_DEBUGGING_ENABLED */

#define qFree(PTR)                      NAMESPACE_NEX MemoryManager::Free(PTR, NAMESPACE_NEX MemoryManager::qAllocFree)
#define qSpecialDelete(PTR)             NAMESPACE_NEX SpecialDelete(PTR)
#define qSpecialDeleteArray(PTR)        NAMESPACE_NEX SpecialDeleteArray(PTR)

/* Please see man pages for details 
 
 */
#define qDelete                         delete


#ifdef HEAP_DEBUGGING_ENABLED
#define STARTSNAPSHOT(Var) \
    Log *pLog##Var = qNew Log(); \
    LogDeviceFile *pLogDevice##Var = qNew LogDeviceFile(); \
    pLogDevice##Var->SetFile(_T(#Var) _T(".log")); \
    pLog##Var->SetLogDevice(pLogDevice##Var); \
    Platform::Sleep(1); \
    Time tiStart##Var = Time::GetTime(); \
 

#define STOPSNAPSHOT(Var) \
    MemoryManager::GetDefaultMemoryManager()->LogSnapshot(pLog##Var,tiStart##Var,Time::GetTime(),true); \
    qDelete pLog##Var;

#define SNAPSHOT(Var) \
    Time tiEnd##Var = Time::GetTime(); \
    Platform::Sleep(1); \
    Log *pLog##Var = qNew Log(); \
    LogDeviceFile *pLogDevice##Var = qNew LogDeviceFile(); \
    pLogDevice##Var->SetFile(_T(#Var) _T(".log")); \
    pLog##Var->SetLogDevice(pLogDevice##Var); \
    MemoryManager::GetDefaultMemoryManager()->LogSnapshot(pLog##Var,0,tiEnd##Var,true); \
    qDelete pLog##Var;

#define STARTLOG(Var) \
    Log *pLog##Var = qNew Log(); \
    LogDeviceFile *pLogDevice##Var = qNew LogDeviceFile(); \
    pLogDevice##Var->SetFile(_T(#Var) _T(".log")); \
    pLog##Var->SetLogDevice(pLogDevice##Var); \
    MemoryManager::GetDefaultMemoryManager()->SetLog(pLog##Var);


#define STOPLOG(Var) \
    MemoryManager::GetDefaultMemoryManager()->SetLog(NULL); \
    qDelete pLog##Var;
#endif /* HEAP_DEBUGGING_ENABLED */

/*  

*/


END_NAMESPACE_NEX

#else /* __cplusplus */
//  C allocation routines

//  void *malloc( size_t size );
//{secret}
void*  QuazalCRTAlloc( size_t stSize );

//  void *realloc(void *memblock, size_t size );
//{secret}
void* QuazalCRTRealloc( void *pPtr, size_t stSize );

//  void* calloc( size_t stNum, size_t stSize );
//{secret}
void* QuazalCRTCalloc( size_t stNum, size_t stSize );

//  void free( void* pPtr )
//{secret}
void QuazalCRTFree( void* pPtr );

#endif // #ifdef __cplusplus

#endif /* _MemoryManager_H_ */
