/*---------------------------------------------------------------------------*
  (C) Copyright 1998-2014, Technologies Quazal Inc. All rights reserved.
  Quazal Net-Z, Quazal Rendez-Vous and Quazal are trademarks of Quazal.

  Copyright (C)2011-2014 Nintendo  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo and/or its licensors and are
  protected by national and international copyright laws.
  They may not be disclosed to third parties or copied or duplicated
  in any form, in whole or in part, without the prior written consent
  of Nintendo.
  The content herein is highly confidential and should be handled
  accordingly.
 *---------------------------------------------------------------------------*/

#ifndef _ProfilingUnit_H_
#define _ProfilingUnit_H_

// Group=Platform Subsystem

#include <OnlineCore/src/Platform/Core/PlatformDecl.h>
#include <OnlineCore/src/Platform/Core/HighResolutionClock.h>
#include <OnlineCore/src/Platform/Core/HighResolutionChrono.h>
#include <OnlineCore/src/Platform/Core/Time.h>
#include <OnlineCore/src/Platform/Core/qChain.h>
#include <OnlineCore/src/Platform/Core/CriticalSection.h>
#include <OnlineCore/src/Platform/Core/qList.h>
#include <OnlineCore/src/Platform/Core/String.h>


BEGIN_NAMESPACE_NEX

//Class used to keep track of statistics (total time, average time, max time, min time) on the time by an event which is repeated multiple times
//Useful to gather stats on the efficiently of a function calls
/* Please see man pages for details 
    

    
    

    

        

        

        

    

    
    
 */
class ProfilingUnit: public RootObject
{
public:

    CHAINABLE(ProfilingUnit);

    /* Please see man pages for details 
    

    
     */
    class Scope: public RootObject
    {
        // Group=
    public:
        Scope(ProfilingUnit &pPU)
        {
            m_pPU=&pPU;
            m_bRecorded = false;
        }

        virtual ~Scope()
        {
            if (!m_bRecorded)
            {
                Record();
            }
        }

        qBool Record()
        {
            if (m_bRecorded)
            {
                return false;   // Record only once.
            }
            m_pPU->IncTime(m_oChronoTime.GetElapsed(true));
            m_bRecorded = true;
            return true;
        }

        void Reset()
        {
            m_bRecorded = false;
            m_oChronoTime.GetElapsed(true);
        }

    private:
        ProfilingUnit   *m_pPU;
        HighResolutionChrono   m_oChronoTime;
        qBool m_bRecorded;
    };

    // Group=
public:
    ProfilingUnit(const qChar *szName=_T("<unknown>"), qUnsignedInt32 uiLoggingFrequency = 0, const qChar* description = _T(""));
    virtual ~ProfilingUnit();

    //n1769
    /* Please see man pages for details 
    

    

    
    */
    void           IncAmount(qUnsignedInt32 amount)
    {
        m_ui32Amount += amount;
    }

    //n1769
    /* Please see man pages for details 
    

    

    
    */
    qUnsignedInt32 GetAmount(void) const
    {
        return m_ui32Amount;
    }

    //n1769
    /* Please see man pages for details 
    

    
    */
    void   SetOrder(qInt32 order)
    {
        m_i32Order = order;
    }

    //n1769
    /* Please see man pages for details 
    

    
    */
    qInt32 GetOrder(void) const
    {
        return m_i32Order;
    }

    //n1769
    /* Please see man pages for details 
    
    */
    static void Enable(void);

    /* Please see man pages for details 
    
    */
    static void Disable(void);

    /* Please see man pages for details 
    

    
    */
    static bool IsEnabled(void);

    //Adds another function call which took oTimeIncrement to the data
    void IncTime(HighResolutionTime oTimeIncrement);
    void IncTime(HighResolutionTime oTimeIncrement, const qChar* pName);
    //Gets the total time (sum) of all the event times that were declared in t he ProfilingUnit object
    HighResolutionTime GetTotalTime() const;  //
    //Gets the time taken by the quickest event
    HighResolutionTime GetMinTime() const;    //
    //Gets the time taken by the slowest event
    HighResolutionTime GetMaxTime() const;    //
    //Gets the time taken by the last event
    HighResolutionTime GetLastTime() const;   //
    //Gets the average time for all events recorded
    qDouble GetAvgTime() const;               //
    //Gets the variance time for all events recorded
    qDouble GetVarTime() const;               //
    //Gets the number of events recorded
    qUnsignedInt32 GetNbOccurences() const;   //

    void EnableLogging(qBool bAutoLogEnabled);
    void SetLoggingFrequency(qUnsignedInt32 uiLoggingFrequency);

    //Logs information about the ProfilingUnit object (its name, statistics, time elapsed since the object was last reset)
    void Log();

    //Gets the name of the ProfilingUnit object
    const qChar* GetName() const;              //

    //Gets the ID of the ProfilingUnit object
    qUnsignedInt32 GetID() const;              //

    //Sets the name of the ProfilingUnit object to be szName
    void SetName(const qChar* szName);         //

    void SetDescription(const qChar* /*description*/) {};
    const qChar* GetDescription() const
    {
        return GetName();
    }

    typedef qChain<ProfilingUnit*>  List;

    //Returns the list of all ProfilingUnit objects
    static List& GetList();

    //Returns class critical section
    static NAMESPACE_NEX CriticalSection& CS();

    //Resets all the statistics for the ProfilingUnit object and sets the recorded time for the reset to be tNow
    void Reset(Time tNow = Time::GetTime());
    //Calls the Reset(...) function on all the ProfilingUnit objects in the list
    static void ResetAll();

    // {secret}
    //Returns a list containing the ID of all the ProfilingUnits in the list
    static qList<qUnsignedInt32> GetAllIDs();

    // {secret}
    // Finds a ProfilingUnit with ID uiID in the list and return it.  This is not meant to be efficient, and will iterate across the entire
    //list of ProfilingUnits.
    static ProfilingUnit* FindByID(const qUnsignedInt32 uiID);

    /* Please see man pages for details 
        
        
        
        
    */
    static const qList<ProfilingUnit*> FindByName(const String& name);

    // {secret}
    //  Print logging of all PUs whose name is in given list
    static void Report(const qList<String>& oListPUs);

#ifdef TEST_ENABLED
    //Executes the unit tests for the ProfilingUnit class
    static qBool Test();
#endif


    // {secret}
    // n2370
    const qChar* GetMaxTimeName() const
    {
        return m_strMaxTimeName.CStr();
    }
    const qChar* GetMinTimeName() const
    {
        return m_strMinTimeName.CStr();
    }

private:
    //Stores the sum of squares
    HighResolutionTime  m_oSumOfSquaredTime;
    //Stores the total time (sum) of all the event times that were declared in t he ProfilingUnit object
    HighResolutionTime  m_oTotalTime;
    //Stores the time taken by the quickest event
    HighResolutionTime  m_oMinTime;
    //Stores the time taken by the slowest event
    HighResolutionTime  m_oMaxTime;
    //Stores the time taken by the last event
    HighResolutionTime  m_oLastTime;
    Time                m_tiStartTime;
    //Stores the number of occurrences
    qUnsignedInt32      m_uiNbOccurence;
    qUnsignedInt32      m_uiLoggingFrequency;
    String              m_strName;
    qBool               m_bAutoLogEnabled;
    qInt16              m_i16LastTracePeriod;

    //Unique ID for this ProfilingUnit object
    qUnsignedInt32 m_uiID;

    //n1769
    // IncAmount() increments this variable.
    qUnsignedInt32      m_ui32Amount;

    //n1769
    // Used for ordering of ProfilingUnit instance.
    qInt32              m_i32Order;

    // n2370 (Mark the ones that take the most time and those that do not take much time)
    String              m_strMaxTimeName;
    String              m_strMinTimeName;

    //First valid ID to assign to the next ProfilingUnit object that will be created
    static qUnsignedInt32 s_uiNextFreeID;
    //List of all ProfilingUnit objects
    static List                 s_lstProfilingUnits;
    static NAMESPACE_NEX CriticalSection      s_cs;

    //n1769
    static qBool s_bEnableProfiling;
};


// For use via SWIG, since inner classes are not supported.

class ProfilingScope: public ProfilingUnit::Scope
{
    // Group=
public:
    ProfilingScope(ProfilingUnit &pPU): ProfilingUnit::Scope(pPU) {}
    virtual ~ProfilingScope() {}

    qBool Record()
    {
        return ProfilingUnit::Scope::Record();
    }
    void Reset()
    {
        ProfilingUnit::Scope::Reset();
    }
};

END_NAMESPACE_NEX

#endif
