﻿/*---------------------------------------------------------------------------*
  (C) Copyright 1998-2014, Technologies Quazal Inc. All rights reserved.
  Quazal Net-Z, Quazal Rendez-Vous and Quazal are trademarks of Quazal.

  Copyright (C) 2011-2014 Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo and/or its licensors and are
  protected by national and international copyright laws.
  They may not be disclosed to third parties or copied or duplicated
  in any form, in whole or in part, without the prior written consent
  of Nintendo.
  The content herein is highly confidential and should be handled
  accordingly.
 *---------------------------------------------------------------------------*/

#ifndef _Result_H_
#define _Result_H_

#include <OnlineCore/src/Platform/Core/PlatformDeclInternal.h>
#include <OnlineCore/src/Platform/Core/StringStream.h>

///////////////////////////////////////////////////////////////////////////////
// IMPORTANT NOTE: A filter translates certain portions of this file into a Python
// script. This filter picks up only C-style comments (comments delimited by
// an opening slash/star and a closing star/slash) where the opening and closing
// delimiters must be alone on their line as seen below. The Python filter ignores
// C++ style comments (as seen on this line).
// Therefore, please note that the comments in this file have been carefully inserted
// so that the translation works as expected.
// See the results in ...src\Services\Constants\qResultErrors.py
///////////////////////////////////////////////////////////////////////////////

BEGIN_NAMESPACE_NEX

class String;

// group=Error Codes

// Description: A 32 bit integer that describes the return result.
// Remarks: The breakdown of the significance of each bit in the integer is as follows:
//
//  @ReturnResultTable
//  Bit number   Value
//  ----------   -----
//  0-15         The return code, which may be either a SuccessCode or an ErrorCode and is specific to a given facility
//  16-24        The facility (refer to qResultFacility for possible values)
//  31           0 if success, 1 if error
//
#define QRESULT_ERROR_BIT           0x80000000

//group=Error Code Enumerations

// Description: Enumerates the facilities from which a return result may originate.
/* Please see man pages for details 
 
 */
#ifdef SWIG
class qResultFacility
{
public:
#else
namespace qResultFacility
{
#endif

    /*
    Description: Enumerates the facilities from which a return result may originate.
    Remarks: A facility is the specific area of the system from which a return result may originate.
    Each facility defines its own success codes and error codes specific for that facility. Therefore,
    an error code of 1, for example, will mean one thing in the Core facility but may mean something
    completely different in the DDL or any other facility.

    You cannot modify the built-in Facilities. However, you may add an almost unlimited number of user-defined
    Facilities.
    */
    /* Please see man pages for details 
     

     
     
     
     

     
     */
    enum Facilities
    {
        /*
        The origin of the return result is the $PRODUCTNAME$ core.
        The error and success codes of this facility are defined in CoreErrorCode and
        CoreSuccessCode respectively.
        */
        Core              = 1,    //
        /*
        The origin of the return result is the DDL compiler.
        The error codes of this facility are defined in DDLErrorCode. Currently
        no success codes are defined.
        */
        DDL               = 2,    //
        // $ifdef RENDEZVOUS$
        /*
        The origin of the return result is Rendez-Vous specific code.
        The error codes of this facility are defined in RendezVousErrorCode. Currently
        no success codes are defined.
        */
        RendezVous        = 3,    //
        /*
        The origin of the return result is the Python scripts.
        The error codes of this facility are defined in PythonCoreErrorCode. Currently
        no success codes are defined.
        */
        PythonCore        = 4,    //
        //$endif$

        /*
        The origin of the return result is the network transport.
        The error codes of this facility are defined in TransportErrorCode. Currently
        no success codes are defined.
        */
        Transport         = 5,    //
        //$ifdef NETZ$
        /*
        The origin of the return result is the duplicated object core.
        The error and success codes of this facility are defined in DOCoreErrorCode and
        DOCoreSuccessCode respectively.
        */
        DOCore            = 6,    //
        //$endif$

        //$ifdef RENDEZVOUS$
        /*
        The origin of the return result is Spark! specific code. The error codes
        of this facility are defined in SparkErrorCode which can be fond in the Spark! Reference Manual.
        Currently the success codes are defined in CoreSuccessCode.
        */
        Spark             = 8,    //
        //$endif$

        //n1769
        TransportSocket   = 9,    //


        ZZZAuthentication = 10,    //
        Tracking = 11,    //
        Telemetry = 12,    //
        TitleStorage = 13,    //
        Privilege = 14,    //
        OnlineConfig = 15,    //
        GameSession = 16,    //
        Localization = 17,    //
        News = 18,    //
        PlayerStats = 19,    //
        UserStorage = 20,    //
        SkillRating = 21,    //
        ZZZFriends = 22,    //
        SubAccountManagement = 23,    //
        Friends = 24,    //


        LAST_QUAZAL_INTERNAL = 50,    //

        /*
        The end of Quazal defined error.
        */
        FirstNintendoFacility = 101,    //

        /*
        For Nintendo Friend and Presence
        The error codes of this facility are defined in FPDErrorCode.
        */
        FPD = 101,    //

        /*
        For Nintendo Matchmaking
        */
        Matchmaking = 102,    //

        /*
        For Nintendo Ranking APIs
        */
        Ranking = 103,    //

        /*
        For Nintendo Authentication Error (NAS and Nintendo specific RV login error)
        defined in AuthenticationErrorCode(AuthenticationErrors.h)
         */
        Authentication = 104,    //

        /*
        For Nintendo DataStore APIs defined in DataStoreErrorCode.
        */
        DataStore = 105,    //

        //{secret}
        Subscription = 106,

        //{secret}
        Rating = 107,

        /*
        For Nintendo ServiceItem APIs defined in ServiceItemErrorCode.
        */
        ServiceItem = 108,    //

        //{secret}
        Shop = 109,

        //{secret}
        MusicRequest = 110,

        /*
        For Nintendo MatchmakeReferee APIs defined in MatchmakeRefereeErrorCode.
        */
        MatchmakeReferee = 111, //

        /*
        For Nintendo Subscriber APIs defined in SubscriberErrorCode.
        */
        Subscriber = 112, //

        //{secret}
        LAST_INTERNAL     = 112,    //

        /*
        The ID of the first user-defined facility, which by default is equal to 200.
        */
        FirstUserFacility = 200    //
    };
#ifdef SWIG
};
#else
}
#endif

//TODO: This macro is misused (only once) and misnamed and does the same as QERROR
// {secret}
// Internal macro used to change the facility of a qResult.
#define QRESULT_SET_FACILITY(code,facilityID) (code & 0x8000FFFF) | (facilityID << 16)

//group=Error Code Enumerations

// Description: This enumeration defines the success codes that may be returned for the Core facility.
/* Please see man pages for details 
 
 */
#ifdef SWIG
class CoreSuccessCode
{
public:
#else
namespace CoreSuccessCode
{
#endif

    /*
    Description: Enumerates the success codes for the Core facility that an operation may return.
    Remarks: Details as to the cause of each SuccessCode are detailed in the documentation
    of each individual $PRODUCTNAME$ method that may return a result.
    */
    /* Please see man pages for details 
     
     */
    enum SuccessCode
    {
        /*
        The operation completed successfully.
        */
        Success                     = 1,    //
        /*
        Success of the operation is pending; that is, the operation is still in progress and has not
        yet returned one of the other SuccessCode s nor a CoreErrorCode.
        */
        SuccessPending              = 2    //
    };
#ifdef SWIG
};
#else
}
#endif

//group=Error Code Enumerations

// Description: This enumeration defines the error codes that may be returned for the Core facility.
/* Please see man pages for details 
 
 */
#ifdef SWIG
class CoreErrorCode
{
public:
#else
namespace CoreErrorCode
{
#endif

    /*
    Description: Enumerates the error codes for the Core facility that a call may return.
    Remarks: Details as to the cause of each ErrorCode are detailed in the documentation
    of each individual $PRODUCTNAME$ method that may return a result.
    */
    /* Please see man pages for details 
     
     */
    enum ErrorCode
    {
        /*
        The reason for the error is unknown.
        */
        Unknown                     = 1,    //
        /*
        The operation is currently not implemented.
        */
        NotImplemented              = 2,    //
        /*
        The operation specifies or accesses an invalid pointer.
        */
        InvalidPointer              = 3,    //
        /*
        The operation was aborted.
        */
        OperationAborted            = 4,    //
        /*
        The operation threw an exception. Note that Rendez-Vous does not log traces for exceptions
        that are registered with the python script engine (qResult and data base exceptions).
        On the other hand, if an unexpected exception occurs (an assert, for example), Rendez-Vous
        dumps the call stack before returning the exception.
        */
        Exception                   = 5,    //
        /*
        An attempt was made to access data in an incorrect manner. This may be due to
        inadequate permission or the data, file, etc., not existing.
        */
        AccessDenied                = 6,    //
        /*
        The operation specifies or accesses an invalid DOHandle.
        */
        InvalidHandle               = 7,    //
        /*
        The operation specifies or accesses an invalid index.
        */
        InvalidIndex                = 8,    //
        /*
        The system could not allocate or access enough memory or disk space to perform the specified operation.
        */
        OutOfMemory                 = 9,    //
        /*
        Invalid argument were passed with the operation. The argument(s) may be out of range or invalid.
        */
        InvalidArgument             = 10,    //
        /*
        The operation did not complete within the specified timeout for that operation.
        */
        Timeout                     = 11,    //
        /*
        Initialization of the component failed.
        */
        InitializationFailure       = 12,    //
        /*
        The call failed to initialize.
        */
        CallInitiationFailure       = 13,    //
        /*
        An error occurred during registration.
        */
        RegistrationError           = 14,    //
        /*
        The buffer is too large to be sent.
        */
        BufferOverflow              = 15,    //
        /*
        A lock is needed for a particular process but it is already locked (the lock
        should be released beforehand) or not already locked (we cannot take the
        lock now because it might cause deadlocks).
        */
        InvalidLockState            = 16,    //
        /* Please see man pages for details */
        InvalidSequence             = 17     //
    };

    //note: When adding new results, it is necessary to change `NEX_QRESULT_TO_STR` in `OnlineCore/src/Platform/Core/Result.cpp`.
    //See operations done when adding error codes to the wiki.
    //If the errors the application should handle have increased, add them to the list of errors the application should handle in the`overall.rst` of the `common manual.`
#ifdef SWIG
};
#else
}
#endif

//group=Error Code Macros

/* Please see man pages for details 
 
 
 */
//Description: Tests if the operation result code is a success.
//Remarks: This macro is typically used in an "if" statement.

/* Please see man pages for details 
 

 
 */
#define QSUCCEEDED(RETURNCODE)  ((RETURNCODE & QRESULT_ERROR_BIT) == 0)
//Description: Tests if the operation result code is an error.
//Remarks: This macro is typically used in an "if" statement.
/* Please see man pages for details 
 

 
 */
#define QFAILED(RETURNCODE)     ((RETURNCODE & QRESULT_ERROR_BIT) != 0)
//Description: Builds a specific error code.
//Remarks: This macro is typically used in the "case" statement of a "switch" statement.
/* Please see man pages for details 
 

 
 
 */
#define QERROR(FACILITY,RETURNCODE) (NAMESPACE_NEX qInt32)(QRESULT_ERROR_BIT | NAMESPACE_NEX qResultFacility::FACILITY<<16 | NAMESPACE_NEX FACILITY##ErrorCode::RETURNCODE)
//Description: Builds a specific success code.
//Remarks: This macro is typically used in the "case" statement of a "switch" statement.
/* Please see man pages for details 
 

 
 
 */
#define QSUCCESS(FACILITY,RETURNCODE) (NAMESPACE_NEX qInt32)(NAMESPACE_NEX qResultFacility::FACILITY<<16 | NAMESPACE_NEX FACILITY##SuccessCode::RETURNCODE)


#ifdef DEBUG_VERSION
//Description: Accesses an error return result as a qResult.
// Parameters:
// FACILITY: The qResultFacility from which the error originates.
// RETURNCODE: The error Return Codes.
//Remarks: The filename and line number where the error occurred are only returned when this macro is used in
//Debug mode.
#define QRESULT_ERROR(FACILITY,RETURNCODE) NAMESPACE_NEX qResult(QERROR(FACILITY,RETURNCODE), _T(__FILE__), __LINE__)
//Description: Accesses a success return result as a qResult.
// Parameters:
// FACILITY: The qResultFacility from which the success result originates.
// RETURNCODE: The success Return Codes.
//Remarks: The filename and line number where the error occurred are only returned when this macro is used in
//Debug mode.
#define QRESULT_SUCCESS(FACILITY,RETURNCODE) NAMESPACE_NEX qResult(QSUCCESS(FACILITY,RETURNCODE), _T(__FILE__), __LINE__)
#else
/* Please see man pages for details 
 

 
 
 */
#define QRESULT_ERROR(FACILITY,RETURNCODE) NAMESPACE_NEX qResult(QERROR(FACILITY,RETURNCODE))
/* Please see man pages for details 
 

 
 
 */
#define QRESULT_SUCCESS(FACILITY,RETURNCODE) NAMESPACE_NEX qResult(QSUCCESS(FACILITY,RETURNCODE))
#endif // DEBUG_VERSION

/*  

*/

//group=Return Results Classes

// Description: This class defines the result that a call may return.
//
//   Remarks: The results that are returned by this class are defined in the ErrorCode and SuccessCode
//   classes defined for each specific facility as detailed in the table below. Refer to these classes
//   for details as to the individual error and success codes that each facility may return. A return code
//   is a 32 bit integer defined by QRESULT_ERROR_BIT. Note that the return result is accessed via one of
//   the QRESULT macros.
//
//  $ifdef RENDEZVOUS$
//    @table
//    qResultFacility       ErrorCode               SuccessCode
//    /Core/                CoreErrorCode           CoreSuccessCode
//    /DDL/                 DDLErrorCode            ---
//    /PythonCore/          PythonCoreErrorCode     ---
//    /RendezVous/          RendezVousErrorCode     ---
//    /Transport/           TransportErrorCode      ---
//
//  $else$
//    @table
//    qResultFacility       ErrorCode               SuccessCode
//    /Core/                CoreErrorCode           CoreSuccessCode
//    /DDL/                 DDLErrorCode            ---
//    /DOCore/              DOCoreErrorCode         DOCoreSuccessCode
//    /Transport/           TransportErrorCode      ---
//  $endif$
//
// Note that as the qResult class uses some explicit constructors the following call is invalid:
//
//              // This will cause an error
//              qResult result1 = true;
//
// Instead, you may call either of the  following:
//
//              // One possibility
//              qResult result1(true);
//              // Another way to do the same thing
//              qResult result1;
//              result1 = true;
//
//  In addition, as the qResult class does not support implicit conversion, the macros that return a qResult
// always need to be prefixed with QRESULT. For example:
//
//             // This call will produce an error
//             return QSUCCESS(Core,Success);
//             // This is the correct way to call the macro
//             return QRESULT_SUCCESS(Core,Success);
//
//
// Example:
//
//          qResult AccountManagementHelper::CreatePlayer(const String &strPlayerName, const String &strPassword,
//                                                        const String &strEmail) {
//                  AccountManagementClient oAccountManagementClient;
//                  oAccountManagementClient.Bind(pGuest);
//                  ProtocolCallContext ccCreateAccount;
//
//                  if (!oAccountManagementClient.CreateAccount(&ccCreateAccount, &rResult, strPlayerName,
//                                                              strPassword, BuiltInGroups::Players, strEmail)) {
//                      return QRESULT_ERROR(Core, CallInitiationFailure);
//                  }
//                  if (!ccCreateAccount.Wait(5000)) {
//                      return QRESULT_ERROR(Core, Timeout);
//                  }
//                  return rResult;
//          }
//
/* Please see man pages for details 









*/
class qResult : public RootObject
{
public:
    //group= Methods

    /* Please see man pages for details 
     
     
     */
    qResult();
    /* Please see man pages for details 
     
     
     
     */
    explicit qResult(const qInt32& iResult);
    /* Please see man pages for details 
     
     
     
     */
    explicit qResult(const qBool& bResult);

    qResult(const qInt32& iResult, const qChar *szFilename, const qInt32& iLineNumber);

    /* Please see man pages for details 
     
     
     */
    qResultFacility::Facilities GetFacility() const;

    // {secret}
    qBool Equals(const qInt32& iResult) const;
    // {secret}
    qBool Equals(const qBool& bResult) const;

    /* Please see man pages for details 
     
     
     
     */
    operator qBool () const;

    /* Please see man pages for details 
       
       
     */
    qResult& operator=(const qBool& bEq);
    /* Please see man pages for details 
       
       
     */
    qResult& operator=(const qInt32& uiEq);
    /* Please see man pages for details 
       
       
     */
    qResult& operator=(const qUnsignedInt32& uiEq);
    /* Please see man pages for details 
       
       
     */
    qResult& operator=(const qResult& oEq);

    /* Please see man pages for details 
     

     
     
     
     */
    bool operator== (const qResult& rhs) const
    {
        return (m_iReturnCode == rhs.m_iReturnCode);
    }

    //Description: Returns the result return code as a qInt32.
    //Returns: The result return code as a qInt32.
    //Remarks: Because any method can return a variety of error or success codes, test
    // test the qResult of a method against success or error (that is, the last bit of the
    // QRESULT_ERROR_BIT) rather than specific Return Codes. This ensures that if the return code of the method
    // is changed, your code will still function correctly.
    /* Please see man pages for details 
     

     

     
     */
    qInt32 GetReturnCode() const
    {
        return m_iReturnCode;
    }

    //n1769
    const qChar* GetFileName() const
    {
        return m_cszFilename;
    }

    //n1769
    qInt32 GetLineNumber() const
    {
        return m_iLineNumber;
    }

    //1769
    /* Please see man pages for details 
     

     


     
     */
    const qChar* GetErrorName() const;

    //n1769: Function to determine the severity of qResult.
    // However, these functions basically only cover the qResult values related to NEX P2P.
    // 
    qBool IsError(void) const;
    qBool IsWarning(void) const;
    qBool IsReinitializeRequired(void) const;
    qBool IsResetRequired(void) const;
    qBool IsProgrammingError(void) const;

    /* Description: Traces the content of this Result object.
       Remarks: This method traces the result return code, file origin, and line number. Note that Rendez-Vous
       does not log traces for exceptions that are registered with the python script engine
       (qResult and data base exceptions). If an unexpected exception occurs
       (an assert, for example), however, Rendez-Vous dumps the call stack before returning the exception.
    */
    /* Please see man pages for details 
        

            
            

        
    */
#ifdef TRACE_ENABLED    // n2370
    void Trace(qUnsignedInt64 uiTraceFlags=TRACE_ALWAYS) const;
#else
    void Trace(qUnsignedInt64 uiTraceFlags=TRACE_ALWAYS) const { (void)uiTraceFlags; }
#endif  // TRACE_ENABLED

    // {secret}
    void Format(String* pstrDescription) const;

    friend StringStream& operator<<(StringStream& oStream, const qResult& rhs);

#ifdef TEST_ENABLED
    // {Secret}
    static qBool Test();
#endif
private:
    qInt32      m_iReturnCode;
    const qChar *m_cszFilename;
    qInt32      m_iLineNumber;
};

// String streaming.
// {secret}
extern StringStream& operator<<(StringStream& oStream, const qResult& rhs);

END_NAMESPACE_NEX

//{secret}
template <typename T>
bool operator == (const T& lhs, const NAMESPACE_NEX qResult& rhs)
{
    return rhs.Equals(lhs);
}
//{secret}
template <typename T>
bool operator == (const NAMESPACE_NEX qResult& lhs, const T& rhs)
{
    return lhs.Equals(rhs);
}
//{secret}
template <typename T>
bool operator != (const T& lhs, const NAMESPACE_NEX qResult& rhs)
{
    return !rhs.Equals(lhs);
}
//{secret}
template <typename T>
bool operator != (const NAMESPACE_NEX qResult& lhs, const T& rhs)
{
    return !lhs.Equals(rhs);
}

#endif // _Result_H_

