﻿/*---------------------------------------------------------------------------*
  (C) Copyright 1998-2014, Technologies Quazal Inc. All rights reserved.
  Quazal Net-Z, Quazal Rendez-Vous and Quazal are trademarks of Quazal.

  Copyright (C)2011-2014 Nintendo  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo and/or its licensors and are
  protected by national and international copyright laws.
  They may not be disclosed to third parties or copied or duplicated
  in any form, in whole or in part, without the prior written consent
  of Nintendo.
  The content herein is highly confidential and should be handled
  accordingly.
 *---------------------------------------------------------------------------*/

#ifndef _SystemClock_H_
#define _SystemClock_H_

#include <OnlineCore/src/Platform/Core/PlatformDecl.h>
#include <OnlineCore/src/Platform/Core/CriticalSection.h>
#include <OnlineCore/src/Platform/Core/Time.h>

BEGIN_NAMESPACE_NEX

// Group=Time Management

/*
Description: Sets, resets and determines the time.

Remarks:

$ifdef NETZ$
The time provided by SystemClock is not synchronized across the stations that belong to the session.
For synchronized times, refer to SessionClock.
$endif$

The SystemClock time is the time elapsed, in ms, since the process started.
The time never wraps (it is calculated and stored using 64 bits).

The precision of the Time value is subject to the capabilities of the time provider
implemented on the various platforms. $PRODUCTNAME$ provides
a default time provider, but you can also use your own (see
SystemClock::RegisterTimeProvider)

$PRODUCTNAME$ provides two time providers on windows, each with its own
restrictions, see SystemClock::UsePerformanceCounter and
SystemClock::UseMultimediaTimer for details.

See Also: Time, SessionClock
*/
/* Please see man pages for details 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 */
class SystemClock: public RootObject
{
public:
    /*
     Remarks: Do not instantiate this object. It is created by
     $PRODUCTNAME$ when a station joins or creates a session and is initialized to 0 ms.
    */
    /* Please see man pages for details 
     
     
     
     
     */
    SystemClock();
    /*
      Remarks: Do not destroy this object. $PRODUCTNAME$ destroys it when the local station leaves the session or the session terminates.
    */
    /* Please see man pages for details 
     
     
     
     
     */
    virtual ~SystemClock();

    /*
    Description: Registers the specified TimeProvider to an object of this class.

    Parameters:
    pTimeProvider: A pointer to the TimeProvider to registered with this object.
    By default, this class uses the $PRODUCTNAME$ internal, platform-specific TimeProvider.

    Remarks: A custom TimeProvider (such as a motherboard timing
    chip providing quality high-resolution timing information, or time
    information retrieved from the game physics engine) may be used to provide the time. Before this class can retrieve
    the time from such a custom TimeProvider, you must register it through this method.

    Note that you must call RegisterTimeProvider before creating the $PRODUCTNAME$ object as
    follows:

            // Call these two lines before you create the NetZ object
            MyTimeProvider oTimeProvider;
            SystemClock::RegisterTimeProvider(&oTimeProvider);

    See Also:
    */
    /* Please see man pages for details 
     
     
     
     
     
     
     
     
     
     
     
     
     */
    static void RegisterTimeProvider(TimeProvider *pTimeProvider, qBool needCorrection = false);

    // Description: Returns the time in millisecs.
    /* Please see man pages for details 
     
     
     
     
     
     
     
     
     
     */
    inline static Time GetTime()
    {
        Time now = s_needCorrection ? GetTimeImpl(false) : GetTimeImplCorrectless();
        Time::SetCachedTime(now); //n2343: Update the cache time for the Time class every time you call SystemClock::GetTime
        return now;
    }

#ifdef WIN32_TARGET

    /* Description: Uses QueryPerformanceCounter to retrieve the time (with a high degree of precision).

      Remarks: Not required on most platforms, this method is required on Windows because there existes a timer bug in multicore CPU
      motherboards.  That is, $PRODUCTNAME$ can use QueryPerformanceCounter to access a very high resolution time, but a major drawback
      is that it is subject to bugs when run on some dual core systems.

      See http://msdn.microsoft.com/library/default.asp?url=/library/en-us/directx9_c/Game_Timing_and_Multicore_Processors.asp
       for details.

      You can resolve this problem in one of two ways:

      One option to resolve this issue consists of locking the entire process to one single CPU using
      SetProcessAffinityMask(GetCurrentProcess(), 1);

      Another option is to lock NetZ threads to one CPU.  When doing so, you must ensure that all threads calling into NetZ are
      run from the same CPU. To do this, set the thread affinity using Core::SetDefaultThreadAffinity(1);
       This sets the thread affinity of the calling thread, and of any threads NetZ is using. You may also need to call
       SetThreadAffinityMask as well for any game thread that calls into NetZ.
    */
    static void UsePerformanceCounter();


    /* Description: Resets the function to retrieve the system time to the default Windows timeGetTime(). The resolution is not
       guranteed to be 1ms.

       Remarks: If you set this method, the system clock precision returned by Windows system clock functions can vary beteween
       1 and 5 milliseconds depending on the system.  Setting the windows time precision with this function, however, may result in system
       hangs due to threading synchronization problems. For more detailed information about this problem,
       see http://msdn.microsoft.com/library/default.asp?url=/library/en-us/multimed/htm/_win32_timegettime.asp

       See also http://msdn.microsoft.com/library/default.asp?url=/library/en-us/multimed/htm/_win32_obtaining_and_setting_timer_resolution.asp
       for an example of what to do to increase the timer resolution.  This is not done automatically, as it is intrusive. But it is an option
       you can decide to use.
    */
    static void UseMultimediaTimer();

    // Description: Returns true when the system is using the Performance Counter. The default is false.
    static qBool IsUsingPerformanceCounter();
#endif

    /* Description: Resets the state of the SystemClock.
       Remarks: Sets the state of the SystemClock to uninitialized. You may Reset the SystemClock at anytime.
       It is useful to reset the System Clock before providing it with a new time provider..
       See Also: RegisterTimeProvider
    */
    /* Please see man pages for details 
     
     
     
     
     
     
     
     */
    static void Reset();

private:
    static Time ProtectedGetTime();
    static Time GetTimeImpl(qBool bProtected);
    static Time GetTimeImplCorrectless();

    static void ApplyCorrection(Time tNow, Time tDesired);
    inline static Time GetUncorrectedTime();

public:

#ifdef TEST_ENABLED
    //{secret}
    static qBool Test();
#endif

private:
    static qBool s_bUseHighPerformanceCounter;
    static TimeProvider *s_pTimeProvider;

    static qBool s_bInitialized;
    static LargeTimeInterval s_tiCorrection;
    static NAMESPACE_NEX CriticalSection s_csTimeCorrection;
    static qBool s_needCorrection;

};

END_NAMESPACE_NEX

#endif
