/*---------------------------------------------------------------------------*
  (C) Copyright 1998-2014, Technologies Quazal Inc. All rights reserved.
  Quazal Net-Z, Quazal Rendez-Vous and Quazal are trademarks of Quazal.

  Copyright (C)2011-2014 Nintendo  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo and/or its licensors and are
  protected by national and international copyright laws.
  They may not be disclosed to third parties or copied or duplicated
  in any form, in whole or in part, without the prior written consent
  of Nintendo.
  The content herein is highly confidential and should be handled
  accordingly.
 *---------------------------------------------------------------------------*/

#ifndef _ThreadVariable_H_
#define _ThreadVariable_H_

#include <OnlineCore/src/Platform/Core/PlatformDecl.h>
#include <OnlineCore/src/Platform/Core/ObjectThread.h>
#include <OnlineCore/src/Platform/Core/ScopedCS.h>
#include <OnlineCore/src/Platform/Core/ForcedCriticalSection.h>
#include <OnlineCore/src/Platform/Core/ObjectPointer.h>

#include <OnlineCore/src/Platform/Core/qMap.h>

#include <OnlineCore/src/Platform/Core/qChain.h>

BEGIN_NAMESPACE_NEX

class GlobalVariables;

// The ThreadVariableRoot is a root class from which we derive the
// ThreadVariable template.  It's main purpose it to provide a common
// interface to all ThreadVariable, and a way to chain them all
// together.

class ThreadVariableRoot: public RootObject
{
public:
    ThreadVariableRoot();
    virtual ~ThreadVariableRoot();
    virtual void ResetValues() {};
    virtual void ClearValue() {};
    virtual void ClearValue(qUnsignedInt32) {};

    CHAINABLE(ThreadVariableRoot);
protected:
    static qUnsignedInt32 GetThreadID();
};

// The ThreadVariable is a template that implements a variable (of type
// T) that has a different value per thread.

template<class VariableType> class ThreadVariable: public ThreadVariableRoot
{
public:
    typedef qMap<qUnsignedInt32, VariableType> VariableTypeMap;

    ThreadVariable(const VariableType & oInitialValue):
        m_oInitialValue(oInitialValue),
        m_csValues(LockMasks::Ignored)
    {
        m_mapValues[GetThreadID()]=oInitialValue;
    }

    virtual ~ThreadVariable()
    {
        ResetValues();
    }

    virtual void ResetValues()
    {
        ScopedCS oCS(m_csValues);
        typename qMap<qUnsignedInt32, VariableType>::iterator it;
        while (!m_mapValues.empty())
        {
            it=m_mapValues.begin();
            m_mapValues.erase(it);
        }
    }

    virtual void ClearValue()
    {
        ScopedCS oCS(m_csValues);
        typename qMap<qUnsignedInt32, VariableType>::iterator it;
        it = m_mapValues.find(GetThreadID());
        if (it != m_mapValues.end())
        {
            m_mapValues.erase(it);
        }
    }

    //n1769: Specify the thread ID explicitly, and delete the corresponding element.
    virtual void ClearValue(qUnsignedInt32 threadID)
    {
        ScopedCS oCS(m_csValues);
        typename qMap<qUnsignedInt32, VariableType>::iterator it;
        it = m_mapValues.find(threadID);
        if (it != m_mapValues.end())
        {
            m_mapValues.erase(it);
        }
    }

    //n1769: Gets the number of accumulated elements.
    size_t GetSize(void) const
    {
        ScopedCS oCS(m_csValues);
        return m_mapValues.size();
    }

    ThreadVariable& operator=(const VariableType& oValue)
    {
        ScopedCS oCS(m_csValues);
        m_mapValues[GetThreadID()]=oValue;
        return *this;
    }

    VariableType& GetRef()
    {
        ScopedCS oCS(m_csValues);
        typename qMap<qUnsignedInt32, VariableType>::iterator it = m_mapValues.find(GetThreadID());
        if (it==m_mapValues.end())
        {
            m_mapValues[GetThreadID()]=m_oInitialValue;
            return m_mapValues[GetThreadID()];
        }
        else
        {
            return (*it).second;
        }
    }

    VariableType& GetRef() const
    {
        // Note.  The const version can still adds an entry to the
        // map, and the map was made mutable.  A version where we
        // return a reference to the m_oInitialValue would not be
        // correct: although the value is correct, if the caller
        // keeps a pointer to the variable, then the location of
        // the pointer might change in the future, the first time
        // we access the value for a write access.  Thus the code
        // must stay as it is.
        ScopedCS oCS(m_csValues);
        typename qMap<qUnsignedInt32, VariableType>::iterator it = m_mapValues.find(GetThreadID());
        if (it==m_mapValues.end())
        {
            m_mapValues[GetThreadID()]=m_oInitialValue;
            return m_mapValues[GetThreadID()];
        }
        else
        {
            return (*it).second;
        }
    }

    operator VariableType()
    {
        return GetRef();
    }

    operator const VariableType() const
    {
        return GetRef();
    }

    ForcedCriticalSection& GetLock()
    {
        return m_csValues;
    }

    VariableTypeMap& GetValues()
    {
        SYSTEMCHECK(m_csValues.ValidateOwnership(true));
        return m_mapValues;
    }

private:
    VariableType m_oInitialValue;
    // TODO.  We could make the CriticalSection static to reduce the number of
    // critical sections.  There might be side effects of
    // doing this, however - I'm not sure.
    // Note - mutable since locking the value is correct,
    // even for read operations.
    mutable ForcedCriticalSection m_csValues;
    // Note - the map itself is mutable since a read to the
    // ThreadVariable might have the side effect of adding
    // an entry to the map.  Since we must be able to
    // return a reference to the variable, we have to be
    // able to create a new variable when required, even
    // just for reads.
    mutable VariableTypeMap m_mapValues;
};


template<class VariableType> class ThreadVariablePointer: public ThreadVariable< ObjectPointer<VariableType> >
{
public:
    ThreadVariablePointer(VariableType* pInitialValue=NULL): ThreadVariable< ObjectPointer<VariableType> >(pInitialValue)
    {
    };
    // Inherits all default methods... plus
    VariableType* GetInstance()
    {
        return ThreadVariable< ObjectPointer<VariableType> >::GetRef().GetInstance();
    }
};



template<class VariableType> void ResetThreadVariable(VariableType& oVar)
{
    (void)oVar;
}
template<class VariableType> void ResetThreadVariable(ThreadVariable<VariableType>& oVar)
{
    oVar.ResetValues();
}


class ThreadVariableList: public RootObject
{
public:
    ThreadVariableList();
    virtual ~ThreadVariableList();

    static ThreadVariableList& GetInstanceRef();

    void ResetValues();
    void ClearValue();
    void ClearValue(qUnsignedInt32 threadID);
    void AddVariable(ThreadVariableRoot *pThreadVariable);
    void RemoveVariable(ThreadVariableRoot *pThreadVariable);

private:

    qChain<ThreadVariableRoot*>     m_tvList;
    ForcedCriticalSection m_csList;

    static qBool    s_isValid;
};


#ifdef TEST_ENABLED
qBool ThreadVariableTest();
#endif


END_NAMESPACE_NEX

#endif
