﻿/*---------------------------------------------------------------------------*
  (C) Copyright 1998-2014, Technologies Quazal Inc.  All rights reserved.
  Quazal Net-Z, Quazal Rendez-Vous and Quazal are trademarks of Quazal.

  Copyright (C)2011-2014 Nintendo  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo and/or its licensors and are
  protected by national and international copyright laws.
  They may not be disclosed to third parties or copied or duplicated
  in any form, in whole or in part, without the prior written consent
  of Nintendo.
  The content herein is highly confidential and should be handled
  accordingly.
 *---------------------------------------------------------------------------*/

#ifndef _UnitTesting_H_
#define _UnitTesting_H_

#include <OnlineCore/src/Platform/Core/PlatformDecl.h>
#include <OnlineCore/src/Platform/Core/String.h>
#include <OnlineCore/src/Platform/Core/StringStream.h>
#include <OnlineCore/src/Platform/Core/qVector.h>
#include <OnlineCore/src/Platform/Core/qMap.h>
#include <OnlineCore/src/Platform/Core/MainWrapper.h>
#if defined(CTR_TARGET)
#define NN_SWITCH_ENABLE_HOST_IO 1
#include <nn/hio.h>
#endif

// Group=Platform Subsystem
BEGIN_NAMESPACE_NEX

class UnitTesting: public RootObject
{
public:
    // The method accessible to the user...

    //;<unittest>
    //; The user's unittest process must implement a C++ main function that calls
    //; the following static method of UnitTesting class:
    //;<&verbatim($arg)>

    static qBool PerformTests(qInt argc, qChar** argv,
                              qBool (*fTest)(qInt iTestID, qChar** pstrDescription, qBool bDoIt));

    static qBool PerformTests(qChar* strCmdLine,
                              qBool (*fTest)(qInt iTestID, qChar** pstrDescription, qBool bDoIt));

#if defined(NEX_CLIENT_OPTIMIZATION) && defined(NN_SWITCH_ENABLE_HOST_IO) && NN_SWITCH_ENABLE_HOST_IO
    static qBool LoadArgsFromHio(qInt *argc, qChar ***argv);
#endif

    qBool InternalPerformTests(qInt argc, qChar** argv,
                               qBool (*fTest)(qInt iTestID, qChar** pstrDescription, qBool bDoIt));
    //;</&>
    //; The first two arguments are those passed to the main, and third argument is
    //; a pointer to the main test function.  If the fourth argument is true,
    //; then it is expected that a TestServer is running.
    //;</unittest>

    // And those called by the TEST macros...
    static void PrintTestResult(qBool bTestResult, qChar* strDescription,
                                qChar* strFile, qUnsignedInt32 uiLine, qBool bResult);

    // Returns true when we are expecting timing to vary a lot.
    // This is the case if we start the tests on a machine under
    // load (e.g. compiling something else, for instance), or when
    // we are using expensive debug checks (e.g. Crt checks).
    // TODO - we should have an option to trigger this.  For now,
    // only -s options triggers it (strict mem checks)
    static qBool LooseTiming();

public:
    UnitTesting();
    virtual ~UnitTesting();

    static UnitTesting *Instance();
    static UnitTesting *GetInstance()
    {
        return _Instance;
    }
    static void DeleteInstance();


protected:

    qBool Test(qInt iTestID, qChar** pstrDescription, qBool bDoIt=true);
    void Usage();

protected:
    virtual void PrepareTest() {};
    virtual qBool GetAdditionalTestResult(qChar* strResult)
    {
        DUMMY_PARAM(strResult);
        return true;
    }
    virtual qBool AdditionalTestResultAvailable()
    {
        return false;
    }
    virtual qBool CallTest(qInt iTestID)
    {
        qChar *strDescription;
        return Test(iTestID,&strDescription);
    }
    qBool PerformTest(qInt iTestID);

private:
    qBool (*m_fTest)(qInt iTestID, qChar** pstrDescription, qBool bDoIt);

protected:
    static UnitTesting* _Instance;

private:
    // Behavior modifiers...
    qBool m_bTraceSuccess;
    qBool m_bMightLeak;
    qBool m_bWillLeak;
    qBool m_bMightRaiseException;
    qBool m_bWillRaiseException;
    qBool m_bForceDump;
    qBool m_boShowMemAllocStack;
    qBool m_bLooseTiming;
    qUnsignedInt32 m_uiUserFirstArg;

    qBool m_bTrapExceptions;

    qBool m_bGlobalSuccessConditions;
    qChar m_strUserCommandLineArgs[256];

    qInt m_iTestRepetitionAllowed;

    typedef qMap<String,String> KeyValueMap;
    KeyValueMap m_mapCustomArgs;

    void UnmaskFloatingPointExc();

public:
    static qBool Testing()
    {
        return _Instance!=NULL;
    }

    void SetLooseTiming(qBool bLoose=true);
    static void StrongMemoryChecks();
    static void MightLeak()
    {
        UnitTesting::Instance()->m_bMightLeak=true;
        UnitTesting::Instance()->m_bWillLeak=false;
    }
    static void WillLeak()
    {
        UnitTesting::Instance()->m_bMightLeak=true;
        UnitTesting::Instance()->m_bWillLeak=true;
    }
    static void WillNotLeak()
    {
        UnitTesting::Instance()->m_bMightLeak=false;
        UnitTesting::Instance()->m_bWillLeak=false;
    }

    static void MightRaiseException()
    {
        UnitTesting::Instance()->m_bMightRaiseException=    true;
        UnitTesting::Instance()->m_bWillRaiseException=     false;
    }

    static void WillRaiseException()
    {
        UnitTesting::Instance()->m_bMightRaiseException=    true;
        UnitTesting::Instance()->m_bWillRaiseException=     true;
    }

    static void WillNotRaiseException()
    {
        UnitTesting::Instance()->m_bMightRaiseException=    false;
        UnitTesting::Instance()->m_bWillRaiseException=     false;
    }

    static void ForceDump()
    {
        UnitTesting::Instance()->m_bForceDump=true;
    }
    static void TrapExceptions(qBool bTrapExceptions);
    static qUnsignedInt32 UserFirstArg()
    {
        return UnitTesting::Instance()->m_uiUserFirstArg;
    }
    static String GetCustomArg(const String& strKey,const String& strDefault=_T(""));
    static qInt GetCustomIntArg(const String& strKey,const qInt& iDefault=0);
    static void ShowMemAllocStack()
    {
        UnitTesting::Instance()->m_boShowMemAllocStack=true;
    }

    static void AddCustomArg(const String& var, const String& val)
    {
        //TODO_CTR_PORT: temp bug in RVCT's STL, see <map> line 207:
        // note: temporary is necessary to avoid bug with xlC 5.0
#ifdef CTR_TARGET
        GetInstance()->m_mapCustomArgs.insert(GetInstance()->m_mapCustomArgs.begin(),
                                              KeyValueMap::value_type(var,val));
#else
        GetInstance()->m_mapCustomArgs[var] = val;
#endif
    }

    static void AllowTestRepetition(qInt iTestRepetitionAllowed)
    {
        UnitTesting::Instance()->m_iTestRepetitionAllowed=iTestRepetitionAllowed;
    }

    static void UnitTestFailed(qChar* strDescription=NULL);
};



#define PRINTTESTRESULT(R,S) NAMESPACE_NEX UnitTesting::PrintTestResult(R, S, _T(__FILE__), __LINE__, bResult);

// Macros used in the client's main Test function.

#define TEST_NOTDEFINED _T("test not defined")
#define SetTestDescription(A) if (pstrDescription) {*pstrDescription=A;}

// Macros used in the individual test functions.

//;<individual-testfunction>
//; An individual test function performs a set of tests on a specific
//; component, and returns either true, if all tests succeeded, or false, if one failed.

//; To ease the management and tracing of the current status of the
//; tests results, a set of macros are provided.

//;</individual-testfunction>

//;<individual-testfunction>
//; All individual test functions starts with the macro \verb+BEGIN_TEST+, and
//; ends with the line \verb+return TESTRESULT;+.
//;</individual-testfunction>

#define BEGIN_TESTS qBool bResult=true;
#define TESTRESULT bResult

//;<individual-testfunction>
//;
//; Between the calls to \verb+BEGIN_TEST+ and \verb+return TESTRESULT+, the function
//; will test the component.  Every time aa particular result is expected, the following
//; macros are used:
//;
//; \begin{itemize}
//;

//; \item \verb+TEST(C)+: Tests the condition C.  If it is false,
//; then result of the individual test function will be false.  All false results are
//; always printed, while true results are only printed when the
//; unittesting are ran in verbose mode (see command line arguments).
//;
//; \item \verb+TEST(C,S)+: Similar to TEST, but the second argument is the string used
//; to describe the test, when the result is traced.

#define TEST2(C,S) {qBool bTmp=(C); if (!bTmp) {bResult=false;} PRINTTESTRESULT(bTmp,S);}
#define TEST(C) TEST2(C,_T(#C))

// Used when we really do not want to trace anything - in a loop where
// timing is important, for example.
#define SILENTTEST(C) {qBool bTestResult=(C); if (!bTestResult) {TEST2(bTestResult,_T(#C));}}

// Used for tests that are timing critical.  When LooseTiming is set,
// they will be skipped.
#define TIMINGTEST(C) TEST(UnitTesting::LooseTiming() || (C))

// String 'PASSED' is indented to align with 'FAILED' in the logs.
#define STR_TEST_PASS      _T("   *** Test [PASSED]")
#define STR_TEST_FAILED    _T("*** Test [FAILED]")

namespace
{

#ifdef TURN_MATHWARN_OFF
#pragma mathwarn off
#endif

template<typename LHS,typename RHS>
qBool CompareValues(const qChar* szL, const LHS& lhs, const qChar* szR, const RHS& rhs, qBool bTestEquality, const qChar* szFile, qUnsignedInt32 uiLine)
{
    // We perform a cast here.  This makes it easier to compare a
    // uint32 to a constant, for example.  The compiler passes the
    // constant as a int, which gives a warning when we test for
    // equality.
    if ((lhs==static_cast<const LHS&>(rhs)) == bTestEquality)
    {
        QLOG(EventLog::Info,STR_TEST_PASS << _T(" [") << szL << (bTestEquality? _T(" == "): _T(" != ")) << szR << _T("] ") << _T("[") << szFile << _T(":") << uiLine << _T("] "));
        return true;
    }
    else
    {
        QLOG(EventLog::Warning,STR_TEST_FAILED << _T(" [") << szL << (bTestEquality? _T(" == "): _T(" != ")) << szR << _T("] ") << _T("[") << szFile << _T(":") << uiLine << _T("] "));

        QLOG(EventLog::Warning,_T("     With ") << szL << _T(" = ") << lhs);
        QLOG(EventLog::Warning,_T("      and ") << szR << _T(" = ") << rhs);
        return false;
    }
}

#ifdef TURN_MATHWARN_OFF
#pragma mathwarn on
#endif

}

#define TESTEQUAL(A,B) if (!CompareValues(_T(#A),A,_T(#B),B,true,_T(__FILE__),__LINE__)) {bResult = false;}
#define TESTNOTEQUAL(A,B) if (!CompareValues(_T(#A),A,_T(#B),B,false,_T(__FILE__),__LINE__)) {bResult = false;}

#define _SPINTEST(C,DELAY,ACTION)   {               \
        SpinTest    oST(100,(DELAY)*1000);  \
        oST.LeaveOnTimeout();               \
        SpinOnAction(oST,!(C),ACTION);                  \
        PRINTTESTRESULT(!oST.HasTimedOut(), _T(#C)); \
        if (!oST.HasTimedOut()) { \
            QLOG(EventLog::Info,_T("Spinned ") << dec << (oST.GetElapsed()) << _T("/") << dec << (oST.GetTimeout()) << _T(" on ") << (_T(#C))); \
        } else { \
            bResult = false;        \
            QLOG(EventLog::Info,_T("SpinTest Timeout occured")); \
        } \
    }

#define SPINTEST(C, DELAY) _SPINTEST(C, DELAY, qInt iDummy=0; iDummy++)
#define STSPINTEST(C, DELAY) _SPINTEST(C, DELAY, Scheduler::GlobalSingleThreadDispatch());

#define CRITICAL_TEST2(C,S) {qBool bTmp=(C); if (!bTmp) {bResult=false;} PRINTTESTRESULT(bTmp,S); if (!bTmp) return false;}
#define CRITICAL_TEST(C) CRITICAL_TEST2(C,_T(#C))

#define TESTDIFF(A,B,D) {   \
qReal fDiff=qReal(fabs(qReal(qReal(A)-qReal(B)))); qBool bTestOutcome=fDiff <= D;   \
if (!bTestOutcome) {bResult=false;}                                     \
StringStream oStrTestDiff;                      \
oStrTestDiff << _T("Diff (") << _T(#A) << _T(",") << _T(#B) << _T(") < ") << qReal(D) << _T(" (diff was ") << fDiff << _T(")"); \
PRINTTESTRESULT(bTestOutcome,const_cast<qChar*>(oStrTestDiff.CStr()));                          \
}

#define STANDARD_TESTMAIN(TESTFAMILY, TESTMETHOD)                           \
    qInt TestMain(qInt argc, qChar** argv) {                                    \
        QTRACE(TRACE_ALWAYS,_T("Running ") << (TESTFAMILY));                    \
        return UnitTesting::PerformTests(argc, argv, &TESTMETHOD)?0:-1;     \
    }                                                                       \
    MAIN_WRAPPER(TestMain);

//; \end{itemize}
//;
//;</individual-testfunction>

END_NAMESPACE_NEX

#endif
