﻿/*---------------------------------------------------------------------------*
  (C) Copyright 1998-2014, Technologies Quazal Inc. All rights reserved.
  Quazal Net-Z, Quazal Rendez-Vous and Quazal are trademarks of Quazal.

  Copyright (C)2011-2014 Nintendo  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo and/or its licensors and are
  protected by national and international copyright laws.
  They may not be disclosed to third parties or copied or duplicated
  in any form, in whole or in part, without the prior written consent
  of Nintendo.
  The content herein is highly confidential and should be handled
  accordingly.
 *---------------------------------------------------------------------------*/

#ifndef _Network_H_
#define _Network_H_

#include <OnlineCore/src/Core/CoreDecl.h>
#include <OnlineCore/src/Core/PseudoSingleton.h>
#include <OnlineCore/src/Core/Scheduler.h>
#include <OnlineCore/src/Platform/Core/qVector.h>
#include <OnlineCore/src/Platform/Core/qProtectedList.h>
#include <OnlineCore/src/Platform/Core/RefCountedObject.h>
#include <OnlineCore/src/Transport/Interface/StationURL.h>
#include <OnlineCore/src/Transport/Interface/ConnectivityManager.h>
#include <OnlineCore/src/Core/Compression/Interface/CompressionAlgorithm.h>
#include <OnlineCore/src/Transport/Interface/TransportAdapter.h>
#include <OnlineCore/src/Core/Encryption/RC4/RC4Encryption.h>
#include <OnlineCore/src/Platform/Stack/Core/InetAddress.h>

#include <OnlineCore/src/Core/LiteBuffer.h>
//n2208
#include <OnlineCore/src/Core/Compression/ZLib/ZLibCompression.h>
//n2967
#include <OnlineCore/src/Transport/PRUDP/PRUDPMessageSelector.h>

BEGIN_NAMESPACE_NEX

//group=Transport Protocol
// $ifdef WII_COMMON$
// group=Net-Z API
// $endif$

// The instances it manages.

class RootTransport;
class NATTraversalEngine;
class qBuffer;

//n2690
class KeyedChecksumAlgorithm;
/*
    Description: This class encapsulates a network object. It is created by the system when you create the $FACADENAME$ object.
    Remarks:  This singleton can be used to compose a list of StationURL s for the local station, to
    register a custom TransportAdapter and to retrieve pointers to either of the former. Use GetInstance to retrieve
    a pointer to the object.
*/
/* Please see man pages for details 
  
 */
class Network: public RefCountedObject
{
    // Group=
public:

    /*
    Description: Creates a network object.
    Remarks: This object should never be directly instantiated. Instead, it is created
    by the system when you create $FACADENAME$.
    */
    Network();

    /*
    Remarks: This object should never be directly destroyed. It is deleted when you delete
    $FACADENAME$.
    */
    virtual ~Network();

    // Group=Methods

    //  Note: Both AcquireInstance and ReleaseInstance are reserved for product facade objects, only.
    //  Calling either one of them directly may result into undefined behavior.

    // {secret}
    static void AcquireInstance();
    // {secret}
    static void ReleaseInstance();

    /*
    Description: Returns a pointer to the Network instance.
    Returns: A pointer to the Network object.
    */
    /* Please see man pages for details 
     
     
     */
    static Network* GetInstance()
    {
        InstanceControl* pInstance = InstanceControl::FindInstance(IC_TYPE_NETWORK, PseudoSingleton::GetCurrentContext());
        if (pInstance != NULL)
        {
            return static_cast<Network*>(pInstance->GetDelegatorInstance());
        }
        else
        {
            //n1589
            //              QTRACE(D_TRACE_CONSTRUCT, _T("Network::GetInstance:pInstance == NULL"));
            return NULL;
        }
    }

    //n1589:todo2:Scheduler::GetNetworkLock
    // TODO - We should remove this and directly use the Scheduler
    // lock - this makes to code cleaner.
    // {secret}
    static CriticalSection& GetLock()
    {
        //n1589
#if NEX_ENABLE_TRANSPORT_THREAD_MODE
        if(Core::IsUsingTransportThreads())
        {
            return Scheduler::GetInstance()->GetTransportLock();
        }
        else
        {
            return Scheduler::GetInstance()->SystemLock();
        }
#else
        return Scheduler::GetInstance()->SystemLock();
#endif
    }

    //n1589
#ifdef LOCK_DEBUG_ENABLED
    static qBool IsLockedByCurrentThread()
    {
#if NEX_ENABLE_TRANSPORT_THREAD_MODE
        if(Core::IsUsingTransportThreads())
        {
            return Scheduler::GetInstance()->IsTransportLockedByCurrentThread();
        }
        else
        {
            return Scheduler::GetInstance()->IsLockedByCurrentThread();
        }
#else
        return Scheduler::GetInstance()->IsLockedByCurrentThread();
#endif
    }
#endif

    // {secret}
    NATTraversalEngine* GetNATTraversalEngine();
    // {secret}
    ConnectivityManager* GetConnectivityManager()
    {
        return &m_oConnectivityManager;
    }


    PRUDPMessageSelector* GetPRUDPMessageSelector()
    {
        return &m_oPRUDPMessageSelector;
    }

    // {secret}
    void SortAndFilterTargetURLs(qList<StationURL> *plstTargetURLs);

public:
    // {secret}
    RootTransport* GetRootTransport()
    {
        return m_pUDPTransport;
    }

    // Description: Populates purlLocal with the list of
    // StationURL s for the local station.
    // Parameters:
    // purlLocal: List to be populated
    void GetLocalURLs(qList<StationURL> * purlLocal) const;

    // {secret}
    // Add URL to list of local stations.
    void AddLocalURL(const StationURL &oURL);
    // {secret}
    // Delete one entry in the list of local stations where the port number and address match.
    qBool RemoveLocalURL(const StationURL &oURL);
    // {secret}
    // If the address matches any of the entries in the list of local stations, update oURL with the first one.
    qBool ReplaceLocalURL(const StationURL &oURL);

    // {secret}
    // If the oTargetURL, port number, and address match any of the entries in the list of local stations, update oURL with the first one.
    qBool ReplaceLocalURL(const StationURL &oTargetUrl, const StationURL &oURL);

    // {secret}
    typedef enum
    {
        NullComp        = 0,    //  Default
        ZLibComp        = 1,
        LastCompAlgo
    } ECompAlgo;

    //  Note: to disable compression, call: Network::UseCompressionAlgorithm(NullComp);

    // {secret}
    static void UseCompressionAlgorithm(ECompAlgo eAlgo);

    // n2208
    static void UseCompressionAlgorithmZLib(qInt windowBits = ZLibCompression::DEFAULT_WINDOW_BITS, qInt memLevel = ZLibCompression::DEFAULT_MEM_LEVEL);

    // {secret}
    static ECompAlgo GetCompressionAlgorithm();

    // {secret}
    static String GetCompressionAlgorithmName();

    // n2619:begin
    // {secret}
    typedef enum
    {
        P2PStateDefault     = 0,
        P2PStateTerminating
    } P2PState;

    // {secret}
    static P2PState GetP2PState()
    {
        return s_eP2PState;
    }

    // {secret}
    static void SetP2PState(P2PState state)
    {
        s_eP2PState = state;
    }

    // {secret}
    const qProtectedList<StationURL>& GetLocalURLs() const;
    // {secret}
    qProtectedList<StationURL>& GetLocalURLs();

    // Description: Registers a custom TransportAdapter to override the system's default one.
    // Parameters:
    // pAdapter: Pointer to the TransportAdapter to register
    // Returns: True if the TransportAdapter was registered succesfully.
    // See Also: TransportAdapter
    static qBool RegisterTransportAdapter(TransportAdapter *pAdapter);

    //n1589
    static qBool UnregisterTransportAdapter();

    // Description: Returns a pointer to the currently registered TransportAdapter.
    // Returns: A pointer to the currently registered TransportAdapter.
    // See Also: TransportAdapter
    static TransportAdapter* GetTransportAdapter()
    {
        //n1589: memo: GetTransportAdapter will not be NULL after the Network instance is created
        //n1589: Used by the UdsNode::PrepareTransportInterface function so commented out
        //          SYSTEMCHECK(s_pTransportAdapter != NULL);
        return s_pTransportAdapter;
    }

    //  Encryption is ON by default
    // {secret}
    static void EnableEncryption();
    // {secret}
    static void DisableEncryption();

    //n1769
    /* Please see man pages for details 
     
     
     
     
     
     
     
     
     
                               
     




*/
    static void SetP2PDataPacketDefaultSignatureKey(const qByte *pKeyData, qUnsignedInt32 uiKeyLen);


    static qByte *GetP2PDataPacketDefaultSignatureKey();
    static qUnsignedInt32 GetP2PDataPacketDefaultSignatureKeyLength()
    {
        return s_P2PSignatureKeyLen;
    }

    /* Please see man pages for details 
     
     
     
     
     
     
     
     
     
     
     
     
     
     
     
     
     
     
     






*/
    void SetP2PDataPacketSessionSignatureKey(const qVector<qByte>& sessionKey);


    /* Please see man pages for details 
     
     
     
     
     
     
     



*/
    void ClearP2PDataPacketSessionSignatureKey();


    /* Please see man pages for details 
     
     
     
     
     
     
     */
    const void* GetP2PDataPacketSessionSignatureKeyPaylod();
    /* Please see man pages for details 
     
     
     
     
     
     
     */
    qUnsignedInt32 GetP2PDataPacketSessionSignatureKeyLength()
    {
        return m_P2PSessionSignatureKey.GetContentSize();
    }

    //n:2690
    /* Please see man pages for details 
     
     
     
     
     
     



*/
    static void SetP2PChecksumAlgorithm(KeyedChecksumAlgorithm* pAlgorithm);

    /* Please see man pages for details 
     
     
     
     
     
     



*/
    static void EnableP2PDataPacketSignature();

    /* Please see man pages for details 
     
     
     
     
     
     



*/
    static void DisableP2PDataPacketSignature();

    /* Please see man pages for details 
     
     
    */
    CriticalSection& GetAckedIDLock()
    {
        return m_ackedIDLock;
    }
    /* Please see man pages for details 
     
     
     */
    qVector<qUnsignedInt32>& GetAckedID()
    {
        return m_ackedID;
    }
    /* Please see man pages for details 
     
     
    */
    CriticalSection& GetDisconnectedIDLock()
    {
        return m_disconnectedIDLock;
    }
    /* Please see man pages for details 
     
     
     */
    qVector< std::pair<void*, ConnectionID> >& GetDisconnectedID()
    {
        return m_disconnectedID;
    }
private:
    void InitializeLocalURLs();

public:

#ifdef TEST_ENABLED
    // {secret}
    static qBool Test();
#endif

private:

    void SetRC4Encryption();
    void ResetEncryption();

    //n1589
public:
    void SetRC4Encryption(Stream::Type stream);
    void ResetEncryption(Stream::Type stream);
    void SetCompression(Stream::Type stream);
    void ResetCompression(Stream::Type stream);

    void SetDefaultChecksumKeyToSessionKey();

private:
    void InitCompression(ECompAlgo eCompAlgo);
    void InitChecksumAlgorithm();
    void InitP2PChecksumAlgorithmKey();

    //n1589
#ifdef NEX_CLIENT_OPTIMIZATION
    TransportAdapter* m_pVirtualTransportAdapter;
#endif

    CriticalSection     m_csNetworkLock;
    PseudoSingleton     m_psInstance;
    qProtectedList<StationURL>  m_lstStationURL;

    //n1589:todo1:rename to m_pTransport
    RootTransport       *m_pUDPTransport;
    ConnectivityManager m_oConnectivityManager;

    NATTraversalEngine      *m_pNATTraversalEngine;

    PRUDPMessageSelector    m_oPRUDPMessageSelector;


    RC4Encryption           m_oRC4Encryption;

    static ECompAlgo        s_eCompAlgo;
    CompressionAlgorithm*   m_pCompressionAlgorithm;

    static TransportAdapter *s_pTransportAdapter;
    static qBool            s_bEncryptionEnabled;

    // n2690 Signature support for data packets
    static qBool            s_EnableRendezVousChecksumAlgorithm;
    KeyedChecksumAlgorithm* m_pRendezVousChecksumAlgorithm;
    static qBool            s_EnableP2PChecksumAlgorithm;
    KeyedChecksumAlgorithm* m_pP2PChecksumAlgorithm;

    //n1869 Signature support for data packets
public:
    static const size_t   MAX_P2P_SIGNATURE_KEY_LEN = 32; //
    static const size_t   MIN_P2P_SIGNATURE_KEY_LEN = 8; //
private:
    static qByte          s_P2PSignatureKeyData[];
    static qUnsignedInt32 s_P2PSignatureKeyLen;

    static qByte s_byDefaultChecksumKey[];
    LiteBuffer<MAX_P2P_SIGNATURE_KEY_LEN> m_P2PSessionSignatureKey;

    // n2208
    static qInt s_ZLibWindowBits;
    static qInt s_ZLibMemLevel;

    // n2619
    static P2PState s_eP2PState;

    // CallContext CallID associated with ack-received reliable packet
    qVector<qUnsignedInt32> m_ackedID;
    // Lock for m_ackedID access
    CriticalSection m_ackedIDLock;

    // Disconnection information (EndPoint* and connection ID)
    qVector< std::pair<void*, ConnectionID> > m_disconnectedID;
    // Lock for m_disconnectedID access
    CriticalSection m_disconnectedIDLock;
};

END_NAMESPACE_NEX

#endif
