﻿/*---------------------------------------------------------------------------*
  (C) Copyright 1998-2014, Technologies Quazal Inc. All rights reserved.
  Quazal Net-Z, Quazal Rendez-Vous and Quazal are trademarks of Quazal.

  Copyright (C) 2011-2014 Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo and/or its licensors and are
  protected by national and international copyright laws.
  They may not be disclosed to third parties or copied or duplicated
  in any form, in whole or in part, without the prior written consent
  of Nintendo.
  The content herein is highly confidential and should be handled
  accordingly.
 *---------------------------------------------------------------------------*/

#ifndef _RootTransport_H_
#define _RootTransport_H_

#include <OnlineCore/src/Transport/Interface/TransportDecl.h>
#include <OnlineCore/src/Transport/Interface/NATProperties.h>
#include <OnlineCore/src/Transport/Interface/Network.h>
#include <OnlineCore/src/Transport/Interface/PacketRecognizer.h>
#include <OnlineCore/src/Transport/Interface/TransportPerfCounters.h>
#include <OnlineCore/src/Transport/Interface/OutputEmulationDevice.h>
#include <OnlineCore/src/Transport/Interface/InputEmulationDevice.h>
#include <OnlineCore/src/Transport/Interface/TransportStreamManager.h>
#include <OnlineCore/src/Transport/Interface/UPnPProperties.h>
#include <OnlineCore/src/Platform/Core/ForcedCriticalSection.h>
#include <OnlineCore/src/Core/SystemPorts.h>

BEGIN_NAMESPACE_NEX

class Stream;
class Buffer;
class PacketOut;
class Packet;
class VirtualNATDevice;
#if NEX_ENABLE_VIRTUAL_FILTER_DEVICE
class VirtualFilterDevice;
#endif

#ifdef ROUTING_SUPPORT
class Router;
#endif
class Relay;

#if defined(WINSOCK2_STACK_SUPPORT)
#define SOCKET_HANDLE_SUPPORT
#endif

//group=Transport Protocol

/*
Description: This singleton is the base class of the $PRODUCTNAME$ transport layer.
$ifdef NETZ$
Remarks: Sets various transport parameters such as the connection timeout and well-known
port number. Can also set and retrieve various network parameters for third party
networking engines, for example.
$else$
Remarks: Sets various Network and NAT parameters. Can also set and retrieve various
network parameters for third party networking engines, for example.
See Also: NATTraversalClient, NATProperties, ConnectivityManager,
$endif$
*/

/* Please see man pages for details 
 
 
 */
class RootTransport: public RootObject
{
public:
    friend class Network;

    //n1589: Declare common stream communication parameters here
public:

    enum
    {
        DEFAULT_TIMEOUT_NAT_TRAVERSAL = (60 * 1000),
        DEFAULT_TIMEOUT_CONNECT_END_POINT = (15 * 1000),
        DEFAULT_TIMEOUT_DISCONNECT_END_POINT = (15 * 1000),
        /*
        n1589: This terminate timeout is used to forcibly interrupt processing that terminates a duplicated object when some fatal error has prevented that process from continuing.
        This is to perform a forcible interrupt.
        In this case, a fatal error occurs.
        This timeout must be sufficiently longer than the disconnection timeout above.
        */
        DEFAULT_TIMEOUT_TERMINATE_DO_CORE = (30 * 1000),

        //n1769: Moved constants, originally #defined by RootTransport.cpp, here.
        //n1589
#if NEX_ENABLE_MTU_MAX
        DEFAULT_MTU = 1364,
#else
        //n2361: Restored this to 1000 because a value of 1364 caused problems.
        DEFAULT_MTU = 1000,
#endif

#if 0 && defined(TEST_ENABLED)
        DEFAULT_MAX_NUMBER_LISTENING_PORTS = 1,
#else
        DEFAULT_MAX_NUMBER_LISTENING_PORTS = 10,
#endif

        //n1589
        //        DEFAULT_WELL_KNOWN_PORT_NUMBER = 9103,

        //n2781: Use 9103 for an FPD build
#ifdef FPD_BUILD
        DEFAULT_WELL_KNOWN_PORT_NUMBER = 9103,
#else
        DEFAULT_WELL_KNOWN_PORT_NUMBER = 9203,
#endif

        DEFAULT_TIMEOUT_TERMINATE_IMMEDIATELY = 2000,

#ifdef NON_SUPPORT_B_TARGET
        DEFAULT_IO_WAIT_TIME = 50,
#else
        //n1589:todo1:DEFAULT_IO_WAIT_TIME = 10
        DEFAULT_IO_WAIT_TIME = 10,
#endif
#ifdef WIN32_TARGET
        DEFAULT_PACKET_TTL = 128,
#else // WIN32_TARGET
        DEFAULT_PACKET_TTL = 64,
#endif // WIN32_TARGET

#if NEX_ENABLE_HYBRID_DETECT_LIMITATION
        /* Please see man pages for details */
        DEFAULT_ROUTING_HISTORY_UPDATE_SPAN = 0,
        DEFAULT_ROUTING_HISTORY_TTL = 10000,
        DEFAULT_ROUTING_LIMIT = 0,
#endif

        DEFAULT_RECEIVE_BUFFER_SIZE = 8192,
        DEFAULT_SEND_BUFFER_SIZE    = 8192,

        DEFAULT_IDLE_SLEEP_TIME = 0,

        /* Please see man pages for details */
        DEFAULT_TRANSPORT_THREAD_SLEEP_SPAN = 0,
        /* Please see man pages for details */
        DEFAULT_TRANSPORT_THREAD_PRIORITY = 24,

        /* Please see man pages for details */
        DEFAULT_TRANSPORT_BUFFER_POOL_NUM = 64,
        /* Please see man pages for details */
        DEFAULT_TRANSPORT_BUFFER_THREAD_SLEEP_SPAN = 2,
        /* Please see man pages for details */
        DEFAULT_TRANSPORT_BUFFER_THREAD_PRIORITY = 10,
        /* Please see man pages for details */
        DEFAULT_TRANSPORT_BUFFER_RECEIVE_COUNT = 32,
        /* Please see man pages for details */
        DEFAULT_TRANSPORT_BUFFER_SEND_COUNT = 32,
        /* Please see man pages for details */
        DEFAULT_TRANSPORT_BUFFER_CORE_NO = QINT32_MIN,
        /* Please see man pages for details */
        DEFAULT_NAME_RESOLVE_THREAD_PRIORITY = 10,
        /* Please see man pages for details */
        DEFAULT_NAME_RESOLVE_THREAD_CORE_NO = QINT32_MIN,
        /* Please see man pages for details */
        DEFAULT_PACKET_HISTORY_BUNDLING_HOLDING_TIME = 182,
        /* Please see man pages for details */
        DEFAULT_PACKET_HISTORY_BUNDLING_MAX_HOLDING_PACKETS = 32,
        /* Please see man pages for details */
        DEFAULT_PACKET_HISTORY_BUNDLING_MAX_RETRANSMIT_PACKTES = QINT16_MAX,
        /* Please see man pages for details */
        DEFAULT_PACKET_HISTORY_BUNDLING_MAX_RETRANSMIT_COUNTS = 2,
        /* Please see man pages for details */
        DEFAULT_PACKET_HISTORY_BUNDLING_FORCED_RETRANSMIT_TIMEOUT = 150,
        /* Please see man pages for details */
        DEFAULT_PACKET_HISTORY_BUNDLING_MAXIMUM_PACKET_BYTES = RootTransport::DEFAULT_MTU
    };

    /* Please see man pages for details 
     
     
     
     
     */
    enum PacketHistoryCount
    {
        /* Please see man pages for details */
        PACKET_HISTORY_UNBUNDLING_COUNT = 0,
        /* Please see man pages for details */
        PACKET_HISTORY_BUNDLING_COUNT_NONE,
        /* Please see man pages for details */
        PACKET_HISTORY_BUNDLING_COUNT_ONE_AND_MORE,
        /* Please see man pages for details */
        PACKET_HISTORY_BUNDLING_COUNT_A_HALF_AND_MORE,
        /* Please see man pages for details */
        PACKET_HISTORY_BUNDLING_COUNT_MAX,
        /* Please see man pages for details 
         
         */
        PACKET_HISTORY_TOTAL_COUNT,
        /* Please see man pages for details */
        MAX_PACKET_HISTORY_COUNT = PACKET_HISTORY_TOTAL_COUNT
    };

    static const qUnsignedInt32 ADDRESS_CHAR_SIZE=128;

    static const qUnsignedInt16 DEFAULT_RELAY_SERVER_PORT;

    /* Please see man pages for details 
     
     
     */
    static qUnsignedInt32 GetNATTraversalTimeout()
    {
        return s_uiNATTraversalTimeout;
    }

    /* Please see man pages for details 
     
     
     
     
     */
    static void SetNATTraversalTimeout(qUnsignedInt32 o)
    {
        s_uiNATTraversalTimeout = o;
    }

    static qUnsignedInt32 GetConnectEndPointTimeout()
    {
        return s_uiConnectEndPointTimeout;
    }

    static void SetConnectEndPointTimeout(qUnsignedInt32 o)
    {
        s_uiConnectEndPointTimeout = o;
    }

    static qUnsignedInt32 GetDisconnectEndPointTimeout()
    {
        return s_uiDisconnectEndPointTimeout;
    }

    static void SetDisconnectEndPointTimeout(qUnsignedInt32 o)
    {
        if(o >= s_uiTerminateDOCoreTimeout - 5000)
        {
            _QLOG(
                EventLog::Warning,
                _T("DisconnectEndPointTimeout(%d) should be shorter than TerminateDOCoreTimeout(%d) - 5000"),
                o,
                s_uiTerminateDOCoreTimeout
            );
        }
        s_uiDisconnectEndPointTimeout = o;
    }

    /* Please see man pages for details 
     
     
     
     
     
     */
    static qUnsignedInt32 GetTerminateDOCoreTimeout()
    {
        return s_uiTerminateDOCoreTimeout;
    }


    /* Please see man pages for details 
     
     
     
     
     
     
     */
    static void SetTerminateDOCoreTimeout(qUnsignedInt32 o)
    {
        if(o <= s_uiDisconnectEndPointTimeout + 5000)
        {
            _QLOG(
                EventLog::Warning,
                _T("TerminateDOCoreTimeout(%d) should be longer than DisconnectEndPointTimeout(%d) + 5000"),
                o,
                s_uiDisconnectEndPointTimeout
            );
        }
        s_uiTerminateDOCoreTimeout = o;
    }


    static void SetTerminateImmediatelyTimeout(qUnsignedInt32 o)
    {
        s_uiTerminateImmediatelyTimeout = o;
    }
    static qUnsignedInt32 GetTerminateImmediatelyTimeout()
    {
        return s_uiTerminateImmediatelyTimeout;
    }

    /* Please see man pages for details 
     
     
     
     
     */
    static qUnsignedInt GetTransportThreadSleepSpan()
    {
        return s_TransportThreadSleepSpan;
    }

    /* Please see man pages for details 
     
     
     
     
     */
    static void SetTransportThreadSleepSpan(qUnsignedInt sleepSpan)
    {
        s_TransportThreadSleepSpan = sleepSpan;
    }

    /* Please see man pages for details 
     
     
     
     
     */
    static qUnsignedInt GetTransportThreadPriority()
    {
        return s_TransportThreadPriority;
    }

    /* Please see man pages for details 
     
     
     
     
     */
    static void SetTransportThreadPriority(qUnsignedInt priority)
    {
        s_TransportThreadPriority = priority;
    }

    static void EnableTransportBufferThread(qBool o)
    {
        Core::UseTransportBufferThread(o);
    }

    static qBool IsEnabledTransportBufferThread()
    {
        return Core::IsUsingTransportBufferThread();
    }

    /* Please see man pages for details 
     
     
     
     
     
     
     
     
     
     */
    static void EnableTransportBufferThreadBufferPool(qBool o)
    {
        s_bEnableTransportBufferThreadBufferPool = o;
    }

    /* Please see man pages for details 
     
     
     
     
     
     */
    static qBool IsEnabledTransportBufferThreadBufferPool()
    {
        return s_bEnableTransportBufferThreadBufferPool;
    }

    /* Please see man pages for details 
     
     
     
     
     */
    static qUnsignedInt GetTransportBufferPoolNum()
    {
        return s_TransportBufferPoolNum;
    }

    /* Please see man pages for details 
     
     
     
     
     
     
     
     
     
     
     */
    static void SetTransportBufferPoolNum(qUnsignedInt poolNum)
    {
        s_TransportBufferPoolNum = poolNum;
    }

    /* Please see man pages for details 
     
     
     
     */
    static qUnsignedInt GetTransportBufferThreadSleepSpan()
    {
        return s_TransportBufferThreadSleepSpan;
    }

    /* Please see man pages for details 
     
     
     
     
     
     
     
     
     */
    static void SetTransportBufferThreadSleepSpan(qUnsignedInt sleepSpan)
    {
        s_TransportBufferThreadSleepSpan = sleepSpan;
    }

    /* Please see man pages for details 
     
     
     
     
     
     
     */
    static qUnsignedInt GetTransportBufferThreadPriority()
    {
        return s_TransportBufferThreadPriority;
    }

    /* Please see man pages for details 
     
     
     
     
     
     
     */
    static void SetTransportBufferThreadPriority(qUnsignedInt priority)
    {
        s_TransportBufferThreadPriority = priority;
        s_TransportBufferThreadSendPriority = priority;
    }

    /* Please see man pages for details 
     
     
     
     
     
     */
    static qUnsignedInt GetTransportBufferThreadRecvPriority()
    {
        return s_TransportBufferThreadPriority;
    }

    /* Please see man pages for details 
     
     
     
     
     
     */
    static void SetTransportBufferThreadRecvPriority(qUnsignedInt priority)
    {
        s_TransportBufferThreadPriority = priority;
    }

    /* Please see man pages for details 
     
     
     
     
     
     */
    static qUnsignedInt GetTransportBufferThreadSendPriority()
    {
        return s_TransportBufferThreadSendPriority;
    }

    /* Please see man pages for details 
     
     
     
     
     
     */
    static void SetTransportBufferThreadSendPriority(qUnsignedInt priority)
    {
        s_TransportBufferThreadSendPriority = priority;
    }


    /* Please see man pages for details 
     
     
     
     */
    static qUnsignedInt GetTransportBufferReceiveCount()
    {
        return s_TransportBufferReceiveCount;
    }

    /* Please see man pages for details 
     
     
     
     */
    static void SetTransportBufferReceiveCount(qUnsignedInt receiveCount)
    {
        s_TransportBufferReceiveCount = receiveCount;
    }

    /* Please see man pages for details 
     
     
     */
    static qUnsignedInt GetTransportBufferSendCount()
    {
        return s_TransportBufferSendCount;
    }

    /* Please see man pages for details 
     
     
     */
    static void SetTransportBufferSendCount( qUnsignedInt sendCount )
    {
        s_TransportBufferSendCount = sendCount;
    }

    /* Please see man pages for details 
     
     
     
     
     
     
     
     
     
     
     
     
     
     
     
     */
    static void SetTransportBufferThreadCoreNo(qInt32 no)
    {
        s_TransportBufferThreadCoreNo = no;
    }

    /* Please see man pages for details 
     
     
     
     
     
     */
    static qInt32 GetTransportBufferThreadCoreNo()
    {
        return s_TransportBufferThreadCoreNo;
    }

    /* Please see man pages for details 
     
     
     
     
     
     */
    static qUnsignedInt GetNameResolveThreadPriority()
    {
        return s_NameResolveThreadPriority;
    }

    /* Please see man pages for details 
     
     
     
     
     
     
     */
    static void SetNameResolveThreadPriority(qUnsignedInt priority)
    {
        s_NameResolveThreadPriority = priority;
    }

    /* Please see man pages for details 
     
     
     
     
     
     
     
     
     
     
     
     
     
     
     
     
     
     
     */
    static void SetNameResolveThreadCoreNo(qInt32 no)
    {
        s_NameResolveThreadCoreNo = no;
    }

    /* Please see man pages for details 
     
     
     
     
     
     
     
     
     */
    static qInt32 GetNameResolveThreadCoreNo()
    {
        return s_NameResolveThreadCoreNo;
    }


#if NEX_ENABLE_LATENCY_LIMIT_TRANSPORT_BUFFER_SEND
    /* Please see man pages for details 
     
     
     
     
     
     */
    static void SetTransportBufferSendLatencyLimit(qUnsignedInt32 limit)
    {
        s_uiTransportBufferSendLatencyLimit = limit;
    }

    static qUnsignedInt32 GetTransportBufferSendLatencyLimit()
    {
        return s_uiTransportBufferSendLatencyLimit;
    }
#endif

#if NEX_ENABLE_LATENCY_LIMIT_PACKET_QUEUE_SEND
    /* Please see man pages for details 
     
     
     
     
     
     
     
     
     */
    static void SetOutgoingPacketQueueLatencyLimit(qUnsignedInt32 limit)
    {
        s_uiOutgoingPacketQueueLatencyLimit = limit;
    }

    static qUnsignedInt32 GetOutgoingPacketQueueLatencyLimit()
    {
        return s_uiOutgoingPacketQueueLatencyLimit;
    }
#endif

    /* Please see man pages for details 
     
     
     
     
     */
    static qBool IsEnablePacketHistoryBundling()
    {
        return s_bEnablePacketHistoryBundling;
    }


    /* Please see man pages for details 
     
     
     
     
     */
    static qBool IsEnableReliablePacketHistoryBundling()
    {
        return s_bEnableReliablePacketHistoryBundling;
    }

    /* Please see man pages for details 
     
     
     
     
     */
    static qBool IsEnableUnreliablePacketHistoryBundling()
    {
        return s_bEnableUnreliablePacketHistoryBundling;
    }

    /* Please see man pages for details 
     
     
     
     
     */
    static qBool IsEnableDOPacketHistoryBundling()
    {
        return s_bEnableDOPacketHistoryBundling;
    }

    /* Please see man pages for details 
     
     
     
     
     */
    static TimeInterval GetPacketHistoryBundlingHoldingTime()
    {
        return s_packetHistoryBundlingHoldingTime;
    }

    /* Please see man pages for details 
     
     
     
     
     */
    static qUnsignedInt16 GetPacketHistoryBundlingMaxHoldingPackets()
    {
        return s_packetHistoryBundlingMaxHoldingPackets;
    }

    /* Please see man pages for details 
     
     
     
     
     */
    static qUnsignedInt16 GetPacketHistoryBundlingMaxRetransmitPackets()
    {
        return s_packetHistoryBundlingMaxRetransmitPackets;
    }

    /* Please see man pages for details 
     
     
     
     
     */
    static qUnsignedInt16 GetPacketHistoryBundlingMaxRetransmitCounts()
    {
        return s_packetHistoryBundlingMaxRetransmitCounts;
    }


    /* Please see man pages for details 
     
     
     
     */
    static TimeInterval GetPacketHistoryBundlingForcedRetransmitTimeout()
    {
        return s_packetHistoryBundlingForcedRetransmitTimeout;
    }

    /* Please see man pages for details 
     
     
     
     */
    static qUnsignedInt32 GetPacketHistoryBundlingMaximumPacketBytes()
    {
        return s_packetHistoryBundlingMaximumPacketBytes;
    }

    /* Please see man pages for details 
     
     
     
     
     
     */
    static const qUnsignedInt32* GetPacketHistoryBundlingPacketCounts()
    {
        return s_packetHistoryBundlingPacketCounts;
    }

    /* Please see man pages for details 
     
     
     
     
     
     */
    static qUnsignedInt32 GetPacketHistoryBundlingPacketCount(PacketHistoryCount phc)
    {
        return s_packetHistoryBundlingPacketCounts[phc];
    }

    /* Please see man pages for details 
     
     
     
     
     
     
     
     
     
     
     
     
     
     
     
     
     
     
     */
    static void EnablePacketHistoryBundling(qBool b)
    {
        s_bEnablePacketHistoryBundling = b;
    }

    /* Please see man pages for details 
     
     
     
     
     */

    static void EnableReliablePacketHistoryBundling(qBool b)
    {
        s_bEnableReliablePacketHistoryBundling = b;
    }

    /* Please see man pages for details 
     
     
     
     
     */
    static void EnableUnreliablePacketHistoryBundling(qBool b)
    {
        s_bEnableUnreliablePacketHistoryBundling = b;
    }


    /* Please see man pages for details 
     
     
     
     
     
     
     
     
     
     */
    static void EnableDOPacketHistoryBundling(qBool b)
    {
        s_bEnableDOPacketHistoryBundling = b;
    }


    /* Please see man pages for details 
     
     
     
     
     */
    static void SetPacketHistoryBundlingHoldingTime(TimeInterval t)
    {
        s_packetHistoryBundlingHoldingTime = t;
    }

    /* Please see man pages for details 
     
     
     
     
     */
    static void SetPacketHistoryBundlingMaxHoldingPackets(qUnsignedInt16 n)
    {
        s_packetHistoryBundlingMaxHoldingPackets = n;
    }

    /* Please see man pages for details 
     
     
     
     
     
     
     
     
     
     */
    static void SetPacketHistoryBundlingMaxRetransmitPackets(qUnsignedInt16 n)
    {
        s_packetHistoryBundlingMaxRetransmitPackets = n;
    }

    /* Please see man pages for details 
     
     
     
     
     
     
     
     */
    static void SetPacketHistoryBundlingMaxRetransmitCounts(qUnsignedInt16 n)
    {
        s_packetHistoryBundlingMaxRetransmitCounts = n;
    }

    /* Please see man pages for details 
     
     
     
     
     
     
     
     */
    static void SetPacketHistoryBundlingForcedRetransmitTimeout(TimeInterval timeout)
    {
        s_packetHistoryBundlingForcedRetransmitTimeout = timeout;
    }

    /* Please see man pages for details 
     
     
     
     */
    static void SetPacketHistoryBundlingMaximumPacketBytes(qUnsignedInt32 bytes)
    {
        s_packetHistoryBundlingMaximumPacketBytes = bytes;
    }

    /* Please see man pages for details 
     
     
     
     
     
     
     
     
     
     
     */
    static void EnableDropDuplicateReorderingUnreliablePackets(qBool b)
    {
        s_bEnableDropDuplicateReorderingUnreliablePackets = b;
    }

    /* Please see man pages for details 
     
     
     
     
     
     
     
     
     
     
     */
    static qBool IsEnableDropDuplicateReorderingUnreliablePackets()
    {
        return s_bEnableDropDuplicateReorderingUnreliablePackets;
    }

    static void UpdatePacketHistoryBundlingPacketCounts(PacketHistoryCount phc)
    {
        s_packetHistoryBundlingPacketCounts[phc]++;
    }

#if NEX_ENABLE_HYBRID_DETECT_LIMITATION
    /* Please see man pages for details 
     
     
     
     */
    static void SetRoutingHistoryUpdateSpan(qUnsignedInt32 ui)
    {
        s_uiRoutingHistoryUpdateSpan = ui;
    }
    /* Please see man pages for details 
     
     
     */
    static qUnsignedInt32 GetRoutingHistoryUpdateSpan()
    {
        return s_uiRoutingHistoryUpdateSpan;
    }

    /* Please see man pages for details 
     
     
     
     
     
     
     */
    static void SetRoutingHistoryTTL(qUnsignedInt32 ui)
    {
        s_uiRoutingHistoryTTL = ui;
    }
    /* Please see man pages for details 
     
     
     */
    static qUnsignedInt32 GetRoutingHistoryTTL()
    {
        return s_uiRoutingHistoryTTL;
    }

    /* Please see man pages for details 
     
     
     
     
     
     
     
     
     */
    static void SetRoutingLimit(qUnsignedInt32 ui)
    {
        s_uiRoutingLimit = ui;
    }
    /* Please see man pages for details 
     
     
     */
    static qUnsignedInt32 GetRoutingLimit()
    {
        return s_uiRoutingLimit;
    }

    /* Please see man pages for details 
     
     
     
     
     
     
     
     
     
     
     */
    static void SetEnabledRelay(qBool flag) { s_bEnableRelay = flag; }

    /* Please see man pages for details 
     
     
     
     
     */
    static qBool IsEnabledRelay() { return s_bEnableRelay; }


    /* Please see man pages for details 
     
     
     
     
     
     */
    static qBool ExistsRelayTable();
#endif

protected:
    static qUnsignedInt32 s_uiNATTraversalTimeout;
    static qUnsignedInt32 s_uiConnectEndPointTimeout;
    static qUnsignedInt32 s_uiDisconnectEndPointTimeout;
    static qUnsignedInt32 s_uiTerminateDOCoreTimeout;

    static qBool s_bEnableDropDuplicateReorderingUnreliablePackets;

    static qBool s_bEnablePacketHistoryBundling;
    static qBool s_bEnableReliablePacketHistoryBundling;
    static qBool s_bEnableUnreliablePacketHistoryBundling;
    static qBool s_bEnableDOPacketHistoryBundling;
    static TimeInterval s_packetHistoryBundlingHoldingTime;
    static qUnsignedInt16 s_packetHistoryBundlingMaxHoldingPackets;
    static qUnsignedInt16 s_packetHistoryBundlingMaxRetransmitPackets;
    static qUnsignedInt16 s_packetHistoryBundlingMaxRetransmitCounts;
    static TimeInterval s_packetHistoryBundlingForcedRetransmitTimeout;
    static qUnsignedInt32 s_packetHistoryBundlingMaximumPacketBytes;
    static qUnsignedInt32 s_packetHistoryBundlingPacketCounts[MAX_PACKET_HISTORY_COUNT + 1];

#if NEX_ENABLE_LATENCY_LIMIT_PACKET_QUEUE_SEND
    static qUnsignedInt32 s_uiOutgoingPacketQueueLatencyLimit;
#endif
#if NEX_ENABLE_LATENCY_LIMIT_TRANSPORT_BUFFER_SEND
    static qUnsignedInt32 s_uiTransportBufferSendLatencyLimit;
#endif

    static qUnsignedInt32 s_uiTerminateImmediatelyTimeout;
    static qBool s_bEnableRelay;

public:
    // n2370: why public? this class is singleton.
    //group=Network Interface
    RootTransport();
    virtual ~RootTransport();
    //{secret}
    virtual qResult         Initialize()
    {
        return QRESULT_SUCCESS(Core, Success);
    }

    //{secret}
    qBool                   RegisterStream(Stream *pStream,
                                           Stream::ID idDesiredStream,
                                           Stream::ID *pAcquiredStreamID = NULL)
    {
        return m_oTransportStreamManager.RegisterStream(pStream,idDesiredStream,pAcquiredStreamID);
    }
    //{secret}
    qBool                   UnregisterStream(Stream::Type oType, Stream::ID idStream)
    {
        return m_oTransportStreamManager.UnregisterStream(oType,idStream);
    }
    //{secret}
    qUnsignedInt32          GetNbStream()
    {
        return m_oTransportStreamManager.GetNbStream();
    }

    //{secret}
    virtual qBool           StartListen(qUnsignedInt16 ui16Port,
                                        qUnsignedInt16 *pui16ListeningPort = NULL,
                                        qBool bPreferred = true, qUnsignedInt uiTrafficType = 0, qBool usePacketBufferPool=true) = 0;
    //{secret}
    virtual qBool           StartListenReuse(qUnsignedInt16 ui16Port,
                                             qUnsignedInt16 *pui16ListeningPort = NULL,
                                             qBool bPreferred = true, qUnsignedInt uiTrafficType = 0, qBool usePacketBufferPool=true) = 0;
    //{secret}
    virtual qBool           StopListen(qUnsignedInt16 ui16Port) = 0;
    //{secret}
    virtual qBool           StopListen() = 0;
    //{secret}
    virtual qUnsignedInt32  GetNbListeningPorts() = 0;


    // {secret}
    void StopTransportThread();
    //{secret}
    virtual qBool           Send(qUnsignedInt16 ui16Port, Stream::Type oType,
                                 Stream::ID idSource, Stream::ID idDestination,
                                 PacketOut *pPacket, qBool bQueue = false) = 0;
    //{secret}
    virtual qBool           Receive(qUnsignedInt16 uiPort, Buffer *pBuffer, const InetAddress *pSourceAddress) = 0;

#ifdef ROUTING_SUPPORT
    //{secret}
    virtual Router         *GetRouter();
#endif

    virtual Relay          *GetRelay();

    //{secret}
    void                    SetMTU(qUnsignedInt32 uiMTU)
    {
        m_uiMTU = uiMTU;
    }
    //{secret}
    qUnsignedInt32          GetMTU()
    {
        return m_uiMTU;
    }

    //{secret}
    void                    SetPreferredPortNumber(qUnsignedInt16 uiPreferredPortNumber)
    {
        SYSTEMCHECK(GetNbListeningPorts()>0);
        m_uiPreferredPortNumber = uiPreferredPortNumber;
    }
    //{secret}
    qUnsignedInt16          GetPreferredPortNumber()
    {
        return m_uiPreferredPortNumber;
    }

    //group=NAT Traversal

    /*
    $ifdef RENDEZVOUS$
    Description: Sets the network interface that $PRODUCTNAME$ will bind to.
    Parameters:
    pszInterfaceAddress: The network interface address to use. The default value is 0.0.0.0.
    Remarks: The network interface address to use. The default value is 0.0.0.0.
    Remarks: By default, $PRODUCTNAME$ will bind on a specific port on 0.0.0.0 and listen
    on that port on all configured addresses on all interfaces. That is, if a station has two
    configured network interfaces, $PRODUCTNAME$ will capture all packets addressed to its
    port on any of the two interfaces. However, if $PRODUCTNAME$ is bound to port 9103 (the
    default port) on the network interface specified here, it will only capture data sent
    to that interface/port and will discard all others. This method was created to facilitate
    testing. It must be called before the $PRODUCTNAME$ object is created.
    See Also: GetInterfaceAddress
    $else$
    {secret}
    $endif$
    */
    static void SetInterfaceAddress(qChar* pszInterfaceAddress);

    /*
    $ifdef RENDEZVOUS$
    Description: Returns the network interface bound to $PRODUCTNAME$ by SetInterfaceAddress.
    Returns     : The network interface used by $PRODUCTNAME$.
    Remarks: The network interface to bind to, set by SetInterfaceAddress.
    See Also: SetInterfaceAddress
    $else$
    {secret}
    $endif$
    */
    static qChar* GetInterfaceAddress();

    //group=Network Interface

    /*
    Description: Sets the well-known port number used by $PRODUCTNAME$.
    Parameters:
    uiWellKnownPort: The well-known port number that will be used. The default is port 9103.
    Remarks: By default, $PRODUCTNAME$ will use port 9103 as the default well-known port.
    If the well-known port cannot be used, the system will open a dynamic port. The default well-known
    port can be changed via this method, which must be called before the $PRODUCTNAME$ object is created.
    In general you should not need to change the well-know port number, except where necessary
    in order to avoid a clash with another application that uses the same port number, such as if you wish
    to create more than one session on the same station.
    See Also: GetWellKnownPortNumber
    */
    /* Please see man pages for details 
     
     
     
     
     
     
     
     
     
     
     */
    static void             SetWellKnownPortNumber(qUnsignedInt16 uiWellKnownPort);

    /*
    Description: Returns the well-known port number used by $PRODUCTNAME$.
    Returns     : The well-known port number used by $PRODUCTNAME$.
    Remarks: The well-known port number is set via SetWellKnownPortNumber.
    See Also: SetWellKnownPortNumber
    */
    /* Please see man pages for details 
     
     
     
     
     
     */
    static qUnsignedInt16   GetWellKnownPortNumber();

    static qUnsignedInt16   GetNextPortNumber();

    // group=Emulation Devices

    //Description: Returns a pointer to an instance of the OutputEmulationDevice.
    //Returns     : A pointer to an instance of the OutputEmulationDevice.
    //See Also: GetInputEmulationDevice, OutputEmulationDevice, InputEmulationDevice
    /* Please see man pages for details 
     
     
     
     
     
     */
    OutputEmulationDevice   *GetOutputEmulationDevice()
    {
        return &m_oOutputEmulationDevice;
    }

    //Description: Returns a pointer to an instance of the InputEmulationDevice.
    //Returns     : A pointer to an instance of the InputEmulationDevice.
    //See Also: GetOutputEmulationDevice, OutputEmulationDevice, InputEmulationDevice
    /* Please see man pages for details 
     
     
     
     
     
     */
    InputEmulationDevice    *GetInputEmulationDevice()
    {
        return &m_oInputEmulationDevice;
    }

    //group=Transport Parameters

    /*
    Description: Returns a pointer to an instance of the transport performance counters.
    Returns     : A pointer to an instance of TransportPerfCounters.
    Example:
            TransportPerfCounters *pCounters = RootTransport::GetInstance()->GetPerfCounters();
            // Return and print the BPS sent and received on a station
            qUnsignedInt32 uiBPSRecv = pCounters->GetValue(PERF_TRANSPORT_BPS_RECV, PERF_FREQUENCY_LAST);
            qUnsignedInt32 uiBPSSent = pCounters->GetValue(PERF_TRANSPORT_BPS_SENT, PERF_FREQUENCY_LAST);
            printf("BPS Sent: %d, Recv %d\n",uiBPSRecv,uiBPSSent);
    See Also: TransportPerfCounters,
    */
#ifdef PERFCOUNTER_ENABLED
    TransportPerfCounters   *GetPerfCounters()
    {
        return &m_oPerfCounters;
    }
#endif // PERFCOUNTER_ENABLED

    //{secret}
    void                    AttachVNATDevice(VirtualNATDevice *pDevice);
    //{secret}
    void                    DetachVNATDevice(VirtualNATDevice *pDevice);
    //{secret}
    VirtualNATDevice*       GetVNATDevice()
    {
        //n1589: mark: TransportLock: A TransportLock is unnecessary because the value is simply NULL or it is not.
        return m_pVNATDevice;
    }

#if NEX_ENABLE_VIRTUAL_FILTER_DEVICE
    VirtualFilterDevice*     GetVirtualFilterDevice()
    {
        return m_pFilterDevice;
    }
#endif

    //{secret}
    TransportStreamManager *GetTransportStreamManager()
    {
        return &m_oTransportStreamManager;
    }

    //group=Methods

    /*
    Description:  Returns a pointer to the RootTransport instance.

    */
    /* Please see man pages for details 
     
     
     */
    static RootTransport* GetInstance()
    {
        Network* pNetwork = Network::GetInstance();
        if (pNetwork == NULL)
        {
            //n1589
            QTRACE(D_TRACE_CONSTRUCT, _T("RootTransport::GetInstance:pNetwork == NULL"));
            return NULL;
        }
        else
        {
            return pNetwork->GetRootTransport();
        }
    }

    //{secret}
    static void SetMaxNbListeningPorts(qUnsignedInt uiMaxNbListeningPorts)
    {
        s_uiMaxNbListeningPorts = uiMaxNbListeningPorts;
    }

    //{secret}
    static qUnsignedInt GetMaxNbListeningPorts()
    {
        return s_uiMaxNbListeningPorts;
    }

    //{secret}
    static void SetRecvBufferSize(qUnsignedInt32 uiSize)
    {
        s_uiRecvBufferSize = uiSize;
    }

    //{secret}
    static qUnsignedInt32 GetRecvBufferSize()
    {
        return s_uiRecvBufferSize;
    }

    //{secret}
    static void SetSendBufferSize(qUnsignedInt32 uiSize)
    {
        s_uiSendBufferSize = uiSize;
    }

    //{secret}
    static qUnsignedInt32 GetSendBufferSize()
    {
        return s_uiSendBufferSize;
    }

    //{secret}
    static qUnsignedInt8 GetDefaultPacketTTL()
    {
        return DEFAULT_PACKET_TTL;
    }

#if defined(SOCKET_HANDLE_SUPPORT)

    //group=Network Interface

    /*
    $ifdef RENDEZVOUS$
    {secret}
    $else$
    Description: Returns the UDP socket that $PRODUCTNAME$ uses.
    Parameters:
    uiFromPort: Optional parameter. If zero (the default), returns the handle to the socket opened on the
    well-known port. In the case that you are using both NetZ and Rendez-Vous and if you expect you may not
    be able to listen on the well-known port, then you can specify a previously established Rendez-Vous
    listening port for this handle (see the Example below).
    Returns     : The UDP socket that $PRODUCTNAME$ uses. It returns NULL if this method is not supported for
    the current platform.
    Remarks: This method should only be called after the creation of the $PRODUCTNAME$ object. The returned
    UDP socket is platform specific and may be used with any appropriate platform specific TCP/IP
    socket calls. If you wish to send non-Net-Z packets, you need to send them using this socket and receive them
    through your registered PacketRecognizer (RegisterPacketRecognizer).
    Example:

            //if credentials valid and using both Net-Z and Rendez-Vous ...
            void* pHandle =
                    RootTransport::GetInstance()->GetSocketHandle(pRendezVous->GetStreamManager()->GetPort());

    See Also: RegisterPacketRecognizer
    $endif$
    */
    virtual void *GetSocketHandle(qUnsignedInt16 uiFromPort=0) = 0;

#endif

    //group=Transport Parameters

    /*
    Description: Returns the number of packets queued in the transport layer.
    Returns     : The number of packets queued in the transport layer.
    Remarks: This method can be used with a selection of heuristics to adapt the game to the
    current network condition. For example, when the packet queue grows, a game could decide to decrease
    the quality of the game, or the number of updates per frame or simply stop the game.
    See Also: GetEstimatedPacketQueueMemoryUsage
    */
    virtual qUnsignedInt32 GetPacketQueueSize() const = 0;

    /*
    Description: Returns an estimation of the memory usage of the packets queued in the transport layer.
    Returns     : An estimation of the memory usage of the packets queued in the transport layer.
    Remarks: This method is only an estimation of the memory usage. For a more accurate value, you can
    create your own memory allocation functions with counters to track the memory you allocate and deallocate
    and then register them with the MemoryManager to override the default Net-Z memory allocation functions.
    See Also: GetPacketQueueSize, MemoryManager
    */
    virtual qUnsignedInt32 GetEstimatedPacketQueueMemoryUsage() const = 0;

    //{secret}
    virtual qResult Send(StationURL *pStationURL, Buffer *pBuffer)
    {
        DUMMY_PARAM(pStationURL);
        DUMMY_PARAM(pBuffer);
        return QRESULT_ERROR(Core, NotImplemented);
    }

    //{secret}
    static qUnsignedInt32 GetIOWaitTime()
    {
        return s_uiIOWaitTime;
    }
    //{secret}
    static void SetIOWaitTime(qUnsignedInt32 uiIOWaitTime)
    {
        s_uiIOWaitTime = uiIOWaitTime;
    }
    //{secret}
    static qUnsignedInt32 GetIdleSleepTime()
    {
        return s_uiIdleSleepTime;
    }
    //{secret}
    static void SetIdleSleepTime(qUnsignedInt32 uiIdleSleepTime)
    {
        s_uiIdleSleepTime = uiIdleSleepTime;
    }
    //{secret}
    qBool IsUsingUserThreadContext()
    {
        return m_bUseUserThreadContext;
    }
    //{secret}
    void UseUserThreadContext(qBool bUse)
    {
        m_bUseUserThreadContext = bUse;
    }

    //group=Network Interface

    //n1589:todo2:remove GetDefaultConnectionTimeout
    static qUnsignedInt32 GetDefaultConnectionTimeout()
    {
        return GetConnectEndPointTimeout();
    }

    /*
    Description: Sets the default connection timeout.
    Parameters:
    uiDefaultConnectionTimeout: The new default connection timeout in milliseconds. The default value is 30 000 ms (30 s).
    Remarks: The default connection timeout is the time that a Station will keep trying to connect to another station
    before giving up. This timeout is enforced any time a connection via a StationURL is attempted.
    */
    //n1589:todo2:remove SetDefaultConnectionTimeout
    static void SetDefaultConnectionTimeout(qUnsignedInt32 uiDefaultConnectionTimeout)
    {
        SetConnectEndPointTimeout(uiDefaultConnectionTimeout);
    }

    /*
    $ifdef RENDEZVOUS$
    {secret}
    $else$
    Description: Registers a PacketRecognizer to this RootTransport.
    Parameters:
    pPacketRecognizer: A pointer to the PacketRecognizer that will be registered to this RootTransport.
    Returns     : true if the specified PacketRecognizer was successfully registered to this RootTransport,
    false if not or an error occurred.
    Remarks: For non-$PRODUCTNAME$ packets to be accepted by the system rather than automatically being dropped
    they must pass via a registered PacketRecognizer. Refer to the PacketRecognizer class for further details.
    See Also: UnregisterPacketRecognizer, PacketRecognizer
    $endif$
    */
    qBool RegisterPacketRecognizer(PacketRecognizer *pPacketRecognizer);

    /*
    $ifdef RENDEZVOUS$
    {secret}
    $else$
    Description: Registers a PacketRecognizer to this RootTransport.
    Parameters:
    pPacketRecognizer: A pointer to the PacketRecognizer that will be unregistered from this RootTransport.
    Returns     : true if the specified PacketRecognizer was successfully unregistered from this RootTransport,
    false if not or an error occurred. This method will return false if the specified PacketRecognizer is
    not currently registered to this RootTransport.
    See Also: RegisterPacketRecognizer, PacketRecognizer
    $endif$
    */
    qBool UnregisterPacketRecognizer(PacketRecognizer *pPacketRecognizer);

    //{secret}
    qList<PacketRecognizer*> *GetPacketRecognizerList()
    {
        return &m_lstPacketRecognizers;
    }
    //{secret}
    qBool                     RejectPacket(Buffer *pBuffer, const InetAddress * const pInetAddress);

    //group=NAT Traversal

    /* Please see man pages for details 
     
     
     
     
     
     */
    qBool SetUpDefaultNATCheckServer() { SYSTEMCHECK(false); return true;}


    /*
    $ifdef RENDEZVOUS$
    Description: Sets the primary NATCheckServer address and ports.
    Parameters:
    pszNATCheckServer: The IP address of the primary NATCheckServer.
    ui16PrimaryPort: First port of the NATCheckServer to contact. The secondary NATCheckServer must use the same primary
    port.
    ui16SecondaryPort: Second port of the NATCheckServer to contact. The secondary NATCheckServer must use the same
    secondary port.
    Returns     : True if the NATCheckServer address and ports have been set correctly; false otherwise.
    Remarks: Set the primary NATCheckServer address and ports before logging in to Rendez-Vous.
    NAT detection requires two NATCheck servers which are configured in parallel as they are both
    contacted during the detection of the station's NAT device Properties.
    See Also: SetSecondaryNATCheckServer
    $else$
    {secret}
    $endif$
    */
    qBool SetPrimaryNATCheckServer(const qChar *pszNATCheckServer,
                                   qUnsignedInt16 ui16PrimaryPort,
                                   qUnsignedInt16 ui16SecondaryPort);

    /*
    $ifdef RENDEZVOUS$
    Description: Sets the secondary NATCheckServer address and ports.
    Parameters:
    pszNATCheckServerAddress: The IP address of the secondary NATCheckServer.
    ui16PrimaryPort: First contact port for this secondary NATCheckServer. The primary NATCheckServer must use the same
    primary port.
    ui16SecondaryPort: Second contact port for this secondary NATCheckServer. The primary NATCheckServer must use the same
    secondary port.
    Returns     : True if the NATCheckServer address and ports have been set correctly; false otherwise.
    Remarks: Set the secondary NATCheckServer address and ports before logging in to Rendez-Vous.
    NAT detection requires two NATCheck servers which are configured in parallel as they are both
    contacted during the detection of the station's NAT device Properties.
    See Also: SetPrimaryNATCheckServer
    $else$
    {secret}
    $endif$
    */
    qBool SetSecondaryNATCheckServer(const qChar *pszNATCheckServerAddress,
                                     qUnsignedInt16 ui16PrimaryPort,
                                     qUnsignedInt16 ui16SecondaryPort);



    //{secret}
    const qVector<StringStream *> &GetNATCheckServerAddresses() const;

    //{secret}
    qBool IsNATCheckServerAddressesSet() const
    {
        return m_isNATCheckServerAddressesSet;
    }

    qBool IsRelayServerAddressesSet() const
    {
        return m_isRelayServerAddressesSet;
    }

    qBool IsSocketAlreadyClosed() const
    {
        return m_isSocketAlreadyClosed;
    }


    //{secret}
    void SetNATProperties(const NATProperties &rNATProperties);

    /*
    $ifdef RENDEZVOUS$
    Description: Retrieves the NAT properties of the router between the station and the Internet.
    Returns     : A NATProperties object containing the NAT properties (based on those outlined in RFC 4787) of the station's router.
    Remarks: To retrieve a valid reading of the station's NAT Properties, it is necessary to call this method
    after logging into Rendez-Vous. Otherwise, the contents of the NATProperties object will be invalid.
    See Also: NATProperties
    $else$
    {secret}
    $endif$
    */
    NATProperties GetNATProperties() const;

    //Use in a separate thread from JobNameResolve. Implement so it does not cause a qNew or launch both simultaneously.
    qBool SetUpDefaultNATCheckServerInternal();
    qBool SetUpRelayServer();

    //{secret}
    void SetUPnPProperties(const UPnPProperties &rUPnPProperties);

    /*
    $ifdef RENDEZVOUS$
    Description: Retrieves the UPnP properties of the router between the station and the Internet.
    Returns     : A UPnPProperties object.
    Remarks: To get a valid reading of the router's UPnP properties, it is necessary to call this method
    after logging into Rendez-Vous. Otherwise, the contents of the UPnPProperties object will be invalid.
    See Also: EnableUPnPSupport, UPnPProperties
    $else$
    {secret}
    $endif$
    */
    UPnPProperties GetUPnPProperties() const;

    //{secret}
    void EnableNATCheckSupport(qBool bEnableNATCheck);

    //{secret}
    qBool IsNATCheckEnabled() const;

    /*
    $ifdef RENDEZVOUS$
    Description: Enables or disables UPnP
    Remarks: UPnP allows an application behind a NAT router to configure the router
    to allow peers outside the private network to contact it.
    Note, enabling UPnP results in a longer login time.
    Parameters:
    bEnableUPnP: True to enable UPnP; false otherwise.
    See Also: GetUPnPProperties, UPnPProperties
    $else$
    {secret}
    $endif$
    */
    void EnableUPnPSupport(qBool bEnableUPnP);

    //{secret}
    qBool IsUPnPSupported() const;

    /*
    $ifdef RENDEZVOUS$
    Description: Enables or disables NAT-PMP.
    Remarks: The NAT Port Mapping Protocol allows an application behind a NAT router to
    configure the router to allow peers outside the private network to contact it.
    Note, enabling NAT-PMP results in a longer login time.
    See Also: IsNATPMPPortOpen
    $else$
    {secret}
    $endif$
    */
    void EnableNATPMPSupport(qBool bEnableNatPMP);

    //{secret}
    qBool IsNATPMPSupported() const;

    //{secret}
    void SetNATPMPPortOpen(qBool bFlag = true);

    /*
    $ifdef RENDEZVOUS$
    Description: Returns whether a port mapping has been added using NAT-PMP.
    Returns     : True if a port has been opened in this way; false otherwise.
    See Also: EnableNATPMPSupport
    $else$
    {secret}
    $endif$
    */
    qBool IsNATPMPPortOpen() const;


    static void SetBlockedAddresses(const qList<InetAddress>& dropList);

    //n1589: Depending on when the ProtectedPacketQueue arguments are passed, RootTransport cannot be generated and is thus unusable.
#if 0
    static CriticalSection& GetTransportLock()
    {
        SYSTEMCHECK(GetInstance() != NULL);
        return GetInstance()->m_csTransportLock;
    }
#endif

    //n1589
#ifdef TRACE_ENABLED
    void Trace(qUnsignedInt64 uiTraceFlags);
#endif

    qUnsignedInt GetVsocketCount()
    {
        return m_oTransportStreamManager.GetNBStreamByStreamType(Stream::Game);
    }
#ifdef NEX_CLIENT_PBPOOL
    virtual void ClearPacketInPacketBufferPool() {}
#endif // NEX_CLIENT_PBPOOL

    const InetAddress & GetRelayAddress() { return m_relayAddress;}
    void SetRelayNameAddress(String &str);

protected:
    mutable CriticalSection m_csPacketRecognizerList;
    mutable CriticalSection m_csNATDevice;

    qBool GetRelayServerAddress(const qChar *address);

    virtual void StopTransportThreadImpl() = 0;

    void SetIsNATCheckServerAddressesSet();

    qUnsignedInt32          m_uiMTU;
    qBool                   m_bUseUserThreadContext;

#ifdef PERFCOUNTER_ENABLED
    TransportPerfCounters       m_oPerfCounters;
#endif // PERFCOUNTER_ENABLED
    OutputEmulationDevice       m_oOutputEmulationDevice;
    InputEmulationDevice        m_oInputEmulationDevice;
    TransportStreamManager      m_oTransportStreamManager;
    VirtualNATDevice           *m_pVNATDevice;

    qUnsignedInt16              m_uiPreferredPortNumber;

    static qUnsignedInt32       s_uiMaxNbListeningPorts;
    static qUnsignedInt32       s_uiMaxNbConnections;
    static qUnsignedInt32       s_uiIdleSleepTime;
    static qUnsignedInt32       s_uiIOWaitTime;
    static qUnsignedInt32       s_uiRecvBufferSize;
    static qUnsignedInt32       s_uiSendBufferSize;
    static qBool                s_bActiveFaultDetection;

    static qUnsignedInt         s_TransportThreadSleepSpan;
    static qUnsignedInt         s_TransportThreadPriority;

    //2342  Flag to reuse buffers allocated using "new" (without throwing them away) in a transport buffer thread
    static qBool                s_bEnableTransportBufferThreadBufferPool;
    static qUnsignedInt         s_TransportBufferPoolNum;
    static qUnsignedInt         s_TransportBufferThreadSleepSpan;
    static qUnsignedInt         s_TransportBufferThreadPriority;
    static qUnsignedInt         s_TransportBufferThreadSendPriority;
    static qUnsignedInt         s_TransportBufferReceiveCount;
    static qUnsignedInt         s_TransportBufferSendCount;
    static qInt32               s_TransportBufferThreadCoreNo;


    static qUnsignedInt         s_NameResolveThreadPriority;
    static qInt32               s_NameResolveThreadCoreNo;


#if NEX_ENABLE_HYBRID_DETECT_LIMITATION
    static qUnsignedInt32 s_uiRoutingHistoryUpdateSpan;
    static qUnsignedInt32 s_uiRoutingHistoryTTL;
    static qUnsignedInt32 s_uiRoutingLimit;
#endif

    qList<PacketRecognizer*>    m_lstPacketRecognizers;
    static qVector<qUnsignedInt32>  s_dropIPList;

    NATProperties m_oNATProperties;
    UPnPProperties m_oUPnPProperties;
    qBool m_bNATCheckSupport;
    qBool m_bUPnPsupport;
    qBool m_bNATPMPSupport;
    qBool m_bNATPMPPortOpen;
    volatile qBool m_isNATCheckServerAddressesSet;
    volatile qBool m_isRelayServerAddressesSet;
    volatile qBool m_isSocketAlreadyClosed;

#if NEX_ENABLE_VIRTUAL_FILTER_DEVICE
    VirtualFilterDevice        *m_pFilterDevice;
#endif

    qVector< StringStream* > m_vecNATCheckAddresses;

    InetAddress m_relayAddress;
    String m_relayNameAddress;
};

END_NAMESPACE_NEX

#endif
