﻿/*---------------------------------------------------------------------------*
  (C) Copyright 1998-2014, Technologies Quazal Inc. All rights reserved.
  Quazal Net-Z, Quazal Rendez-Vous and Quazal are trademarks of Quazal.

  Copyright (C) 2011-2014 Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo and/or its licensors and are
  protected by national and international copyright laws.
  They may not be disclosed to third parties or copied or duplicated
  in any form, in whole or in part, without the prior written consent
  of Nintendo.
  The content herein is highly confidential and should be handled
  accordingly.
 *---------------------------------------------------------------------------*/

#ifndef _StationURL_H_
#define _StationURL_H_

#include <OnlineCore/src/Platform/Core/qList.h>
#include <OnlineCore/src/Platform/Core/PlatformDecl.h>
#include <OnlineCore/src/Transport/Interface/NATProperties.h>
#include <OnlineCore/src/Transport/Interface/TransportDecl.h>
#include <OnlineCore/src/Transport/Interface/Stream.h>
#include <OnlineCore/src/Platform/Core/String.h>
#include <OnlineCore/src/Platform/Core/ConnectionID.h>
#include <OnlineCore/src/Platform/Core/PrincipalID.h>

#ifdef NON_SUPPORT_C_TARGET
#endif

#ifdef WII_TARGET
//  {secret}
typedef unsigned char DWC_AID;
// {secret}
static const DWC_AID aidINVALID = static_cast<DWC_AID>(-1);
#endif

//n1589
#ifdef NEX_CLIENT_OPTIMIZATION
//typedef qUnsignedInt16 NODE_ID_TYPE;
typedef unsigned short NODE_ID_TYPE;
static const NODE_ID_TYPE INVALID_NODE_ID = 0;
#endif

#define URL_TYPE_DELIMITER          _T(":/")
#define URL_PARAM_DELIMITER         _T(";")
#define URL_USER_PARAM_DELIMITER    _T("#")
#define URL_TAG_ASSIGN              _T("=")

BEGIN_NAMESPACE_NEX

class InetAddress;


/*
Description: Constructs a station URL.

Remarks: Use this class to construct a Rendez-Vous URL for any given Station. A station URL should contain
at least one URL and a port and, for NAT Traversal, the Rendez-Vous Connection ID.



$ifdef WII_DWC$
Although this class is part of the standard $PRODUCTNAME$ distribution, in this
distribution, several additional methods are included to enable
you to set the Wii AID, which uniquely identifies a specific Wii Console.
$endif$

*/
/* Please see man pages for details 




*/
class StationURL: public RootObject
{

    //{secret}
    class Param;

    //{secret}
    friend class Param;

public:
    //group=Type Definitions

    // Description: Enumerates the various transport types that a StationURL may specify.
    // See Also: URLType
    /* Please see man pages for details 
    
    */
    enum URLType
    {
        // The transport type is unknown.
        unknown,  //
        // The transport type is PRUDP, Quazal's proprietary, reliable UDP transport protocol.
        prudp,    //
        // The transport type is PRUDPS, Quazal's proprietary, reliable, secure UDP transport protocol
        // that is used by Rendez-Vous, Quazal's lobby solution.
        prudps,   //
        // The transport type is UDP.  This represents a raw UDP protocol URL.  This can be used by developers
        // for feeding the game socket URL to Rendez-Vous.
        udp       //
    };

    // Description: Enumerates the various Flags that may be set for a StationURL.
    // See Also: Flags
    /* Please see man pages for details 
    

    
    */
    enum Flags
    {
        // The Station that this StationURL pertains to is behind a Network Address Translator (NAT).
        BehindNAT   = (1 << 0),   //
        // The StationURL refers to a public IP address and port number.
        Public      = (1 << 1)    //
    };

    enum PlatformType
    {
        PLATFORM_TYPE_UNKNOWN = 0,
        PLATFORM_TYPE_CTR = 1,
        PLATFORM_TYPE_WUP = 2,

        PLATFORM_TYPE_WINDOWS = 254,
        PLATFORM_TYPE_LINUX = 255 //Maximum number
    };

public:

    //Group=Methods

    /*
    Parameters:
    strURL: The StationURL as a string.
    szURL: The StationURL as a qChar. This value must be null-terminated.
    Remarks: A StationURL that uses the prudp URLType is expressed in the form:
    "prudp:/address=xx;port=xxxx", where the address may be specified in either text or numeric
    form, "www.quazal.com" or "192.168.10.1". For example:

            StationURL oStationURL(_T("prudp:/address=www.quazal.com;port=9103"));
            StationURL oStationURL(_T("prudp:/address=192.168.10.1;port=9103"));

    On the other hand a StationURL that uses the prudps URLType is expressed in the form:
    "prudps:/PID=xx". For example:

            StationURL oStationURL(_T("prudps:/PID=1005"));

    */
    /* Please see man pages for details 
     
     */
    StationURL();
    /* Please see man pages for details 
     

     
     */
    StationURL(const String &strURL);
    /* Please see man pages for details 
     

     
     */
    StationURL(const qChar* szURL);
    /* Please see man pages for details 
     

     
     */
    StationURL(const StationURL &oStationURL);
    /* Please see man pages for details 
     
     */
    virtual ~StationURL();

    /*
    Description: Sets the URLType that this StationURL will use.
    Parameters:
    eType: The URLType that this StationURL will use.
    See Also: GetURLType, URLType
    */
    /* Please see man pages for details 
    

    
    */
    void SetURLType(const URLType eType);

    /*
    Description: Returns the transport type specified by this StationURL.
    Returns     : The URLType specified by this StationURL.
    See Also: SetURLType, URLType
    */
    /* Please see man pages for details 
    

    
    */
    URLType GetURLType() const;

    /*
    Description: Sets the address that this StationURL will use.
    Parameters:
    strAddress: The address that this StationURL will use, which may be specified in either text
    or numeric form, for example, "www.quazal.com" or "192.168.10.1".
    Remarks: If you do not specify the address and port number of the StationURL in the constructor
    you can set it afterwards as follows:

                {
                    $PRODUCTNAME$ *g_p$PRODUCTNAME$ = qNew $PRODUCTNAME$;
                    StationURL urlSession;
                    urlSession.SetAddress(_T("192.168.10.1"));
                    // Or you could use an address in text form
                    // StationURL::SetAddress(_T("www.quazal.com"));
                    urlSession.SetPortNumber(9103);
                }

    Note: Addresses are only valid if the URLType is PRUDP.

    See Also: GetAddress, URLType
    */
    /* Please see man pages for details 
    

    

    
    */
    qBool SetAddress(const qChar *strAddress);
    //{secret}
    qBool SetAddress(const String &strAddress)
    {
        return SetAddress(strAddress.CStr());
    }

    /*
    Description: Returns the address specified by this StationURL.
    Returns     : The address specified by this StationURL.

    Note: Addresses are only valid if the URLType is PRUDP.

    $ifdef NON_SUPPORT_C$
    Note: (NON_SUPPORT_C Specific) If you are using $PRODUCTNAME$ for NON_SUPPORT_C with the NPTransportAdapter,
     do not use this method. Use SetNpId instead.
    $endif$

    See Also: SetAddress
    */
    /* Please see man pages for details 
     

     
     */
    String GetAddress() const;

    /*
    Description: Sets the port number that this StationURL will use.
    Parameters:
    ui16Port: The port number that this StationURL will use.
    Remarks: Refer to SetAddress for an example of this method in use.

    $ifdef NON_SUPPORT_C$
    Note: (NON_SUPPORT_C Specific) If you are using $PRODUCTNAME$ for the NON_SUPPORT_C with the NPTransportAdapter,
     do not use this method. Use SetNpIdGetConnectionID instead.
    $endif$

    Note: Port numbers are only valid if the URLType is PRUDP.
    See Also: GetPortNumber, URLType
    */
    /* Please see man pages for details 
    

    
    */
    void SetPortNumber(const qUnsignedInt16 ui16Port);

    /*
    Description: Returns the port number specified by this StationURL.
    Returns     : The port number specified by this StationURL.

    Note: Port numbers are only valid if the URLType is PRUDP.
    See Also: SetPortNumber
    */
    /* Please see man pages for details 
    

    

    
    */
    qUnsignedInt16 GetPortNumber() const;

    // {secret}
    void SetStreamID(const Stream::ID idStream);
    // {secret}
    Stream::ID GetStreamID() const;

    // {secret}
    void SetStreamType(const Stream::Type oStreamType);
    // {secret}
    Stream::Type GetStreamType() const;

#ifdef NON_SUPPORT_C_TARGET
#error "Not supported platform."
#endif

#ifdef WII_TARGET

    /*

    {group: Wii-DWC Specific Methods}

    $ifdef WII_DWC$

    Description: Sets this StationURL to the Wii ID (AID) which identifies a specific Wii console
    to the Nintendo matchmaking server.
    Parameters:
    ucAID: The AID that this StationURL will use.
    Remarks: In the $PRODUCTNAME$ distribution only, this method
    enables you to set the StationURL to the Nintendo matchmaking ID (AID) which
    identifies a specific Wii console.

    See Also: URLType
    $else$
    {secret}
    $endif$
    */
    void SetAID(DWC_AID ucAID);

    /*
    {group: Wii-DWC Specific Methods}

    $ifdef WII_DWC$
    Description: Retrieves the Wii ID (AID) contained in the StationURL.
    Parameters:
    ucAID: The AID that this StationURL uses.
    Remarks: In the $PRODUCTNAME$ distribution only, this method
    enables you to get the Nintendo matchmaking information (AID) contained
    in the StationURL.

    See Also: URLType
    $else$
    {secret}
    $endif$
    */
    DWC_AID GetAID() const;
#endif

#ifdef NEX_CLIENT_OPTIMIZATION
    void SetNodeId(qUnsignedInt16 nodeId);
    qUnsignedInt16 GetNodeId() const;
#endif

    /*
    {group: Rendez-Vous Specific Methods}

    Description: Sets the PrincipalID that this StationURL will use.
    Parameters:
    pid: The PrincipalID that this StationURL will use.
    Remarks: This method is only valid when $PRODUCTNAME$ is used in conjunction with
    Rendez-Vous, Quazal's lobby service.

    Note: PrincipalIDs are only valid if the URLType is PRUDPS.
    See Also: GetPrincipalID, URLType
    */
    /* Please see man pages for details 
    

    
    */
    void SetPrincipalID(const PrincipalID pid);

    /*
    {group: Rendez-Vous Specific Methods}

    Description: Returns the PrincipalID specified by this StationURL.
    Returns     : The PrincipalID specified by this StationURL.
    Remarks: This method is only valid when Net-Z is used in conjunction with
    Rendez-Vous, Quazal's lobby service.

    Note: PrincipalIDs are only valid if the URLType is PRUDPS.
    See Also: SetPrincipalID
    */
    /* Please see man pages for details 
    

    
    

    

    
    */
    PrincipalID GetPrincipalID() const;

    /*
    {group: Rendez-Vous Specific Methods}

    Description: Sets the ConnectionID that this StationURL will use.
    Parameters:
    cid: The ConnectionID that this StationURL will use.
    Remarks: This method is only valid when $PRODUCTNAME$ is used in conjunction with
    Rendez-Vous, Quazal's lobby service.

    Note: ConnectionIDs are only valid if the URLType is PRUDPS.
    See Also: GetConnectionID, URLType
    */
    void SetConnectionID(const ConnectionID cid);

    /*
    {group: Rendez-Vous Specific Methods}

    Description: Returns the ConnectionID specified by this StationURL.
    Returns     : The ConnectionID specified by this StationURL.
    Remarks: This method is only valid when $PRODUCTNAME$ is used in conjunction with
    Rendez-Vous, Quazal's lobby service.

    Note: ConnectionIDs are only valid if the URLType is PRUDPS.
    See Also: SetConnectionID
    */
    ConnectionID GetConnectionID() const;

    /*
    {group: Rendez-Vous Specific Methods}

    Description:  Appends the RVConnectionID to this StationURL.

    Remarks: This must be the unique ID assigned to each client
    when it connects to a Rendez-Vous lobby. The MatchMakingClient
    automatically adds this value to Station URLS.
    */

    /*
    

    
    

    
    */
    void SetRVConnectionID(const ConnectionID cid);

    /*
    {group: Rendez-Vous Specific Methods}

    Description:  Gets the unique RVConnectionID of this StationURL.

    Remarks: This is the unique ID assigned to each client when it connects to a Rendez-Vous
    lobby. The MatchMakingClient automatically adds this value to Station URLS.
    */
    /*
    

    
    */
    ConnectionID GetRVConnectionID() const;

    // {secret}
    void SetProbeRequestID(const qUnsignedInt32 prid);
    // {secret}
    qUnsignedInt32 GetProbeRequestID() const;
    // {secret}
    void SetFastProbeResponse(const qBool bFastProbeResponse);
    // {secret}
    qBool GetFastProbeResponse() const;

    // {secret}
    void SetNATMapping(NATProperties::MappingProperties eMappingType);
    // {secret}
    NATProperties::MappingProperties GetNATMapping() const;

    // {secret}
    void SetNATFiltering(NATProperties::FilteringProperties eFilteringType);
    // {secret}
    NATProperties::FilteringProperties GetNATFiltering() const;

    // {secret}
    void SetProbeRequestInitiation(qBool bProbeRequestInitiation);
    // {secret}
    qBool GetProbeRequestInitiation() const;

    // {secret}
    void SetUPnPSupport(qBool bSupportUPnP);
    // {secret}
    qBool GetUPnPSupport() const;

    // {secret}
    void SetNatPMPSupport(qBool bSupportNatPMP);
    // {secret}
    qBool GetNatPMPSupport() const;

    // {secret}
    void SetType(const qUnsignedInt32 uiFlags);
    // {secret}
    qUnsignedInt32 GetType() const;

    // {secret}
    void SetRelayServerAddress(const String& relayServerAddress);
    // {secret}
    String GetRelayServerAddress() const;

    // {secret}
    void SetRelayServerPort(qUnsignedInt16 relayServerPort);
    // {secret}
    qUnsignedInt16 GetRelayServerPort() const;

    void SetRelayAddress(const String& relayServerAddress);
    // {secret}
    String GetRelayAddress() const;

    // {secret}
    void SetRelayPort(qUnsignedInt16 relayServerPort);
    // {secret}
    qUnsignedInt16 GetRelayPort() const;

    // {secret}
    void SetUseRelayServer(qBool bUseRelayServer);
    // {secret}
    qBool GetUseRelayServer() const;

    /*
     * Returns a platform identifier.
     */
    qUnsignedInt32 GetPlatformType() const;

    void SetPlatformType();



    /*
    Description: Returns whether or not the Station that this StationURL pertains to is behind a NAT.
    Returns     : true if the Station that this StationURL pertains to is behind a Network Address Translator (NAT),
    false if not or an error occurred.
    See Also: Flags
    */
    /* Please see man pages for details 
    

    
    

    
    */
    qBool IsBehindNAT() const
    {
        return (GetType() & BehindNAT)==BehindNAT;
    }

    /*
    Description: Returns whether or not this StationURL refers to a public IP address and port number.
    Returns     : true if this StationURL refers to a public IP address and port number, false if not or an
    error occurred.
    See Also: Flags
    */
    /* Please see man pages for details 
    

    
    
    */
    qBool IsPublic() const
    {
        return (GetType() & Public)==Public;
    }

    //{secret}
    qBool SetInetAddress(const InetAddress *pInetAddress);
    //{secret}
    const InetAddress &GetInetAddress() const;

    /* Please see man pages for details 
     

     
     */
    StationURL &operator=(const StationURL &oStationURL);
    /* Please see man pages for details 
     

     
     */
    StationURL &operator=(const String &strURL);
    /* Please see man pages for details 
     

     
     */
    StationURL &operator=(const qChar* szURL);
    /* Please see man pages for details 
     
     */
    qBool operator==(const StationURL &oStationURL) const;
    /* Please see man pages for details 
     
     */
    qBool operator!=(const StationURL &oStationURL) const;

    /*
    Description: Returns the StationURL as a string.
    Returns     : The StationURL as a string.
    */
    /* Please see man pages for details 
    

    
    */
    String GetStringRepresentation() const
    {
        return GetURL();
    }

#ifdef FOR_DOC_PURPOSE
    /*
    Description: Sets user-defined attributes in the StationURL.
    Parameters:
    strKey: The Key variable, as a String, of the attribute to set.
    strValue: The Value variable, as a String, of the attribute to set.
    Remarks: An attribute is a user-defined Key-Value pair of String variables. Attributes
    are used to add additional information to the StationURL. Key and Value strings
    cannot contain the following characters since they are reserved for delimiters:
    '=',':','/','#',';'.

    See Also: Get, Remove
    */
    /* Please see man pages for details 
    

    

    
    

    
    */
    qBool Set(const String &strKey, const String &strValue);
#else
    // {secret}
    qBool Set(const String &strKey, const String &strValue, qBool bUserParam = false);
#endif

#ifdef FOR_DOC_PURPOSE
    /*
    Description: Returns the value associated with the specified key.
    Parameters:
    strKey: The Key, as a String, of the value to get.
    pstrValue: A pointer to store the Value for the given Key.
    Returns     : true if the given attribute exists, false otherwise.
    See Also: Set, Remove
    */

    /* Please see man pages for details 
    

    

    
    

    
    */
    qBool Get(const String &strKey, String *pstrValue) const;
#else
    // {secret}
    qBool Get(const String &strKey, String *pstrValue, qBool bUserParam = false) const;
#endif

#ifdef FOR_DOC_PURPOSE
    /*
    Description: Removes user-defined attributes from the StationURL.
    Parameters:
    strKey: The Key, as a String, of the attribute to remove.
    Returns     : true if the given attribute was removed, false is the attribute was not found.
    szKey: The Key variable, as a string, of the attribute to remove.
    See Also: Set, GetAttribute, FindAttribute, FindFirstAttribute, FindNextAttribute
    */
    /* Please see man pages for details 
    

    

    

    
    */
    qBool Remove(const String &strKey);
#else
    // {secret}
    qBool Remove(const String &strKey, qBool bUserParam = false);
#endif

    // Description: Traces the content of this StationURL object.
    // Remarks: This method traces the URL of this StationURL.
    /* Please see man pages for details 
    

    
    */
    void Trace(qUnsignedInt64 uiTraceFlags = TRACE_ALWAYS) const;

    /*
    Description: Returns the StationURL as a const C-style version of the string.
    Returns     : The StationURL as a const C-style version of the string.
    */
    /* Please see man pages for details 
    

    
    */
    const qChar *CStr() const
    {
        return GetURL();
    }

    //{secret}
    qResult  SetParamValue(const qChar* szKey, const qByte *pBuffer,
                           const qUnsignedInt uiBufferSize);
    //{secret}
    qResult  SetParamValue(const qChar* szKey, const qUnsignedInt uiValue);

    //{secret}
    qUnsignedInt32 GetParamValue(const qChar* szKey, qUnsignedInt32 uiNotFoundValue=0) const;

    //{secret}
    qBool  GetParamValue(const qChar* szKey, qByte *pBuffer, const qUnsignedInt uiBufferSize) const;

    //{secret}
    qResult GetURLType(qChar *szURLType, qUnsignedInt32 uiSize) const;

    //{secret}
    qUnsignedInt GetNbParam(qBool bUserParam = false);

    // {secret}
    qBool IsMatching(const StationURL& urlStation) const;

    // {secret}
    qBool IsAddressMatching(const StationURL &rurlStation) const;

    //{secret}
    qBool  IsValid() const;

    //{secret}
    const qChar *GetURL() const;
    //{secret}
    void SetURL(const qChar* szURL);


#ifdef TRACE_ENABLED
    static void TraceList(qUnsignedInt64 uiTraceFlags, const qList<StationURL>& oURLList, const qChar* szContext=NULL);
#endif

#ifdef TEST_ENABLED
    //{secret}
    static qBool Test();
#endif

private:

    typedef qMap<String,String> URLStrMap;
    typedef qMap<String,qUnsignedInt> URLIntMap;

    static qResult CopyInSuppliedBuffer(qChar *szDest, qUnsignedInt32 uiSize, const qChar *szSource,
                                        qUnsignedInt32 uiNeededSize);

    void Format();

    // Parsing functions
    void Parse();
    URLType ParseURLType() const;
    qBool ParseParams(qChar *ptr, qBool bUserParam );
    qBool ParseParam(qChar *ptr, qBool bUserParam );

    void Copy(const StationURL &urlStation);
    void ValidateData() const;
    void ValidateStr() const;

    // Internal version of Get functions that do not copy strings
    void GetIterator( const String &strKey, URLStrMap::iterator &it )
    {
        ValidateData();
        it = m_StrMap.find(strKey);
    }

private:
    // Special cases
    InetAddress *m_pAddr;
    URLType     m_urlType;

    // Parsed data
    URLIntMap  m_IntMap;
    URLStrMap  m_StrMap;
    URLStrMap  m_UserMap;

    qChar   *m_szURL;

    // State flags
    qBool m_bDataValid;
    qBool m_bStrValid;
    qBool m_bURLValid;
};


// String streaming.
// {secret}
extern StringStream& operator<<(StringStream& oStream, const StationURL& strValue);

END_NAMESPACE_NEX

#endif
