/*---------------------------------------------------------------------------*
  (C) Copyright 1998-2014, Technologies Quazal Inc. All rights reserved.
  Quazal Net-Z, Quazal Rendez-Vous and Quazal are trademarks of Quazal.

  Copyright (C)2011-2014 Nintendo  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo and/or its licensors and are
  protected by national and international copyright laws.
  They may not be disclosed to third parties or copied or duplicated
  in any form, in whole or in part, without the prior written consent
  of Nintendo.
  The content herein is highly confidential and should be handled
  accordingly.
 *---------------------------------------------------------------------------*/

#ifndef _TransportAdapter_H_
#define _TransportAdapter_H_

#include <OnlineCore/src/Transport/Interface/TransportDecl.h>
//n1589:todo3:class SocketDriver;
#include <OnlineCore/src/Platform/Stack/Core/SocketDriver.h>
#include <OnlineCore/src/Transport/Interface/StationURL.h>
#include <OnlineCore/src/Platform/Core/RefCountedObject.h>
#include <OnlineCore/src/Platform/Core/qProtectedList.h>
#include <OnlineCore/src/Core/CallContext.h>

//n1589
#include <OnlineCore/src/Platform/Stack/Core/InetAddress.h>

//n1589: Use BerkeleySocketDriver::GetSendMultiPort to eliminate header dependency
//#include <OnlineCore/src/Platform/Stack/Core/BerkeleySocketDriver.h>
//#define TRANSPORT_ADAPTER_SEND_MULTI_PORT BERKLEY_SOCKET_DRIVER_SEND_MULTI_PORT

BEGIN_NAMESPACE_NEX

class InterfaceInfo;

//group=Transport Protocol
// $ifdef WII_COMMON$
// group=Net-Z API
// $endif$

/*
Description: Base class for implementing all transport adapters.
Remarks: A transport adapter is responsible for interfacing the platform-independent Transport layer
with platform-specific or custom transport layers. For example, one could implement a shared memory
based (IPC) or file based transport layer. Derived TransportAdpaters should be registered to the
system by calling Network::RegisterTransportAdapter before creating the $FACADENAME$ object.
*/
class TransportAdapter: public RootObject
{
public:
    //n1589
    typedef enum
    {
        Basic = 0,
        UnreliableBroadcast,
        //n1589: memo: UnreliableSendMulti is supported only for NEX_CLIENT_OPTIMIZATION so far
        UnreliableSendMulti,

        LastTransportMode
    } ETransportMode;

    //n1589
    static void TraceInetAddress(qUnsignedInt64 uiTraceFlags, InetAddress* address);

    //n1589
    static qBool IsSendMultiAddress(const InetAddress* address);

    //group=
    TransportAdapter();
    virtual ~TransportAdapter();

    //n1589: Appropriate TransportMode is automatically selected
    void SetTransportMode(ETransportMode eMode);

    ETransportMode GetTransportMode()
    {
        return m_eTransportMode;
    }

    /*
    Description: Returns the list of URLs of the local machine.
    Remarks: These URLs should contain enough information to uniquely identify the local machine.
    The default implementation returns a StationURL for each IP addresses found on the local machine
    (for example, prudp:/address=203.132.83.23). An example specialization of this method for an Ethernet-based
    TransportAdapter could, for example, return the MAC address of the local machine (for example, prudp:/MAC=6A:33:D3:9A:6F:51).
    The StationURLs returned should always be of type prudp since this is ultimately the type of the
    upper-level transport that will use this URL. This method gets called after SocketDriver::Socket::Bind()
    to let the TransportAdapter know the URLs that can be used to talk to the local station.
    Parameters:
    plstLocalURLs: A pointer to a qList of StationURL that will be filled by the method.
    Returns: True is the list of Local URLs was retrieved successfully.
        */
    virtual qBool GetLocalURLs(qList<StationURL> *plstLocalURLs);

    /*
    Description: Resolves the address attribute of the given StationURL.
    Remarks: Ultimately, the underlying SocketDriver only deals with IP addresses, so this method should
    set the address attribute of the StationURL so that the address can be passed to the SocketDriver.
    The address should be a valid IP address in the form of 255.255.255.255. For example resolving
    prudp:/MAC=6A:33:D3:9A:6F:51 would lead to something like prudp:/address=192.168.10.12;MAC=6A:33:D3:9A:6F:51.
    This will be called when the prudp transport tries to establish a connection to a given station.
    Parameters:
    pCallContext: A pointer to a CallContext that will be used for asynchronous calls
    pURLtoResolve: A pointer to a StationURL that needs to be resolved.
    Returns: True if the address could be resolved, false otherwise.
    */
    virtual void ResolveAddress(CallContext *pCallContext, StationURL *pURLtoResolve);

    /*
    Description: Releases every resource (like connections) that may have been acquired during the ResolveAddress.
    Parameters:
    oStationToRelease: station which we want to release resources.
    Remarks: This should be called when we close an EndPoint.
    */
    virtual void ReleaseAddress(const StationURL &oStationToRelease);

    /*
    Description: Sorts and filters the Target URLs given the local URLs.
    Remarks: The method should filter out target URLs that cannot be used by the local station to connect to the target station.
    It should also sort the list from the best to the worse URL to use. This is mainly an optimization to
    reduce/improve connection time. For example, in NAT scenarios, if the local station and the target station
    both have the same public IP address (thus meaning they are behind the same NAT device), then this method
    will return the list of private (LAN) addresses first. Inversely, if the public IP addresses are different,
    the method will filter out the private (LAN) addresses.
    This will be called when the prudp transport tries to connect to a given station,
    just after ResolveAddress.
    Parameters:
    lstLocalURLs: StationURL qList of the local station.
    plstTargetURLs: Pointer to the qList of the target StationURLs. The method is responsible to sort and
    filter that list.
    */
    virtual void SortAndFilterTargetURLs(const qProtectedList<StationURL> &lstLocalURLs,
                                         qList<StationURL> *plstTargetURLs);

    /*
    Description: Identifies an exportable StationURL.
    Remarks: An exportable StationURL can be defined as a StationURL containing attributes that can uniquely
    and globally identify a station. An exportable StationURL can safely be sent over the network to
    another station and the later could use that StationURL to establish a connection. To make a parallel
    with distributed object systems, an object ID is exportable, but an object pointer is not.
    Parameters:
    refURL: The StationURL to test for exportability.
    Returns: True if the URL is exportable, false otherwise.
    */
    virtual qBool IsExportableURL(const StationURL &refURL);

    //  Given a valid URL, this method deduces an unique machine ID. See also ObjDup/StationInfo.h.

    // {secret}
    virtual qUnsignedInt32 GetUniqueMachineID(const StationURL& oURL) const;

    // {secret}
    virtual void Load() {};

    // {secret}
    virtual void Unload() {};

    // {secret}
    SocketDriver* GetSocketDriver()
    {
        return m_pSocketDriver;
    }

    //n1589:begin
    virtual qBool SetNetworkInterfaceInfo(InterfaceInfo* info);

    virtual qBool SetLocalAddress(const String& string);

    virtual qBool SetPortNumber(qUnsignedInt16 port);

#ifdef NEX_CLIENT_OPTIMIZATION
    InetAddress* GetBroadcastAddress()
    {
        return &m_oBroadcastAddress;
    }

    InetAddress* GetSendMultiAddress()
    {
        return &m_oSendMultiAddress;
    }
#endif // NEX_CLIENT_OPTIMIZATION

    InetAddress* GetLocalAddress()
    {
        return &m_oLocalAddress;
    }

    void Trace(qUnsignedInt64 uiTraceFlags);
    //n1589:end

protected:
    /*
    Description: Sets the SocketDriver instance to be used by this TransportAdapter.
    Parameters:
    pSocketDriver: Pointer to the custom SocketDriver to use.
    See Also: SocketDriver
    */
    void SetSocketDriver(SocketDriver* pSocketDriver)
    {
        m_pSocketDriver = pSocketDriver;
    }

    //n1589
    ETransportMode m_eTransportMode;
#ifdef NEX_CLIENT_OPTIMIZATION
    InetAddress m_oBroadcastAddress;
    InetAddress m_oSendMultiAddress;
#endif // NEX_CLIENT_OPTIMIZATION
    InetAddress m_oLocalAddress;

    //n1589:moved to protected
    SocketDriver* m_pSocketDriver;

private:
    // This map is only present to validate the *mechanism* for the management of udpp2p connections
    // with other targets (like win32). This map should be defined only when TEST_DEFINED, but it
    // raises SYSTEMCHECK and we do not have time to investigate. So for now, deactivate this
    // validation.
#ifdef VALIDATE_UDPP2P_CONNECTION_LIFETIME
    qMap<String, qUnsignedInt32> m_oMapAddressWithConnectionId;
#endif
};


END_NAMESPACE_NEX

#endif
