/*---------------------------------------------------------------------------*
  (C) Copyright 1998-2014, Technologies Quazal Inc. All rights reserved.
  Quazal Net-Z, Quazal Rendez-Vous and Quazal are trademarks of Quazal.

  Copyright (C)2011-2014 Nintendo  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo and/or its licensors and are
  protected by national and international copyright laws.
  They may not be disclosed to third parties or copied or duplicated
  in any form, in whole or in part, without the prior written consent
  of Nintendo.
  The content herein is highly confidential and should be handled
  accordingly.
 *---------------------------------------------------------------------------*/

#ifndef _PRUDPEndPoint_H_
#define _PRUDPEndPoint_H_

#include <OnlineCore/src/Transport/PRUDP/PRUDPDecl.h>
#include <OnlineCore/src/Transport/Interface/EndPoint.h>
#include <OnlineCore/src/Transport/PRUDP/IORequestContext.h>
#include <OnlineCore/src/Platform/Core/Time.h>
#include <OnlineCore/src/Transport/PRUDP/RTT.h>
#include <OnlineCore/src/Platform/Stack/Core/InetAddress.h>
#include <OnlineCore/src/Platform/Core/ProfilingUnit.h>
#include <OnlineCore/src/Core/LogicalClock.h>
#include <OnlineCore/src/Transport/Interface/PacketEncDec.h>
#include <OnlineCore/src/Core/Checksum/Interface/SignatureBytes.h>
#include <OnlineCore/src/Transport/UDP/PacketDefines.h>

//n1589: Make it possible to get the fault reason from the application
#define NEX_ENABLE_PRUDP_RECORD_FAULT_REASON 1

//n1589: Make it possible to do profiling for the transmission state of PRUDP, so profiling will be necessary in the product build for application development, with only a slight processing load
#ifdef NEX_CLIENT_OPTIMIZATION
#define NEX_ENABLE_PRUDP_PROFILING 1
#endif
#include <OnlineCore/src/Transport/Interface/EndPointInfo.h>

//n1589: Make it possible for the application to adjust to the re-send packet timeout
#define NEX_ENABLE_CALC_RETRANSMISSION_RTO_CALLBACK 1

BEGIN_NAMESPACE_NEX

class PRUDPStream;
class PacketDispatchQueue;
class SlidingWindow;
class PacketOut;
class PacketIn;
class Packet;
class PRUDPMessageSelector;

/* Please see man pages for details 
 
 
 
 
 
 
 
 
 
 
 
 */
class PRUDPEndPoint: public EndPoint
{
    friend class PRUDPStream;
    friend class TimeoutManager;

public:


    static const qUnsignedInt32 SupportedFunctionsMinorVersionMask = 0xFF;
    static const qUnsignedInt32 SupportedFunctionsMinorVersionMaskShift = 0;
    static const qUnsignedInt32 SupportedFunctionsUsingOptionMask = 0xFFFFFF00;
    static const qUnsignedInt32 SupportedFunctionsUsingOptionMaskShift = 8;

    enum SupportedFunctionsFlags
    {
        // option mask (32bit)
        ENABLE_FAST_SORRY                                  = 0x000001ULL,
        //over 0xFFFFFF is created from minor vesion (64bit)
        SUPPROTED_FLAG_AGGREGATE_ACK                    = 0x001000000ULL,  //minor version 1
        SUPPROTED_FLAG_NEGOTIATION_CONNECTION_SIGNATURE = 0x002000000ULL,  //minor version 2
        SUPPROTED_FLAG_MULTI_SUBSTREAM                  = 0x004000000ULL,  //minor version 2
        SUPPROTED_FLAG_RVDDLCLASSREVISION               = 0x008000000ULL,  //minor version 3
        SUPPROTED_FLAG_FAST_SORRY                       = 0x010000000ULL   //minor version 4
    };


#if NEX_ENABLE_CALC_RETRANSMISSION_RTO_CALLBACK
    typedef qUnsignedInt32 (*PRUDPCalcRetransmissionTimeoutCallback)(qUnsignedInt32 uiRto, qUnsignedInt16 uiNbTransmission);
    static void SetCalcRetransmissionTimeoutCallback(PRUDPCalcRetransmissionTimeoutCallback callback);
#endif

    PRUDPEndPoint(ConnectionOrientedStream *pStream, const StationURL *pStationURL,
                  LogicalClock16 *p_oNextUnreliableBroadcastSeqIdToSend, qByte version);
    virtual ~PRUDPEndPoint();

    virtual qResult _Connect(Buffer *pOutBuffer, Buffer *pInBuffer,
                             const EndPoint::Callback pCallback, const UserContext &oCallbackContext,
                             qUnsignedInt32 uiTimeout);

    virtual qResult _Disconnect(const EndPoint::Callback pCallback, const UserContext &oCallbackContext,
                                qUnsignedInt32 uiTimeout);

    virtual qResult         _Send(Buffer *pBuffer, qUnsignedInt32 uiFlags, qBool qEncrypt=false, qByte substream=0, qUnsignedInt32 uiCallID=0);

    /* Please see man pages for details 
     
     
     
     
     
     
     
     
     */
    virtual qUnsignedInt32  GetRTT()  const
    {
        return m_oRTT.GetLastRTT();
    }

    /* Please see man pages for details 
     
     
     
     
     
     
     
     
     */
    virtual qUnsignedInt32  GetRTTAverage() const
    {
        return m_oRTT.GetRTTSmoothedAvg();
    }

    qUnsignedInt32  GetNbReliablePending(qByte substream);

    RTT             *GetRTTPtr()
    {
        return &m_oRTT;
    }

    virtual qBool IsNotConnected()
    {
        return (m_eConnectionState == NotConnected);
    }
    virtual qBool IsConnecting()
    {
        return (m_eConnectionState == Connecting);
    }
    virtual qBool IsDisconnecting()
    {
        return (m_eConnectionState == Disconnecting);
    }
    virtual qBool IsFaulty()
    {
        return (m_eConnectionState == Faulty);
    }
    virtual qBool IsConnected()
    {
        return (m_eConnectionState == Connected);
    }
    virtual ConnectionState GetConnectionState() const
    {
        return m_eConnectionState;
    }
    virtual qBool PeerIsConnected()
    {
        return m_bPeerConnected;
    }
    virtual qBool PeerIsDisconnected()
    {
        return !m_bPeerConnected;
    }

    virtual void            SetKeepAliveTimeout(qUnsignedInt32 uiKeepAliveTimeout);
    virtual qUnsignedInt32  GetKeepAliveTimeout()
    {
        return m_uiKeepAliveTimeout;
    }

    virtual qUnsignedInt32 GetMaxSilenceTime()
    {
        return m_uiMaxSilenceTime;
    }

    virtual void SetMaxSilenceTime(qUnsignedInt32 uiMaxSilenceTime)
    {
        m_uiMaxSilenceTime = uiMaxSilenceTime;
    }

    //  This is called when the endpoint receives an ack on a CONNECT request from the peer.
    virtual void SetPeerConnected()
    {
        m_bPeerConnected = true;
#ifdef TRACE_ENABLED    // n2370
        Trace();
#endif
    }

    virtual void SetPeerDisconnected()
    {
        m_bPeerConnected = false;
#ifdef TRACE_ENABLED    // n2370
        Trace();
#endif
    }

    const InetAddress *GetInetAddress()
    {
        return &m_oStationURL.GetInetAddress();
    }

    //n2690
    void SetIsOldRVSecure(qBool flag)
    {
        m_IsOldRVSecure = flag;
    }
    qBool IsOldRVSecure() const
    {
        return m_IsOldRVSecure;
    }

    qBool IsDuplicateReorderingPacket(PacketIn *pPacket);

    virtual qBool IsAnyPacketProcessed();

    //n1589:moved to public
    Stream::ID GetStreamID() const
    {
        return m_oStationURL.GetStreamID();
    }

    qByte GetPeerSessionID() const
    {
        return m_byPeerSessionID;
    }

    qByte GetSessionID() const
    {
        return m_bySessionID;
    }

    qByte           GetVersion() const
    {
        return m_version;
    }

    qUnsignedInt32  GetSupportedFunctions() const
    {
        return  m_uiSupportedFunctions;
    }

#if NEX_ENABLE_PRUDP_RECORD_FAULT_REASON
    qByte GetFaultReason() const
    {
        return m_byFaultReason;
    }
#endif

#if NEX_ENABLE_USE_SECURE_KEY
    virtual void SetEncryptionKey(const Key &oKey)
    {
        m_oPacketEncDec.SetEncryptionKey(oKey);
    }
    virtual Key *GetEncryptionKey()
    {
        return m_oPacketEncDec.GetEncryptionKey();
    }
#endif

#if NEX_ENABLE_HYBRID
    static void SetHybridPartialFaultProcessingMode(qBool bRecover);
    static qBool s_bRecoverPartialFault;
#endif

    static PseudoGlobalVariable<qByte> s_sessionID;

    /*
     
     
     
     */
    virtual void ResetEndPointInfo();

    /*
     
     
     
     
     
     
     */
    virtual qBool GetEndPointInfo(EndPointInfo* info) const;


    /* Please see man pages for details 
     
     
     
     
     
     
     */
    virtual qUnsignedInt32 GetReliableBufferPacketCount() const;

    qUnsignedInt32 GetReliableBufferPacketCount(qByte substream) const;

    /* Please see man pages for details 
     
     
     
     
     
     
     */
    virtual void GetReliableBufferPacketCounts(qVector<qUnsignedInt32>& counts) const;

    /* Please see man pages for details 
     
     
     
     
     
     */
    virtual qByte GetMaxSubStreamID() const
    {
        return static_cast<qByte>(m_uiMaxUserReliableSubStreams);
    }

    void AnalyzeRttRto(qUnsignedInt32 rto);

    EndPointInfo* GetAnalyzeInfo()
    {
        return &m_AnalyzeInfo;
    }

    /*
    n1589: The API that calculates these packet loss rates will not show the proper values because the sequence number is skipped in stations that receive broadcast packets addressed to other station addresses.
   
    */
    //n2788: Packet loss rate calculation function
    //      Returns the packet loss rate at the time the EndPoint instance was created or from the time ResetAnalyzeInfo final call up to the instant the API was called.
    //      
    qFloat PacketLossRateInReceiveUnreliable();
    //n2788: Packet loss rate calculation function (different specs)
    //      Returns the packet loss rate and the total number of packets that should have been received from either the creation of the EndPoint instance or the last call to the ResetAnalyzeInfo function to the instant when this API function was invoked.
    //      
    void PacketLossRateInReceiveUnreliable(qFloat *packetLossRate, qUnsignedInt32 *totalPackets);
    //n2788: Supplementary packet loss rate calculation function
    //      Returns 'true' if the total number of packets that should have been received since either the creation of the EndPoint instance or the last call to the ResetAnalyzeInfo function exceeds the specified value. Returns 'false' otherwise.
    //      
    qBool TotalPacketsInReceiveUnreliableIsOver(const qUnsignedInt32 thresholdPackets);

    // Get the maximum size of a packet.
    qUnsignedInt32 GetMaxPacketSize();

#ifdef TEST_ENABLED
    static qBool Test();
#endif


    static inline qUnsignedInt32 SetSupportedFunctionsMinorVersion(qUnsignedInt32 supportedFunctions, qByte version)
    {
        qUnsignedInt32 v = (static_cast<qUnsignedInt32>(version) << SupportedFunctionsMinorVersionMaskShift)
                           & SupportedFunctionsMinorVersionMask;
        return (supportedFunctions & (~SupportedFunctionsMinorVersionMask)) | v;
    }
    static inline qByte GetSupportedFunctionsMinorVersion(qUnsignedInt32 supportedFunctions)
    {
        return static_cast<qByte>((supportedFunctions & SupportedFunctionsMinorVersionMask) >>  SupportedFunctionsMinorVersionMaskShift);
    }

    static inline qUnsignedInt32 SetSupportedFunctionsOptionMask(qUnsignedInt32 supportedFunctions, qUnsignedInt32 mask)
    {
        qUnsignedInt32 m = (mask << SupportedFunctionsUsingOptionMaskShift)
                           & SupportedFunctionsUsingOptionMask;
        return (supportedFunctions & (~SupportedFunctionsUsingOptionMask)) | m;
    }


    static inline qUnsignedInt32 GetSupportedFunctionsOptionMask(qUnsignedInt32 supportedFunctions)
    {
        return (supportedFunctions & SupportedFunctionsUsingOptionMask) >>  SupportedFunctionsUsingOptionMaskShift;
    }

    static inline qUnsignedInt32 SetSupportedFunctionsOptionFlag(qUnsignedInt32 supportedFunctions, qUnsignedInt32 flag)
    {
        return SetSupportedFunctionsOptionMask(supportedFunctions, GetSupportedFunctionsOptionMask(supportedFunctions) | flag);
    }

    static inline qUnsignedInt32 ClearSupportedFunctionsOptionFlag(qUnsignedInt32 supportedFunctions, qUnsignedInt32 flag)
    {
        return SetSupportedFunctionsOptionMask(supportedFunctions, GetSupportedFunctionsOptionMask(supportedFunctions) & (~flag));
    }

    static inline qBool DoesSetSupportedFunctionsOptionFlag(qUnsignedInt32 supportedFunctions, qUnsignedInt32 flag)
    {
        return (GetSupportedFunctionsOptionMask(supportedFunctions) & flag) != 0;
    }

    static qBool IsNegotiationConnectionSignatureVersion(const Packet *pPacket);

    virtual qUnsignedInt64 GetSupportedFunctionsFlags() const
    {
        return m_uiSupportedFunctionsFlags;
    }

    virtual qUnsignedInt32 GetSignaledFaultError() const
    {
        return m_uiSignaledFaultError;
    }

protected:
#if NEX_ENABLE_CALC_RETRANSMISSION_RTO_CALLBACK
    static PRUDPCalcRetransmissionTimeoutCallback s_pfCalcRetransmissionTimeoutCallback;
#endif

    qResult Frag(
        Buffer *pBuffer,
        qUnsignedInt32 uiMTU,
        qUnsignedInt32 uiMaxPacketSize,
        PacketType uiPacketType,
        qUnsignedInt16 uiPacketFlags,
        qBool bExpress,
        qUnsignedInt32 uiCallID,
        qByte subStreamID
    );

    qBool           Defrag(PacketIn *pDispatchPacket);
    PacketOut       *GetNextSendPacket();
    PacketIn        *GetNextDispatchPacket(qByte substream);
    void            Dispatch(PacketIn *pPacket, Time oIOTime);
    //n1589
    void HandlerDispatch(Buffer* pBuffer, PacketIn* pPacket);

    void            ServiceIncomingPacket(PacketIn *pPacket);

    void            ProcessConnect(Packet *pPacket);
    void            ServiceTimeout(PacketOut *pPacket);
    virtual void    SignalFaultEvent(qUnsignedInt32 uiError, qBool bProcessNow);
    void            ProcessFaultEvent(qUnsignedInt32 uiError);

    qBool           EarlyConnectAllowed();
    void            AllowEarlyConnect();

    void            ConnectCompleted();

    void            PacketAggregateAcknoledged(Packet *pPacket);
    void            PacketAcknowledged(const Time &recvIoTime, const LogicalClock16& ackSeq, qByte subStreamID);
    qBool           CancelPacketTimeout(PacketOut *pPacket);
    void            StartPing();
    void            StopPing();
    void            TimeToPing();

    void            SendAggregateACK();

    qBool           ReadyToDisconnect();

    virtual qBool   SetConnectionState(ConnectionState eNewState);
    qBool           CompleteIO(qResult result);

    void            AdjustRTT(const Packet *pPacket, const PacketOut *pAckedPacket);
    void            AdjustRTT(const Time &ioTime, const PacketOut *pAckedPacket);
    qUnsignedInt    ComputeRetransmitTimeout(const PacketOut* pPacket);
    qReal           GetTimeoutMultiplier(qUnsignedInt16 ui16NbTransmission, StreamSettings &rSettings)const;

    void            SendNextReliable(qByte subStreamID);
    void            SendReliable(PacketOut *pPacket);
    qResult         SendPacket(PacketOut *pPacket);
    qResult         HandleSlidingWindowPushError();

    void            SetPreferredPort(qUnsignedInt16 uiPort)
    {
        m_uiPort = uiPort;
    }
    qUnsignedInt32  GetPreferredPort()
    {
        return m_uiPort;
    }

    qBool           IsIdle(qByte substream);
    qBool           IsFlooded();

    qBool IsSendAggregateACK();

    PRUDPStream *GetStream()
    {
        return (PRUDPStream*)m_pStream;
    }

    static const qChar* GetNetFaultString(qUnsignedInt32 uiError);

    qUnsignedInt16              m_uiPort;

    qVector<PacketDispatchQueue *> m_dispatchQueue;
    qVector<SlidingWindow *>    m_slidingWindow;

    PacketOut                   *m_pPingPacket;
    PacketOut                   *m_pSynPacket;
    qBool                       m_bSynPhaseCompleted;
    qUnsignedInt32              m_uiMaxSilenceTime;
    qUnsignedInt32              m_uiKeepAliveTimeout;
    ConnectionState             m_eConnectionState;
    qBool                       m_bPeerConnected;
    qUnsignedInt32              m_uiSignaledFaultError;

    RTT                         m_oRTT;

    Time                        m_oLastPacketSentTime;
    Time                        m_oLastPacketRecvTime;

    LogicalClock16              m_lcNextUnreliableSeqIdToSend;
    LogicalClock16              m_lcNextUnreliableSeqIdToRecv;
    LogicalClock16              *m_pNextUnreliableBroadcastSeqIdToSend; //n2967: Shared among all PRUDPEndPoints from the PRUDPStream
    LogicalClock16              m_lcNextUnreliableBroadcastSeqIdToRecv;

    qUnsignedInt16              m_uiMaxUserReliableSubStreams;

    SignatureBytes         m_oSignature;
    SignatureBytes         m_oSendConnectionSignature;

    SignatureBytes* GetConnectionSignature()
    {
        return &m_oSignature;
    }
    SignatureBytes* GetSendConnectionSignature()
    {
        return &m_oSendConnectionSignature;
    }

    qVector<Buffer *>           m_currentPartialBuffer;
    qVector<qByte>              m_nextPartToRecv;

    //n1589:memo:this is not joining session id but endpoint session id
    qByte                       m_byPeerSessionID;
    qByte                       m_bySessionID;
#if NEX_ENABLE_PRUDP_RECORD_FAULT_REASON
    qByte m_byFaultReason;
#endif

    //m2967: Broadcast packets sometimes may have a different SessionID than the normal one since they are sent from the representative uppermost StationID.
    qByte                       m_byBroadcastPeerSessionID;

    qByte                       m_version;

    qUnsignedInt32              m_uiSupportedFunctions;

    qUnsignedInt64              m_uiSupportedFunctionsFlags;

    //n2690 Client using the old RVSecure
    qBool m_IsOldRVSecure;
    qVector<qBool> m_needAggregateACK; //Note: qBool


    ProfilingUnit               m_puSend;

    IORequestContext            m_oCurrentIORequestContext;

    PRUDPMessageSelector*       m_pPRUDPMessageSelector;


#if NEX_ENABLE_USE_SECURE_KEY
    PacketEncDec m_oPacketEncDec;
#endif

#if NEX_ENABLE_HYBRID
    //n2781 Do not fault when receiving a connection from a destination during connection processing
    qBool m_bAvoidFaultWhileConnecting;
#endif

    EndPointInfo m_AnalyzeInfo;
private:
    // Number of unreceived acks for CallContext
    qMap<qUnsignedInt32, qUnsignedInt32> m_unackedCount;

    void InitUnreliableSeq(PacketOut *packet);

private:
    void SendACK(PacketIn *pPacket);
    void SendAggregateACK(Buffer* pBuffer, qUnsignedInt16 highestSeriesSeq, qByte subStreamID);
    void SetSupportedFunctionsFlags();
    void SendAggregateACKOld();
    void SendAggregateACKNew();
    void DecideMaxUserReliableSubStreams(qUnsignedInt16 maxUserReliableSubStreams);
    void PacketAggregateAcknoledgedOld(Packet *pPacket);
    void PacketAggregateAcknoledgedNew(Packet *pPacket);
    void SendResetDisconnectPacket();



    PRUDPMessageInterface::SignatureMethod DecideSignatureMethod(const Packet *pPacket, Stream::Type type, qBool hasEndPoint);
};


END_NAMESPACE_NEX

#endif
