﻿/*---------------------------------------------------------------------------*
  (C) Copyright 1998-2014, Technologies Quazal Inc.  All rights reserved.
  Quazal Net-Z, Quazal Rendez-Vous and Quazal are trademarks of Quazal.

  Copyright (C)2011-2014 Nintendo  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo and/or its licensors and are
  protected by national and international copyright laws.
  They may not be disclosed to third parties or copied or duplicated
  in any form, in whole or in part, without the prior written consent
  of Nintendo.
  The content herein is highly confidential and should be handled
  accordingly.
 *---------------------------------------------------------------------------*/

#ifndef _SecureStream_H_
#define _SecureStream_H_

#include <OnlineCore/src/Transport/Interface/TransportDecl.h>
#include <OnlineCore/src/Transport/Interface/ConnectionOrientedStream.h>
#include <OnlineCore/src/Transport/Interface/TransportError.h>
#include <RendezVous/Core/src/Client/SecureTransport/SecureEndPoint.h>
#include <RendezVous/Core/src/Client/Credentials.h>

#define SECURE_TRANSPORT_TYPE   _T("prudps")

#define CID_URL_KEY     _T("CID")
#define PID_URL_KEY     _T("PID")

BEGIN_NAMESPACE_NEX

class SecureConnectionClient;
class AuthenticationClient;

/*
\ingroup group_rdv_core_services_securetransport_client
\brief Use this class to specify specific attributes of a transport plug-in.
\remarks This class is used to specify specific attributes of a transport plug-in.
*/
class OSDK_API SecureStream: public ConnectionOrientedStream
{
public:
    SecureStream();
    virtual ~SecureStream();

    AuthenticationClient *GetAuthenticationClient()
    {
        return m_pAuthenticationClient;
    }
    SecureConnectionClient *GetSecureConnectionClient()
    {
        return m_pSecureConnectionClient;
    }

    ConnectionOrientedStream * GetAssociatedStream()
    {
        return m_pAssociatedStream;
    }

    /** \internal */
    void SetAssociatedStream(ConnectionOrientedStream *pTransport);
    ConnectionOrientedStream *GetAssociatedStream() const;

    /** \internal */
    virtual qBool           ResponsibleForURL(const StationURL *pStationURL);
    /** \internal */
    virtual qChar           *GetURLType();

    /** \internal */
    virtual EndPoint        *OpenEndPoint(const StationURL *pStationURL);
    /** \internal */
    virtual EndPoint        *OpenEndPoint(const ConnectionID cidStation);
    /** \internal */
    virtual qBool           OpenEndPoint(EndPoint *pEndPoint)
    {
        return ConnectionOrientedStream::OpenEndPoint(pEndPoint);
    }
    /** \internal */
    virtual EndPoint        *FindEndPointByPID(const PrincipalID pidStation);
    /** \internal */
    virtual EndPoint        *FindEndPointByCID(const ConnectionID cidStation);
    /** \internal */

    virtual const qVector<ConnectionID> GetAllConnectionIDs() const;
    /** \internal */
    virtual qUnsignedInt32  GetEndPointNumber();
    /** \internal */
    virtual EndPoint        *CreateEndPoint(const StationURL *pStationURL);
    /** \internal */
    virtual void            CloseEndPoint(EndPoint *pEndPoint);

    virtual qBool ReceiveIncomingPacket(qUnsignedInt16 uiPort, Stream::ID idStream, Packet *pPacket);
    virtual void DoWork();

    virtual void LockEndPointMap();
    virtual void UnlockEndPointMap();


    /**
    \brief Returns the number of local URLs.
    \returns The number of local URLs. */
    virtual qUnsignedInt32  GetNbLocalURL();

    /**
    \brief Returns the local URL.
    \returns The URL of the local station. */
    virtual StationURL      *GetLocalURL(qUnsignedInt32 uiIndex);

    /**
    \brief Deletes the station URL obtained via GetIPURL or GetLocalURL. */
    virtual void            ReleaseURL(StationURL *pStationURL);

    virtual StationURL      *BuildURL(qUnsignedInt32 uiCID, qUnsignedInt32 uiPID);

    virtual qResult         Send(StationURL *pStationURL, Buffer *pBuffer)
    {
        DUMMY_PARAM(pStationURL);
        DUMMY_PARAM(pBuffer);
        return QRESULT_ERROR(Transport, Unknown);
    }
    virtual qResult         Send(StationURL *pStationURL, Buffer *pBuffer,
                                 const ConnectionOrientedStream::Callback pCallback, const UserContext &oContext)
    {
        DUMMY_PARAM(pStationURL);
        DUMMY_PARAM(pBuffer);
        DUMMY_PARAM(pCallback);
        DUMMY_PARAM(oContext);
        return QRESULT_ERROR(Transport, Unknown);
    }
    virtual qBool           IsCapable(qUnsignedInt32 uiCapability)
    {
        DUMMY_PARAM(uiCapability);
        return false;
    }
    virtual qResult         SetOption(qUnsignedInt32 uiOption, const UserContext &oOptionSetting)
    {
        DUMMY_PARAM(uiOption);
        DUMMY_PARAM(oOptionSetting);
        return QRESULT_ERROR(Transport, Unknown);
    }
    virtual qResult         GetOption(qUnsignedInt32 uiOption, UserContext &oOptionSetting)
    {
        DUMMY_PARAM(uiOption);
        DUMMY_PARAM(oOptionSetting);
        return QRESULT_ERROR(Transport, Unknown);
    }

    /** \internal */
    void SetCredentials(Credentials* pCredentials);
    Credentials* GetCredentials()
    {
        return m_pCredentials;
    }

    qBool FilterIncomingConnection(Buffer* pInBuffer, Buffer* pOutBuffer, EndPoint* pEndPoint);

private:
    ConnectionOrientedStream *m_pAssociatedStream;

    AuthenticationClient *m_pAuthenticationClient;
    SecureConnectionClient *m_pSecureConnectionClient;
    Credentials *m_pCredentials;
};

extern qBool SecureStreamConnectionFilter(const UserContext& oContext, Buffer* pInBuffer, Buffer* pOutBuffer, EndPoint* pEndPoint);

END_NAMESPACE_NEX

#endif
