﻿/*---------------------------------------------------------------------------*
  (C) Copyright 1998-2014, Technologies Quazal Inc. All rights reserved.
  Quazal Net-Z, Quazal Rendez-Vous and Quazal are trademarks of Quazal.

  Copyright (C) 2011-2014 Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo and/or its licensors and are
  protected by national and international copyright laws.
  They may not be disclosed to third parties or copied or duplicated
  in any form, in whole or in part, without the prior written consent
  of Nintendo.
  The content herein is highly confidential and should be handled
  accordingly.
 *---------------------------------------------------------------------------*/

#ifndef _AnyDataAdapter_H
#define _AnyDataAdapter_H

#include <OnlineCore/src/Platform/Core/PlatformDecl.h>
#include <RendezVous/Core/src/Common/Data.h>
#include <RendezVous/Core/src/Common/AnyObjectAdapter.h>

BEGIN_NAMESPACE_NEX

/*
\ingroup group_rdv_core_common
\brief This Adapter enables information defined in custom Data classes to be correctly
streamed between a client and the \productname back end.
\remarks By default, the data streamed by an AnyDataAdapter is stored as a BLOB in the \productname
database. Therefore, if you want the service to be able to interpret the data you need to implement
a type extension script, as detailed in the \productname User's Guide.
*/
class OSDK_API AnyDataAdapter: public AnyObjectAdapter<Data, String>
{
public:
    /*
    \brief Constructor
    \remarks An AnyDataAdapter object should never be directly instantiated.
    Instead, it is created by \productname when a client logs into the lobby. To access the AnyDataAdapter
    object from within your game use the GetInstance method.
    AnyDataAdapter itself is initialized statically. It must be initialized using the AnyDataAdapter::Register function before it can be used.
    */
    AnyDataAdapter();

    /*
    \brief Destructor
    \remarks An AnyDataAdapter object should never be directly destroyed. It is destroyed by \productname
    when a given client logs off from the lobby.
    AnyDataAdapter itself is initialized statically. It must be finalized using the AnyDataAdapter::Unregister function after being used.
    */
    virtual ~AnyDataAdapter();

    /*
    \brief Registers a specific custom Data class to this AnyDataAdapter.
    \param pNewPrototype A pointer to the custom Data class that will be registered to this AnyDataAdapter.
    \returns true if the specified Data class was successfully registered, false if not, if the specified prototype
    is already currently registered, or an error occurred.
    \remarks In order for a custom Data class to be used by \productname, it must first be registered
    via an AnyDataAdapter. For example, the MyPrivateData class would be registered as follows:
    \code
            AnyDataAdapter::GetInstance()->RegisterPrototype(qNew MyPrivateData);
    \endcode
    \note Note that if your custom data is implemented via a hierarchy, then you will need to register all of the derived classes
    in your custom hierarchy that you wish to use.
    \sa GetInstance
    */
    qBool RegisterPrototype(Data* pNewPrototype);

    /* Please see man pages for details */
    qBool UnregisterPrototype(const String& strDataType);
    /* Please see man pages for details */
    void UnregisterAllPrototypes();
    /* Please see man pages for details */
    const Data* FindPrototype(const String& strDataType) const;
    /* Please see man pages for details */
    Data* CreateObject(const String& strType, qBool isOldVersion) const;
    /* Please see man pages for details */
    virtual void StreamIn(Message* pMessageToSend, const Data& oValue) const;
    /* Please see man pages for details */
    virtual void StreamOut(Message* pReceivedMessage, Data **ppValue) const;

    /*
    \brief Returns a pointer to the AnyDataAdapter instance.
    \returns A reference to the AnyDataAdapter object.
    \remarks Use the reference returned by this member function to control how this Adapter operates. For example,
    register the MyPrivateData class as follows:
    \code
            AnyDataAdapter::GetInstance()->RegisterPrototype(qNew MyPrivateData);
    \endcode
    */
    static AnyDataAdapter* GetInstance()
    {
        return &s_oInstance;
    }

    /* Please see man pages for details */
    static void Register();

    //{secret}
    static void Unregister();


private:
    static AnyDataAdapter s_oInstance;
};

END_NAMESPACE_NEX

#endif
